#!/bin/bash
#
# Run gnulib tests under Syd.
#
# Copyright 2024 Ali Polatel <alip@chesswob.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# TEST_LIMIT: Number of tests to run at a time.
# TEST_PATTERN: A Regex (PCRE) for tests to run.
case "$1" in
    '')
        TEST_LIMIT=250
        TEST_PATTERN='.*'
        ;;
    [0-9]*)
        TEST_LIMIT="${1}"
        TEST_PATTERN='.*'
        ;;
    *)
        TEST_LIMIT=250
        TEST_PATTERN="${1}"
        ;;
esac

# A Regex (PCRE) for tests to skip.
SKIP_PATTERN='(bison|datetime|exception|gettext|link-warning|printf|trapping|windows|^(array|java|uni)|-ieee$|(^(execinfo|gnumakefile|havelib|timevar)$))'

# Tests that have failed in the past.
# truncate: https://gitlab.exherbo.org/sydbox/sydbox/-/jobs/83046
# copy-file, utimensat: https://builds.sr.ht/~alip/job/1257729
FAIL_HISTORY=(
    copy-file
    truncate
    utimensat
)

# Make sure we don't trigger TPE.
umask 077

# Disable coredumps.
ulimit -c 0

SYD="${CARGO_BIN_EXE_syd:-syd}"

set -ex
DIR="$(mktemp -d --tmpdir=/tmp syd-gnulib.XXXXX)"
set +ex

function finish() {
    rm -rf "${DIR}"
}

trap finish EXIT

edo() {
    echo >&2 "$*"
    "$@"
}

elog() {
    echo >&2 "$*"
    "$@" >.log 2>&1
    r=$?
    [[ $r == 0 ]] || cat .log
    rm -f .log
    return $r
}

run_test() {
    local name="$1"
    local tdir="$(readlink -f "${DIR}")"
    local tnam="$(echo "${name}" | sed -e 's|/|-|g')"
    elog "${SYD}" -ppaludis ./gnulib-tool --dir "${tdir}/${tnam}" --test "${name}"
    r=$?
    if [[ $r == 0 ]]; then
        rm -fr "${tdir}/${tnam}"
    else
        echo '--8<-- TEST LOG BEGIN -->8--'
        find "${tdir}/${tnam}" -name test-suite.log -exec cat '{}' '+'
        echo '-->8-- TEST LOG END   --8<--'
        echo "Keeping test directory of failed test: ${tdir}/${tnam}"
    fi
    return $r
}

set -ex
pushd "${DIR}"
#git clone --depth 1 https://git.savannah.gnu.org/git/gnulib.git || exit 0
git clone --depth 1 https://github.com/coreutils/gnulib.git || exit 0
pushd gnulib
sed -i \
    -e 's|"0.0.0.0"|"127.0.0.1"|' \
    tests/test-getsockname.c
set +ex

PASS=0
FAIL=0
SKIP=0
TESTS=( $(./gnulib-tool --list | grep -P "${TEST_PATTERN}" | grep -vP "${SKIP_PATTERN}" | shuf ) )
CTEST=${#TESTS[@]}
NTEST=${TEST_LIMIT}
if [[ ${NTEST} -gt ${CTEST} ]]; then
    NTEST=${CTEST}
fi
TESTS=( "${FAIL_HISTORY[@]}" "${TESTS[@]:0:${NTEST}}" )
NTEST=${#TESTS[@]}

idx=0
for name in "${TESTS[@]}"; do
    : $(( idx++ ))
    echo -e "\033[92m*** $name ($idx of $NTEST: $PASS ok, $FAIL notok, $SKIP todo) ***\033[0m"
    if echo "${name}" | grep -qP "${SKIP_PATTERN}"; then
        echo "ok ${idx} - ${name} # TODO"
        : $(( SKIP++ ))
    elif run_test "${name}"; then
        echo "ok ${idx} - ${name}"
        : $(( PASS++ ))
    else
        echo "not ok ${idx} - ${name} - FAIL: $?"
        : $(( FAIL++ ))
    fi
done

echo "# $PASS tests passed."
echo "# $FAIL tests failed."
echo "# $SKIP tests skipped."
exit $FAIL
