//! Endorphin is key-value based in-memory cache library with custom expiration policies.
//!
//! Basically, endorphin provides four pre-defined policies.
//! - [`LazyFixedTTLPolicy`] : A Lazy TTL Policy that checks TTL when accessing values.
//! - [`MixedPolicy`] : TTL & TTI mixed policy.
//! - [`TTIPolicy`] : Time To Idle Policy.
//! - [`TTLPolicy`] : Time To Live policy .
//!
//! You can also define new custom policies by using [`ExpirePolicy`].
//!
//! [`ExpirePolicy`]: policy::ExpirePolicy
//! [`LazyFixedTTLPolicy`]: policy::LazyFixedTTLPolicy
//! [`MixedPolicy`]: policy::MixedPolicy
//! [`TTIPolicy`]: policy::TTIPolicy
//! [`TTLPolicy`]: policy::TTLPolicy

// for internal use.
pub(crate) mod hash;
pub(crate) mod instrinsic;
pub(crate) mod storage;

// for external use.

/// A hash map implemented with [`hashbrown`] internal.
/// [`hashbrown`]: https://docs.rs/hashbrown/latest/hashbrown/index.html
pub(crate) mod map;

/// A hash set implemented with [`hashbrown`] internal.
/// [`hashbrown`]: https://docs.rs/hashbrown/latest/hashbrown/index.html
pub(crate) mod set;

/// An expiration policy including four pre-defined policies.
pub(crate) mod policy;

mod entry;

pub(crate) use entry::*;

#[doc(inline)]
pub(crate) use crate::endorphin::map::HashMap;
#[doc(inline)]
pub(crate) use crate::endorphin::set::HashSet;
pub(crate) use crate::endorphin::{entry::EntryId, storage::Storage};
