//
// Syd: rock-solid unikernel
// src/hash.rs: Utilities for hashing
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    collections::{hash_map::DefaultHasher, HashSet},
    hash::{BuildHasher, Hash, Hasher},
    io::{BufRead, IoSlice},
    os::fd::{AsFd, AsRawFd, FromRawFd, OwnedFd},
    path::Path,
};

use nix::{
    errno::Errno,
    fcntl::{open, splice, OFlag, SpliceFFlags},
    sys::{
        socket::{
            accept4, bind, sendmsg, setsockopt, socket, sockopt::AlgSetKey, AddressFamily, AlgAddr,
            ControlMessage, MsgFlags, SockFlag, SockType,
        },
        stat::Mode,
    },
    unistd::{lseek, pipe2, read, write, Whence},
};
use secure_string::SecureBytes;

use crate::{
    config::PIPE_BUF,
    fs::{
        create_memfd, fgetxattr, fillrandom, fsetxattr, lock_fd, retry_on_eintr, seal_memfd,
        set_pipemax, MFD_NOEXEC_SEAL,
    },
    proc::proc_pipemax,
};

/// Avoid duplicate hashing while using `HashSet` with u64 keys.
pub struct NoHasher {
    value: u64,
}

impl Hasher for NoHasher {
    fn write(&mut self, _bytes: &[u8]) {
        unreachable!("NoHasher should only be used for u64 keys");
    }

    fn write_u64(&mut self, i: u64) {
        self.value = i;
    }

    fn finish(&self) -> u64 {
        self.value
    }
}

/// A builder for creating instances of `NoHasher`.
#[derive(Clone)]
pub struct NoHasherBuilder;

impl Default for NoHasherBuilder {
    fn default() -> Self {
        Self
    }
}

impl BuildHasher for NoHasherBuilder {
    type Hasher = NoHasher;

    fn build_hasher(&self) -> Self::Hasher {
        NoHasher { value: 0 }
    }
}

/// A `HashSet` with no hashers.
pub type NoHashSet = HashSet<u64, NoHasherBuilder>;

/// Hash a string slice.
pub fn hash_str(s: &str) -> u64 {
    let mut hasher = DefaultHasher::new();
    s.hash(&mut hasher);
    hasher.finish()
}

/// Defines hash functions supported by syd.
#[derive(Debug, Clone, Copy)]
pub enum HashAlgorithm {
    /// Md5
    Md5,
    /// SHA-1.
    Sha1,
    /// SHA3-256
    Sha256,
    /// SHA3-384
    Sha384,
    /// SHA3-512
    Sha512,
}

impl TryFrom<usize> for HashAlgorithm {
    type Error = Errno;

    fn try_from(len: usize) -> Result<Self, Self::Error> {
        match len {
            16 => Ok(HashAlgorithm::Md5),
            20 => Ok(HashAlgorithm::Sha1),
            32 => Ok(HashAlgorithm::Sha256),
            48 => Ok(HashAlgorithm::Sha384),
            64 => Ok(HashAlgorithm::Sha512),
            _ => Err(Errno::EINVAL),
        }
    }
}

/// AES-CTR encryption key size: 256 bits.
pub const KEY_SIZE: usize = 32;

/// AES-CTR IV size: 128 bits.
pub const IV_SIZE: usize = 16;

/// AES-CTR block size: 128 bits.
pub const BLOCK_SIZE: usize = 16;

/// Key holds the AES encryption key.
///
/// This struct ensures that the key doesn't get swapped out and is
/// securely zeroized when it is dropped.
#[derive(Debug)]
pub struct Key(SecureBytes);

impl Key {
    /// Creates a new Key with the given key data.
    ///
    /// # Arguments
    ///
    /// * `key` - A 64-byte array containing the AES key.
    pub fn new(key: [u8; KEY_SIZE]) -> Self {
        Self(SecureBytes::from(key))
    }

    /// Creates a random Key using the OS random number generator.
    pub fn random() -> Result<Self, Errno> {
        let mut bytes = SecureBytes::new(vec![0; KEY_SIZE]);
        fillrandom(bytes.unsecure_mut())?;
        Ok(Self(bytes))
    }

    /// Creates an IV from a hex-encoded string.
    pub fn from_hex(hex: &[u8]) -> Result<Self, Errno> {
        let key: [u8; KEY_SIZE] = const_hex::FromHex::from_hex(hex).map_err(|_| Errno::EINVAL)?;
        Ok(Self::new(key))
    }

    /// Returns a hex-encoded string of the KEY.
    pub fn as_hex(&self) -> String {
        const_hex::encode(self.0.unsecure())
    }

    /// Check if the KEY is all zeros.
    pub fn is_zero(&self) -> bool {
        self.as_ref().iter().all(|&byte| byte == 0)
    }
}

impl AsRef<[u8]> for Key {
    fn as_ref(&self) -> &[u8] {
        self.0.unsecure()
    }
}

/// Key holds the AES IV
///
/// This struct ensures that the IV doesn't get swapped out and is
/// securely zeroized when it is dropped.
#[derive(Debug)]
pub struct IV(SecureBytes);

impl IV {
    /// Creates a new IV with the given key data.
    ///
    /// # Arguments
    ///
    /// * `iv` - A 16-byte array containing the AES IV.
    pub fn new(iv: [u8; IV_SIZE]) -> Self {
        Self(SecureBytes::from(iv))
    }

    /// Creates a random IV using the OS random number generator.
    pub fn random() -> Result<Self, Errno> {
        let mut bytes = SecureBytes::new(vec![0; IV_SIZE]);
        fillrandom(bytes.unsecure_mut())?;
        Ok(Self(bytes))
    }

    /// Creates an IV from a hex-encoded string.
    pub fn from_hex(hex: &[u8]) -> Result<Self, Errno> {
        let iv: [u8; IV_SIZE] = const_hex::FromHex::from_hex(hex).map_err(|_| Errno::EINVAL)?;
        Ok(Self::new(iv))
    }

    /// Returns a hex-encoded string of the IV.
    pub fn as_hex(&self) -> String {
        const_hex::encode(self.0.unsecure())
    }

    /// Check if the IV is all zeros.
    pub fn is_zero(&self) -> bool {
        self.as_ref().iter().all(|&byte| byte == 0)
    }

    /// Add the given counter to the IV in AES-CTR mode.
    ///
    /// In AES-CTR (Counter) mode, encryption and decryption are done by
    /// generating a keystream using the AES block cipher and a counter
    /// value. The IV (Initialization Vector) is combined with a counter
    /// to generate unique input blocks for encryption. This function
    /// updates the IV by adding a given counter value, effectively
    /// updating the nonce for the next encryption block. The counter is
    /// incremented in a block-aligned manner.
    ///
    /// # Parameters
    /// - `ctr`: The counter value to be added to the IV. This counter
    ///          is divided by the block size to ensure correct block-aligned
    ///          increments.
    #[allow(clippy::arithmetic_side_effects)]
    pub fn add_counter(&mut self, ctr: i64) {
        // Convert the counter to a u128 and divide by the block size.
        // This aligns the counter to the size of an AES block (16 bytes).
        #[allow(clippy::cast_sign_loss)]
        let mut ctr = ctr as u128 / BLOCK_SIZE as u128;

        // If the counter is zero after adjustment,
        // there's no need to update the IV.
        if ctr == 0 {
            return;
        }

        // Access the IV bytes for modification.
        let val = self.0.unsecure_mut();

        // Process each byte of the IV from least significant to most
        // significant. This is because we are effectively treating the
        // IV as a large integer counter.
        for i in (0..IV_SIZE).rev() {
            // Add the least significant byte of the counter to the
            // current byte of the IV. `overflowing_add` handles byte
            // overflow, which is equivalent to a carry in multi-byte
            // addition.
            let (new_byte, overflow) = val[i].overflowing_add((ctr & 0xFF) as u8);

            // Update the IV byte with the new value.
            val[i] = new_byte;

            // Shift the counter right by 8 bits to process the next
            // byte. If there was an overflow, carry the overflow to
            // the next byte.
            ctr = (ctr >> 8) + if overflow { 1 } else { 0 };

            // Return if counter is zero and there is no overflow.
            if ctr == 0 {
                break;
            }
        }
    }
}

impl Clone for IV {
    fn clone(&self) -> Self {
        IV(SecureBytes::from(self.0.unsecure()))
    }
}

impl AsRef<[u8]> for IV {
    fn as_ref(&self) -> &[u8] {
        self.0.unsecure()
    }
}

impl AsMut<[u8]> for IV {
    fn as_mut(&mut self) -> &mut [u8] {
        self.0.unsecure_mut()
    }
}

/// Represents crypt secrets.
pub(crate) enum Secret {
    Fd(OwnedFd),
    Key(Key),
}

/// Calculate sha{1,256,512} of the given buffered reader.
/// Returns a byte array.
pub fn hash<R: BufRead>(mut reader: R, func: HashAlgorithm) -> Result<Vec<u8>, Errno> {
    match func {
        HashAlgorithm::Md5 => {
            let mut context = md5::Context::new();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_raw(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    context.consume(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            let digest = context.compute();
            let dbytes: [u8; 16] = digest.into();
            Ok(dbytes.to_vec())
        }
        HashAlgorithm::Sha1 => {
            let mut hasher = rs_sha1::Sha1Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_raw(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha1::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
        HashAlgorithm::Sha256 => {
            let mut hasher = rs_sha3_256::Sha3_256Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_raw(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha3_256::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
        HashAlgorithm::Sha384 => {
            let mut hasher = rs_sha3_384::Sha3_384Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_raw(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha3_384::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
        HashAlgorithm::Sha512 => {
            let mut hasher = rs_sha3_512::Sha3_512Hasher::default();
            loop {
                let consumed = {
                    let buf = reader.fill_buf().map_err(|e| {
                        Errno::from_raw(e.raw_os_error().unwrap_or(nix::libc::EINVAL))
                    })?;
                    if buf.is_empty() {
                        break;
                    }
                    hasher.write(buf);
                    buf.len()
                };
                reader.consume(consumed);
            }
            Ok(rs_sha3_512::HasherContext::finish(&mut hasher)
                .as_ref()
                .to_vec())
        }
    }
}

/// Check if fd is marked syd-aes=3.
#[allow(clippy::type_complexity)]
pub(crate) fn aes_getxattr<F: AsRawFd>(fd: &F) -> Result<Option<IV>, Errno> {
    // SAFETY:
    // 1. Acquire a read lock on the file descriptor.
    // 2. Wait until the lock can be acquired.
    lock_fd(fd, false, true)?;

    let mut val = [0u8; 1];
    match fgetxattr(fd, "user.syd.crypt.api", Some(&mut val)) {
        Ok(_) if val[0] == b'3' => {
            // If the API check succeeded, check for the IV attribute.
            let mut buf = vec![0; IV_SIZE * 2];
            match fgetxattr(fd, "user.syd.crypt.iv", Some(&mut buf)) {
                Ok(_) => {
                    // If IV attribute exists, decode it and return.
                    Ok(Some(IV::from_hex(&buf)?))
                }
                Err(Errno::ERANGE) => Err(Errno::EINVAL), // IV corrupt.
                Err(Errno::ENODATA | Errno::EOPNOTSUPP) => Err(Errno::ENOKEY),
                Err(errno) => Err(errno),
            }
        }
        Ok(_) | Err(Errno::ERANGE) => Ok(None), // API mismatch.
        Err(Errno::ENODATA | Errno::EOPNOTSUPP) => Ok(None),
        Err(errno) => Err(errno),
    }
}

/// Set encryption extended attributes to the file descriptor.
pub(crate) fn aes_setxattr<F: AsRawFd>(fd: &F, append: bool) -> Result<IV, Errno> {
    // SAFETY:
    // 1. Acquire a write lock on the file descriptor.
    // 2. Wait until the lock can be acquired.
    lock_fd(fd, true, true)?;

    match fsetxattr(fd, "user.syd.crypt.api", &[b'3'], nix::libc::XATTR_CREATE) {
        Ok(_) => {
            // If setting the attribute succeeded, create a random IV and set it.
            let iv = IV::random()?;
            fsetxattr(fd, "user.syd.crypt.iv", iv.as_hex().as_bytes(), 0)?;
            Ok(iv)
        }
        Err(Errno::EEXIST) => {
            // If the encryption attribute already exists, check for the IV attribute.
            let mut buf = vec![0; IV_SIZE * 2];
            match fgetxattr(fd, "user.syd.crypt.iv", Some(&mut buf)) {
                Ok(_) => {
                    // If IV attribute exists, decode it and return.
                    let iv = IV::from_hex(&buf)?;
                    Ok(iv)
                }
                Err(Errno::ENODATA) if !append => {
                    // If IV attribute does not exist, and we're creating a new file:
                    // Create a random IV and set it.
                    let iv = IV::random()?;
                    fsetxattr(fd, "user.syd.crypt.iv", iv.as_hex().as_bytes(), 0)?;
                    Ok(iv)
                }
                Err(errno) => Err(errno),
            }
        }
        Err(errno) => Err(errno),
    }
}

/// Sets up the AES-CTR encryption/decryption using the Kernel crypto API.
///
/// # Arguments
///
/// * `key` - A reference to the AES key.
///
/// # Returns
///
/// * `Result<OwnedFd, Errno>` - The file descriptor for the socket on success, or an error.
pub fn aes_ctr_setup(key: &Key) -> Result<OwnedFd, Errno> {
    // Create the socket for the AF_ALG interface
    let sock = socket(
        AddressFamily::Alg,
        SockType::SeqPacket,
        SockFlag::empty(),
        None,
    )?;

    // Set up the sockaddr_alg structure
    let addr = AlgAddr::new("skcipher", "ctr(aes)");

    // Bind the socket
    bind(sock.as_raw_fd(), &addr)?;

    // Set the encryption key and authentication tag size.
    setsockopt(&sock, AlgSetKey::default(), &key.as_ref())?;

    Ok(sock)
}

/// Initializes the AES-CTR encryption/decryption using an existing socket.
///
/// # Arguments
///
/// * `fd`       - The file descriptor of the existing socket.
/// * `nonblock` - True if socket should be set non-blocking.
///
/// # Returns
///
/// * `Result<OwnedFd, Errno>` - The file descriptor for the new socket on success, or an error.
pub fn aes_ctr_init<F: AsRawFd>(fd: &F, nonblock: bool) -> Result<OwnedFd, Errno> {
    let mut flags = SockFlag::SOCK_CLOEXEC;
    if nonblock {
        flags |= SockFlag::SOCK_NONBLOCK;
    }

    let fd = retry_on_eintr(|| accept4(fd.as_raw_fd(), flags))?;

    // SAFETY: accept4 returns a valid FD.
    Ok(unsafe { OwnedFd::from_raw_fd(fd) })
}

/// Encrypts a chunk of data using the initialized AES-CTR socket.
///
/// # Arguments
///
/// * `sock` - The file descriptor for the socket.
/// * `chunk` - A reference to the chunk of data to be encrypted.
/// * `iv` - A mutable reference to the initialization vector.
/// * `more` - A boolean flag indicating whether more data is expected.
///
/// # Returns
///
/// * `Result<IV, Errno>` - Ok(IV) on success, or an error.
pub fn aes_ctr_enc<S: AsRawFd>(
    sock: &S,
    chunk: &[u8],
    iv: &IV,
    more: bool,
) -> Result<usize, Errno> {
    let data = chunk.to_vec();

    // Prepare the IoSlice for the data
    let iov = [IoSlice::new(&data)];

    // Prepare the control message for the IV
    let cmsgs = [
        ControlMessage::AlgSetOp(&nix::libc::ALG_OP_ENCRYPT),
        ControlMessage::AlgSetIv(iv.as_ref()),
    ];

    // nix does not define MSG_MORE yet.
    const MSG_MORE: MsgFlags = MsgFlags::from_bits_retain(0x8000);

    // Determine the flags for the sendmsg operation.
    let flags = if more { MSG_MORE } else { MsgFlags::empty() };

    // Send the message with the IV and data
    retry_on_eintr(|| sendmsg::<AlgAddr>(sock.as_raw_fd(), &iov, &cmsgs, flags, None))
}

/// Decrypts a chunk of data using the initialized AES-CTR socket.
///
/// # Arguments
///
/// * `sock` - The file descriptor for the socket.
/// * `chunk` - A reference to the chunk of data to be decrypted.
/// * `iv` - A mutable reference to the initialization vector.
/// * `more` - A boolean flag indicating whether more data is expected.
///
/// # Returns
///
/// * `Result<usize, Errno>` - Ok(number of bytes sent) on success, or an error.
pub fn aes_ctr_dec<S: AsRawFd>(
    sock: &S,
    chunk: &[u8],
    iv: &IV,
    more: bool,
) -> Result<usize, Errno> {
    // Prepare the IoSlice for the data
    let iov = [IoSlice::new(chunk)];

    // Prepare the control message for the IV
    let cmsgs = [
        ControlMessage::AlgSetOp(&nix::libc::ALG_OP_DECRYPT),
        ControlMessage::AlgSetIv(iv.as_ref()),
    ];

    // nix does not define MSG_MORE yet.
    const MSG_MORE: MsgFlags = MsgFlags::from_bits_retain(0x8000);

    // Determine the flags for the sendmsg operation
    let flags = if more { MSG_MORE } else { MsgFlags::empty() };

    // Send the message with the IV and data
    retry_on_eintr(|| sendmsg::<AlgAddr>(sock.as_raw_fd(), &iov, &cmsgs, flags, None))
}

/// Finishes the AES-CTR {en,de}cryption and reads the {en,de}crypted data.
///
/// # Arguments
///
/// * `sock` - The file descriptor for the socket.
/// * `size` - The total size of the data to read.
///
/// # Returns
///
/// * `Result<Vec<u8>, Errno>` - The {en,de}crypted data on success, or an error.
pub fn aes_ctr_fini<S: AsRawFd>(sock: &S, size: usize) -> Result<SecureBytes, Errno> {
    let mut data = SecureBytes::new(vec![0u8; size]);
    let buf = data.unsecure_mut();

    let mut nread = 0;
    while nread < size {
        #[allow(clippy::arithmetic_side_effects)]
        match read(sock.as_raw_fd(), &mut buf[nread..]) {
            Ok(0) => return Err(Errno::EINVAL),
            Ok(n) => nread += n,
            Err(Errno::EINTR) => continue,
            Err(errno) => return Err(errno),
        }
    }

    Ok(data)
}

/// Decrypt the given file into a temporary fd with zero-copy.
///
/// # Arguments
///
/// * `sock`    - The file descriptor for the socket to handle decryption.
/// * `fd`      - The file descriptor to which the data should be read from.
/// * `tmp`     - Path to the temporary directory, if None a memory fd is used.
/// * `zeroize` - Zeroize memory buffers that hold plaintext.
///
/// # Returns
///
/// * `Result<OwnedFd, Errno>` - Ok(fd) on success, or an error.
pub fn aes_ctr_tmp<S: AsFd, T: AsFd, P: AsRef<Path>>(
    sock: &S,
    fd: &T,
    tmp: Option<P>,
    zeroize: bool,
) -> Result<OwnedFd, Errno> {
    let (pipe_rd, pipe_wr) = pipe2(OFlag::O_CLOEXEC)?;

    // Set pipe buffer size to the max allowed by the system.
    // Handle errors gracefully.
    let pipe_max = if let Ok(pipe_max) = proc_pipemax() {
        set_pipemax(&pipe_wr, pipe_max).unwrap_or(*PIPE_BUF)
    } else {
        *PIPE_BUF
    };

    let dst_fd = if let Some(ref tmp) = tmp {
        open(
            tmp.as_ref(),
            OFlag::O_RDWR | OFlag::O_TMPFILE | OFlag::O_CLOEXEC,
            Mode::empty(),
        )
        .map(|fd| {
            // SAFETY: open returns a valid FD.
            unsafe { OwnedFd::from_raw_fd(fd) }
        })
    } else {
        // Open memory fd early and mark non-executable.
        create_memfd(MFD_NOEXEC_SEAL)
    }?;

    let mut nflush = 0;
    loop {
        // Use splice to move data from fd to pipe
        let nread = splice(
            fd,
            None,
            &pipe_wr,
            None,
            pipe_max,
            SpliceFFlags::SPLICE_F_MORE,
        )?;

        if nread == 0 {
            // EOF.
            break;
        }

        // Use splice to move data from pipe to sock (AF_ALG socket)
        // Pipe buffer size is typically bigger than the socket
        // receive buffer size so we have to loop and splice
        // PIPE_BUF at a time to ensure we don't clog the pipes.
        let mut nwrite = 0;
        #[allow(clippy::arithmetic_side_effects)]
        while nwrite < nread {
            let n = splice(
                &pipe_rd,
                None,
                sock,
                None,
                *PIPE_BUF,
                SpliceFFlags::SPLICE_F_MORE,
            )?;

            nflush += n;
            while nflush >= BLOCK_SIZE {
                let nbytes = (nflush / BLOCK_SIZE) * BLOCK_SIZE;
                nflush -= aes_ctr_flush(sock, &dst_fd, nbytes, zeroize)?;
            }

            nwrite += n;
        }
    }

    // Signal the AF_ALG socket the end of the process (No MSG_MORE in flags).
    let cmsgs = [ControlMessage::AlgSetOp(&nix::libc::ALG_OP_DECRYPT)];
    retry_on_eintr(|| {
        sendmsg::<AlgAddr>(
            sock.as_fd().as_raw_fd(),
            &[],
            &cmsgs,
            MsgFlags::empty(),
            None,
        )
    })?;

    while nflush > 0 {
        // Flush the final batch.
        #[allow(clippy::arithmetic_side_effects)]
        match aes_ctr_flush(sock, &dst_fd, nflush, zeroize)? {
            0 => return Err(Errno::EINVAL),
            n => nflush -= n,
        }
    }

    // Seek the file offset to the beginning.
    lseek(dst_fd.as_raw_fd(), 0, Whence::SeekSet)?;

    if tmp.is_none() {
        // Seal memory fd.
        seal_memfd(&dst_fd)?;
    }
    Ok(dst_fd)
}

/// Flush data in the AF_ALG socket into the given file descriptor.
pub fn aes_ctr_flush<S: AsFd, F: AsFd>(
    sock: &S,
    fd: &F,
    size: usize,
    zeroize: bool,
) -> Result<usize, Errno> {
    let (mut bufz, mut bufu) = if zeroize {
        (Some(SecureBytes::from(vec![0u8; size])), None)
    } else {
        (None, Some(vec![0u8; size]))
    };
    let buf = if let Some(ref mut bufz) = bufz {
        bufz.unsecure_mut()
    } else if let Some(ref mut bufu) = bufu {
        bufu
    } else {
        unreachable!()
    };
    let mut nread = 0;
    while nread < size {
        #[allow(clippy::arithmetic_side_effects)]
        match read(sock.as_fd().as_raw_fd(), &mut buf[nread..]) {
            Ok(0) => break,
            Ok(n) => nread += n,
            Err(Errno::EINTR) => continue,
            Err(errno) => return Err(errno),
        }
    }

    // Write output data to the file descriptor.
    let mut nwrite = 0;
    while nwrite < nread {
        #[allow(clippy::arithmetic_side_effects)]
        match write(fd, &buf[nwrite..nread]) {
            Ok(n) => nwrite += n,
            Err(Errno::EINTR) => continue,
            Err(errno) => return Err(errno),
        }
    }

    Ok(nwrite)
}

#[cfg(test)]
mod tests {
    use std::{io::Cursor, path::PathBuf};

    use const_hex::encode_upper;
    use nix::unistd::{lseek, Whence};

    use super::*;
    use crate::fs::create_memfd; // Import everything from the outer scope

    struct TestCase(&'static [u8], &'static str, HashAlgorithm);

    // Source: https://www.di-mgt.com.au/sha_testvectors.html
    // MD5 test vectors were calculated with python-3.11.8's hashlib.md5
    const TEST_CASES: &[TestCase] = &[
    TestCase(
        b"",
        "D41D8CD98F00B204E9800998ECF8427E",
        HashAlgorithm::Md5,
    ),
    TestCase(
        b"",
        "DA39A3EE5E6B4B0D3255BFEF95601890AFD80709",
        HashAlgorithm::Sha1,
    ),
    TestCase(
        b"",
        "A7FFC6F8BF1ED76651C14756A061D662F580FF4DE43B49FA82D80A4B80F8434A",
        HashAlgorithm::Sha256,
    ),
    TestCase(
        b"",
        "0C63A75B845E4F7D01107D852E4C2485C51A50AAAA94FC61995E71BBEE983A2AC3713831264ADB47FB6BD1E058D5F004",
        HashAlgorithm::Sha384,
    ),
    TestCase(
        b"",
        "A69F73CCA23A9AC5C8B567DC185A756E97C982164FE25859E0D1DCC1475C80A615B2123AF1F5F94C11E3E9402C3AC558F500199D95B6D3E301758586281DCD26",
        HashAlgorithm::Sha512,
    ),
    TestCase(
        b"abc",
        "900150983CD24FB0D6963F7D28E17F72",
        HashAlgorithm::Md5,
    ),
    TestCase(
        b"abc",
        "A9993E364706816ABA3E25717850C26C9CD0D89D",
        HashAlgorithm::Sha1,
    ),
    TestCase(
        b"abc",
        "3A985DA74FE225B2045C172D6BD390BD855F086E3E9D525B46BFE24511431532",
        HashAlgorithm::Sha256,
    ),
    TestCase(
        b"abc",
        "EC01498288516FC926459F58E2C6AD8DF9B473CB0FC08C2596DA7CF0E49BE4B298D88CEA927AC7F539F1EDF228376D25",
        HashAlgorithm::Sha384,
    ),
    TestCase(
        b"abc",
        "B751850B1A57168A5693CD924B6B096E08F621827444F70D884F5D0240D2712E10E116E9192AF3C91A7EC57647E3934057340B4CF408D5A56592F8274EEC53F0",
        HashAlgorithm::Sha512
    ),
    TestCase(
        b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
        "8215EF0796A20BCAAAE116D3876C664A",
        HashAlgorithm::Md5,
    ),
    TestCase(
        b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
        "84983E441C3BD26EBAAE4AA1F95129E5E54670F1",
        HashAlgorithm::Sha1,
    ),
    TestCase(
        b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
        "41C0DBA2A9D6240849100376A8235E2C82E1B9998A999E21DB32DD97496D3376",
        HashAlgorithm::Sha256,
    ),
    TestCase(
        b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
        "991C665755EB3A4B6BBDFB75C78A492E8C56A22C5C4D7E429BFDBC32B9D4AD5AA04A1F076E62FEA19EEF51ACD0657C22",
        HashAlgorithm::Sha384,
    ),
    TestCase(
        b"abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq",
        "04A371E84ECFB5B8B77CB48610FCA8182DD457CE6F326A0FD3D7EC2F1E91636DEE691FBE0C985302BA1B0D8DC78C086346B533B49C030D99A27DAF1139D6E75E",
        HashAlgorithm::Sha512,
    ),
    TestCase(
        b"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
        "03DD8807A93175FB062DFB55DC7D359C",
        HashAlgorithm::Md5,
    ),
    TestCase(
        b"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
        "A49B2446A02C645BF419F995B67091253A04A259",
        HashAlgorithm::Sha1,
    ),
    TestCase(
        b"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
        "916F6061FE879741CA6469B43971DFDB28B1A32DC36CB3254E812BE27AAD1D18",
        HashAlgorithm::Sha256,
    ),
    TestCase(
        b"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
        "79407D3B5916B59C3E30B09822974791C313FB9ECC849E406F23592D04F625DC8C709B98B43B3852B337216179AA7FC7",
        HashAlgorithm::Sha384,
    ),
    TestCase(
        b"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu",
        "AFEBB2EF542E6579C50CAD06D2E578F9F8DD6881D7DC824D26360FEEBF18A4FA73E3261122948EFCFD492E74E82E2189ED0FB440D187F382270CB455F21DD185",
        HashAlgorithm::Sha512,
    ),
    ];

    #[test]
    fn test_hash_simple() {
        let mut errors = Vec::new();

        for case in TEST_CASES {
            let input_cursor = Cursor::new(case.0);
            let result = match hash(input_cursor, case.2) {
                Ok(hash) => encode_upper(&hash),
                Err(e) => {
                    errors.push(format!(
                        "Hashing failed for {:?} with error: {:?}",
                        case.2, e
                    ));
                    continue;
                }
            };

            if result != case.1 {
                errors.push(format!(
                    "Mismatch for {:?}: expected {}, got {}",
                    case.2, case.1, result
                ));
            }
        }

        assert!(errors.is_empty(), "Errors encountered: {:?}", errors);
    }

    #[test]
    fn test_hash_long() {
        let mut errors = Vec::new();

        let input = b"a".repeat(1_000_000);
        let cases = &[
            (HashAlgorithm::Md5, "7707D6AE4E027C70EEA2A935C2296F21"),
            (HashAlgorithm::Sha1, "34AA973CD4C4DAA4F61EEB2BDBAD27316534016F"),
            (HashAlgorithm::Sha256, "5C8875AE474A3634BA4FD55EC85BFFD661F32ACA75C6D699D0CDCB6C115891C1"),
            (HashAlgorithm::Sha384, "EEE9E24D78C1855337983451DF97C8AD9EEDF256C6334F8E948D252D5E0E76847AA0774DDB90A842190D2C558B4B8340"),
            (HashAlgorithm::Sha512, "3C3A876DA14034AB60627C077BB98F7E120A2A5370212DFFB3385A18D4F38859ED311D0A9D5141CE9CC5C66EE689B266A8AA18ACE8282A0E0DB596C90B0A7B87"),
        ];

        for case in cases {
            let input_cursor = Cursor::new(input.clone());
            let result = match hash(input_cursor, case.0) {
                Ok(hash) => encode_upper(&hash),
                Err(e) => {
                    errors.push(format!(
                        "Hashing failed for {:?} with error: {:?}",
                        case.0, e
                    ));
                    continue;
                }
            };

            if result != case.1 {
                errors.push(format!(
                    "Mismatch for {:?}: expected {}, got {}",
                    case.0, case.1, result
                ));
            }
        }

        assert!(errors.is_empty(), "Errors encountered: {:?}", errors);
    }

    #[test]
    #[ignore] // it is too expensive.
    fn test_hash_extremely_long() {
        let mut errors = Vec::new();

        let input =
            b"abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmno".repeat(16_777_216);
        let cases = &[
            (HashAlgorithm::Md5, "D338139169D50F55526194C790EC0448"),
            (HashAlgorithm::Sha1, "7789F0C9EF7BFC40D93311143DFBE69E2017F592"),
            (HashAlgorithm::Sha256, "ECBBC42CBF296603ACB2C6BC0410EF4378BAFB24B710357F12DF607758B33E2B"),
            (HashAlgorithm::Sha384, "A04296F4FCAAE14871BB5AD33E28DCF69238B04204D9941B8782E816D014BCB7540E4AF54F30D578F1A1CA2930847A12"),
            (HashAlgorithm::Sha512, "235FFD53504EF836A1342B488F483B396EABBFE642CF78EE0D31FEEC788B23D0D18D5C339550DD5958A500D4B95363DA1B5FA18AFFC1BAB2292DC63B7D85097C"),
        ];

        for case in cases {
            let input_cursor = Cursor::new(input.clone());
            let result = match hash(input_cursor, case.0) {
                Ok(hash) => encode_upper(&hash),
                Err(e) => {
                    errors.push(format!(
                        "Hashing failed for {:?} with error: {:?}",
                        case.0, e
                    ));
                    continue;
                }
            };

            if result != case.1 {
                errors.push(format!(
                    "Mismatch for {:?}: expected {}, got {}",
                    case.0, case.1, result
                ));
            }
        }

        assert!(errors.is_empty(), "Errors encountered: {:?}", errors);
    }

    #[test]
    fn test_aes_ctr_setup() {
        let key = Key::random().unwrap();
        assert!(!key.is_zero(), "key is all zeros!");

        let result = aes_ctr_setup(&key);
        assert!(result.is_ok());
    }

    #[test]
    fn test_aes_ctr_init() {
        let key = Key::random().unwrap();
        assert!(!key.is_zero(), "key is all zeros!");

        let setup_fd = aes_ctr_setup(&key).unwrap();

        let result = aes_ctr_init(&setup_fd, false);
        assert!(result.is_ok());
    }

    #[test]
    fn test_aes_ctr_enc_and_dec() {
        let key = Key::random().unwrap();
        assert!(!key.is_zero(), "key is all zeros!");

        let iv = IV::random().unwrap();
        assert!(!iv.is_zero(), "iv is all zeros!");

        let setup_fd = aes_ctr_setup(&key).unwrap();
        let sock_enc = aes_ctr_init(&setup_fd, false).unwrap();

        let data = b"Heavy is the root of light. Still is the master of moving.";
        let enc_result = aes_ctr_enc(&sock_enc, data, &iv, false);
        assert!(enc_result.is_ok());
        let encrypted_size = enc_result.unwrap();
        assert_eq!(encrypted_size, data.len());

        let encrypted_data = aes_ctr_fini(&sock_enc, encrypted_size).unwrap();
        assert_eq!(encrypted_data.unsecure().len(), data.len());
        drop(sock_enc);

        let sock_dec = aes_ctr_init(&setup_fd, false).unwrap();
        let dec_result = aes_ctr_dec(&sock_dec, &encrypted_data.unsecure(), &iv, false);
        assert!(dec_result.is_ok());

        let decrypted_size = dec_result.unwrap();
        let decrypted_data = aes_ctr_fini(&sock_dec, decrypted_size).unwrap();
        assert_eq!(decrypted_data.unsecure(), data);
    }

    #[test]
    fn test_aes_ctr_enc_with_more_flag() {
        let key = Key::random().unwrap();
        assert!(!key.is_zero(), "key is all zeros!");

        let iv = IV::random().unwrap();
        assert!(!iv.is_zero(), "iv is all zeros!");

        let setup_fd = aes_ctr_setup(&key).unwrap();
        let sock = aes_ctr_init(&setup_fd, false).unwrap();

        let data_chunks = vec![
            b"Heavy is ".to_vec(),
            b"the root of light. ".to_vec(),
            b"Still is ".to_vec(),
            b"the master of moving.".to_vec(),
        ];

        let mut total_encrypted_size = 0;
        for (i, chunk) in data_chunks.iter().enumerate() {
            let more = if i < data_chunks.len() - 1 {
                true
            } else {
                false
            };
            let enc_result = aes_ctr_enc(&sock, chunk, &iv, more);
            assert!(enc_result.is_ok());
            total_encrypted_size += enc_result.unwrap();
        }

        let encrypted_data = aes_ctr_fini(&sock, total_encrypted_size).unwrap();
        drop(sock);

        let sock_dec = aes_ctr_init(&setup_fd, false).unwrap();
        let dec_result = aes_ctr_dec(&sock_dec, &encrypted_data.unsecure(), &iv, false);
        assert!(dec_result.is_ok());

        let decrypted_size = dec_result.unwrap();
        let decrypted_data = aes_ctr_fini(&sock_dec, decrypted_size).unwrap();
        let original_data: Vec<u8> = data_chunks.concat();
        assert_eq!(decrypted_data.unsecure(), original_data.as_slice());
    }

    #[test]
    fn test_aes_ctr_enc_and_dec_zero_copy() {
        let key = Key::random().unwrap();
        assert!(!key.is_zero(), "key is all zeros!");

        let iv = IV::random().unwrap();
        assert!(!iv.is_zero(), "iv is all zeros!");

        let setup_fd = aes_ctr_setup(&key).unwrap();
        let sock_enc = aes_ctr_init(&setup_fd, false).unwrap();

        let data = b"Heavy is the root of light. Still is the master of moving.";
        let total_size = data.len();
        let enc_result = aes_ctr_enc(&sock_enc, data, &iv, false);
        assert!(enc_result.is_ok());
        let encrypted_size = enc_result.unwrap();
        assert_eq!(encrypted_size, data.len());

        // Use a memfd to hold the encrypted data.
        let encrypted_memfd = create_memfd(0).unwrap();
        aes_ctr_flush(&sock_enc, &encrypted_memfd, encrypted_size, false).unwrap();
        drop(sock_enc);

        // Read the encrypted data back to decrypt it.
        let mut encrypted_data = vec![0u8; encrypted_size];
        lseek(encrypted_memfd.as_raw_fd(), 0, Whence::SeekSet).unwrap();
        read(encrypted_memfd.as_raw_fd(), &mut encrypted_data).unwrap();

        // Decrypt the data directly into a memfd with zero-copy.
        let sock_dec = aes_ctr_init(&setup_fd, false).unwrap();
        aes_ctr_dec(&sock_dec, &[], &iv, true).unwrap();
        lseek(encrypted_memfd.as_raw_fd(), 0, Whence::SeekSet).unwrap();
        let decrypted_memfd =
            aes_ctr_tmp(&sock_dec, &encrypted_memfd, None::<PathBuf>, true).unwrap();
        drop(sock_dec);

        // Verify the decrypted data matches the original data.
        let mut decrypted_data = vec![0u8; total_size];
        read(decrypted_memfd.as_raw_fd(), &mut decrypted_data).unwrap();
        assert_eq!(
            decrypted_data,
            data,
            "mismatch: {decrypted_data:?} != {data:?} ({} != {}, {} != {})",
            String::from_utf8_lossy(&decrypted_data),
            String::from_utf8_lossy(data),
            decrypted_data.len(),
            data.len()
        );
    }
}
