//
// Syd: rock-solid unikernel
// src/log.rs: Simple logging on standard error using JSON lines
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

/* Simple logging with JSON lines */
use std::{
    env,
    ffi::CString,
    fs::{File, OpenOptions},
    io::{self, Write},
    os::{
        fd::{FromRawFd, RawFd},
        unix::fs::OpenOptionsExt,
    },
    path::PathBuf,
    time::{SystemTime, UNIX_EPOCH},
};

use anyhow::{Context, Result};
use env_logger::{Builder, Logger};
use lexis::ToName;
use log::Level;
use nix::unistd::{Pid, Uid};
use once_cell::sync::Lazy;
use parking_lot::Mutex;
use serde_json::{json, Map, Value};

use crate::{config::*, fs::mask_path, proc::*};

/// info! logging macro
#[macro_export]
macro_rules! info {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Info) {
            #[allow(clippy::disallowed_methods)]
            if let Some(map) = serde_json::json!({$($key: $value),+}).as_object().cloned() {
                $crate::log::log_with_data(log::Level::Info, map);
            } else {
                log::error!("syd-log: Failed to create JSON object for logging info.");
            }
        }
    }
}

/// error! logging macro
#[macro_export]
macro_rules! error {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Error) {
            #[allow(clippy::disallowed_methods)]
            if let Some(map) = serde_json::json!({$($key: $value),+}).as_object().cloned() {
                $crate::log::log_with_data(log::Level::Error, map);
            } else {
                log::error!("syd-log: Failed to create JSON object for logging error.");
            }
        }
    }
}

/// warn! logging macro
#[macro_export]
macro_rules! warn {
    ($($key:literal : $value:expr),+) => {
        if log::log_enabled!(log::Level::Warn) {
            #[allow(clippy::disallowed_methods)]
            if let Some(map) = serde_json::json!({$($key: $value),+}).as_object().cloned() {
                $crate::log::log_with_data(log::Level::Warn, map);
            } else {
                log::error!("syd-log: Failed to create JSON object for logging warn.");
            }
        }
    }
}

/// debug! logging macro
#[macro_export]
macro_rules! debug {
    ($($key:literal : $value:expr),+) => {
        #[cfg(feature = "log")]
        if log::log_enabled!(log::Level::Debug) {
            #[allow(clippy::disallowed_methods)]
            if let Some(map) = serde_json::json!({$($key: $value),+}).as_object().cloned() {
                $crate::log::log_with_data(log::Level::Debug, map);
            } else {
                log::error!("syd-log: Failed to create JSON object for logging debug.");
            }
        }
    }
}

/// trace! logging macro
#[macro_export]
macro_rules! trace {
    ($($key:literal : $value:expr),+) => {
        #[cfg(feature = "log")]
        if log::log_enabled!(log::Level::Trace) {
            #[allow(clippy::disallowed_methods)]
            if let Some(map) = serde_json::json!({$($key: $value),+}).as_object().cloned() {
                $crate::log::log_with_data(log::Level::Trace, map);
            } else {
                log::error!("syd-log: Failed to create JSON object for logging trace.");
            }
        }
    }
}

type Writer = Box<dyn Write + Send>;

/// Simple logging on standard error using JSON lines
pub struct JsonLinesLogger {
    filter: Logger,
    writer: Mutex<Writer>,
}

impl JsonLinesLogger {
    fn new(log: &Option<PathBuf>, default_level: &str) -> Result<Self> {
        let mut builder = Builder::new();

        let level = env::var(ENV_LOG).unwrap_or(default_level.to_string());
        builder.parse_filters(&level);

        let writer: Box<dyn io::Write + Send> = if let Some(log) = log {
            Box::new(
                OpenOptions::new()
                    .append(true)
                    .create(true)
                    .mode(0o600)
                    .custom_flags(nix::libc::O_CLOEXEC)
                    .open(log)
                    .map_err(anyhow::Error::msg)?,
            )
        } else {
            match env::var(ENV_LOG_FD) {
                Ok(fd) => {
                    // SAFETY: We panic on parse errors.
                    #[allow(clippy::disallowed_methods)]
                    let fd: RawFd = fd.parse().expect("Invalid log file descriptor");
                    // SAFETY: We trust the user to pass a valid FD.
                    Box::new(unsafe { File::from_raw_fd(fd) })
                }
                Err(_) => Box::new(io::stderr()),
            }
        };

        Ok(Self {
            filter: builder.build(),
            writer: Mutex::new(writer),
        })
    }

    /// Initialize logging
    pub fn init(log: &Option<PathBuf>, default_level: &str) -> Result<()> {
        let logger = Self::new(log, default_level).context("Failed to initialize logging")?;

        log::set_max_level(logger.filter.filter());
        log::set_boxed_logger(Box::new(logger))
            .map_err(anyhow::Error::msg)
            .context("Failed to set logger")
    }
}

impl log::Log for JsonLinesLogger {
    fn enabled(&self, metadata: &log::Metadata) -> bool {
        self.filter.enabled(metadata)
    }

    fn log(&self, record: &log::Record) {
        if self.enabled(record.metadata()) {
            let mut writer = self.writer.lock();
            let _ = writeln!(writer, "{}", record.args());
        }
    }

    fn flush(&self) {}
}

/// Helper for logging using JSON lines.
#[allow(clippy::cognitive_complexity)]
pub fn log_with_data(level: Level, data: Map<String, Value>) {
    static NO_SYSLOG: Lazy<bool> = Lazy::new(|| std::env::var(ENV_NO_SYSLOG).is_ok());

    // FIXME: json! uses unwrap()!
    #[allow(clippy::disallowed_methods)]
    let mut msg = json!({
        "c": tick(),
        "t": now(),
        "l": level as usize,
        "uid": Uid::current().as_raw(),
    });

    let obj = if let Some(obj) = msg.as_object_mut() {
        obj.extend(data);
        obj
    } else {
        log::error!("Failed to create log message object.");
        return;
    };

    if let Some(pid) = obj.get("pid") {
        // Process ID is given.
        // If log feature is enabled enrichen the message with cmdline and cwd.
        // If log feature is not enabled enrichen the message with comm only.
        #[allow(clippy::cast_possible_truncation)]
        let pid = Pid::from_raw(pid.as_i64().unwrap_or(0) as nix::libc::pid_t);
        if pid.as_raw() != 0 {
            #[allow(clippy::cast_sign_loss)]
            obj.insert(
                "id".to_string(),
                Value::String((pid.as_raw() as u64).to_name()),
            );

            #[cfg(not(feature = "log"))]
            {
                if let Ok(cmd) = proc_comm(pid).map(mask_path) {
                    obj.insert("cmd".to_string(), Value::String(cmd));
                } else {
                    obj.insert("cmd".to_string(), Value::Null);
                }
            }

            #[cfg(feature = "log")]
            {
                if let Ok(cmd) = proc_cmdline(pid).map(mask_path) {
                    obj.insert("cmd".to_string(), Value::String(cmd));
                } else {
                    obj.insert("cmd".to_string(), Value::Null);
                }
                if let Ok(dir) = proc_cwd(pid).map(mask_path) {
                    obj.insert("cwd".to_string(), Value::String(dir));
                } else {
                    obj.insert("cwd".to_string(), Value::Null);
                }
            }
        }
    }

    let msg = msg.to_string();
    match level {
        Level::Error => {
            if !*NO_SYSLOG {
                syslog(&msg);
            }
            log::error!("{msg}");
        }
        Level::Warn => {
            if !*NO_SYSLOG {
                syslog(&msg);
            }
            log::warn!("{msg}");
        }
        Level::Info => log::info!("{msg}"),
        Level::Debug => log::debug!("{msg}"),
        Level::Trace => log::trace!("{msg}"),
    }
}

/// Logs a message to the system's syslog.
///
/// # Arguments
///
/// * `message` - The message string to be logged.
fn syslog(message: &str) {
    let c_msg = CString::new(message).unwrap_or_else(|_|
        // SAFETY: We ensure the string has no null bytes and the vector
        // we pass into the function is an actual byte vector.
        unsafe { CString::from_vec_unchecked(b"?".to_vec()) });

    // SAFETY: Use the syslog interface provided by libc.
    unsafe {
        nix::libc::openlog(std::ptr::null(), nix::libc::LOG_PID, nix::libc::LOG_USER);
        nix::libc::syslog(
            nix::libc::LOG_WARNING,
            "%s\0".as_ptr().cast(),
            c_msg.as_ptr(),
        );
        nix::libc::closelog();
    }
}

/// Returns a current value of the tick counter to use as a staring point.
#[cfg(any(target_arch = "aarch64", target_arch = "x86_64"))]
#[inline]
fn tick() -> u64 {
    tick_counter::start()
}

/// Return 0 for compat.
#[cfg(not(any(target_arch = "aarch64", target_arch = "x86_64")))]
#[inline]
fn tick() -> u64 {
    0u64
}

/// Return the current time in seconds since the unix epoch.
#[inline]
fn now() -> u64 {
    SystemTime::now()
        .duration_since(UNIX_EPOCH)
        .unwrap_or_default()
        .as_secs()
}
