//
// Syd: rock-solid unikernel
// src/pool.rs: Self growing / shrinking `ThreadPool` implementation
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon rusty_pool which is:
//     Copyright (c) Robin Friedli <robinfriedli@icloud.com>
//     SPDX-License-Identifier: Apache-2.0
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Last sync with rusty_pool:
// Version 0.7.0
// Commit:d56805869ba3cbe47021d5660bbaf19ac5ec4bfb

use std::{
    option::Option,
    sync::{
        atomic::{AtomicUsize, Ordering},
        Arc,
    },
    thread,
    time::Duration,
};

use parking_lot::{Condvar, Mutex};

const BITS: usize = std::mem::size_of::<usize>() * 8;
/// The absolute maximum number of workers. This corresponds to the maximum value that can be stored within half the bits of usize,
/// as two counters (total workers and idle workers) are stored in one AtomicUsize.
pub(crate) const MAX_SIZE: usize = (1 << (BITS / 2)) - 1;

type Job = Box<dyn FnOnce() + Send + 'static>;

/// Trait to implement for all items that may be executed by the
/// `ThreadPool`.
pub(crate) trait Task<R: Send>: Send {
    /// Execute this task and return its result.
    fn run(self) -> R;

    /// Transform this `Task` into a heap allocated `FnOnce` if
    /// possible.
    ///
    /// Used by
    /// [`ThreadPool::execute`](struct.ThreadPool.html#method.execute)
    /// to turn this `Task` into a `Job` directly without having to
    /// create an additional `Job` that calls this `Task`.
    fn into_fn(self) -> Option<Box<dyn FnOnce() -> R + Send + 'static>>;

    /// Return `true` if calling [`Task::into_fn`] on this `Task` returns `Some`.
    fn is_fn(&self) -> bool;
}

/// Implement the `Task` trait for any FnOnce closure that returns a
/// thread-safe result.
impl<R, F> Task<R> for F
where
    R: Send,
    F: FnOnce() -> R + Send + 'static,
{
    fn run(self) -> R {
        self()
    }

    fn into_fn(self) -> Option<Box<dyn FnOnce() -> R + Send + 'static>> {
        Some(Box::new(self))
    }

    fn is_fn(&self) -> bool {
        true
    }
}

/// Self growing / shrinking `ThreadPool` implementation based on
/// crossbeam's multi-producer multi-consumer channels that enables
/// awaiting the result of a task and offers async support.
///
/// This `ThreadPool` has two different pool sizes; a core pool size
/// filled with threads that live for as long as the channel and a max
/// pool size which describes the maximum amount of worker threads that
/// may live at the same time. Those additional non-core threads have a
/// specific keep_alive time described when creating the `ThreadPool`
/// that defines how long such threads may be idle for without receiving
/// any work before giving up and terminating their work loop.
///
/// This `ThreadPool` does not spawn any threads until a task is
/// submitted to it. Then it will create a new thread for each task
/// until the core pool size is full. After that a new thread will only
/// be created upon an `execute()` call if the current pool is lower
/// than the max pool size and there are no idle threads.
///
/// Functions like `evaluate()` and `complete()` return a `JoinHandle`
/// that may be used to await the result of a submitted task or future.
/// JoinHandles may be sent to the thread pool to create a task that
/// blocks a worker thread until it receives the result of the other
/// task and then operates on the result. If the task panics the
/// `JoinHandle` receives a cancellation error. This is implemented
/// using a futures oneshot channel to communicate with the worker
/// thread.
///
/// When creating a new worker this `ThreadPool` tries to increment the
/// worker count using a compare-and-swap mechanism, if the increment
/// fails because the total worker count has been incremented to the
/// specified limit (the core_size when trying to create a core thread,
/// else the max_size) by another thread, the pool tries to create a
/// non-core worker instead (if previously trying to create a core
/// worker and no idle worker exists) or sends the task to the channel
/// instead. Panicking workers are always cloned and replaced.
///
/// Locks are only used for the join functions to lock the `Condvar`,
/// apart from that this `ThreadPool` implementation fully relies on
/// crossbeam and atomic operations. This `ThreadPool` decides whether
/// it is currently idle (and should fast-return join attempts) by
/// comparing the total worker count to the idle worker count, which are
/// two values stored in one `AtomicUsize` (both half the size of usize)
/// making sure that if both are updated they may be updated in a single
/// atomic operation.
///
/// The thread pool and its crossbeam channel can be destroyed by using
/// the shutdown function, however that does not stop tasks that are
/// already running but will terminate the thread the next time it will
/// try to fetch work from the channel. The channel is only destroyed
/// once all clones of the `ThreadPool` have been shut down / dropped.
#[derive(Clone)]
pub(crate) struct ThreadPool {
    core_size: usize,
    max_size: usize,
    keep_alive: Duration,
    channel_data: Arc<ChannelData>,
    worker_data: Arc<WorkerData>,
}

impl ThreadPool {
    /// Construct a new `ThreadPool` with the specified core pool size,
    /// max pool size and keep_alive time for non-core threads. This
    /// function does not spawn any threads. This `ThreadPool` will
    /// receive a default name in the following format: "syd_pool_" +
    /// pool number.
    ///
    /// `core_size` specifies the amount of threads to keep alive for as
    /// long as the `ThreadPool` exists and its channel remains
    /// connected.
    ///
    /// `max_size` specifies the maximum number of worker threads that
    /// may exist at the same time.
    ///
    /// `keep_alive` specifies the duration for which to keep non-core
    /// pool worker threads alive while they do not receive any work.
    ///
    /// # Panics
    ///
    /// This function will panic if max_size is 0, lower than core_size
    /// or exceeds half the size of usize. This restriction exists
    /// because two counters (total workers and idle counters) are
    /// stored within one AtomicUsize.
    pub(crate) fn new(core_size: usize, max_size: usize, keep_alive: Duration) -> Self {
        static POOL_COUNTER: AtomicUsize = AtomicUsize::new(1);
        let name = format!("syd_pool_{}", POOL_COUNTER.fetch_add(1, Ordering::Relaxed));
        ThreadPool::new_named(name, core_size, max_size, keep_alive)
    }

    /// Construct a new `ThreadPool` with the specified name, core pool
    /// size, max pool size and keep_alive time for non-core threads.
    /// This function does not spawn any threads.
    ///
    /// `name` the name of the `ThreadPool` that will be used as prefix
    /// for each thread.
    ///
    /// `core_size` specifies the amount of threads to keep alive for as
    /// long as the `ThreadPool` exists and its channel remains
    /// connected.
    ///
    /// `max_size` specifies the maximum number of worker threads that
    /// may exist at the same time.
    ///
    /// `keep_alive` specifies the duration for which to keep non-core
    /// pool worker threads alive while they do not receive any work.
    ///
    /// # Panics
    ///
    /// This function will panic if max_size is 0, lower than core_size
    /// or exceeds half the size of usize. This restriction exists
    /// because two counters (total workers and idle counters) are
    /// stored within one AtomicUsize.
    pub(crate) fn new_named(
        name: String,
        core_size: usize,
        max_size: usize,
        keep_alive: Duration,
    ) -> Self {
        let (sender, receiver) = crossbeam_channel::unbounded();

        if max_size == 0 || max_size < core_size {
            panic!("max_size must be greater than 0 and greater or equal to the core pool size");
        } else if max_size > MAX_SIZE {
            panic!(
                "max_size may not exceed {}, the maximum value that can be stored within half the bits of usize ({} -> {} bits in this case)",
                MAX_SIZE,
                BITS,
                BITS / 2
            );
        }

        let worker_data = WorkerData {
            pool_name: name,
            worker_count_data: WorkerCountData::default(),
            worker_number: AtomicUsize::new(1),
            join_notify_condvar: Condvar::new(),
            join_notify_mutex: Mutex::new(()),
        };

        let channel_data = ChannelData { sender, receiver };

        Self {
            core_size,
            max_size,
            keep_alive,
            channel_data: Arc::new(channel_data),
            worker_data: Arc::new(worker_data),
        }
    }

    /// Send a new task to the worker threads. This function is
    /// responsible for sending the message through the channel and
    /// creating new workers if needed. If the current worker count is
    /// lower than the core pool size this function will always create a
    /// new worker. If the current worker count is equal to or greater
    /// than the core pool size this function only creates a new worker
    /// if the worker count is below the max pool size and there are no
    /// idle threads.
    ///
    /// When attempting to increment the total worker count before
    /// creating a worker fails due to the counter reaching the provided
    /// limit (core_size when attempting to create core thread, else
    /// max_size) after being incremented by another thread, the pool
    /// tries to create a non-core worker instead (if previously trying
    /// to create a core worker and no idle worker exists) or sends the
    /// task to the channel instead. If incrementing the counter
    /// succeeded, either because the current value of the counter
    /// matched the expected value or because the last observed value
    /// was still below the limit, the worker starts with the provided
    /// task as initial task and spawns its thread.
    ///
    /// # Panics
    ///
    /// This function might panic if `try_execute` returns an error when
    /// the crossbeam channel has been closed unexpectedly.
    /// This should never occur under normal circumstances using safe
    /// code, as shutting down the `ThreadPool` consumes ownership and
    /// the crossbeam channel is never dropped unless dropping the
    /// `ThreadPool`.
    pub(crate) fn execute<T: Task<()> + 'static>(&self, task: T) {
        if self.try_execute(task).is_err() {
            panic!("the channel of the thread pool has been closed");
        }
    }

    /// Send a new task to the worker threads. This function is
    /// responsible for sending the message through the channel and
    /// creating new workers if needed. If the current worker count is
    /// lower than the core pool size this function will always create a
    /// new worker. If the current worker count is equal to or greater
    /// than the core pool size this function only creates a new worker
    /// if the worker count is below the max pool size and there are no
    /// idle threads.
    ///
    /// When attempting to increment the total worker count before
    /// creating a worker fails due to the counter reaching the provided
    /// limit (core_size when attempting to create core thread, else
    /// max_size) after being incremented by another thread, the pool
    /// tries to create a non-core worker instead (if previously trying
    /// to create a core worker and no idle worker exists) or sends the
    /// task to the channel instead. If incrementing the counter
    /// succeeded, either because the current value of the counter
    /// matched the expected value or because the last observed value
    /// was still below the limit, the worker starts with the provided
    /// task as initial task and spawns its thread.
    ///
    /// # Errors
    ///
    /// This function might return `crossbeam_channel::SendError` if the
    /// sender was dropped unexpectedly.
    pub(crate) fn try_execute<T: Task<()> + 'static>(
        &self,
        task: T,
    ) -> Result<(), crossbeam_channel::SendError<Job>> {
        if task.is_fn() {
            #[allow(clippy::disallowed_methods)]
            self.try_execute_task(
                task.into_fn()
                    .expect("Task::into_fn returned None despite is_fn returning true"),
            )
        } else {
            self.try_execute_task(Box::new(move || {
                task.run();
            }))
        }
    }

    #[inline]
    fn try_execute_task(&self, task: Job) -> Result<(), crossbeam_channel::SendError<Job>> {
        // create a new worker either if the current worker count is lower than the core pool size
        // or if there are no idle threads and the current worker count is lower than the max pool size
        let worker_count_data = &self.worker_data.worker_count_data;
        let mut worker_count_val = worker_count_data.worker_count.load(Ordering::Relaxed);
        let (mut curr_worker_count, idle_worker_count) = WorkerCountData::split(worker_count_val);
        let mut curr_idle_count = idle_worker_count;

        // always create a new worker if current pool size is below core size
        if curr_worker_count < self.core_size {
            let witnessed =
                worker_count_data.try_increment_worker_total(worker_count_val, self.core_size);

            // the witnessed value matched the expected value, meaning the initial exchange succeeded, or the final witnessed
            // value is still below the coreSize, meaning the increment eventually succeeded
            if witnessed == worker_count_val
                || WorkerCountData::get_total_count(witnessed) < self.core_size
            {
                let worker = Worker::new(
                    self.channel_data.receiver.clone(),
                    Arc::clone(&self.worker_data),
                    None,
                );

                worker.start(Some(task));
                return Ok(());
            }

            curr_worker_count = WorkerCountData::get_total_count(witnessed);
            curr_idle_count = WorkerCountData::get_idle_count(witnessed);
            worker_count_val = witnessed;
        }

        // create a new worker if the current worker count is below the
        // maxSize and the pool has been observed to be busy (no idle
        // workers) during the invocation of this function.
        if curr_worker_count < self.max_size && (idle_worker_count == 0 || curr_idle_count == 0) {
            let witnessed =
                worker_count_data.try_increment_worker_total(worker_count_val, self.max_size);

            if witnessed == worker_count_val
                || WorkerCountData::get_total_count(witnessed) < self.max_size
            {
                let worker = Worker::new(
                    self.channel_data.receiver.clone(),
                    Arc::clone(&self.worker_data),
                    Some(self.keep_alive),
                );

                worker.start(Some(task));
                return Ok(());
            }
        }

        self.send_task_to_channel(task)
    }

    #[inline]
    fn send_task_to_channel(&self, task: Job) -> Result<(), crossbeam_channel::SendError<Job>> {
        self.channel_data.sender.send(task)?;

        Ok(())
    }
}

impl Default for ThreadPool {
    /// create default ThreadPool with the core pool size being equal to the number of cpus
    /// and the max_size being twice the core size with a 60 second timeout
    fn default() -> Self {
        let num_cpus = num_cpus::get();
        ThreadPool::new(
            num_cpus,
            std::cmp::max(num_cpus, num_cpus.saturating_mul(2)),
            Duration::from_secs(60),
        )
    }
}

/// A helper struct to aid creating a new `ThreadPool` using default values where no value was
/// explicitly specified.
#[derive(Default)]
pub(crate) struct Builder {
    name: Option<String>,
    core_size: Option<usize>,
    max_size: Option<usize>,
    keep_alive: Option<Duration>,
}

impl Builder {
    /// Create a new `Builder`.
    pub(crate) fn new() -> Builder {
        Builder::default()
    }

    /// Specify the name of the `ThreadPool` that will be used as prefix
    /// for the name of each worker thread.
    /// By default the name is "syd_pool_x" with x being a static pool
    /// counter.
    pub(crate) fn name(mut self, name: String) -> Builder {
        self.name = Some(name);
        self
    }

    /// Specify the core pool size for the `ThreadPool`. The core pool
    /// size is the number of threads that stay alive for the entire
    /// lifetime of the `ThreadPool` or, to be more precise, its
    /// channel. These threads are spawned if a task is submitted to the
    /// `ThreadPool` and the current worker count is below the core pool
    /// size.
    pub(crate) fn core_size(mut self, size: usize) -> Builder {
        self.core_size = Some(size);
        self
    }

    /// Specify the maximum pool size this `ThreadPool` may scale up to.
    /// This numbers represents the maximum number of threads that may
    /// be alive at the same time within this pool. Additional threads
    /// above the core pool size only remain idle for the duration
    /// specified by the `keep_alive` parameter before terminating. If
    /// the core pool is full, the current pool size is below the max
    /// size and there are no idle threads then additional threads will
    /// be spawned.
    pub(crate) fn max_size(mut self, size: usize) -> Builder {
        self.max_size = Some(size);
        self
    }

    /// Specify the duration for which additional threads outside the
    /// core pool remain alive while not receiving any work before
    /// giving up and terminating.
    pub(crate) fn keep_alive(mut self, keep_alive: Duration) -> Builder {
        self.keep_alive = Some(keep_alive);
        self
    }

    /// Build the `ThreadPool` using the parameters previously supplied
    /// to this `Builder` using the number of CPUs as default core size
    /// if none provided, twice the core size as max size if none
    /// provided, 60 seconds keep_alive if none provided and the default
    /// naming (syd_{pool_number}) if none provided. This
    /// function calls
    /// [`ThreadPool::new`](struct.ThreadPool.html#method.new) or
    /// [`ThreadPool::new_named`](struct.ThreadPool.html#method.new_named)
    /// depending on whether a name was provided.
    ///
    /// # Panics
    ///
    /// Building might panic if the `max_size` is 0 or lower than
    /// `core_size` or exceeds half the size of usize. This restriction
    /// exists because two counters (total workers and idle counters)
    /// are stored within one AtomicUsize.
    pub(crate) fn build(self) -> ThreadPool {
        use std::cmp::{max, min};

        let core_size = self.core_size.unwrap_or_else(|| {
            let num_cpus = num_cpus::get();
            if let Some(max_size) = self.max_size {
                min(MAX_SIZE, min(num_cpus, max_size))
            } else {
                min(MAX_SIZE, num_cpus)
            }
        });
        // handle potential overflow: try using twice the core_size or
        // return core_size.
        let max_size = self
            .max_size
            .unwrap_or_else(|| min(MAX_SIZE, max(core_size, core_size.saturating_mul(2))));
        let keep_alive = self.keep_alive.unwrap_or_else(|| Duration::from_secs(60));

        if let Some(name) = self.name {
            ThreadPool::new_named(name, core_size, max_size, keep_alive)
        } else {
            ThreadPool::new(core_size, max_size, keep_alive)
        }
    }
}

#[derive(Clone)]
struct Worker {
    receiver: crossbeam_channel::Receiver<Job>,
    worker_data: Arc<WorkerData>,
    keep_alive: Option<Duration>,
}

impl Worker {
    fn new(
        receiver: crossbeam_channel::Receiver<Job>,
        worker_data: Arc<WorkerData>,
        keep_alive: Option<Duration>,
    ) -> Self {
        Worker {
            receiver,
            worker_data,
            keep_alive,
        }
    }

    fn start(self, task: Option<Job>) {
        let worker_name = format!(
            "{}_emu_{}",
            self.worker_data.pool_name,
            self.worker_data
                .worker_number
                .fetch_add(1, Ordering::Relaxed)
        );

        #[allow(clippy::disallowed_methods)]
        thread::Builder::new()
            .name(worker_name)
            .spawn(move || {
                let mut sentinel = Sentinel::new(&self);

                if let Some(task) = task {
                    self.exec_task_and_notify(&mut sentinel, task);
                }

                loop {
                    // the two functions return different error types,
                    // but since the error type doesn't matter it is
                    // mapped to unit to make them compatible.
                    let received_task: Result<Job, _> = match self.keep_alive {
                        Some(keep_alive) => self.receiver.recv_timeout(keep_alive).map_err(|_| ()),
                        None => self.receiver.recv().map_err(|_| ()),
                    };

                    match received_task {
                        Ok(task) => {
                            // mark current as no longer idle and execute task.
                            self.worker_data.worker_count_data.decrement_worker_idle();
                            self.exec_task_and_notify(&mut sentinel, task);
                        }
                        Err(_) => {
                            // either channel was broken because the
                            // sender disconnected or, if can_timeout is
                            // true, the Worker has not received any
                            // work during its keep_alive period and
                            // will now terminate, break working loop.
                            break;
                        }
                    }
                }

                // can decrement both at once as the thread only gets
                // here from an idle state (if waiting for work and
                // receiving an error).
                self.worker_data.worker_count_data.decrement_both();
            })
            .expect("could not spawn thread");
    }

    #[inline]
    fn exec_task_and_notify(&self, sentinel: &mut Sentinel, task: Job) {
        sentinel.is_working = true;
        task();
        sentinel.is_working = false;
        // can already mark as idle as this thread will continue the
        // work loop.
        self.mark_idle_and_notify_joiners_if_no_work();
    }

    #[inline]
    fn mark_idle_and_notify_joiners_if_no_work(&self) {
        let (old_total_count, old_idle_count) = self
            .worker_data
            .worker_count_data
            .increment_worker_idle_ret_both();
        // if the last task was the last one in the current generation,
        // i.e. if incrementing the idle count leads to the idle count
        // being equal to the total worker count, notify joiners.
        if old_total_count == old_idle_count.saturating_add(1) && self.receiver.is_empty() {
            let _lock = self.worker_data.join_notify_mutex.lock();
            self.worker_data.join_notify_condvar.notify_all();
        }
    }
}

/// Type that exists to manage worker exit on panic.
///
/// This type is constructed once per `Worker` and implements `Drop` to
/// handle proper worker exit in case the worker panics when executing
/// the current task or anywhere else in its work loop. If the
/// `Sentinel` is dropped at the end of the worker's work loop and the
/// current thread is panicking, handle worker exit the same way as if
/// the task completed normally (if the worker panicked while executing
/// a submitted task) then clone the worker and start it with an initial
/// task of `None`.
struct Sentinel<'s> {
    is_working: bool,
    worker_ref: &'s Worker,
}

impl Sentinel<'_> {
    fn new(worker_ref: &Worker) -> Sentinel<'_> {
        Sentinel {
            is_working: false,
            worker_ref,
        }
    }
}

impl Drop for Sentinel<'_> {
    fn drop(&mut self) {
        if thread::panicking() {
            if self.is_working {
                // worker thread panicked in the process of executing a
                // submitted task, run the same logic as if the task
                // completed normally and mark it as idle, since a clone
                // of this worker will start the work loop as idle
                // thread.
                self.worker_ref.mark_idle_and_notify_joiners_if_no_work();
            }

            let worker = self.worker_ref.clone();
            worker.start(None);
        }
    }
}

const WORKER_IDLE_MASK: usize = MAX_SIZE;
const INCREMENT_TOTAL: usize = 1 << (BITS / 2);
const INCREMENT_IDLE: usize = 1;

/// Struct that stores and handles an `AtomicUsize` that stores the
/// total worker count in the higher half of bits and the idle worker
/// count in the lower half of bits. This allows to to increment /
/// decrement both counters in a single atomic operation.
#[derive(Default)]
struct WorkerCountData {
    worker_count: AtomicUsize,
}

impl WorkerCountData {
    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn increment_both(&self) -> (usize, usize) {
        let old_val = self
            .worker_count
            .fetch_add(INCREMENT_TOTAL | INCREMENT_IDLE, Ordering::Relaxed);
        WorkerCountData::split(old_val)
    }

    fn decrement_both(&self) -> (usize, usize) {
        let old_val = self
            .worker_count
            .fetch_sub(INCREMENT_TOTAL | INCREMENT_IDLE, Ordering::Relaxed);
        WorkerCountData::split(old_val)
    }

    fn try_increment_worker_total(&self, expected: usize, max_total: usize) -> usize {
        self.try_increment_worker_count(expected, INCREMENT_TOTAL, max_total)
    }

    fn try_increment_worker_count(
        &self,
        mut expected: usize,
        increment: usize,
        max_total: usize,
    ) -> usize {
        loop {
            match self.worker_count.compare_exchange_weak(
                expected,
                expected.saturating_add(increment),
                Ordering::Relaxed,
                Ordering::Relaxed,
            ) {
                Ok(witnessed) => return witnessed,
                Err(witnessed) if WorkerCountData::get_total_count(witnessed) >= max_total => {
                    return witnessed
                }
                Err(witnessed) => expected = witnessed,
            }
        }
    }

    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn increment_worker_total(&self) -> usize {
        let old_val = self
            .worker_count
            .fetch_add(INCREMENT_TOTAL, Ordering::Relaxed);
        WorkerCountData::get_total_count(old_val)
    }

    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn increment_worker_total_ret_both(&self) -> (usize, usize) {
        let old_val = self
            .worker_count
            .fetch_add(INCREMENT_TOTAL, Ordering::Relaxed);
        WorkerCountData::split(old_val)
    }

    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn decrement_worker_total(&self) -> usize {
        let old_val = self
            .worker_count
            .fetch_sub(INCREMENT_TOTAL, Ordering::Relaxed);
        WorkerCountData::get_total_count(old_val)
    }

    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn decrement_worker_total_ret_both(&self) -> (usize, usize) {
        let old_val = self
            .worker_count
            .fetch_sub(INCREMENT_TOTAL, Ordering::Relaxed);
        WorkerCountData::split(old_val)
    }

    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn increment_worker_idle(&self) -> usize {
        let old_val = self
            .worker_count
            .fetch_add(INCREMENT_IDLE, Ordering::Relaxed);
        WorkerCountData::get_idle_count(old_val)
    }

    fn increment_worker_idle_ret_both(&self) -> (usize, usize) {
        let old_val = self
            .worker_count
            .fetch_add(INCREMENT_IDLE, Ordering::Relaxed);
        WorkerCountData::split(old_val)
    }

    fn decrement_worker_idle(&self) -> usize {
        let old_val = self
            .worker_count
            .fetch_sub(INCREMENT_IDLE, Ordering::Relaxed);
        WorkerCountData::get_idle_count(old_val)
    }

    // keep for testing and completion's sake
    #[allow(dead_code)]
    fn decrement_worker_idle_ret_both(&self) -> (usize, usize) {
        let old_val = self
            .worker_count
            .fetch_sub(INCREMENT_IDLE, Ordering::Relaxed);
        WorkerCountData::split(old_val)
    }

    #[inline]
    fn split(val: usize) -> (usize, usize) {
        let total_count = val >> (BITS / 2);
        let idle_count = val & WORKER_IDLE_MASK;
        (total_count, idle_count)
    }

    #[inline]
    fn get_total_count(val: usize) -> usize {
        val >> (BITS / 2)
    }

    #[inline]
    fn get_idle_count(val: usize) -> usize {
        val & WORKER_IDLE_MASK
    }
}

/// struct containing data shared between workers
struct WorkerData {
    pool_name: String,
    worker_count_data: WorkerCountData,
    worker_number: AtomicUsize,
    join_notify_condvar: Condvar,
    join_notify_mutex: Mutex<()>,
}

struct ChannelData {
    sender: crossbeam_channel::Sender<Job>,
    receiver: crossbeam_channel::Receiver<Job>,
}

#[cfg(test)]
mod tests {

    use std::{
        sync::{
            atomic::{AtomicUsize, Ordering},
            Arc,
        },
        thread,
        time::Duration,
    };

    use super::{Builder, ThreadPool};

    #[test]
    fn it_works() {
        let pool = ThreadPool::new(2, 10, Duration::from_secs(5));
        let count = Arc::new(AtomicUsize::new(0));

        let count1 = count.clone();
        pool.execute(move || {
            count1.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        let count2 = count.clone();
        pool.execute(move || {
            count2.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        let count3 = count.clone();
        pool.execute(move || {
            count3.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        let count4 = count.clone();
        pool.execute(move || {
            count4.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        thread::sleep(std::time::Duration::from_secs(20));
        let count5 = count.clone();
        pool.execute(move || {
            count5.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        let count6 = count.clone();
        pool.execute(move || {
            count6.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        let count7 = count.clone();
        pool.execute(move || {
            count7.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        let count8 = count.clone();
        pool.execute(move || {
            count8.fetch_add(1, Ordering::Relaxed);
            thread::sleep(std::time::Duration::from_secs(4));
        });
        thread::sleep(std::time::Duration::from_secs(20));
    }

    #[test]
    #[ignore]
    fn stress_test() {
        let pool = Arc::new(ThreadPool::new(3, 50, Duration::from_secs(30)));
        let counter = Arc::new(AtomicUsize::new(0));

        for _ in 0..5 {
            let pool_1 = pool.clone();
            let clone = counter.clone();
            pool.execute(move || {
                for _ in 0..160 {
                    let clone = clone.clone();
                    pool_1.execute(move || {
                        clone.fetch_add(1, Ordering::Relaxed);
                        thread::sleep(Duration::from_secs(10));
                    });
                }

                thread::sleep(Duration::from_secs(20));

                for _ in 0..160 {
                    let clone = clone.clone();
                    pool_1.execute(move || {
                        clone.fetch_add(1, Ordering::Relaxed);
                        thread::sleep(Duration::from_secs(10));
                    });
                }
            });
        }
    }

    #[should_panic(
        expected = "max_size must be greater than 0 and greater or equal to the core pool size"
    )]
    #[test]
    fn test_panic_on_0_max_pool_size() {
        ThreadPool::new(0, 0, Duration::from_secs(2));
    }

    #[should_panic(
        expected = "max_size must be greater than 0 and greater or equal to the core pool size"
    )]
    #[test]
    fn test_panic_on_smaller_max_than_core_pool_size() {
        ThreadPool::new(10, 4, Duration::from_secs(2));
    }

    #[should_panic(expected = "max_size may not exceed")]
    #[test]
    fn test_panic_on_max_size_exceeds_half_usize() {
        ThreadPool::new(
            10,
            1 << ((std::mem::size_of::<usize>() * 8) / 2),
            Duration::from_secs(2),
        );
    }

    #[test]
    fn test_builder_max_size() {
        Builder::new().max_size(1).build();
    }
}
