//
// Syd: rock-solid unikernel
// src/syd-aes.rs: AES-CTR Encryption and Decryption Utility
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    env,
    os::fd::AsRawFd,
    process::ExitCode,
    time::{Duration, Instant},
};

use anyhow::{Context, Result};
use getargs::{Opt, Options};
use nix::{
    fcntl::{splice, OFlag, SpliceFFlags},
    sys::socket::{sendmsg, AlgAddr, ControlMessage, MsgFlags},
    unistd::{isatty, pipe2},
};
use syd::hash::{
    aes_ctr_dec, aes_ctr_enc, aes_ctr_flush, aes_ctr_init, aes_ctr_setup, Key, BLOCK_SIZE, IV,
    IV_SIZE, KEY_SIZE,
};

fn process_data(encrypting: bool, key: Key, iv: IV, verbose: bool) -> Result<()> {
    let fd = aes_ctr_setup(&key)
        .map_err(anyhow::Error::msg)
        .context("Error setting up Kernel Cryptography API connection.")?;
    let fd = aes_ctr_init(&fd, false)
        .map_err(anyhow::Error::msg)
        .context("Error initializing AES-CTR {en,de}cryption socket.")?;

    if encrypting {
        aes_ctr_enc(&fd, &[], &iv, true)
            .map_err(anyhow::Error::msg)
            .context("Error setting up AES-CTR encryption.")?;
    } else {
        aes_ctr_dec(&fd, &[], &iv, true)
            .map_err(anyhow::Error::msg)
            .context("Error setting up AES-CTR decryption.")?;
    }

    let (pipe_rd, pipe_wr) = pipe2(OFlag::O_CLOEXEC)
        .map_err(anyhow::Error::msg)
        .context("Error creating splice pipe pair.")?;

    let mut nread = 0;
    let stime = Instant::now();
    let mut ltime = stime;
    let mut nbytes = 0;
    let mut nwrite = 0;
    let verbose = if verbose {
        isatty(nix::libc::STDERR_FILENO).unwrap_or(false)
    } else {
        false
    };

    loop {
        // Use splice to move data from standard input to pipe.
        match splice(
            std::io::stdin(),
            None,
            &pipe_wr,
            None,
            nix::libc::PIPE_BUF,
            SpliceFFlags::empty(),
        )
        .map_err(anyhow::Error::msg)
        .context("splice from standard input failed!")?
        {
            0 => break, // EOF
            n => {
                match splice(&pipe_rd, None, &fd, None, n, SpliceFFlags::SPLICE_F_MORE)
                    .map_err(anyhow::Error::msg)
                    .context("splice into kernel crypto socket failed!")?
                {
                    0 => break, // EOF
                    n => nread += n,
                };

                while nread >= BLOCK_SIZE {
                    let nblock = (nread / BLOCK_SIZE) * BLOCK_SIZE;
                    let n = aes_ctr_flush(&fd, &std::io::stdout(), nblock, !encrypting)
                        .map_err(anyhow::Error::msg)
                        .context("Failed to write result to standard output!")?;
                    nread -= n;
                    nbytes += n;
                    nwrite += 1;
                }

                if verbose {
                    let now = Instant::now();
                    if now.duration_since(ltime) >= Duration::from_millis(500) {
                        let elapsed = stime.elapsed();
                        let speed = nbytes as f64 / elapsed.as_secs_f64();
                        let output = format!(
                            "{} bytes ({:.2} GB, {:.2} GiB) processed, {:.2?} s, {:.2} MB/s",
                            nbytes,
                            nbytes as f64 / 1_000_000_000.0,
                            nbytes as f64 / (1 << 30) as f64,
                            elapsed,
                            speed / (1 << 20) as f64
                        );
                        eprint!("\r\x1B[K{output}");
                        ltime = now;
                    }
                }
            }
        }
    }

    // Signal the AF_ALG socket the end of the process (No MSG_MORE in flags).
    let cmsgs = [ControlMessage::AlgSetOp(if encrypting {
        &nix::libc::ALG_OP_ENCRYPT
    } else {
        &nix::libc::ALG_OP_DECRYPT
    })];
    sendmsg::<AlgAddr>(fd.as_raw_fd(), &[], &cmsgs, MsgFlags::empty(), None)
        .map_err(anyhow::Error::msg)?;

    if nread > 0 {
        // {En,De}crypt the final batch.
        // SAFETY: Zero-out memory if decrypting.
        aes_ctr_flush(&fd, &std::io::stdout(), nread, !encrypting).map_err(anyhow::Error::msg)?;
        if verbose {
            nbytes += nread;
            nwrite += 1;
        }
    }

    if verbose {
        let elapsed = stime.elapsed();
        eprintln!(
            "\n{} records of each {} bytes processed.\n{} bytes ({:.2} GB, {:.2} GiB) processed, {:.5?} s, {:.2} MB/s",
            nwrite,
            nix::libc::PIPE_BUF,
            nbytes,
            nbytes as f64 / 1_000_000_000.0,
            nbytes as f64 / (1 << 30) as f64,
            elapsed,
            nbytes as f64 / elapsed.as_secs_f64() / (1 << 20) as f64
        );
    }

    Ok(())
}

#[allow(clippy::disallowed_methods)]
fn main() -> Result<ExitCode> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut encrypt: Option<bool> = None;
    let mut key_hex: Option<&str> = None;
    let mut iv_hex: Option<&str> = None;
    let mut verbose = false;

    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('v') => {
                verbose = true;
            }
            Opt::Short('e') => {
                encrypt = Some(true);
            }
            Opt::Short('d') => {
                encrypt = Some(false);
            }
            Opt::Short('k') => {
                key_hex = opts.value().ok();
            }
            Opt::Short('i') => {
                iv_hex = opts.value().ok();
            }
            _ => {
                eprintln!("Unknown option: {opt:?}");
                help();
                return Ok(ExitCode::FAILURE);
            }
        }
    }

    if encrypt.is_none() || key_hex.is_none() || iv_hex.is_none() {
        eprintln!("Error: -e or -d and both -k and -i options are required.");
        help();
        return Ok(ExitCode::FAILURE);
    }

    let key = match const_hex::decode(key_hex.unwrap()) {
        Ok(key) => {
            if key.len() != KEY_SIZE {
                eprintln!("Error: Key must be 512 bits (64 bytes) in length.");
                return Ok(ExitCode::FAILURE);
            }
            Key::new(key.try_into().unwrap())
        }
        Err(err) => {
            eprintln!("Error: Key must be valid hex: {err}");
            return Ok(ExitCode::FAILURE);
        }
    };

    let iv = match const_hex::decode(iv_hex.unwrap()) {
        Ok(iv) => {
            if iv.len() != IV_SIZE {
                eprintln!("Error: IV must be 128 bits (16 bytes) in length.");
                return Ok(ExitCode::FAILURE);
            }
            iv
        }
        Err(err) => {
            eprintln!("Error: IV must be valid hex: {err}");
            return Ok(ExitCode::FAILURE);
        }
    };
    let iv = IV::new(iv.try_into().unwrap());

    process_data(encrypt.unwrap(), key, iv, verbose)
        .map(|_| ExitCode::SUCCESS)
        .context("Error processing data!")
}

fn help() {
    println!("Usage: syd-aes [-h] -e|-d -k <hex-encoded key> -i <hex-encoded iv>");
    println!("AES-CTR Encryption and Decryption Utility");
    println!("Reads from standard input and writes to standard output.");
    println!("  -h        Print this help message and exit.");
    println!("  -v        Enable verbose mode.");
    println!("  -e        Encrypt the input data.");
    println!("  -d        Decrypt the input data.");
    println!("  -k <key>  Hex-encoded key (256 bits for AES-CTR)");
    println!("  -i <iv>   Hex-encoded IV (128 bits for AES-CTR)");
}
