//
// Syd: rock-solid unikernel
// src/syd-cat.rs: Print out the rules of the given sandbox profile and exit
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use anyhow::{Context, Result};
use nix::unistd::{Gid, Uid};
use syd::sandbox::Sandbox;

fn main() -> Result<()> {
    #[allow(clippy::disallowed_methods)]
    syd::set_sigpipe_dfl().expect("SIGPIPE");

    let mut args = std::env::args();
    match args.nth(1).as_deref() {
        None | Some("-h") => {
            println!("Usage: syd-cat profile-name|path");
            println!("Print out the rules of the given sandbox profile and exit.");
            println!("Use list as name to get a list of sandboxing profiles.");
            println!("Given a path, validate the rules in the configuration.");
        }
        Some("list") => {
            println!("container");
            println!("immutable");
            println!("landlock");
            println!("lib");
            println!("noipv4");
            println!("noipv6");
            println!("oci");
            println!("paludis");
            println!("pandora");
            println!("silent");
            println!("user");
        }
        Some("container") => {
            println!("# Syd profile: Container");
            println!(
                "# Number of rules: {}",
                syd::config::PROFILE_CONTAINER.len()
            );
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_CONTAINER {
                println!("{command}");
            }
        }
        Some("immutable") => {
            println!("# Syd profile: Immutable Container");
            println!(
                "# Number of rules: {}",
                syd::config::PROFILE_IMMUTABLE.len()
            );
            println!("# Copyright (c) 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_IMMUTABLE {
                println!("{command}");
            }
        }
        Some("landlock") => {
            println!("# Syd profile: LandLock");
            println!("# Number of rules: {}", syd::config::PROFILE_LANDLOCK.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_LANDLOCK {
                println!("{command}");
            }
        }
        Some("paludis") => {
            println!("# Syd profile: Paludis");
            println!("# Number of rules: {}", syd::config::PROFILE_PALUDIS.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_PALUDIS {
                println!("{command}");
            }
        }
        Some("pandora") => {
            println!("# Syd profile: Pand☮ra");
            println!("# Number of rules: {}", syd::config::PROFILE_PANDORA.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_PANDORA {
                println!("{command}");
            }
        }
        Some("noipv4") => {
            println!("# Syd profile: NoIpv4");
            println!("# Number of rules: {}", syd::config::PROFILE_NOIPV4.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_NOIPV4 {
                println!("{command}");
            }
        }
        Some("noipv6") => {
            println!("# Syd profile: NoIpv6");
            println!("# Number of rules: {}", syd::config::PROFILE_NOIPV6.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_NOIPV6 {
                println!("{command}");
            }
        }
        Some("p") | Some("priv") | Some("privileged") => {
            println!("# Syd profile: Privileged");
            println!(
                "# Number of rules: {}",
                syd::config::PROFILE_PRIVILEGED.len()
            );
            println!("# Copyright (c) 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_PRIVILEGED {
                println!("{command}");
            }
        }
        Some("silent") => {
            println!("# Syd profile: Silent");
            println!("# Number of rules: {}", syd::config::PROFILE_SILENT.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_SILENT {
                println!("{command}");
            }
        }
        Some("lib") => {
            println!("# Syd profile: LibSyd");
            println!("# Number of rules: {}", syd::config::PROFILE_LIB.len());
            println!("# Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_LIB {
                println!("{command}");
            }
        }
        Some("oci") => {
            println!("# Syd profile: OCI");
            println!("# Number of rules: {}", syd::config::PROFILE_OCI.len());
            println!("# Copyright (c) 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_OCI {
                println!("{command}");
            }
        }
        Some("trace") => {
            println!("# Syd profile: Trace");
            println!("# Number of rules: {}", syd::config::PROFILE_TRACE.len());
            println!("# Copyright (c) 2024 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");
            for command in syd::config::PROFILE_TRACE {
                println!("{command}");
            }
        }
        Some("user") => {
            let uid = Uid::effective();
            let gid = Gid::effective();
            let name = syd::get_user_name(uid);
            let mut home = syd::get_user_home(&name);

            println!("# Syd profile: User \"{name}\"");
            println!(
                "# Number of rules: {}",
                syd::config::PROFILE_USER.len().saturating_add(9)
            );
            println!("# Copyright (c) 2023 Ali Polatel <alip@chesswob.org>");
            println!("# SPDX-License-Identifier: GPL-3.0-or-later");

            // Step 1: Static configuration defined at compile time.
            for command in syd::config::PROFILE_USER {
                println!("{command}");
            }

            // Step 2: Dynamic, user-specific configuration.
            // Keep in sync with Sandbox::parse_profile()
            if !home.ends_with('/') {
                home.push('/');
            }

            println!("allow/lock/write+{home}");
            println!("allow/read+{home}***");
            println!("allow/stat+{home}***");
            println!("allow/write+{home}**");
            println!("allow/exec+{home}**");
            println!("allow/net/bind+{home}**");
            println!("allow/net/connect+{home}**");
            println!("allow/read+/run/user/{uid}/**");
            println!("allow/write+/run/user/{uid}/**");
            println!("tpe/negate:1");
            println!("tpe/user_owned:1");
            println!("tpe/gid:{gid}");
        }
        Some(path) => {
            let mut syd = Sandbox::new();
            syd.parse_config_file(path)
                .with_context(|| format!("Error parsing file {path}!"))?;
            eprintln!("Success parsing file {path}.");
        }
    }

    Ok(())
}
