//
// Syd: rock-solid unikernel
// src/syd-cp.rs: Utility to copy files using splice(2)
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    env,
    fs::{File, OpenOptions},
    os::fd::{AsRawFd, BorrowedFd},
    path::Path,
    process::ExitCode,
    sync::atomic::{AtomicBool, Ordering},
    time::{Duration, Instant},
};

use anyhow::{Context, Result};
use getargs::{Opt, Options};
use nix::{
    errno::Errno,
    fcntl::{splice, SpliceFFlags},
    sys::{
        signal,
        signal::{SigAction, SigHandler, Signal},
    },
    unistd::isatty,
};
use syd::config::PIPE_BUF;

static INTERRUPTED: AtomicBool = AtomicBool::new(false);

extern "C" fn handle_sigint(_: nix::libc::c_int) {
    INTERRUPTED.store(true, Ordering::SeqCst);
}

fn process_data(src: &dyn AsRawFd, dst: &dyn AsRawFd, verbose: bool) -> Result<()> {
    let stime = Instant::now();
    let mut ltime = stime;
    let mut nbytes = 0;
    let mut nwrite = 0;
    let verbose = if verbose {
        isatty(nix::libc::STDERR_FILENO).unwrap_or(false)
    } else {
        false
    };

    let src = unsafe { BorrowedFd::borrow_raw(src.as_raw_fd()) };
    let dst = unsafe { BorrowedFd::borrow_raw(dst.as_raw_fd()) };

    while !INTERRUPTED.load(Ordering::Relaxed) {
        let n = match splice(src, None, dst, None, *PIPE_BUF, SpliceFFlags::SPLICE_F_MORE) {
            Ok(0) | Err(Errno::EINTR | Errno::EPIPE) => break,
            Ok(n) => n,
            Err(errno) => return Err(anyhow::Error::msg(errno)),
        };

        if !verbose {
            continue;
        }

        nbytes += n;
        nwrite += 1;

        let now = Instant::now();
        if now.duration_since(ltime) >= Duration::from_millis(500) {
            let elapsed = stime.elapsed();
            let speed = nbytes as f64 / elapsed.as_secs_f64();
            let output = format!(
                "{} bytes ({:.2} GB, {:.2} GiB) spliced, {:.2?} s, {:.2} MB/s",
                nbytes,
                nbytes as f64 / 1_000_000_000.0,
                nbytes as f64 / (1 << 30) as f64,
                elapsed,
                speed / (1 << 20) as f64
            );
            eprint!("\r\x1B[K{output}");
            ltime = now;
        }
    }

    if verbose {
        let elapsed = stime.elapsed();
        eprintln!(
            "\n{} records of each {} bytes spliced.\n{} bytes ({:.2} GB, {:.2} GiB) spliced, {:.5?} s, {:.2} MB/s",
            nwrite,
            *PIPE_BUF,
            nbytes,
            nbytes as f64 / 1_000_000_000.0,
            nbytes as f64 / (1 << 30) as f64,
            elapsed,
            nbytes as f64 / elapsed.as_secs_f64() / (1 << 20) as f64
        );
    }

    Ok(())
}

fn main() -> Result<ExitCode> {
    let args = env::args().skip(1).collect::<Vec<_>>();
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut force = false;
    let mut keeperm = false;
    let mut verbose = false;

    #[allow(clippy::disallowed_methods)]
    while let Some(opt) = opts.next_opt().expect("next opt") {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('f') => {
                force = true;
            }
            Opt::Short('a' | 'p') => {
                keeperm = true;
            }
            Opt::Short('v') => {
                verbose = true;
            }
            _ => {
                eprintln!("Unknown option: {opt:?}");
                help();
                return Ok(ExitCode::FAILURE);
            }
        }
    }

    let mut positionals = opts.positionals();
    let (src, dst) = if let Some(first) = positionals.next() {
        if let Some(second) = positionals.next() {
            (Some(Path::new(first)), Path::new(second))
        } else {
            if keeperm {
                eprintln!("Error: Cannot preserve permissions when source is standard input.");
                help();
                return Ok(ExitCode::FAILURE);
            }
            let dst = Path::new(first);
            if dst.is_dir() {
                eprintln!("Error: Cannot copy from standard input to a directory.");
                help();
                return Ok(ExitCode::FAILURE);
            }
            (None, dst)
        }
    } else {
        eprintln!("Error: Destination file is required.");
        help();
        return Ok(ExitCode::FAILURE);
    };

    #[allow(clippy::type_complexity)]
    let (src, dst): (Box<dyn AsRawFd>, Box<dyn AsRawFd>) = if let Some(src) = src {
        let dst_path = if dst.is_dir() {
            let src_file_name = src
                .file_name()
                .ok_or_else(|| anyhow::Error::msg("Error getting source file name."))?;
            dst.join(src_file_name)
        } else {
            dst.to_path_buf()
        };
        let src = File::open(src)
            .map_err(anyhow::Error::msg)
            .context("Error opening source file.")?;
        let mut opt = OpenOptions::new();
        if force {
            opt.create(true).truncate(true);
        } else {
            opt.create_new(true);
        }
        let dst = opt
            .write(true)
            .open(&dst_path)
            .map_err(anyhow::Error::msg)
            .context("Error opening destination file.")?;

        if keeperm {
            // Get the file permissions of the source file
            let src_permissions = src
                .metadata()
                .map_err(anyhow::Error::msg)
                .context("Error getting metadata for source file.")?
                .permissions();

            // Set the same permissions for the destination file
            dst.set_permissions(src_permissions.clone())
                .map_err(anyhow::Error::msg)
                .context("Error setting permissions for destination file.")?;
        }

        (Box::new(src), Box::new(dst))
    } else {
        let src = std::io::stdin();
        let mut opt = OpenOptions::new();
        if force {
            opt.create(true).truncate(true);
        } else {
            opt.create_new(true);
        }
        let dst = opt
            .write(true)
            .open(dst)
            .map_err(anyhow::Error::msg)
            .context("Error opening destination file.")?;
        (Box::new(src), Box::new(dst))
    };

    // Set up the SIGINT handler
    unsafe {
        signal::sigaction(
            Signal::SIGINT,
            &SigAction::new(
                SigHandler::Handler(handle_sigint),
                signal::SaFlags::empty(),
                signal::SigSet::empty(),
            ),
        )
    }
    .map_err(anyhow::Error::msg)
    .context("Error setting up SIGINT handler")?;

    // SAFETY: Ignore SIGPIPE, this way when the read end of the pipe is
    // closed we'll get an EPIPE errno on the subsequent splice call.
    unsafe {
        signal::sigaction(
            Signal::SIGPIPE,
            &SigAction::new(
                SigHandler::SigIgn,
                signal::SaFlags::empty(),
                signal::SigSet::empty(),
            ),
        )
    }
    .map_err(anyhow::Error::msg)
    .context("Error ignoring SIGPIPE")?;

    process_data(&*src, &*dst, verbose)
        .map(|_| ExitCode::SUCCESS)
        .context("Error processing data")
}

fn help() {
    println!("Usage: syd-cp [-hapfv] <src-file> <dst-file-or-dir>");
    println!("       syd-cp [-hfv] <dst-file>");
    println!("Utility to copy files using splice(2).");
    println!("Copies from stdin when only destination file is given.");
    println!("  -h        Display help.");
    println!("  -f        Force overwriting destination file.");
    println!("  -p, -a    Preserve file permissions.");
    println!("  -v        Enable verbose mode.");
}
