//
// Syd: rock-solid unikernel
// src/syd.rs: Main entry point
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
// Proxying code is based in part upon socksns crate which is:
//     Copyright (c) 2020 Steven Engler
//     SPDX-License-Identifier: MIT
//
// SPDX-License-Identifier: GPL-3.0-or-later

//! Syd: rock-solid unikernel
//! Main entry point.

// We like clean and simple code with documentation.
#![deny(missing_docs)]
#![deny(clippy::allow_attributes_without_reason)]
#![deny(clippy::arithmetic_side_effects)]
#![deny(clippy::as_ptr_cast_mut)]
#![deny(clippy::as_underscore)]
#![deny(clippy::assertions_on_result_states)]
#![deny(clippy::borrow_as_ptr)]
#![deny(clippy::branches_sharing_code)]
#![deny(clippy::case_sensitive_file_extension_comparisons)]
#![deny(clippy::cast_lossless)]
#![deny(clippy::cast_possible_truncation)]
#![deny(clippy::cast_possible_wrap)]
#![deny(clippy::cast_precision_loss)]
#![deny(clippy::cast_ptr_alignment)]
#![deny(clippy::cast_sign_loss)]
#![deny(clippy::checked_conversions)]
#![deny(clippy::clear_with_drain)]
#![deny(clippy::clone_on_ref_ptr)]
#![deny(clippy::cloned_instead_of_copied)]
#![deny(clippy::cognitive_complexity)]
#![deny(clippy::collection_is_never_read)]
#![deny(clippy::copy_iterator)]
#![deny(clippy::create_dir)]
#![deny(clippy::dbg_macro)]
#![deny(clippy::debug_assert_with_mut_call)]
#![deny(clippy::decimal_literal_representation)]
#![deny(clippy::default_trait_access)]
#![deny(clippy::default_union_representation)]
#![deny(clippy::derive_partial_eq_without_eq)]
#![deny(clippy::doc_link_with_quotes)]
#![deny(clippy::doc_markdown)]
#![deny(clippy::explicit_into_iter_loop)]
#![deny(clippy::explicit_iter_loop)]
#![deny(clippy::fallible_impl_from)]
#![deny(clippy::missing_safety_doc)]
#![deny(clippy::undocumented_unsafe_blocks)]

use std::{
    collections::HashMap,
    env,
    ffi::CString,
    fs::File,
    io::Write,
    net::{IpAddr, Ipv4Addr, TcpListener},
    os::{
        fd::{AsRawFd, RawFd},
        unix::{ffi::OsStrExt, net::UnixStream, process::CommandExt},
    },
    path::Path,
    process::{ExitCode, Stdio},
    str::FromStr,
};

use anyhow::{anyhow, bail, Context};
use getargs::{Opt, Options};
use libseccomp::{scmp_cmp, ScmpAction, ScmpFilterContext, ScmpSyscall};
use nix::{
    errno::Errno,
    libc::setdomainname,
    sched::{unshare, CloneFlags},
    sys::wait::{Id, WaitPidFlag},
    time::{clock_gettime, ClockId},
    unistd::{close, fork, getgid, getpid, getuid, sethostname, ForkResult},
    NixPath,
};
use sendfd::SendWithFd;
use syd::{
    bring_up_loopback, caps,
    compat::{waitid, WaitStatus},
    config::*,
    error,
    fs::{err2no, mask_path, set_cloexec},
    hash::{hash_str, NoHashSet},
    hook::{ExportMode, Supervisor},
    info,
    log::JsonLinesLogger,
    nsflags_name,
    sandbox::Sandbox,
    unshare::{GidMap, UidMap},
};

// This line will include the content of "esyd.sh" at compile-time
const ESYD: &str = include_str!("esyd.sh");

#[allow(clippy::cognitive_complexity)]
fn main() -> anyhow::Result<ExitCode> {
    // Set SIGPIPE handler to default.
    syd::set_sigpipe_dfl()
        .map_err(anyhow::Error::msg)
        .context("Failed to set SIGPIPE signal handler to default.")?;

    // Note: If someone ever needs/wants non-UTF-8 CLI arguments,
    // we can consider changing this, but till then we consider
    // this safe and sound.
    #[allow(clippy::disallowed_methods)]
    let mut argv = argv::iter()
        .map(|os| os.to_str().expect("Argument is not valid UTF-8!"))
        .peekable();
    // SAFETY: Safe to assume argv has at least one element.
    // On the off chance that it's empty, we panic.
    #[allow(clippy::disallowed_methods)]
    let is_login = argv.next().expect("arg0").starts_with('-');

    // Handle quick options early before reexecution for convenience.
    if !is_login {
        if let Some(arg) = argv.peek() {
            match *arg {
                "-h" | "--help" => {
                    help();
                    return Ok(ExitCode::SUCCESS);
                }
                "-V" | "--version" => {
                    syd::syd_info();
                    return Ok(ExitCode::SUCCESS);
                }
                "--sh" => {
                    println!("{ESYD}");
                    return Ok(ExitCode::SUCCESS);
                }
                _ => {}
            }
        }
    }

    // Guard against CVE-2019-5736:
    // Copy /proc/self/exe in an anonymous fd (created via memfd_create), seal it and re-execute it.
    // See:
    // - https://github.com/opencontainers/runc/commit/0a8e4117e7f715d5fbeef398405813ce8e88558b
    // - https://github.com/lxc/lxc/commit/6400238d08cdf1ca20d49bafb85f4e224348bf9d
    // Note: syd's procfs protections is another layer of defense against this.
    pentacle::ensure_sealed()
        .map_err(anyhow::Error::msg)
        .context("Failed to seal /proc/self/exe. Is /proc mounted?")?;

    // Initialize logging.
    JsonLinesLogger::init(&None, "warn")?;

    // Parse CLI arguments
    let mut export: Option<ExportMode> = None;
    let mut sandbox: Sandbox = Sandbox::default();
    let mut cmd_arg0: Option<String> = None;
    let mut cmd_argv: Vec<String> = vec![];
    let mut cmd_envp: NoHashSet = NoHashSet::default();

    // Initialize Options.
    let user_done = if is_login || argv.peek().is_none() {
        sandbox
            .parse_profile("user")
            .context("Error parsing user profile!")?;
        true
    } else {
        false
    };
    let mut user_parse = false;
    let mut opts = Options::new(argv);

    // Local options handled by this function.
    let sh = env::var(ENV_SH).unwrap_or(SYD_SH.to_string());

    while let Some(opt) = opts
        .next_opt()
        .map_err(anyhow::Error::msg)
        .context("Failed to call Options::next!")?
    {
        match opt {
            /*
             * Basic options
             */
            Opt::Short('h') | Opt::Long("help") => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('V') | Opt::Long("version") => {
                syd::syd_info();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Long("sh") => {
                println!("{ESYD}");
                return Ok(ExitCode::SUCCESS);
            }

            /*
             * Sandbox options
             */
            Opt::Short('E') => {
                export = Some(ExportMode::from_str(
                    opts.value()
                        .map_err(anyhow::Error::msg)
                        .context("-E requires an argument!")?,
                )?);
                sandbox
                    .parse_profile("lib")
                    .context("Error parsing lib profile!")?;
                if export == Some(ExportMode::PseudoFiltercode) {
                    // This variable makes setup_seccomp_parent print rules.
                    env::set_var("SYD_SECX", "1");
                } else {
                    env::remove_var("SYD_SECX");
                }
            }
            Opt::Short('x') => {
                env::set_var(ENV_NO_SYSLOG, "1");
                sandbox
                    .parse_profile("trace")
                    .context("Error parsing trace profile")?;
            }
            Opt::Short('m') => {
                let cmd = opts
                    .value()
                    .map_err(anyhow::Error::msg)
                    .context("-m requires an argument!")?;
                if sandbox.locked() {
                    bail!("Failed to execute magic command `{cmd}': sandbox locked!");
                } else {
                    sandbox
                        .config(cmd)
                        .map_err(anyhow::Error::msg)
                        .context(format!("Failed to execute magic command `{cmd}'."))?;
                }
            }
            Opt::Short('f') => {
                // Login shell compatibility:
                // Parse user profile as necessary.
                user_parse = true;
            }
            Opt::Short('l') | Opt::Long("login") => {
                // Login shell compatibility:
                // Parse user profile as necessary.
                user_parse = true;
            }
            Opt::Short('c') => {
                // When multiple -c arguments are given,
                // only the first one is honoured and
                // the rest is ignored in consistency
                // with how bash and dash behaves.
                user_parse = true;
                if cmd_argv.is_empty() {
                    cmd_argv.push(sh.clone());
                    cmd_argv.push("-c".to_string());
                    cmd_argv.push(
                        opts.value()
                            .map_err(anyhow::Error::msg)
                            .context("-c requires an argument!")?
                            .to_string(),
                    );
                }
            }
            Opt::Short('P') => {
                let path = opts
                    .value()
                    .map_err(anyhow::Error::msg)
                    .context("-f requires an argument!")?;
                if sandbox.locked() {
                    error!("ctx": "init", "op": "configure_sandbox",
                        "path": mask_path(path),
                        "error": "sandbox is locked", "errno": Errno::EAGAIN as i32);
                } else {
                    sandbox
                        .parse_config_file(path)
                        .with_context(|| format!("Failed to parse configuration file `{path}'."))?;
                }
            }
            /* We keep --profile for syd-1 compatibility.
             * It's undocumented. */
            Opt::Short('p') | Opt::Long("profile") => {
                let profile = opts
                    .value()
                    .map_err(anyhow::Error::msg)
                    .context("-p requires an argument!")?;
                if sandbox.locked() {
                    error!("ctx": "init", "op": "configure_sandbox",
                        "profile": mask_path(profile),
                        "error": "sandbox is locked", "errno": Errno::EAGAIN as i32);
                } else {
                    sandbox.parse_profile(profile).with_context(|| {
                        format!("Failed to parse configuration profile `{profile}'.")
                    })?;
                }
            }

            /*
             * Unshare options
             */
            Opt::Short('a') => {
                let val = opts
                    .value()
                    .map_err(anyhow::Error::msg)
                    .context("-a requires an argument!")?
                    .to_string();
                cmd_arg0 = Some(val);
            }
            Opt::Short('e') => {
                let value = opts
                    .value()
                    .map_err(anyhow::Error::msg)
                    .context("-e requires an argument!")?;
                match value.split_once('=') {
                    Some((var, val)) => {
                        cmd_envp.insert(hash_str(var));
                        if !val.is_empty() {
                            // This way we give the user the chance to pass-through
                            // denylisted environment variables e.g.
                            //      syd -eLD_LIBRARY_PATH= cmd
                            // is equivalent to
                            //      syd -eLD_LIBRARY_PATH=$LD_LIBRARY_PATH cmd
                            env::set_var(var, val);
                        }
                    }
                    None => {
                        cmd_envp.remove(&hash_str(value));
                        env::remove_var(value);
                    }
                }
            }

            Opt::Short(c) => {
                bail!("Invalid option `-{c}'!");
            }
            Opt::Long(c) => {
                bail!("Invalid option `--{c}'!");
            }
        }
    }

    if user_parse && !user_done && !sandbox.locked() {
        sandbox
            .parse_profile("user")
            .context("Error parsing user profile!")?;
    }

    // Prepare the command to execute, which may be a login shell.
    cmd_argv.extend(opts.positionals().map(|s| s.to_string()));
    let argv0 = if !cmd_argv.is_empty() {
        Some(cmd_argv.remove(0))
    } else {
        None
    };

    let argv0 = match (export.is_some(), argv0, is_login) {
        (true, _, _) => "true".to_string(),
        (false, Some(argv0), false) => argv0.to_string(),
        (false, None, false) | (false, _, true) => {
            if cmd_arg0.is_none() {
                // Allow user to override with -a
                // SAFETY: unwrap is safe as `sh` is UTF-8.
                #[allow(clippy::disallowed_methods)]
                let sh = Path::new(&sh)
                    .file_name()
                    .map(|s| s.to_str().unwrap())
                    .unwrap_or("-");
                cmd_arg0 = Some(format!("-{sh}"));
            }
            sh.to_string()
        }
    };

    // Keep standard output open if we're exporting.
    if export.is_some() {
        env::set_var("SYD_KEEP_STDOUT", "1");
    }

    // time implies unshare/time:1
    if sandbox.time.is_some() {
        sandbox.set_unshare_time(true);
    }

    // root/map:1 implies unshare/user:1
    // We do this here rather than in sandbox.rs,
    // because the user may toggle this flag many times.
    let map_root = sandbox.map_root();
    if map_root {
        sandbox.set_unshare_user(true);
    }

    // Set unshare-mount if we've any bind mounts.
    if sandbox.has_bind_mounts() {
        sandbox.set_unshare_mount(true);
    }

    // root: implies unshare/mount:1 and unshare/pid:1.
    if sandbox.root.is_some() {
        sandbox.set_unshare_mount(true);
        sandbox.set_unshare_pid(true);
    }

    // SAFETY: We cannot support NEWPID without NEWNS.
    // ie, pid namespace must have its own private /proc.
    if sandbox.unshare_pid() {
        sandbox.set_unshare_mount(true);
    }

    let proxy = if sandbox.has_proxy() {
        // sandbox/proxy:on implies unshare/net:1
        sandbox.set_unshare_net(true);

        // Step 1: Create a PIDFd of this process.
        // SAFETY: nix has no wrapper for pidfd_open.
        let pidfd = match unsafe {
            nix::libc::syscall(
                nix::libc::SYS_pidfd_open,
                getpid().as_raw(),
                nix::libc::PIDFD_NONBLOCK,
            )
        } {
            e if e < 0 => {
                return Err(anyhow::Error::msg(Errno::last())
                    .context("Failed to create a PIDFd of the current process for proxying!"))
            }
            fd => fd as RawFd,
        };

        // Step 2: Create a UNIX socket pair.
        let (stream_parent, stream_child) = UnixStream::pair()
            .map_err(anyhow::Error::msg)
            .context("Failed to create UNIX socket-pair for proxying!")?;

        // Step 3: Unset the CLOEXEC flags on the file descriptors.
        // PIDFds and Rust sockets are automatically CLOEXEC.
        set_cloexec(&pidfd, false)
            .map_err(anyhow::Error::msg)
            .context("Failed to unset CLOEXEC flag on the PIDFd for proxying!")?;
        let fd = stream_parent.as_raw_fd(); // Borrows FD.
        set_cloexec(&fd, false)
            .map_err(anyhow::Error::msg)
            .context("Failed to unset CLOEXEC flag on the parent UNIX socket for proxying!")?;

        // Step 4: Prepare environment of the syd-tor process.
        // Filter the environment variables to only include PATH and LD_LIBRARY_PATH.
        let safe_env: HashMap<_, _> = env::vars_os()
            .filter(|(key, _)| matches!(key.as_bytes(), b"PATH" | b"LD_LIBRARY_PATH"))
            .collect();

        // Step 5: Spawn syd-tor process outside the namespace.
        // Pass one end of the socket-pair to it.
        std::process::Command::new("syd-tor")
            .arg(format!("-p{pidfd}"))
            .arg(format!("-i{fd}"))
            .arg(format!(
                "-o{}:{}",
                sandbox.proxy_ext_addr, sandbox.proxy_ext_port
            ))
            .env_clear()
            .envs(safe_env)
            .current_dir("/var/empty")
            .process_group(0)
            .stdin(Stdio::null())
            .stdout(Stdio::null())
            .stderr(Stdio::null())
            //Enable debug (comment stderr ^^)
            //.arg("-d")
            //.arg(format!("-b{}", PIPE_BUF))
            .spawn()
            .map_err(anyhow::Error::msg)
            .context("Failed to spawn syd-tor for proxying!")?;
        close(pidfd)
            .map_err(anyhow::Error::msg)
            .context("Failed to close the PIDFd after passing to syd-tor!")?;

        // Step 4: Pass the other end of the socket-pair
        // to the new namespace.
        Some((stream_child, sandbox.proxy_port))
    } else {
        None
    };

    // Set up Linux namespaces if requested. Note,
    // we set it up here before spawning the child so as to
    // include the syd process into the pid namespace as well
    // such that the sandbox process and syd have the identical
    // view of /proc.
    let namespaces = sandbox.namespaces();
    if namespaces == 0 {
        // No namespace arguments passed, run normally.
        return Supervisor::run(&argv0, cmd_arg0, cmd_argv, cmd_envp, sandbox, export)
            .map(ExitCode::from);
    } else {
        info!("ctx": "init", "op": "unshare",
            "ns": namespaces,
            "namespaces": *nsflags_name(namespaces));
    }

    let id_buf = if sandbox.unshare_user() {
        // create the UID and GID mappings.
        let uid = getuid().as_raw();
        let gid = getgid().as_raw();

        let uid_buf = {
            let uid_maps = vec![
                UidMap {
                    inside_uid: if map_root { 0 } else { uid },
                    outside_uid: uid,
                    count: 1,
                }, // Map the current user.
            ];
            let mut buf = Vec::new();
            for map in uid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_uid, map.outside_uid, map.count
                )
                .map_err(anyhow::Error::msg)
                .context("Failed to format uidmap!")?;
            }
            buf
        };

        let gid_buf = {
            let gid_maps = vec![
                GidMap {
                    inside_gid: if map_root { 0 } else { gid },
                    outside_gid: gid,
                    count: 1,
                }, // Map the current group.
            ];
            let mut buf = Vec::new();
            for map in gid_maps {
                writeln!(
                    &mut buf,
                    "{} {} {}",
                    map.inside_gid, map.outside_gid, map.count
                )
                .map_err(anyhow::Error::msg)
                .context("Failed to format gidmap!")?;
            }
            buf
        };
        Some((uid_buf, gid_buf))
    } else {
        None
    };

    // Tell the kernel to keep the capabilities after the unshare call.
    // This is important because unshare() can change the user
    // namespace, which often leads to a loss of capabilities.
    caps::securebits::set_keepcaps(true)
        .map_err(anyhow::Error::msg)
        .context("Failed to set keep capabilities flag!")?;

    // CLONE_NEWTIME may only be used with unshare(2).
    // CloneFlags don't support CLONE_NEWTIME directly so we use retain.
    unshare(CloneFlags::from_bits_retain(namespaces))
        .map_err(anyhow::Error::msg)
        .with_context(|| {
            format!(
                "Failed to create {} namespace!",
                nsflags_name(namespaces).join(", ")
            )
        })?;

    // Write uid/gid map for user namespace.
    if let Some((ref uid_buf, ref gid_buf)) = id_buf {
        // Write "deny" to /proc/self/setgroups before writing to gid_map.
        File::create("/proc/self/setgroups")
            .and_then(|mut f| f.write_all(b"deny"))
            .map_err(anyhow::Error::msg)
            .context("Failed to write to /proc/self/setgroups")?;
        File::create("/proc/self/gid_map")
            .and_then(|mut f| f.write_all(&gid_buf[..]))
            .map_err(anyhow::Error::msg)
            .context("Failed to write to /proc/self/gid_map")?;
        File::create("/proc/self/uid_map")
            .and_then(|mut f| f.write_all(&uid_buf[..]))
            .map_err(anyhow::Error::msg)
            .context("Failed to write to /proc/self/uid_map")?;

        // Set inheritable mask and ambient caps to retain caps after execve(2).
        caps::securebits::set_keepcaps(true)
            .map_err(anyhow::Error::msg)
            .context("Failed to set keep capabilities flag")?;
        let permitted_caps = caps::read(None, caps::CapSet::Permitted)
            .map_err(anyhow::Error::msg)
            .context("Failed to read permitted capabilities")?;
        caps::set(None, caps::CapSet::Inheritable, &permitted_caps)
            .map_err(anyhow::Error::msg)
            .context("Failed to set inheritable capabilities mask")?;
        // Set the same capabilities as ambient, if necessary.
        for cap in permitted_caps {
            caps::raise(None, caps::CapSet::Ambient, cap)
                .map_err(anyhow::Error::msg)
                .with_context(|| format!("Failed to raise capability {cap} to ambient!"))?;
        }
    }

    // Bring up loopback device for net namespace.
    if sandbox.unshare_net() {
        // Set up the loopback interface.
        // Ignore errors.
        match bring_up_loopback() {
            Ok(_) => info!("ctx": "init", "op": "start_loopback"),
            Err(errno) => {
                error!("ctx": "init", "op": "start_loopback", "errno": errno as i32)
            }
        }

        // Handle proxy sandboxing.
        if let Some((stream_child, proxy_port)) = proxy {
            let proxy_addr = IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1));
            match TcpListener::bind((proxy_addr, proxy_port)).map_err(err2no) {
                Ok(listener) => {
                    let buf = [0u8; 1];
                    let lfd = [listener.as_raw_fd()];
                    if let Err(errno) = stream_child.send_with_fd(&buf, &lfd).map_err(err2no) {
                        error!("ctx": "init", "op": "connect_proxy",
                            "errno": errno as i32);
                    }
                    info!("ctx": "init", "op": "connect_proxy");
                }
                Err(errno) => {
                    error!("ctx": "init", "op": "start_proxy",
                        "port": proxy_port, "errno": errno as i32);
                }
            }
            drop(stream_child);
        }
    }

    // Set host and domain name for uts namespace.
    // Ignore errors.
    if sandbox.unshare_uts() {
        // SAFETY: Domain name must not contain a nul-byte.
        #[allow(clippy::disallowed_methods)]
        let domainname = CString::new(sandbox.domainname.clone()).unwrap();

        // SAFETY: There's no setdomainname wrapper in nix.
        let _ = unsafe { setdomainname(domainname.as_ptr() as *const _, domainname.len()) };
        let _ = sethostname(&sandbox.hostname);
    }

    if sandbox.unshare_time() {
        if let Some(offset) = sandbox.time {
            let mut file = File::create("/proc/self/timens_offsets")
                .map_err(anyhow::Error::msg)
                .context("Failed to open time namespace offsets file!")?;
            let contents = format!("monotonic {offset} 0\nboottime {offset} 0\n");
            file.write_all(contents.as_bytes())
                .map_err(anyhow::Error::msg)
                .context("Failed to write offsets to the timens_offsets file")?;
        } else if let Ok(boottime) = clock_gettime(ClockId::CLOCK_BOOTTIME) {
            // Set uptime to 0 for time namespace. Ignore errors.
            if let Ok(mut file) = File::create("/proc/self/timens_offsets") {
                #[allow(clippy::arithmetic_side_effects)]
                let contents = format!("boottime {} 0\n", -boottime.tv_sec());
                let _ = file.write_all(contents.as_bytes());
            }
        }
    }

    // SAFETY: fork is our safest option here.
    let child = match unsafe { fork() }
        .map_err(anyhow::Error::msg)
        .with_context(|| {
            format!(
                "Failed to clone into new namespaces: {:?}!",
                nsflags_name(namespaces)
            )
        })? {
        ForkResult::Child => {
            let retval =
                match Supervisor::run(&argv0, cmd_arg0, cmd_argv, cmd_envp, sandbox, export) {
                    Ok(retval) => nix::libc::c_int::from(retval),
                    Err(error) => {
                        println!("{error}");
                        127
                    }
                };
            // SAFETY: _exit is safe to call in child.
            unsafe { nix::libc::_exit(retval) };
        }
        ForkResult::Parent { child, .. } => {
            // Ensure we release all unneeded resources.
            drop(cmd_arg0);
            drop(cmd_argv);
            drop(cmd_envp);
            drop(sandbox);
            // SAFETY: nix does not have a close_range wrapper yet.
            let _ = unsafe {
                nix::libc::syscall(
                    nix::libc::SYS_close_range,
                    nix::libc::STDERR_FILENO + 1,
                    nix::libc::c_int::MAX,
                    0,
                )
            };
            child
        }
    };

    // Only parent ends up here and `child` is child pid.
    // SAFETY: Set up a Landlock sandbox to disallow all access.
    let p = vec![];
    let _ = syd::landlock_operation(&p, &p, LANDLOCK_ABI);

    // SAFETY: Set up a seccomp filter which only allows
    // 1. write to standard error.
    // 2. waitid and exit.
    // 3. memory allocation syscalls
    // 4. signal handling syscalls
    let mut filter = ScmpFilterContext::new_filter(ScmpAction::KillProcess)
        .map_err(anyhow::Error::msg)
        .context("Failed to create seccomp filter!")?;
    let allow_call = [
        "exit",
        "exit_group",
        "waitid",
        "brk",
        "madvise",
        "mremap",
        "munmap",
        "sigaction",
        "sigaltstack",
        "sigpending",
        "sigprocmask",
        "sigsuspend",
        "sigreturn",
        "rt_sigaction",
        "rt_sigpending",
        "rt_sigprocmask",
        "rt_sigqueueinfo",
        "rt_sigreturn",
        "rt_sigtimedwait",
        "rt_sigtimedwait_time64",
    ];
    for call in allow_call {
        filter
            .add_rule(ScmpAction::Allow, ScmpSyscall::new(call))
            .map_err(anyhow::Error::msg)?;
    }

    // Allow write to standard error.
    for name in ["write", "writev"] {
        filter
            .add_rule_conditional(
                ScmpAction::Allow,
                ScmpSyscall::new(name),
                &[scmp_cmp!($arg0 == nix::libc::STDERR_FILENO as u64)],
            )
            .map_err(anyhow::Error::msg)
            .with_context(|| format!("Failed to add seccomp filter rule for syscall {name}"))?;
    }

    // Prevent executable memory.
    const PROT_EXEC: u64 = nix::libc::PROT_EXEC as u64;
    for name in ["mmap", "mmap2", "mprotect"] {
        filter
            .add_rule_conditional(
                ScmpAction::Allow,
                ScmpSyscall::new(name),
                &[scmp_cmp!($arg2 & PROT_EXEC == 0)],
            )
            .map_err(anyhow::Error::msg)
            .with_context(|| format!("Failed to add seccomp filter rule for syscall {name}"))?;
    }

    filter
        .load()
        .map_err(anyhow::Error::msg)
        .context("Failed to load seccomp filter!")?;

    loop {
        #[allow(clippy::cast_possible_truncation)]
        #[allow(clippy::cast_sign_loss)]
        break match waitid(Id::Pid(child), WaitPidFlag::WEXITED) {
            Ok(WaitStatus::Exited(_, code)) =>
            {
                #[allow(clippy::cast_possible_truncation)]
                #[allow(clippy::cast_sign_loss)]
                Ok(ExitCode::from(code as u8))
            }
            Ok(WaitStatus::Signaled(_, signal, _)) => {
                Ok(ExitCode::from(128_u8.saturating_add(signal as u8)))
            }
            Ok(WaitStatus::StillAlive) | Err(Errno::EINTR) => continue,
            Ok(status) => Err(anyhow!("Unexpected wait status: {status:?}")),
            Err(error) => Err(anyhow!("Failed to wait: {error}")),
        };
    }
}

fn help() {
    println!(
        "syd [-acefhlmpxEPV] [--] {{command [arg...]}}
syd --sh
syd-cat profile-name|list
syd-chk
syd-env pid [-i] [name=value]... {{command [arg...]}}
syd-err number|name-regex
syd-exec {{command [arg...]}}
syd-ldd
syd-lock
syd-log
syd-ls set
syd-mem [-HV] [pid]
syd-norm path
syd-read path
syd-run pid {{command [arg...]}}
syd-size size|human-size
syd-stat [pid]
syd-sys [-a list|native|x86|x86_64|aarch64...] number|name-regex
syd-test [<name-regex>|<number>|<number>..<number>]..
syd-tty [pid]"
    );
}
