#![allow(clippy::undocumented_unsafe_blocks)]

use std::{
    io::IoSlice,
    mem,
    os::fd::{AsFd, AsRawFd},
    ptr,
};

use libloading::os::unix::Symbol;
use nix::{
    errno::Errno,
    fcntl::{splice, vmsplice, SpliceFFlags},
    libc,
    sys::signal::{kill, signal, SigHandler, Signal},
    unistd::{close, getpid},
};

use crate::{
    caps,
    unshare::{
        error::ErrorCode as Err,
        run::{ChildInfo, Exe},
    },
};

unsafe fn fail_errno(code: Err, errno: i32) -> ! {
    let msg = match code {
        Err::CapSet => b"syd: capset error\0".as_ptr(),
        Err::Exec => b"syd: exec error\0".as_ptr(),
        Err::ParentDeathSignal => b"syd: parent-death-signal error\0".as_ptr(),
        Err::PreExec => b"syd: pre-exec error\0".as_ptr(),
        Err::ProcessStop => b"syd: error stopping process\0".as_ptr(),
        Err::Seccomp => b"syd: seccomp error\0".as_ptr(),
        Err::SeccompSendFd => b"syd: seccomp send notify-fd error\0".as_ptr(),
        Err::SeccompWaitFd => b"syd: seccomp wait for notify-fd error\0".as_ptr(),
        Err::SetTSC => b"syd: set-tsc error\0".as_ptr(),
        Err::SigPipe => b"syd: signal error\0".as_ptr(),
    };
    errno::set_errno(errno::Errno(errno));
    libc::perror(msg as *const libc::c_char);
    libc::_exit(errno);
}

macro_rules! fail_safe {
    ($child:expr, $error:expr) => {
        let errno = Errno::last_raw();
        drop($child);
        unsafe { fail_errno($error, errno) };
    };
}

macro_rules! fail_errno_safe {
    ($child:expr, $error:expr, $errno:expr) => {
        drop($child);
        unsafe { fail_errno($error, $errno) };
    };
}

#[allow(clippy::cognitive_complexity)]
pub extern "C" fn child_after_clone(arg: *mut libc::c_void) -> libc::c_int {
    // SAFETY: arg is a valid ChildInfo structure.
    let child: Box<ChildInfo> = unsafe { Box::from_raw(arg as *mut ChildInfo) };

    if let Some(&sig) = child.cfg.death_sig.as_ref() {
        if unsafe { libc::prctl(libc::PR_SET_PDEATHSIG, sig as libc::c_ulong, 0, 0, 0) } != 0 {
            fail_safe!(child, Err::ParentDeathSignal);
        }
    }

    if child.cfg.restore_sigmask {
        unsafe {
            let mut sigmask: libc::sigset_t = mem::zeroed();
            libc::sigemptyset(&mut sigmask);
            libc::pthread_sigmask(libc::SIG_SETMASK, &sigmask, ptr::null_mut());
            for sig in 1..32 {
                libc::signal(sig, libc::SIG_DFL);
            }
        }
    }

    if let Some(callback) = &child.pre_exec {
        if let Err(e) = callback() {
            fail_errno_safe!(child, Err::PreExec, e as i32);
        }
    }

    #[cfg(target_arch = "x86_64")]
    if child.cfg.deny_tsc && unsafe { libc::prctl(libc::PR_SET_TSC, libc::PR_TSC_SIGSEGV) } != 0 {
        fail_safe!(child, Err::SetTSC);
    }

    let seccomp_fd = if let Some(seccomp_filter) = &child.seccomp_filter {
        // Close the write end of the first pipe
        // Close the read end of the second pipe
        let _ = close(child.seccomp_pipefd.0 .1.as_raw_fd());
        let _ = close(child.seccomp_pipefd.1 .0.as_raw_fd());
        let pipe_ro = &child.seccomp_pipefd.0 .0;
        let pipe_rw = &child.seccomp_pipefd.1 .1;

        // Load the seccomp filter.
        if seccomp_filter.load().is_err() {
            fail_safe!(child, Err::Seccomp);
        }

        // Get seccomp notification fd.
        let fd = match seccomp_filter.get_notify_fd() {
            Ok(fd) => fd,
            Err(_) => {
                fail_safe!(child, Err::Seccomp);
            }
        };

        // SAFETY: Avoid SIGPIPE before I/O to pipes.
        if let Err(errno) = unsafe { signal(Signal::SIGPIPE, SigHandler::SigIgn) } {
            // SAFETY: close(2) may hang here, so we don't drop child.
            unsafe { fail_errno(Err::SigPipe, errno as i32) };
        }

        // Write the value of the fd to the pipe.
        // SAFETY: Avoid read,write,close until tracing is initiated.
        let fd_bytes = fd.to_le_bytes();
        let iov = [IoSlice::new(&fd_bytes)];
        if let Err(errno) = vmsplice(pipe_rw.as_fd(), &iov, SpliceFFlags::SPLICE_F_GIFT) {
            // SAFETY: close(2) may hang here, so we don't drop child.
            unsafe { fail_errno(Err::SeccompSendFd, errno as i32) };
        }

        // Wait for the parent to get the file descriptor.
        // Splice data from pipe_ro to pipe_rw to signal readiness.
        // Careful, the pipe may close mid-way here, which is fine.
        match splice(
            pipe_ro.as_fd(),
            None,
            pipe_rw.as_fd(),
            None,
            1,
            SpliceFFlags::empty(),
        ) {
            Ok(_) | Err(Errno::EPIPE) => {}
            Err(errno) => {
                // SAFETY: close(2) may hang here, so we don't drop child.
                unsafe { fail_errno(Err::SeccompWaitFd, errno as i32) };
            }
        }

        Some(fd)
    } else {
        None
    };

    // Drop CAP_SYS_PTRACE late as syd may need it.
    if !child.cfg.keep {
        for capset in [
            caps::CapSet::Effective,
            caps::CapSet::Ambient,
            caps::CapSet::Inheritable,
            caps::CapSet::Permitted,
        ] {
            if caps::drop(None, capset, caps::Capability::CAP_SYS_PTRACE).is_err() {
                fail_safe!(child, Err::CapSet);
            }
        }
    }
    if caps::securebits::set_keepcaps(true).is_err() {
        fail_safe!(child, Err::CapSet);
    }

    if child.cfg.stop {
        // Stop the process to give the parent a chance to seize us and set ptrace options.
        if let Err(e) = kill(getpid(), Signal::SIGSTOP) {
            fail_errno_safe!(child, Err::ProcessStop, e as i32);
        }
    }

    // Tracing has been successfully initiated.
    // We can go ahead and close(2) the file descriptors
    // without having to worry whether close(2) is hooked.
    if let Some(seccomp_fd) = seccomp_fd {
        let _ = close(seccomp_fd);
        let _ = close(child.seccomp_pipefd.0 .0.as_raw_fd());
        let _ = close(child.seccomp_pipefd.1 .1.as_raw_fd());

        // SAFETY: Set SIGPIPE to its default handler, now that we're done with pipes.
        if let Err(errno) = unsafe { signal(Signal::SIGPIPE, SigHandler::SigDfl) } {
            fail_errno_safe!(child, Err::SigPipe, errno as i32);
        }
    }

    match child.exe {
        Exe::Library(lib) => unsafe {
            let fun: Symbol<unsafe extern "C" fn() -> i32> = match lib.get(b"syd_main") {
                Ok(fun) => fun,
                Err(err) => {
                    eprintln!("syd: {err}"); // FIXME: This is unsafe!
                    nix::libc::_exit(128);
                }
            };
            nix::libc::_exit(fun());
        },
        Exe::Program((filename, ref args)) => {
            let args = &args[..];
            unsafe { libc::execvp(filename, args.as_ptr()) };
            fail_safe!(child, Err::Exec);
        }
    }
}
