//
// Syd: rock-solid unikernel
// src/sandbox.rs: Sandbox configuration
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon HardenedBSD's sys/hardenedbsd/hbsd_pax_segvguard.c which is:
//   Copyright (c) 2006 Elad Efrat <elad@NetBSD.org>
//   Copyright (c) 2013-2017, by Oliver Pinter <oliver.pinter@hardenedbsd.org>
//   Copyright (c) 2014, by Shawn Webb <shawn.webb@hardenedbsd.org>
//   Copyright (c) 2014, by Danilo Egea Gondolfo <danilo at FreeBSD.org>
//   All rights reserved.
//   SPDX-License-Identifier: BSD-3-Clause
//
// SPDX-License-Identifier: GPL-3.0-or-later

use std::{
    clone::Clone,
    collections::{HashMap, HashSet},
    env,
    ffi::OsStr,
    fmt,
    fs::File,
    io::{self, BufRead, BufReader, Read},
    net::{IpAddr, Ipv4Addr},
    ops::{Deref, DerefMut},
    os::{
        fd::{AsRawFd, RawFd},
        unix::{ffi::OsStrExt, process::CommandExt},
    },
    path::{Path, PathBuf},
    process::{Command, Stdio},
    str::FromStr,
    time::Duration,
};

use anyhow::{bail, Context};
use bitflags::bitflags;
use globset::{Glob, GlobBuilder, GlobSet, GlobSetBuilder};
use hashbrown::hash_map::DefaultHashBuilder;
use ipnetwork::IpNetwork;
use lexis::ToName;
use libseccomp::{ScmpAction, ScmpFilterContext, ScmpNotifReq, ScmpSyscall};
use nix::{
    errno::Errno,
    fcntl::OFlag,
    mount::MsFlags,
    sys::stat::{fstat, Mode},
    unistd::{getgroups, getpid, Gid, Pid, Uid},
};
#[cfg(feature = "oci")]
use oci_spec::runtime::Spec;
use once_cell::sync::Lazy;
use parking_lot::{RwLockReadGuard, RwLockUpgradableReadGuard, RwLockWriteGuard};
use regex::{Captures, Regex, RegexBuilder};
use serde::{
    ser::{SerializeMap, SerializeStruct},
    Serialize, Serializer,
};
use smallvec::SmallVec;

use crate::{
    config::*,
    error, extend_ioctl,
    fs::{
        err2no, mask_path, path_exists, path_has_control_chars, safe_open_file, safe_open_path,
        FileInformation,
    },
    hash::{aes_ctr_setup, HashAlgorithm, Key, Secret},
    hook::{RemoteProcess, UNotifyEventRequest},
    info, parse_group, parse_user, seccomp2errno, warn, SydMsFlags, CLONE_NEWTIME,
    NAMESPACE_FLAGS_ALL,
};

const LINE_MAX: usize = 4096;

static RE_BIND: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        bind
        (?P<mod>[-+^])
        (?P<src>[^:]+)
        :
        (?P<dst>[^:]+)
        (?::
            (?P<opt>([^,]+)(,[^,]+)*)
        )?
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid bind mount regex, please file a bug!")
});

static RE_FORCE: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        force
        (?P<mod>[-+^])
        (?P<src>/[^:]*)?
        (?:
            :
            (?P<key>([0-9a-fA-F]{32}|[0-9a-fA-F]{40}|[0-9a-fA-F]{64}|[0-9a-fA-F]{96}|[0-9a-fA-F]{128}))
            :
            (?P<act>warn|stop|kill)
        )?
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid integrity force regex, please file a bug!")
});

static RE_SETID_0: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        set(?P<id>[ug])id
        (?P<mod>[-+])
        (?P<src>[^:]+)
        :
        (?P<dst>[^:]+)
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid integrity setid-0 regex, please file a bug!")
});

static RE_SETID_1: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        set(?P<id>[ug])id
        (?P<mod>\^)
        (?P<src>[^:]+)?
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid integrity setid-1 regex, please file a bug!")
});

static RE_RULE: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        (
            # We have an action with a capability.
            (?P<act>
                allow |
                deny |
                filter |
                warn |
                stop |
                kill |
                exit
            )/
            (
                # Match combinations of read, write, exec, stat, and ioctl
                (?P<cap_many>
                    (
                        read |
                        write |
                        exec |
                        stat |
                        ioctl
                    )
                    (,
                        (
                            read |
                            write |
                            exec |
                            stat |
                            ioctl
                        )
                    )*
                ) |
                # Other capabilities, not allowing combinations
                (?P<cap_single>
                    lock/read |
                    lock/write |
                    net/bind |
                    net/connect
                )
            )
        )
        (?P<mod>[-+:^])
        (
            (?P<addr>[A-Fa-f0-9:\.]+(/[0-9]+)?[!@][0-9]+(-[0-9]+)?) |
            (?P<path>.+)
        )
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid sandbox rule regex, please file a bug!")
});

static RE_NETALIAS: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        (?P<command>
            (
                allow |
                deny |
                filter |
                warn |
                stop |
                kill
            )
            /net/
            (
                bind |
                connect
            )
            [-+^]
            # SAFETY: Every item in the regex group `alias' below,
            # must have a corresponding item in the MAP_NETALIAS hash map!
            (?P<alias>
                ([aA][nN][yY][46]?) |
                ([lL][oO][cC][aA][lL][46]?) |
                ([lL][oO][oO][pP][bB][aA][cC][kK][46]?) |
                ([lL][iI][nN][kK][lL][oO][cC][aA][lL][46]?)
            )
            [!@]
            [0-9]+
            (-[0-9]+)?
        )
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid network alias regex, please file a bug!")
});

type AliasMap<'a> = HashMap<&'a str, Vec<&'a str>>;
static MAP_NETALIAS: Lazy<AliasMap> = Lazy::new(|| {
    let mut map = HashMap::new();
    map.insert("any4", vec!["0.0.0.0/0"]);
    map.insert("any6", vec!["::/0"]);
    map.insert("any", vec!["0.0.0.0/0", "::/0"]);
    map.insert("linklocal4", vec!["fe80::/10"]);
    map.insert("linklocal6", vec!["fe80::/10"]);
    map.insert("linklocal", vec!["169.254.0.0/16", "fe80::/10"]);
    map.insert(
        "local4",
        vec![
            "127.0.0.0/8",
            "10.0.0.0/8",
            "172.16.0.0/12",
            "192.168.0.0/16",
        ],
    );
    map.insert("local6", vec!["::1", "fe80::/7", "fc00::/7", "fec0::/7"]);
    map.insert(
        "local",
        vec![
            "127.0.0.0/8",
            "10.0.0.0/8",
            "172.16.0.0/12",
            "192.168.0.0/16",
            "::1/128",
            "fe80::/7",
            "fc00::/7",
            "fec0::/7",
        ],
    );
    map.insert("loopback4", vec!["127.0.0.0/8"]);
    map.insert("loopback6", vec!["::1/128"]);
    map.insert("loopback", vec!["127.0.0.0/8", "::1/128"]);

    map
});

static RE_RULE_NETLINK: Lazy<Regex> = Lazy::new(|| {
    #[allow(clippy::disallowed_methods)]
    RegexBuilder::new(
        r"
        \A
        allow/net/link
        (
            (?P<clr>^) |
            (
                ((?P<add>[+])|(?P<del>[-]))
                (?P<fml>
                    (
                        route |
                        usersock |
                        firewall |
                        inet_diag |
                        sock_diag |
                        nflog |
                        xfrm |
                        selinux |
                        iscsi |
                        audit |
                        fib_lookup |
                        connector |
                        netfilter |
                        ip6_fw |
                        dnrtmsg |
                        kobject_uevent |
                        generic |
                        scsitransport |
                        ecryptfs |
                        rdma |
                        crypto
                    )
                    (,
                        (
                            route |
                            usersock |
                            firewall |
                            inet_diag |
                            sock_diag |
                            nflog |
                            xfrm |
                            selinux |
                            iscsi |
                            audit |
                            fib_lookup |
                            connector |
                            netfilter |
                            ip6_fw |
                            dnrtmsg |
                            kobject_uevent |
                            generic |
                            scsitransport |
                            ecryptfs |
                            rdma |
                            crypto
                        )
                    )*
                )
            )
        )
        \z
    ",
    )
    .ignore_whitespace(true)
    .build()
    .expect("Invalid sandbox rule regex, please file a bug!")
});

const ROOT_UID: Uid = Uid::from_raw(0);
const ROOT_GID: Gid = Gid::from_raw(0);

#[inline]
fn strbool(s: &str) -> Result<bool, Errno> {
    match s.to_ascii_lowercase().as_str() {
        "1" | "on" | "t" | "tr" | "tru" | "true" | "✓" => Ok(true),
        "0" | "off" | "f" | "fa" | "fal" | "fals" | "false" | "✗" => Ok(false),
        "" => Err(Errno::ENOENT),
        _ => Err(Errno::EINVAL),
    }
}

type LandlockPathPair = (Vec<String>, Vec<String>);

bitflags! {
    /// Represents allowlisted netlink families.
    #[derive(Clone, Copy, Debug, Eq, PartialEq)]
    pub(crate) struct NetlinkFamily: u32 {
        const NETLINK_ROUTE = 1 << 0;
        const NETLINK_USERSOCK = 1 << 1;
        const NETLINK_FIREWALL = 1 << 2;
        // sock_diag = inet_diag
        const NETLINK_SOCK_DIAG = 1 << 3;
        const NETLINK_NFLOG = 1 << 4;
        const NETLINK_XFRM = 1 << 5;
        const NETLINK_SELINUX = 1 << 6;
        const NETLINK_ISCSI = 1 << 7;
        const NETLINK_AUDIT = 1 << 8;
        const NETLINK_FIB_LOOKUP = 1 << 9;
        const NETLINK_CONNECTOR = 1 << 10;
        const NETLINK_NETFILTER = 1 << 11;
        const NETLINK_IP6_FW = 1 << 12;
        const NETLINK_DNRTMSG = 1 << 13;
        const NETLINK_KOBJECT_UEVENT = 1 << 14;
        const NETLINK_GENERIC = 1 << 15;
        const NETLINK_SCSITRANSPORT = 1 << 16;
        const NETLINK_ECRYPTFS = 1 << 17;
        const NETLINK_RDMA = 1 << 18;
        const NETLINK_CRYPTO = 1 << 19;
    }
}

impl FromStr for NetlinkFamily {
    type Err = Errno;

    fn from_str(value: &str) -> Result<Self, Self::Err> {
        let mut families = NetlinkFamily::empty();
        for family in value.split(',') {
            // Let's be as lax as possible but not more.
            let family = family.trim();
            if family.is_empty() {
                continue;
            }
            families |= match family.to_ascii_lowercase().as_str() {
                "route" => Self::NETLINK_ROUTE,
                "usersock" => Self::NETLINK_USERSOCK,
                "firewall" => Self::NETLINK_FIREWALL,
                "inet_diag" | "sock_diag" => Self::NETLINK_SOCK_DIAG,
                "nflog" => Self::NETLINK_NFLOG,
                "xfrm" => Self::NETLINK_XFRM,
                "selinux" => Self::NETLINK_SELINUX,
                "iscsi" => Self::NETLINK_ISCSI,
                "audit" => Self::NETLINK_AUDIT,
                "fib_lookup" => Self::NETLINK_FIB_LOOKUP,
                "connector" => Self::NETLINK_CONNECTOR,
                "netfilter" => Self::NETLINK_NETFILTER,
                "ip6_fw" => Self::NETLINK_IP6_FW,
                "dnrtmsg" => Self::NETLINK_DNRTMSG,
                "kobject_uevent" => Self::NETLINK_KOBJECT_UEVENT,
                "generic" => Self::NETLINK_GENERIC,
                "scsitransport" => Self::NETLINK_SCSITRANSPORT,
                "ecryptfs" => Self::NETLINK_ECRYPTFS,
                "rdma" => Self::NETLINK_RDMA,
                "crypto" => Self::NETLINK_CRYPTO,
                _ => return Err(Errno::EINVAL),
            };
        }

        if !families.is_empty() {
            Ok(families)
        } else {
            Err(Errno::ENOENT)
        }
    }
}

impl NetlinkFamily {
    const ALL: [Self; 20] = [
        Self::NETLINK_ROUTE,
        Self::NETLINK_USERSOCK,
        Self::NETLINK_FIREWALL,
        Self::NETLINK_SOCK_DIAG,
        Self::NETLINK_NFLOG,
        Self::NETLINK_XFRM,
        Self::NETLINK_SELINUX,
        Self::NETLINK_ISCSI,
        Self::NETLINK_AUDIT,
        Self::NETLINK_FIB_LOOKUP,
        Self::NETLINK_CONNECTOR,
        Self::NETLINK_NETFILTER,
        Self::NETLINK_IP6_FW,
        Self::NETLINK_DNRTMSG,
        Self::NETLINK_KOBJECT_UEVENT,
        Self::NETLINK_GENERIC,
        Self::NETLINK_SCSITRANSPORT,
        Self::NETLINK_ECRYPTFS,
        Self::NETLINK_RDMA,
        Self::NETLINK_CRYPTO,
    ];

    pub(crate) fn max() -> nix::libc::c_int {
        nix::libc::NETLINK_CRYPTO.saturating_add(1)
    }

    pub(crate) fn to_vec(self) -> Vec<nix::libc::c_int> {
        let mut vec = Vec::new();

        if self.is_empty() {
            // Do nothing.
        } else {
            for netlink_family in Self::ALL {
                if self.contains(netlink_family) {
                    vec.push(netlink_family.to_raw_int());
                }
            }
        }

        vec
    }

    fn to_raw_int(self) -> nix::libc::c_int {
        match self {
            Self::NETLINK_ROUTE => nix::libc::NETLINK_ROUTE,
            Self::NETLINK_USERSOCK => nix::libc::NETLINK_USERSOCK,
            Self::NETLINK_FIREWALL => nix::libc::NETLINK_FIREWALL,
            Self::NETLINK_SOCK_DIAG => nix::libc::NETLINK_SOCK_DIAG,
            Self::NETLINK_NFLOG => nix::libc::NETLINK_NFLOG,
            Self::NETLINK_XFRM => nix::libc::NETLINK_XFRM,
            Self::NETLINK_SELINUX => nix::libc::NETLINK_SELINUX,
            Self::NETLINK_ISCSI => nix::libc::NETLINK_ISCSI,
            Self::NETLINK_AUDIT => nix::libc::NETLINK_AUDIT,
            Self::NETLINK_FIB_LOOKUP => nix::libc::NETLINK_FIB_LOOKUP,
            Self::NETLINK_CONNECTOR => nix::libc::NETLINK_CONNECTOR,
            Self::NETLINK_NETFILTER => nix::libc::NETLINK_NETFILTER,
            Self::NETLINK_IP6_FW => nix::libc::NETLINK_IP6_FW,
            Self::NETLINK_DNRTMSG => nix::libc::NETLINK_DNRTMSG,
            Self::NETLINK_KOBJECT_UEVENT => nix::libc::NETLINK_KOBJECT_UEVENT,
            Self::NETLINK_GENERIC => nix::libc::NETLINK_GENERIC,
            Self::NETLINK_SCSITRANSPORT => nix::libc::NETLINK_SCSITRANSPORT,
            Self::NETLINK_ECRYPTFS => nix::libc::NETLINK_ECRYPTFS,
            Self::NETLINK_RDMA => nix::libc::NETLINK_RDMA,
            Self::NETLINK_CRYPTO => nix::libc::NETLINK_CRYPTO,
            _ => unreachable!(),
        }
    }
}

/// Represents an integrity error, used by Force Sandboxing.
#[derive(Debug)]
pub enum IntegrityError {
    /// System error, represented by an Errno.
    Sys(Errno),
    /// Hash mismatch.
    Hash {
        /// The action to take on mismatch.
        action: Action,
        /// Expected hash as hex-encoded string.
        expected: String,
        /// Found hash as hex-encoded string.
        found: String,
        /// Offending path.
        path: PathBuf,
    },
}

impl From<Errno> for IntegrityError {
    fn from(errno: Errno) -> Self {
        IntegrityError::Sys(errno)
    }
}

/// Represents a recursive bind mount operation.
#[derive(Debug)]
pub struct BindMount {
    /// Source directory
    pub src: PathBuf,
    /// Target directory, can be the same as source
    pub dst: PathBuf,
    /// The options that are allowed are:
    /// ro, nosuid, nodev, noexec, noatime, nodiratime and relatime
    /// kernel is going to ignore other options.
    pub opt: MsFlags,
    /// Optional filesystem specific data, useful for tmpfs.
    pub dat: Option<PathBuf>,
}

impl PartialEq for BindMount {
    // Flags are not used in equality check.
    fn eq(&self, other: &Self) -> bool {
        self.src == other.src && self.dst == other.dst
    }
}

// Eq can be derived as well since PartialEq is implemented
impl Eq for BindMount {}

impl TryFrom<&Captures<'_>> for BindMount {
    type Error = Errno;

    fn try_from(captures: &Captures) -> Result<Self, Self::Error> {
        let src = PathBuf::from(&captures["src"]);
        let dst = PathBuf::from(&captures["dst"]);
        let mut opt = MsFlags::empty();
        let mut dat = Vec::new();

        if let Some(opt_match) = captures.name("opt") {
            for flag in opt_match.as_str().split(',') {
                match SydMsFlags::from_name(flag) {
                    Some(flag) => opt |= flag.0,
                    None => {
                        if !dat.is_empty() {
                            // Separate unrecognized flags with a comma
                            dat.push(b',');
                        }
                        dat.extend_from_slice(flag.as_bytes());
                    }
                }
            }
        }

        let dat = if dat.is_empty() {
            None
        } else {
            Some(PathBuf::from(OsStr::from_bytes(&dat)))
        };

        Ok(Self { src, dst, opt, dat })
    }
}

impl Serialize for BindMount {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut map = serializer.serialize_map(Some(4))?; // We expect 4 fields.

        // Serialize src field
        map.serialize_entry("src", &self.src.to_string_lossy())?;

        // Serialize dst field
        map.serialize_entry("dst", &self.dst.to_string_lossy())?;

        // Serialize opt field using to_names from SydMsFlags
        let syd_flags = SydMsFlags(self.opt);
        let flags_names = syd_flags.to_names();
        map.serialize_entry("opt", &flags_names)?;

        // Serialize dat field
        match &self.dat {
            Some(dat) => map.serialize_entry("dat", &dat.to_string_lossy())?,
            None => map.serialize_entry("dat", &None::<String>)?,
        };

        map.end()
    }
}

bitflags! {
    /// Sandboxing capabilities
    #[derive(Clone, Copy, Debug, Eq, PartialEq)]
    pub struct Capability: u32 {
        /// Read capability
        const CAP_READ = 1 << 0;
        /// List capability
        const CAP_STAT = 1 << 1;
        /// Write capability
        const CAP_WRITE = 1 << 2;
        /// Execute capability
        const CAP_EXEC = 1 << 3;
        /// Ioctl capability
        const CAP_IOCTL = 1 << 4;
        /// Network connect capability
        const CAP_CONNECT = 1 << 5;
        /// Network bind capability
        const CAP_BIND = 1 << 6;
        /// Memory capability
        const CAP_MEM = 1 << 7;
        /// Pid capability
        const CAP_PID = 1 << 8;
        /// Crypt capability
        const CAP_CRYPT = 1 << 9;
        /// Force capability
        const CAP_FORCE = 1 << 10;
        /// Trusted Path Execution (TPE) capability
        const CAP_TPE = 1 << 11;
        /// Proxy sandboxing
        const CAP_PROXY = 1 << 12;
        /// Landlock read capability
        const CAP_LOCK_RO = 1 << 13;
        /// Landlock read-write capability
        const CAP_LOCK_RW = 1 << 14;
        /// Landlock capability
        const CAP_LOCK = Self::CAP_LOCK_RO.bits() | Self::CAP_LOCK_RW.bits();
    }
}

impl Capability {
    /// Define a static array containing all capabilities with path/glob rules
    const GLOB: [Self; 9] = [
        Self::CAP_READ,
        Self::CAP_STAT,
        Self::CAP_WRITE,
        Self::CAP_EXEC,
        Self::CAP_IOCTL,
        Self::CAP_CONNECT,
        Self::CAP_BIND,
        Self::CAP_LOCK_RO,
        Self::CAP_LOCK_RW,
    ];

    /// Returns true if the Capability relates to filesystem read.
    pub fn can_read(self) -> bool {
        self == Self::CAP_READ
    }

    /// Returns true if the Capability relates to filesystem stat.
    pub fn can_stat(self) -> bool {
        self == Self::CAP_STAT
    }

    /// Returns true if the Capability relates to filesystem write.
    pub fn can_write(self) -> bool {
        // bind is for UNIX sockets whose creation is filesystem write.
        matches!(self, Self::CAP_WRITE | Self::CAP_BIND)
    }

    /// Returns true if the Capability relates to networking.
    pub fn can_net(self) -> bool {
        matches!(self, Self::CAP_BIND | Self::CAP_CONNECT)
    }

    /// Returns true if the Capability relates to the Landlock LSM.
    pub fn can_lock(self) -> bool {
        matches!(self, Self::CAP_LOCK_RO | Self::CAP_LOCK_RW)
    }

    /// Returns true if the Capability must be set at startup.
    pub fn is_startup(self) -> bool {
        self.intersects(Self::CAP_CRYPT | Self::CAP_PROXY | Self::CAP_LOCK)
    }
}

impl FromStr for Capability {
    type Err = Errno;

    fn from_str(value: &str) -> Result<Self, Self::Err> {
        let mut caps = Self::empty();
        for cap in value.split(',') {
            // Let's be as lax as possible but not more.
            let cap = cap.trim();
            if cap.is_empty() {
                continue;
            }
            caps |= match cap.to_ascii_lowercase().as_str() {
                "crypt" => Self::CAP_CRYPT,
                "exec" => Self::CAP_EXEC,
                "force" => Self::CAP_FORCE,
                "lock" => Self::CAP_LOCK,
                "mem" => Self::CAP_MEM,
                "pid" => Self::CAP_PID,
                "proxy" => Self::CAP_PROXY,
                "read" => Self::CAP_READ,
                "stat" => Self::CAP_STAT,
                "ioctl" => Self::CAP_IOCTL,
                "tpe" => Self::CAP_TPE,
                "write" => Self::CAP_WRITE,
                "net" => Self::CAP_BIND | Capability::CAP_CONNECT,
                /* Network sandboxing but only for one of Bind, Connect */
                "net/bind" => Self::CAP_BIND,
                "net/connect" => Self::CAP_CONNECT,
                _ => return Err(Errno::EINVAL),
            };
        }

        if !caps.is_empty() {
            Ok(caps)
        } else {
            Err(Errno::ENOENT)
        }
    }
}

#[allow(clippy::cognitive_complexity)]
impl fmt::Display for Capability {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let mut capabilities: SmallVec<[&str; 10]> = SmallVec::new();

        if self.contains(Self::CAP_READ) {
            capabilities.push("Read");
        }
        if self.contains(Self::CAP_STAT) {
            capabilities.push("Stat");
        }
        if self.contains(Self::CAP_WRITE) {
            capabilities.push("Write");
        }
        if self.contains(Self::CAP_EXEC) {
            capabilities.push("Execute");
        }
        if self.contains(Self::CAP_IOCTL) {
            capabilities.push("Ioctl");
        }
        if self.contains(Self::CAP_CRYPT) {
            capabilities.push("Crypt");
        }
        if self.contains(Self::CAP_FORCE) {
            capabilities.push("Force");
        }
        if self.contains(Self::CAP_TPE) {
            capabilities.push("Trusted Path Execution (TPE)");
        }
        if self.contains(Self::CAP_PROXY) {
            capabilities.push("Proxy");
        }
        if self.contains(Self::CAP_CONNECT) {
            capabilities.push("Connect");
        }
        if self.contains(Self::CAP_MEM) {
            capabilities.push("Memory");
        }
        if self.contains(Self::CAP_PID) {
            capabilities.push("Pid");
        }
        if self.contains(Self::CAP_BIND) {
            capabilities.push("Bind");
        }
        if self.contains(Self::CAP_LOCK) {
            capabilities.push("LandLock");
        }

        write!(f, "{}", capabilities.join(", "))
    }
}

impl Serialize for Capability {
    #[allow(clippy::cognitive_complexity)]
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut capabilities = String::new();

        if self.contains(Self::CAP_READ) {
            capabilities.push('r');
        }
        if self.contains(Self::CAP_STAT) {
            capabilities.push('s');
        }
        if self.contains(Self::CAP_WRITE) {
            capabilities.push('w');
        }
        if self.contains(Self::CAP_EXEC) {
            capabilities.push('x');
        }
        if self.contains(Self::CAP_IOCTL) {
            capabilities.push('i');
        }
        if self.contains(Self::CAP_CRYPT) {
            capabilities.push('*');
        }
        if self.contains(Self::CAP_FORCE) {
            capabilities.push('f');
        }
        if self.contains(Self::CAP_TPE) {
            capabilities.push('t');
        }
        if self.contains(Self::CAP_PROXY) {
            capabilities.push('@');
        }
        if self.contains(Self::CAP_BIND) {
            capabilities.push('b');
        }
        if self.contains(Self::CAP_CONNECT) {
            capabilities.push('c');
        }
        if self.contains(Self::CAP_MEM) {
            capabilities.push('m');
        }
        if self.contains(Self::CAP_PID) {
            capabilities.push('p');
        }
        if self.contains(Self::CAP_LOCK) {
            capabilities.push('l');
        }

        if capabilities.is_empty() {
            serializer.serialize_none()
        } else {
            serializer.serialize_str(&capabilities)
        }
    }
}

impl From<&Captures<'_>> for Capability {
    fn from(captures: &Captures) -> Self {
        // Note, we do not use CAP_FORCE here.
        // Force rules are defined elsewhere.
        // We do not use CAP_PROXY here either,
        // it's irrelevant (has no rules).
        // We do not use CAP_MEM, CAP_PID, or
        // CAP_TPE either as they only have a
        // single default Action.
        if let Some(cap) = captures.name("cap_single") {
            match cap.as_str() {
                "lock/read" => Self::CAP_LOCK_RO,
                "lock/write" => Self::CAP_LOCK_RW,
                "net/bind" => Self::CAP_BIND,
                "net/connect" => Self::CAP_CONNECT,
                _ => unreachable!(),
            }
        } else if let Some(caps) = captures.name("cap_many") {
            caps.as_str()
                .split(',')
                .map(|cap| match cap {
                    "read" => Self::CAP_READ,
                    "write" => Self::CAP_WRITE,
                    "exec" => Self::CAP_EXEC,
                    "stat" => Self::CAP_STAT,
                    "ioctl" => Self::CAP_IOCTL,
                    _ => unreachable!(),
                })
                .fold(Self::empty(), |acc, cap| acc | cap)
        } else {
            unreachable!("Invalid rule regex!");
        }
    }
}

impl TryFrom<(&RemoteProcess, &UNotifyEventRequest, &ScmpNotifReq, &str)> for Capability {
    type Error = Errno;

    // Find out capabilities of the system call using the system call name and seccomp request.
    fn try_from(
        value: (&RemoteProcess, &UNotifyEventRequest, &ScmpNotifReq, &str),
    ) -> Result<Self, Errno> {
        let (proc, request, req, syscall_name) = value;
        match syscall_name {
            name if Self::exec(name) => Ok(Self::CAP_EXEC),
            name if Self::stat(name) => Ok(Self::CAP_STAT),
            "ioctl" => Ok(Self::CAP_IOCTL),
            "bind" => Ok(Self::CAP_BIND),
            "connect" | "sendto" | "sendmsg" | "sendmmsg" => Ok(Self::CAP_CONNECT),
            "open" | "openat" | "openat2" => Self::open(syscall_name, req, proc, request),
            _ => Ok(Self::CAP_WRITE),
        }
    }
}

impl Capability {
    fn open(
        syscall_name: &str,
        req: &ScmpNotifReq,
        proc: &RemoteProcess,
        request: &UNotifyEventRequest,
    ) -> Result<Self, Errno> {
        Ok(match syscall_name {
            "open" | "openat" => {
                let flidx = if syscall_name == "open" { 1 } else { 2 };
                #[allow(clippy::cast_possible_truncation)]
                let flags = OFlag::from_bits_truncate(req.data.args[flidx] as nix::libc::c_int);
                match flags & OFlag::O_ACCMODE {
                    OFlag::O_RDONLY => Self::CAP_READ,
                    _ => Self::CAP_WRITE,
                }
            }
            "openat2" => {
                #[allow(clippy::cast_possible_truncation)]
                let rohow = proc.remote_ohow(
                    req.data.args[2] as usize,
                    req.data.args[3] as usize,
                    request,
                )?;
                #[allow(clippy::cast_possible_truncation)]
                let flags = OFlag::from_bits_truncate(rohow.flags as nix::libc::c_int);
                match flags & OFlag::O_ACCMODE {
                    OFlag::O_RDONLY => Self::CAP_READ,
                    _ => Self::CAP_WRITE,
                }
            }
            _ => unreachable!(),
        })
    }

    fn exec(syscall_name: &str) -> bool {
        matches!(syscall_name, "execve" | "execveat")
    }

    fn stat(syscall_name: &str) -> bool {
        matches!(
            syscall_name,
            "access"
                | "faccessat"
                | "faccessat2"
                | "chdir"
                | "fchdir"
                | "getdents"
                | "getdents64"
                | "stat"
                | "statfs"
                | "statfs64"
                | "statx"
                | "fstat"
                | "fstatfs"
                | "fstatfs64"
                | "lstat"
                | "newfstatat"
                | "readlink"
                | "readlinkat"
                | "getxattr"
                | "lgetxattr"
                | "fgetxattr"
                | "listxattr"
                | "flistxattr"
                | "llistxattr"
                | "fanotify_mark"
                | "inotify_add_watch"
        )
    }
}

bitflags! {
    /// Sandboxing options
    #[derive(Clone, Copy, Debug, Eq, PartialEq)]
    pub struct Flag: u64 {
        /// Force O_CLOEXEC flag for open(2) family calls.
        const FL_FORCE_CLOEXEC = 1 << 0;
        /// Deny .. in path resolution for open(2) family calls.
        const FL_DENY_DOTDOT = 1 << 1;
        /// Apply CONTINUE on O_PATH file descriptors rather than
        /// turning them into O_RDONLY.
        const FL_ALLOW_UNSAFE_OPEN = 1 << 2;
        /// Allow unsafe debug (perf, ptrace, etc) calls
        /// Keeps the capability CAP_SYS_PTRACE.
        /// Allows required paths such as /proc/kallsyms, /proc/self/m{aps,mem}.
        const FL_ALLOW_UNSAFE_DEBUG = 1 << 3;
        /// Deny execution of 32-bit ELF binaries.
        const FL_DENY_ELF32 = 1 << 4;
        /// Deny execution of dynamically linked binaries.
        const FL_DENY_ELF_DYNAMIC = 1 << 5;
        /// Deny execution of statically linked binaries.
        const FL_DENY_ELF_STATIC = 1 << 6;
        /// Deny execution of scripts using an interpreter via #!<path>.
        const FL_DENY_SCRIPT = 1 << 7;
        /// 1. Allow non-elf interpreters via binfmt (wine, java etc.)
        /// 2. Allow executing shared libraries directly (ldd, ld-linux.so.1 etc.)
        const FL_ALLOW_UNSUPP_EXEC = 1 << 8;
        /// Allow successful bind calls for subsequent connect calls
        const FL_ALLOW_SAFE_BIND = 1 << 9;
        /// Allow access to the Linux kernel crypto API
        const FL_ALLOW_SAFE_KCAPI = 1 << 10;
        /// Allow socket families which are unsupported
        const FL_ALLOW_UNSUPP_SOCKET = 1 << 11;
        /// Allow filenames with control characters in them.
        const FL_ALLOW_UNSAFE_FILENAME = 1 << 12;
        /// Allow secret memfds and executable memfds.
        const FL_ALLOW_UNSAFE_MEMFD = 1 << 13;
        /// Negate TPE GID logic.
        const FL_TPE_NEGATE = 1 << 14;
        /// Ensure TPE directory is root owned.
        const FL_TPE_ROOT_OWNED = 1 << 15;
        /// Ensure TPE directory is user owned.
        const FL_TPE_USER_OWNED = 1 << 16;
        /// Wait for all processes before exiting.
        const FL_EXIT_WAIT_ALL = 1 << 17;

        /// If set at startup, all id system calls return 0 in the sandbox.
        const FL_FAKE_ROOT = 1 << 28;
        /// If set at startup, root is mapped to current user in the user namespace.
        const FL_MAP_ROOT = 1 << 29;
        /// Lock personality(2) changes.
        const FL_LOCK_PERSONALITY = 1 << 30;
        /// Allow safe setid calls.
        const FL_ALLOW_SAFE_SETID = 1 << 31;
        /// Retain capability CAP_NET_BIND_SERVICE
        const FL_ALLOW_UNSAFE_BIND = 1 << 32;
        /// Allow system calls for CPU emulation functionality
        const FL_ALLOW_UNSAFE_CPU = 1 << 33;
        /// Disable exec-TOCTOU mitigator, allows strace -f syd
        const FL_ALLOW_UNSAFE_EXEC = 1 << 34;
        /// Allow Kernel keyring access.
        const FL_ALLOW_UNSAFE_KEYRING = 1 << 35;
        /// Allow unsafe memory manipulation.
        const FL_ALLOW_UNSAFE_MEMORY = 1 << 36;
        /// Allow system calls used for memory protection keys.
        const FL_ALLOW_UNSAFE_PKEY = 1 << 37;
        /// Retain _all_ Linux capabilities.
        const FL_ALLOW_UNSAFE_CAPS = 1 << 38;
        /// Allow unsafe environment variables.
        const FL_ALLOW_UNSAFE_ENV = 1 << 39;
        /// Allow unsafe socket families (RAW and PACKET).
        const FL_ALLOW_UNSAFE_SOCKET = 1 << 40;
        /// Allow unsafe syslog calls (reading /proc/kmsg etc.)
        const FL_ALLOW_UNSAFE_SYSLOG = 1 << 41;
        /// Allow unsafe prctl calls
        const FL_ALLOW_UNSAFE_PRCTL = 1 << 42;
        /// Allow unsafe prlimit calls
        const FL_ALLOW_UNSAFE_PRLIMIT = 1 << 43;
        /// Allow unsafe adjtimex and clock_adjtime calls,
        /// and keep the CAP_SYS_TIME capability.
        const FL_ALLOW_UNSAFE_TIME = 1 << 44;
        /// Allow the unsafe io-uring interface
        const FL_ALLOW_UNSAFE_IOURING = 1 << 45;
        /// Allow the deprecated tkill system call.
        const FL_ALLOW_UNSAFE_KILL = 1 << 46;
        /// Deny reading the timestamp counter (x86 only)
        const FL_DENY_TSC = 1 << 47;

        /// Allow mount namespace
        const FL_ALLOW_UNSAFE_UNSHARE_MOUNT = 1 << 48;
        /// Allow uts namespace
        const FL_ALLOW_UNSAFE_UNSHARE_UTS = 1 << 49;
        /// Allow ipc namespace
        const FL_ALLOW_UNSAFE_UNSHARE_IPC = 1 << 50;
        /// Allow user namespace
        const FL_ALLOW_UNSAFE_UNSHARE_USER = 1 << 51;
        /// Allow pid namespace
        const FL_ALLOW_UNSAFE_UNSHARE_PID = 1 << 52;
        /// Allow net namespace
        const FL_ALLOW_UNSAFE_UNSHARE_NET = 1 << 53;
        /// Allow cgroup namespace
        const FL_ALLOW_UNSAFE_UNSHARE_CGROUP = 1 << 54;
        /// Allow time namespace
        const FL_ALLOW_UNSAFE_UNSHARE_TIME = 1 << 55;

        /// Unshare mount namespace
        const FL_UNSHARE_MOUNT = 1 << 56;
        /// Unshare uts namespace
        const FL_UNSHARE_UTS = 1 << 57;
        /// Unshare ipc namespace
        const FL_UNSHARE_IPC = 1 << 58;
        /// Unshare user namespace
        const FL_UNSHARE_USER = 1 << 59;
        /// Unshare pid namespace
        const FL_UNSHARE_PID = 1 << 60;
        /// Unshare net namespace
        const FL_UNSHARE_NET = 1 << 61;
        /// Unshare cgroup namespace
        const FL_UNSHARE_CGROUP = 1 << 62;
        /// Unshare time namespace
        const FL_UNSHARE_TIME = 1 << 63;
    }
}

impl fmt::Display for Flag {
    #[allow(clippy::cognitive_complexity)]
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let mut flags: SmallVec<[&str; 64]> = SmallVec::new();

        if self.contains(Flag::FL_UNSHARE_MOUNT) {
            flags.push("Unshare Mount");
        }
        if self.contains(Flag::FL_UNSHARE_UTS) {
            flags.push("Unshare UTS");
        }
        if self.contains(Flag::FL_UNSHARE_IPC) {
            flags.push("Unshare Ipc");
        }
        if self.contains(Flag::FL_UNSHARE_USER) {
            flags.push("Unshare User");
        }
        if self.contains(Flag::FL_UNSHARE_PID) {
            flags.push("Unshare Pid");
        }
        if self.contains(Flag::FL_UNSHARE_NET) {
            flags.push("Unshare Net");
        }
        if self.contains(Flag::FL_UNSHARE_CGROUP) {
            flags.push("Unshare CGroup");
        }
        if self.contains(Flag::FL_UNSHARE_TIME) {
            flags.push("Unshare Time");
        }
        if self.contains(Flag::FL_EXIT_WAIT_ALL) {
            flags.push("Exit Wait All");
        }
        if self.contains(Flag::FL_TPE_NEGATE) {
            flags.push("Negate TPE GID Logic");
        }
        if self.contains(Flag::FL_TPE_ROOT_OWNED) {
            flags.push("Ensure TPE directory is root-owned");
        }
        if self.contains(Flag::FL_TPE_USER_OWNED) {
            flags.push("Ensure TPE directory is user-owned");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_BIND) {
            flags.push("Allow Unsafe Bind");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_CPU) {
            flags.push("Allow Unsafe CPU Emulation");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_EXEC) {
            flags.push("Allow Unsafe Exec");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_KEYRING) {
            flags.push("Allow Unsafe Kernel keyring");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_MEMORY) {
            flags.push("Allow Unsafe Memory");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_PKEY) {
            flags.push("Allow Unsafe Memory Protection Keys");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_CAPS) {
            flags.push("Allow Unsafe Capabilities");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_ENV) {
            flags.push("Allow Unsafe Environment");
        }
        if self.contains(Flag::FL_FAKE_ROOT) {
            flags.push("Fake Root");
        }
        if self.contains(Flag::FL_MAP_ROOT) {
            flags.push("Map Root");
        }
        if self.contains(Flag::FL_ALLOW_SAFE_BIND) {
            flags.push("Allow Safe Bind");
        }
        if self.contains(Flag::FL_ALLOW_SAFE_KCAPI) {
            flags.push("Allow Safe Kernel Crypto API");
        }
        if self.contains(Flag::FL_ALLOW_SAFE_SETID) {
            flags.push("Allow Safe Set Id");
        }
        if self.contains(Flag::FL_ALLOW_UNSUPP_EXEC) {
            flags.push("Allow Unsupported Exec");
        }
        if self.contains(Flag::FL_ALLOW_UNSUPP_SOCKET) {
            flags.push("Allow Unsupported Socket Families");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_SOCKET) {
            flags.push("Allow Unsafe Socket Families");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_SYSLOG) {
            flags.push("Allow Unsafe Syslog");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_PRCTL) {
            flags.push("Allow Unsafe PRctl");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_PRLIMIT) {
            flags.push("Allow Unsafe PRlimit");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_TIME) {
            flags.push("Allow Unsafe Time");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_DEBUG) {
            flags.push("Allow Unsafe Debug");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_OPEN) {
            flags.push("Allow Unsafe Open");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_FILENAME) {
            flags.push("Allow Unsafe Filename");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_IOURING) {
            flags.push("Allow Unsafe IO_Uring");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_KILL) {
            flags.push("Allow Unsafe Kill");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_MEMFD) {
            flags.push("Allow Unsafe MemFD");
        }
        if self.contains(Flag::FL_FORCE_CLOEXEC) {
            flags.push("Force Close on Exec");
        }
        if self.contains(Flag::FL_DENY_DOTDOT) {
            flags.push("Deny DotDot");
        }
        if self.contains(Flag::FL_DENY_ELF32) {
            flags.push("Deny Elf32");
        }
        if self.contains(Flag::FL_DENY_ELF_DYNAMIC) {
            flags.push("Deny Dynamically Linked Elf");
        }
        if self.contains(Flag::FL_DENY_ELF_STATIC) {
            flags.push("Deny Statically Linked Elf");
        }
        if self.contains(Flag::FL_DENY_SCRIPT) {
            flags.push("Deny Script");
        }
        if self.contains(Flag::FL_DENY_TSC) {
            flags.push("Deny TSC");
        }
        if self.contains(Flag::FL_LOCK_PERSONALITY) {
            flags.push("Lock Personality");
        }

        flags.sort();
        write!(f, "{}", flags.join(", "))
    }
}

impl Serialize for Flag {
    #[allow(clippy::cognitive_complexity)]
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut flags: SmallVec<[&str; 64]> = SmallVec::new();
        if self.is_empty() {
            return serializer.collect_seq(flags);
        }
        if self.contains(Flag::FL_UNSHARE_MOUNT) {
            flags.push("unshare-mount");
        }
        if self.contains(Flag::FL_UNSHARE_UTS) {
            flags.push("unshare-uts");
        }
        if self.contains(Flag::FL_UNSHARE_IPC) {
            flags.push("unshare-ipc");
        }
        if self.contains(Flag::FL_UNSHARE_USER) {
            flags.push("unshare-user");
        }
        if self.contains(Flag::FL_UNSHARE_PID) {
            flags.push("unshare-pid");
        }
        if self.contains(Flag::FL_UNSHARE_NET) {
            flags.push("unshare-net");
        }
        if self.contains(Flag::FL_UNSHARE_TIME) {
            flags.push("unshare-time");
        }
        if self.contains(Flag::FL_FAKE_ROOT) {
            flags.push("fake-root");
        }
        if self.contains(Flag::FL_MAP_ROOT) {
            flags.push("map-root");
        }
        if self.contains(Flag::FL_EXIT_WAIT_ALL) {
            flags.push("exit-wait-all");
        }
        if self.contains(Flag::FL_TPE_NEGATE) {
            flags.push("tpe-negate");
        }
        if self.contains(Flag::FL_TPE_ROOT_OWNED) {
            flags.push("tpe-root-owned");
        }
        if self.contains(Flag::FL_TPE_USER_OWNED) {
            flags.push("tpe-user-owned");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_BIND) {
            flags.push("allow-unsafe-bind");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_CPU) {
            flags.push("allow-unsafe-cpu");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_EXEC) {
            flags.push("allow-unsafe-exec");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_KEYRING) {
            flags.push("allow-unsafe-keyring");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_MEMORY) {
            flags.push("allow-unsafe-memory");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_PKEY) {
            flags.push("allow-unsafe-pkey");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_CAPS) {
            flags.push("allow-unsafe-caps");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_ENV) {
            flags.push("allow-unsafe-env");
        }
        if self.contains(Flag::FL_ALLOW_SAFE_BIND) {
            flags.push("allow-safe-bind");
        }
        if self.contains(Flag::FL_ALLOW_SAFE_KCAPI) {
            flags.push("allow-safe-kcapi");
        }
        if self.contains(Flag::FL_ALLOW_SAFE_SETID) {
            flags.push("allow-safe-setid");
        }
        if self.contains(Flag::FL_ALLOW_UNSUPP_EXEC) {
            flags.push("allow-unsupp-exec");
        }
        if self.contains(Flag::FL_ALLOW_UNSUPP_SOCKET) {
            flags.push("allow-unsupp-socket");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_SOCKET) {
            flags.push("allow-unsafe-socket");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_SYSLOG) {
            flags.push("allow-unsafe-syslog");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_PRCTL) {
            flags.push("allow-unsafe-prctl");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_PRLIMIT) {
            flags.push("allow-unsafe-prlimit");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_DEBUG) {
            flags.push("allow-unsafe-debug");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_TIME) {
            flags.push("allow-unsafe-time");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_OPEN) {
            flags.push("allow-unsafe-open");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_FILENAME) {
            flags.push("allow-unsafe-filename");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_IOURING) {
            flags.push("allow-unsafe-uring");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_KILL) {
            flags.push("allow-unsafe-kill");
        }
        if self.contains(Flag::FL_ALLOW_UNSAFE_MEMFD) {
            flags.push("allow-unsafe-memfd");
        }
        if self.contains(Flag::FL_FORCE_CLOEXEC) {
            flags.push("force-cloexec");
        }
        if self.contains(Flag::FL_DENY_DOTDOT) {
            flags.push("deny-dotdot");
        }
        if self.contains(Flag::FL_DENY_ELF32) {
            flags.push("deny-elf32");
        }
        if self.contains(Flag::FL_DENY_ELF_DYNAMIC) {
            flags.push("deny-elf-dynamic");
        }
        if self.contains(Flag::FL_DENY_ELF_STATIC) {
            flags.push("deny-elf-static");
        }
        if self.contains(Flag::FL_DENY_SCRIPT) {
            flags.push("deny-script");
        }
        if self.contains(Flag::FL_DENY_TSC) {
            flags.push("deny-tsc");
        }
        if self.contains(Flag::FL_LOCK_PERSONALITY) {
            flags.push("lock-personality");
        }

        flags.sort();
        serializer.collect_seq(flags)
    }
}

impl Flag {
    // Returns true if the Flag must be set at startup.
    fn is_startup(self) -> bool {
        self.intersects(
            Self::FL_MAP_ROOT
                | Self::FL_FAKE_ROOT
                | Self::FL_EXIT_WAIT_ALL
                | Self::FL_ALLOW_UNSAFE_BIND
                | Self::FL_ALLOW_UNSAFE_CPU
                | Self::FL_ALLOW_UNSAFE_EXEC
                | Self::FL_ALLOW_UNSAFE_KEYRING
                | Self::FL_ALLOW_UNSAFE_MEMORY
                | Self::FL_ALLOW_UNSAFE_PKEY
                | Self::FL_ALLOW_UNSAFE_CAPS
                | Self::FL_ALLOW_UNSAFE_ENV
                | Self::FL_ALLOW_SAFE_SETID
                | Self::FL_ALLOW_UNSAFE_SOCKET
                | Self::FL_ALLOW_UNSAFE_SYSLOG
                | Self::FL_ALLOW_UNSAFE_PRCTL
                | Self::FL_ALLOW_UNSAFE_PRLIMIT
                | Self::FL_ALLOW_UNSAFE_DEBUG
                | Self::FL_ALLOW_UNSAFE_TIME
                | Self::FL_ALLOW_UNSAFE_IOURING
                | Self::FL_ALLOW_UNSAFE_KILL
                | Self::FL_DENY_TSC
                | Self::FL_LOCK_PERSONALITY,
        )
    }

    fn ns_from_str(value: &str, unsafe_: bool) -> Result<Self, Errno> {
        let mut flags = Flag::empty();
        for flag in value.split(',') {
            // Let's be as lax as possible but not more.
            let flag = flag.trim();
            if flag.is_empty() {
                continue;
            }
            flags |= match flag.to_ascii_lowercase().as_str() {
                "mount" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_MOUNT,
                "uts" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_UTS,
                "ipc" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_IPC,
                "user" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_USER,
                "pid" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_PID,
                "net" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_NET,
                "cgroup" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_CGROUP,
                "time" if unsafe_ => Flag::FL_ALLOW_UNSAFE_UNSHARE_TIME,
                "mount" => Flag::FL_UNSHARE_MOUNT,
                "uts" => Flag::FL_UNSHARE_UTS,
                "ipc" => Flag::FL_UNSHARE_IPC,
                "user" => Flag::FL_UNSHARE_USER,
                "pid" => Flag::FL_UNSHARE_PID,
                "net" => Flag::FL_UNSHARE_NET,
                "cgroup" => Flag::FL_UNSHARE_CGROUP,
                "time" => Flag::FL_UNSHARE_TIME,
                _ => return Err(Errno::EINVAL),
            };
        }

        if !flags.is_empty() {
            Ok(flags)
        } else {
            Err(Errno::ENOENT)
        }
    }
}

/// Represents a network address pattern
#[derive(Debug, Eq, PartialEq)]
pub struct AddressPattern {
    addr: IpNetwork,
    port: [u16; 2],
}

impl fmt::Display for AddressPattern {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        if self.port[0] == self.port[1] {
            write!(f, "{}:{}", self.addr, self.port[0])
        } else {
            write!(f, "{}:{}-{}", self.addr, self.port[0], self.port[1])
        }
    }
}

impl Serialize for AddressPattern {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        let mut map = serializer.serialize_map(Some(2))?;
        map.serialize_entry("addr", &self.addr)?;
        if self.port[0] == self.port[1] {
            // Single port
            map.serialize_entry("port", &self.port[0])?;
        } else {
            // Port range
            map.serialize_entry("port", &self.port)?;
        }
        map.end()
    }
}

impl FromStr for AddressPattern {
    type Err = Errno;

    fn from_str(pat: &str) -> Result<Self, Self::Err> {
        let mut split = pat.splitn(2, ['!', '@']);
        if let (Some(addr), Some(port)) = (split.next(), split.next()) {
            let mut split = port.splitn(2, '-');
            if let Some(port0) = split.next() {
                if let Ok(port0) = port0.parse::<u16>() {
                    let port1 = if let Some(port1) = split.next() {
                        port1.parse::<u16>()
                    } else {
                        Ok(port0)
                    };
                    if let Ok(port1) = port1 {
                        if let Ok(addr) = IpNetwork::from_str(addr) {
                            return Ok(Self {
                                addr,
                                port: [port0, port1],
                            });
                        }
                    }
                }
            }
        }

        Err(Errno::EINVAL)
    }
}

/// Represents a rule action.
#[derive(Debug, Copy, Clone, Eq, PartialEq, Ord, PartialOrd)]
pub enum Action {
    /// Allow system call.
    Allow,
    /// Allow system call and warn.
    Warn,
    /// Deny system call silently.
    Filter,
    /// Deny system call and warn.
    Deny,
    /// Deny system call, warn and stop offending process.
    Stop,
    /// Deny system call, warn and kill offending process.
    Kill,
    /// Warn, and exit Syd immediately with deny errno as exit value.
    Exit,
}

impl FromStr for Action {
    type Err = Errno;

    fn from_str(value: &str) -> Result<Self, Self::Err> {
        Ok(match value.to_ascii_lowercase().as_str() {
            "allow" => Action::Allow,
            "warn" => Action::Warn,
            "filter" => Action::Filter,
            "deny" => Action::Deny,
            "stop" => Action::Stop,
            "kill" => Action::Kill,
            "exit" => Action::Exit,
            _ => return Err(Errno::EINVAL),
        })
    }
}

impl fmt::Display for Action {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match self {
            Action::Allow => write!(f, "Allow"),
            Action::Warn => write!(f, "Warn"),
            Action::Filter => write!(f, "Filter"),
            Action::Deny => write!(f, "Deny"),
            Action::Stop => write!(f, "Stop"),
            Action::Kill => write!(f, "Kill"),
            Action::Exit => write!(f, "Exit"),
        }
    }
}

impl Serialize for Action {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(&self.to_string())
    }
}

impl From<&Captures<'_>> for Action {
    fn from(captures: &Captures) -> Self {
        if let Some(act) = captures.name("act") {
            Action::from_str(act.as_str()).unwrap_or(Action::Deny)
        } else {
            Action::Deny
        }
    }
}

impl Action {
    /// Return true if Action is allowing.
    pub fn is_allowing(self) -> bool {
        matches!(self, Self::Allow | Self::Warn)
    }

    /// Return true if Action is denying.
    pub fn is_denying(self) -> bool {
        !self.is_allowing()
    }

    /// Return true if Action is signaling.
    pub fn is_signaling(self) -> bool {
        matches!(self, Self::Stop | Self::Kill)
    }
}

/// Represents a network address sandboxing rule.
#[derive(Debug, Eq, PartialEq)]
pub struct CidrRule {
    act: Action,
    cap: Capability,
    pat: AddressPattern,
}

/// Represents a glob sandboxing rule.
#[derive(Debug)]
pub struct GlobRule {
    act: Action,
    cap: Capability,
    pat: Glob,
}

/// Represents a simple sandboxing rule.
#[derive(Debug)]
pub struct Rule(Glob);

impl CidrRule {
    /// Create a network address sandboxing rule.
    pub fn new(act: Action, cap: Capability, addr: IpNetwork, port: [u16; 2]) -> Self {
        Self {
            act,
            cap,
            pat: AddressPattern { addr, port },
        }
    }
}

impl fmt::Display for CidrRule {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Action: {}, Capability: {}, Pattern: {}",
            self.act, self.cap, self.pat
        )
    }
}

impl fmt::Display for GlobRule {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "Action: {}, Capability: {}, Pattern: {}",
            self.act,
            self.cap,
            self.pat.glob()
        )
    }
}

impl fmt::Display for Rule {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "Pattern: {}", self.0.glob())
    }
}

impl Serialize for CidrRule {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut state = serializer.serialize_struct("cidr", 3)?;
        state.serialize_field("act", &self.act)?;
        state.serialize_field("cap", &self.cap)?;
        state.serialize_field("pat", &self.pat)?;
        state.end()
    }
}

impl Serialize for GlobRule {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut state = serializer.serialize_struct("glob", 3)?;
        state.serialize_field("act", &self.act)?;
        state.serialize_field("cap", &self.cap)?;
        state.serialize_field("pat", &self.pat.glob())?;
        state.end()
    }
}

impl Serialize for Rule {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut state = serializer.serialize_struct("rule", 1)?;
        state.serialize_field("pat", &self.0.glob())?;
        state.end()
    }
}

/// Represents the state of the magic command lock.
#[derive(Copy, Clone, Debug, Eq, PartialEq)]
pub enum LockState {
    /// Lock is off, sandbox commands are allowed.
    Off,
    /// Sandbox commands are only allowed to the syd execve child.
    Exec,
    /// Lock is set, sandbox commands are not allowed.
    Set,
}

impl Serialize for LockState {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(match self {
            LockState::Exec => "exec",
            LockState::Set => "set",
            LockState::Off => "off",
        })
    }
}

type SandboxGlobSet = (GlobSet, GlobSetBuilder, bool);

/// Sandbox
pub struct Sandbox {
    /// Sandbox options represented using a set of `Flag` flags.
    pub flags: Flag,

    /// Sandbox state represented using a set of `Capability` flags.
    pub state: Capability,

    /// State of the magic lock.
    /// SAFETY: If None, set to LockState::Set the moment cpid is set.
    pub lock: Option<LockState>,

    /// Process ID of the syd execve child.
    /// SAFETY: This is 0 before the sandbox process has been executed.
    cpid: nix::libc::pid_t,

    /// PID file descriptor of the syd execve child.
    fpid: RawFd,

    /// Change root to this directory on startup.
    pub root: Option<PathBuf>,

    /// Mountpoint for the proc filesystem.
    pub proc: Option<PathBuf>,

    /// Mount propagation flags.
    pub propagation: Option<MsFlags>,

    /// Hostname in UTS namespace.
    pub hostname: String,

    /// Domainname in UTS namespace.
    pub domainname: String,

    /// Timestamp for BOOTTIME in Time namespace.
    pub time: Option<i64>,

    /// Per-process memory limit in bytes for memory sandboxing.
    pub mem_max: u64,

    /// Per-process virtual memory limit in bytes for memory sandboxing.
    pub mem_vm_max: u64,

    /// Pid limit for PID sandboxing.
    pub pid_max: usize,

    /// Umask mode to force, None to disable.
    pub umask: Option<Mode>,

    /// Proxy local port (defaults to 9050).
    pub proxy_port: u16,

    /// Proxy external port (defaults to 9050).
    pub proxy_ext_port: u16,

    /// Proxy external address (defaults to 127.0.0.1).
    pub proxy_ext_addr: IpAddr,

    /// Default action for Read sandboxing.
    read_act: Action,

    /// Default action for Write sandboxing.
    write_act: Action,

    /// Default action for Stat sandboxing.
    stat_act: Action,

    /// Default action for Exec sandboxing.
    exec_act: Action,

    /// Default action for Ioctl sandboxing.
    ioctl_act: Action,

    /// Default action for Net/Bind sandboxing.
    bind_act: Action,

    /// Default action for Net/Connect sandboxing.
    connect_act: Action,

    /// Default action for Memory sandboxing.
    mem_act: Action,

    /// Default action for PID sandboxing.
    pid_act: Action,

    /// Default action for TPE sandboxing.
    tpe_act: Action,

    /// Integrity force defalt mode.
    force_act: Action,

    /// SegvGuard default mode.
    segvguard_act: Action,

    // Set of denylisted ioctl(2) requests.
    ioctl_set: HashSet<u64>,

    // TPE GID
    tpe_gid: Option<Gid>,

    // Vector of safe {U,G}ID transitions.
    // source_{u,g}id->target_{u,g}id
    transit_uids: Vec<(Uid, Uid)>,
    transit_gids: Vec<(Gid, Gid)>,

    // SegvGuard entry expiry timeout.
    segvguard_expiry: Duration,

    // SegvGuard entry suspension timeout.
    segvguard_suspension: Duration,

    // SegvGuard max number of crashes before suspension.
    segvguard_maxcrashes: u8,

    // SegvGuard Lists: map is used for expiry, set is used for suspension.
    segvguard_map: crate::endorphin::HashMap<PathBuf, u8, crate::endorphin::policy::TTLPolicy>,
    segvguard_set:
        crate::endorphin::HashSet<PathBuf, crate::endorphin::policy::TTLPolicy, DefaultHashBuilder>,

    // Crypt Sandboxing.
    pub(crate) crypt_id: Option<Secret>,
    pub(crate) crypt_tmp: Option<PathBuf>,

    // Integrity force check map.
    force_map: HashMap<PathBuf, (Action, Vec<u8>)>,

    // List of bind mounts.
    bind_mounts: Option<Vec<BindMount>>,

    // List of allowlisted Netlink families.
    pub(crate) netlink_families: NetlinkFamily,

    // List of network address sandboxing rules.
    cidr_rules: Vec<CidrRule>,
    // List of glob sandboxing rules.
    glob_rules: Vec<GlobRule>,
    // List of crypt rules.
    crypt_rules: Vec<Rule>,
    // List of SegvGuard filter rules.
    segvguard_filter_rules: Vec<Rule>,
    // List of mask rules.
    mask_rules: Vec<Rule>,

    // read glob sets for access and filter.
    globset_access_r: SandboxGlobSet,
    globset_filter_r: SandboxGlobSet,

    // stat glob sets for access and filter.
    globset_access_s: SandboxGlobSet,
    globset_filter_s: SandboxGlobSet,

    // write glob sets for access and filter.
    globset_access_w: SandboxGlobSet,
    globset_filter_w: SandboxGlobSet,

    // execute glob sets for access and filter
    globset_access_x: SandboxGlobSet,
    globset_filter_x: SandboxGlobSet,

    // ioctl glob sets for access and filter
    globset_access_i: SandboxGlobSet,
    globset_filter_i: SandboxGlobSet,

    // network bind glob sets for access and filter.
    globset_access_nb: SandboxGlobSet,
    globset_filter_nb: SandboxGlobSet,

    // network connect glob sets for access and filter.
    globset_access_nc: SandboxGlobSet,
    globset_filter_nc: SandboxGlobSet,

    // crypt glob set
    globset_crypt: SandboxGlobSet,

    // segvguard globset for filter.
    globset_segvguard_filter: SandboxGlobSet,

    // mask glob set
    globset_mask: SandboxGlobSet,
}

/// Sandbox Lock types.
#[derive(Debug)]
pub enum SandboxLock {
    /// Read Lock
    Read,
    /// Upgradable read lock
    UpgradableRead,
    /// Write lock
    Write,
}

/// Sandbox guard to use it practically under a read/write lock.
#[derive(Debug)]
pub enum SandboxGuard<'a> {
    /// Sandbox locked for read
    Read(RwLockReadGuard<'a, Sandbox>),
    /// Sandbox locked for upgradable read
    UpgradableRead(RwLockUpgradableReadGuard<'a, Sandbox>),
    /// Sandbox locked for write
    Write(RwLockWriteGuard<'a, Sandbox>),
}

impl Default for Flag {
    fn default() -> Self {
        Self::empty()
    }
}

impl Default for Capability {
    fn default() -> Self {
        Self::CAP_READ
            | Self::CAP_WRITE
            | Self::CAP_EXEC
            | Self::CAP_STAT
            | Self::CAP_IOCTL
            | Self::CAP_BIND
            | Self::CAP_CONNECT
            | Self::CAP_TPE
    }
}

impl Default for LockState {
    fn default() -> Self {
        Self::Exec
    }
}

impl FromStr for LockState {
    type Err = io::Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s {
            "off" => Ok(Self::Off),
            "on" => Ok(Self::Set),
            "exec" => Ok(Self::Exec),
            _ => Err(io::Error::from_raw_os_error(nix::libc::EINVAL)),
        }
    }
}

impl<'a> Deref for SandboxGuard<'a> {
    type Target = Sandbox;
    fn deref(&self) -> &Self::Target {
        match self {
            SandboxGuard::Read(guard) => guard,
            SandboxGuard::UpgradableRead(guard) => guard,
            SandboxGuard::Write(guard) => guard,
        }
    }
}

impl<'a> DerefMut for SandboxGuard<'a> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        match self {
            SandboxGuard::Write(guard) => guard,
            _ => panic!("Cannot mutate a read-locked Sandbox!"),
        }
    }
}

impl<'a> SandboxGuard<'a> {
    /// Upgrades an upgradable read lock to a write lock.
    /// Panics if not upgradable.
    pub fn upgrade(self) -> SandboxGuard<'a> {
        match self {
            SandboxGuard::UpgradableRead(guard) => {
                SandboxGuard::Write(RwLockUpgradableReadGuard::upgrade(guard))
            }
            _ => panic!("Attempted to upgrade a non-upgradable guard!"),
        }
    }

    /// Downgrades a write or upgradable read lock to a read lock or upgradable read lock.
    /// Panics if not applicable.
    pub fn downgrade(self) -> SandboxGuard<'a> {
        match self {
            SandboxGuard::Write(guard) => SandboxGuard::Read(RwLockWriteGuard::downgrade(guard)),
            SandboxGuard::UpgradableRead(guard) => {
                SandboxGuard::Read(RwLockUpgradableReadGuard::downgrade(guard))
            }
            _ => panic!("Attempted to downgrade a non-downgradable guard!"),
        }
    }
}

impl Default for Sandbox {
    #[allow(clippy::disallowed_methods)]
    fn default() -> Self {
        #[allow(clippy::cast_sign_loss)]
        let mut sbox = Sandbox {
            flags: Flag::default(),
            state: Capability::default(),
            lock: None,
            cpid: 0,
            fpid: nix::libc::AT_FDCWD,
            proc: Some(PathBuf::from("/proc")),
            root: None,
            propagation: Some(MsFlags::MS_PRIVATE | MsFlags::MS_REC),
            hostname: (getpid().as_raw() as u64).to_name(),
            domainname: format!("syd-{API_VERSION}"),
            time: None,
            read_act: Action::Deny,
            write_act: Action::Deny,
            stat_act: Action::Deny,
            exec_act: Action::Deny,
            ioctl_act: Action::Deny,
            bind_act: Action::Deny,
            connect_act: Action::Deny,
            mem_act: Action::Deny,
            mem_max: 128_u64.saturating_mul(1024).saturating_mul(1024),
            mem_vm_max: 4096_u64.saturating_mul(1024).saturating_mul(1024),
            pid_act: Action::Kill,
            pid_max: 128,
            umask: None,
            // TIOCSTI, TIOCCONS and TIOCLINUX are denylisted by default.
            ioctl_set: HashSet::from([0x5412, 0x541c, 0x541d]),
            proxy_port: 9050,
            proxy_ext_port: 9050,
            proxy_ext_addr: IpAddr::V4(Ipv4Addr::new(127, 0, 0, 1)),
            tpe_act: Action::Deny,
            tpe_gid: None,
            transit_uids: Vec::new(),
            transit_gids: Vec::new(),
            segvguard_act: Action::Kill,
            segvguard_expiry: SEGVGUARD_EXPIRY,
            segvguard_suspension: SEGVGUARD_SUSPENSION,
            segvguard_maxcrashes: SEGVGUARD_MAXCRASHES,
            segvguard_map: crate::endorphin::HashMap::new(
                crate::endorphin::policy::TTLPolicy::new(),
            ),
            segvguard_set: crate::endorphin::HashSet::new(
                crate::endorphin::policy::TTLPolicy::new(),
            ),
            crypt_id: None,
            crypt_tmp: None,
            force_act: Action::Kill,
            force_map: HashMap::new(),
            bind_mounts: None,
            netlink_families: NetlinkFamily::empty(),
            cidr_rules: Vec::new(),
            glob_rules: Vec::new(),
            crypt_rules: Vec::new(),
            segvguard_filter_rules: Vec::new(),
            mask_rules: Vec::new(),
            globset_access_r: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_access_s: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_access_w: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_access_x: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_access_i: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_access_nb: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_access_nc: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_r: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_s: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_w: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_x: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_i: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_nb: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_filter_nc: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_crypt: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_segvguard_filter: (GlobSet::empty(), GlobSetBuilder::new(), false),
            globset_mask: (GlobSet::empty(), GlobSetBuilder::new(), false),
        };

        // SAFETY: We provide an empty file in place of /proc/cmdline
        // for compatibility rather than denying outright. In any case,
        // syd does not leak the contents of this file to the sandbox
        // process. See: https://gitlab.exherbo.org/sydbox/sydbox/-/issues/106
        // SAFETY: We add a mask rule here rather than checking it with a
        // hardcoded clause so as to give the user a chance allow access
        // to /proc/cmdline using "mask-/proc/cmdline" should there be a
        // need to pass boot parameters into the sandbox.
        sbox.rule_add_mask("/proc/cmdline")
            .expect("mask /proc/cmdline");

        // SAFETY: We provide an empty file in place of /{dev,proc}/kmsg
        // for convenience rather than denying outright. In any case Syd
        // does not leak kernel log information to the sandbox process.
        // Without this rule in place, the user is going to get a
        // "Function not implemented" error on dmesg(1) as it fallsback
        // to the syslog(2) system call if /{dev,proc}/kmsg is not
        // available and Syd denies this system call.
        // Again, these rules may be removed from Syd using e.g.
        // "mask-/proc/kmsg". This is necessary when containing a
        // program that interacts with kernel's log buffer such as
        // syslogd.
        sbox.rule_add_mask("/dev/kmsg").expect("mask /dev/kmsg");
        sbox.rule_add_mask("/proc/kmsg").expect("mask /proc/kmsg");

        // SAFETY: We allow /dev/null for write and /proc for read by
        // default in the Landlock sandbox, because Syd is included in
        // the Landlock sandbox and Syd requires access to this file to
        // function correctly.
        sbox.rule_add_glob(Action::Allow, Capability::CAP_LOCK_RO, "/proc")
            .expect("allow/lock/read+/proc");
        sbox.rule_add_glob(Action::Allow, Capability::CAP_LOCK_RW, "/dev/null")
            .expect("allow/lock/write+/dev/null");

        sbox
    }
}

impl fmt::Display for Sandbox {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "syd:")?;
        writeln!(f, "  Process ID: {}", self.cpid)?;
        writeln!(f, "  Lock: {:?}", self.lock)?;
        writeln!(f, "  Capabilities: {}", self.state)?;
        writeln!(f, "  Options: {}", self.flags)?;
        writeln!(f, "  Memory Max: {}", self.mem_max)?;
        writeln!(f, "  Virtual Memory Max: {}", self.mem_vm_max)?;
        writeln!(f, "  Pid Max: {}", self.pid_max)?;
        writeln!(f, "  SegvGuard Max Crashes: {}", self.segvguard_maxcrashes)?;
        writeln!(
            f,
            "  SegvGuard Expiry: {} seconds",
            self.segvguard_expiry.as_secs()
        )?;
        writeln!(
            f,
            "  SegvGuard Suspension: {} seconds",
            self.segvguard_suspension.as_secs()
        )?;
        writeln!(
            f,
            "  Allowed UID Transitions: (total: {}, source -> target)",
            self.transit_uids.len()
        )?;
        for (source_uid, target_uid) in &self.transit_uids {
            writeln!(f, "    - {source_uid} -> {target_uid}")?;
        }
        writeln!(
            f,
            "  Allowed GID Transitions: (total: {}, source -> target)",
            self.transit_gids.len()
        )?;
        for (source_gid, target_gid) in &self.transit_gids {
            writeln!(f, "    - {source_gid} -> {target_gid}")?;
        }
        writeln!(
            f,
            "  Cidr Rules: (total {}, highest precedence first)",
            self.cidr_rules.len()
        )?;
        for (idx, rule) in self.cidr_rules.iter().rev().enumerate() {
            // rev() because last matching rule wins.
            let idx = idx.saturating_add(1);
            writeln!(f, "    {idx}. {rule}")?;
        }
        writeln!(
            f,
            "  Glob Rules: (total {}, highest precedence first)",
            self.glob_rules.len()
        )?;
        for (idx, rule) in self.glob_rules.iter().rev().enumerate() {
            // rev() because last matching rule wins.
            let idx = idx.saturating_add(1);
            writeln!(f, "    {idx}. {rule}")?;
        }
        writeln!(f, "  Mask Rules: (total {})", self.mask_rules.len())?;
        for (idx, rule) in self.mask_rules.iter().enumerate() {
            let idx = idx.saturating_add(1);
            writeln!(f, "    {idx}. {rule}")?;
        }

        let mut force_rules: Vec<_> = self.force_map.iter().collect();
        force_rules.sort_by_key(|&(k, _)| k); // Sort by path keys for consistent ordering
        writeln!(
            f,
            "  Force Rules: (total {}, default action: {})",
            force_rules.len(),
            self.force_act,
        )?;
        for (idx, (k, v)) in force_rules.iter().enumerate() {
            let idx = idx.saturating_add(1);
            let (force, bytes) = v;
            let hash = const_hex::encode(bytes);
            let path = k.to_string_lossy();
            writeln!(
                f,
                "    {idx}. Action: {force}, Checksum: {hash}, Path: {path}"
            )?;
        }

        Ok(())
    }
}

impl fmt::Debug for Sandbox {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("Sandbox")
            .field("flags", &self.flags)
            .field("state", &self.state)
            .field("lock", &self.lock)
            .field("cpid", &self.cpid)
            .field("fpid", &self.fpid)
            .field("root", &self.root)
            .field("proc", &self.proc)
            .field("default_read", &self.read_act)
            .field("default_write", &self.write_act)
            .field("default_stat", &self.stat_act)
            .field("default_exec", &self.exec_act)
            .field("default_ioctl", &self.ioctl_act)
            .field("default_net_bind", &self.bind_act)
            .field("default_net_connect", &self.connect_act)
            .field("default_mem", &self.mem_act)
            .field("default_pid", &self.pid_act)
            .field("default_tpe", &self.tpe_act)
            .field("default_force", &self.force_act)
            .field("default_segvguard", &self.segvguard_act)
            .field("propagation", &self.propagation)
            .field("hostname", &self.hostname)
            .field("domainname", &self.domainname)
            .field("mem_max", &self.mem_max)
            .field("mem_vm_max", &self.mem_vm_max)
            .field("pid_max", &self.pid_max)
            .field("umask", &self.umask)
            .field("transit_uids", &self.transit_uids)
            .field("transit_gids", &self.transit_gids)
            .field("segvguard_expiry", &self.segvguard_expiry)
            .field("segvguard_suspension", &self.segvguard_suspension)
            .field("segvguard_maxcrashes", &self.segvguard_maxcrashes)
            //.field("segvguard_map", &self.segvguard_map) // skipped
            //.field("segvguard_set", &self.segvguard_set) // skipped
            .field("force_map", &self.force_map)
            .field("bind_mounts", &self.bind_mounts)
            .field("cidr_rules", &self.cidr_rules)
            .field("glob_rules", &self.glob_rules)
            .field("crypt_rules", &self.crypt_rules)
            .field("mask_rules", &self.mask_rules)
            .field("globset_access_r", &self.globset_access_r)
            .field("globset_filter_r", &self.globset_filter_r)
            .field("globset_access_s", &self.globset_access_s)
            .field("globset_filter_s", &self.globset_filter_s)
            .field("globset_access_w", &self.globset_access_w)
            .field("globset_filter_w", &self.globset_filter_w)
            .field("globset_access_x", &self.globset_access_x)
            .field("globset_filter_x", &self.globset_filter_x)
            .field("globset_access_i", &self.globset_access_i)
            .field("globset_filter_i", &self.globset_filter_i)
            .field("globset_access_nb", &self.globset_access_nb)
            .field("globset_filter_nb", &self.globset_filter_nb)
            .field("globset_access_nc", &self.globset_access_nc)
            .field("globset_filter_nc", &self.globset_filter_nc)
            .field("globset_crypt", &self.globset_crypt)
            .field("globset_mask", &self.globset_mask)
            .finish()
    }
}

impl Serialize for Sandbox {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut map = serializer.serialize_map(None)?;

        map.serialize_entry("flags", &self.flags)?;
        map.serialize_entry("state", &self.state)?;
        map.serialize_entry("lock", &self.lock)?;
        map.serialize_entry("cpid", &self.cpid)?;
        map.serialize_entry("default_read", &self.read_act)?;
        map.serialize_entry("default_write", &self.write_act)?;
        map.serialize_entry("default_stat", &self.stat_act)?;
        map.serialize_entry("default_exec", &self.exec_act)?;
        map.serialize_entry("default_ioctl", &self.ioctl_act)?;
        map.serialize_entry("default_net_bind", &self.bind_act)?;
        map.serialize_entry("default_net_connect", &self.connect_act)?;
        map.serialize_entry("default_mem", &self.mem_act)?;
        map.serialize_entry("default_pid", &self.pid_act)?;
        map.serialize_entry("default_tpe", &self.tpe_act)?;
        map.serialize_entry("default_force", &self.force_act)?;
        map.serialize_entry("default_segvguard", &self.segvguard_act)?;
        map.serialize_entry("mem_max", &self.mem_max)?;
        map.serialize_entry("mem_vm_max", &self.mem_vm_max)?;
        map.serialize_entry("pid_max", &self.pid_max)?;
        map.serialize_entry("segvguard_expiry", &self.segvguard_expiry.as_secs())?;
        map.serialize_entry("segvguard_suspension", &self.segvguard_suspension.as_secs())?;
        map.serialize_entry("segvguard_maxcrashes", &self.segvguard_maxcrashes)?;
        map.serialize_entry(
            "transit_uids",
            &self
                .transit_uids
                .iter()
                .map(|(src, dst)| (src.as_raw(), dst.as_raw()))
                .collect::<Vec<(nix::libc::uid_t, nix::libc::uid_t)>>(),
        )?;
        map.serialize_entry(
            "transit_gids",
            &self
                .transit_gids
                .iter()
                .map(|(src, dst)| (src.as_raw(), dst.as_raw()))
                .collect::<Vec<(nix::libc::gid_t, nix::libc::gid_t)>>(),
        )?;
        map.serialize_entry("cidr_rules", &self.cidr_rules)?;
        map.serialize_entry("glob_rules", &self.glob_rules)?;
        map.serialize_entry("crypt_rules", &self.crypt_rules)?;
        map.serialize_entry("mask_rules", &self.mask_rules)?;
        map.serialize_entry("segvguard_filter_rules", &self.segvguard_filter_rules)?;

        // Serialize the `force_map` as "force_rules"
        let mut force_rules: Vec<_> = self.force_map.iter().collect();
        force_rules.sort_by_key(|&(k, _)| k); // Sort by path keys for consistent ordering
                                              // FIXME: json! calls unwrap().
        #[allow(clippy::disallowed_methods)]
        let force_rules: Vec<_> = force_rules
            .iter()
            .map(|(k, v)| {
                let (force, bytes) = v;
                serde_json::json!({
                    "pat": k.to_string_lossy(),
                    "sha": const_hex::encode(bytes),
                    "act": force
                })
            })
            .collect();
        map.serialize_entry("force_rules", &force_rules)?;

        map.end()
    }
}

#[cfg(feature = "oci")]
impl TryFrom<&Spec> for Sandbox {
    type Error = Errno;

    /// Create a sandbox using saved config /.oci.syd-3.
    /// If the file does not exist, parse the oci profile.
    fn try_from(_spec: &Spec) -> Result<Self, Self::Error> {
        // Initialize sandbox to default state.
        let mut sandbox = Self::default();

        // SAFETY: Let container engine declare capabilities.
        sandbox.flags.insert(Flag::FL_ALLOW_UNSAFE_CAPS);

        // SAFETY: mask /.oci-syd.3 to avoid leaking its contents to the sandbox process.
        sandbox.rule_add_mask("/.oci.syd-3")?;

        // Parse the user specified configuration file.
        // Parse the hardcoded OCI profile if no user config was specified.
        let path = PathBuf::from("/.oci.syd-3");
        if path_exists(&path) {
            sandbox
                .parse_config_file(&path)
                .map_err(|_| Errno::EINVAL)?;
            info!("ctx": "oci", "op": "load_config");
        } else {
            sandbox.parse_profile("oci").map_err(|_| Errno::EINVAL)?;
            info!("ctx": "oci", "op": "load_profile", "profile": "oci");
        }

        Ok(sandbox)
    }
}

impl Sandbox {
    /// Parses a configuration from a given file-like object and applies its configuration to the sandbox.
    ///
    /// This function reads from the given file-like object line by line. It skips lines that are either
    /// empty or start with a '#' (treated as comments). For each valid line, it applies its
    /// configuration to the provided sandbox.
    ///
    /// # Arguments
    ///
    /// * `file` - A file-like object to read the configuration from. This can be any type that
    ///            implements the `Read` and `BufRead` traits.
    /// * `path` - Path name of the file, must be an absolute path and have the file extension ".syd-3".
    /// * `imap` - A HashSet of file information to keep track of included files to prevent loops.
    ///
    /// # Returns
    ///
    /// * A Result indicating the success or failure of the operation.
    ///
    /// # Errors
    ///
    /// This function will return an error if:
    /// * There's an error reading a line from the file.
    /// * There's an issue in parsing and applying a configuration line to the sandbox.
    #[allow(clippy::cognitive_complexity)]
    pub fn parse_config<F: Read + BufRead, P: AsRef<Path>>(
        &mut self,
        mut file: F,
        path: P,
        imap: &mut HashSet<FileInformation>,
    ) -> anyhow::Result<()> {
        if path.as_ref().is_relative() {
            bail!("BUG: Relative path passed to parse_config!");
        }
        let name = path.as_ref().to_string_lossy().to_string();
        let load = name == "/dev/syd/load";
        let fext = format!(".syd-{API_VERSION}");
        if !load && !name.ends_with(&fext) {
            bail!("File has invalid file extension, expected {fext}!");
        }
        let mut line = vec![0; LINE_MAX];
        let mut temp = Vec::new();
        let mut line_count = 1usize;

        loop {
            let bytes_read = file.read(&mut line[..]).map_err(anyhow::Error::msg)?;
            if bytes_read == 0 {
                break; // End of file reached.
            }

            if !line.iter().take(bytes_read).any(|&b| b == b'\n') {
                // If no newline is found in the current chunk and we're
                // reading from a file like /dev/zero, return an error
                // indicating the line count at which this was detected.
                bail!(
                    "{name}: No newline found in the first {LINE_MAX} bytes at line {line_count}!"
                );
            }
            temp.extend_from_slice(&line[..bytes_read]);

            while let Some(pos) = temp.iter().position(|&b| b == b'\n') {
                let line = &temp[..pos]; // excludes the \n character.
                let line = std::str::from_utf8(line)
                    .map_err(anyhow::Error::msg)
                    .context(format!(
                        "{name}: Failed to convert bytes to UTF-8 at line {line_count}",
                    ))?;

                if line.trim_start().starts_with('#') {
                    /* Comment: do nothing */
                } else if line.trim().is_empty() {
                    /* Blank line: do nothing */
                } else if self.locked() {
                    // SAFETY: To be consistent with the way we handle
                    // consequent -m CLI arguments, we check for sandbox
                    // lock before each sandbox.config() call.
                    warn!("ctx": "config", "op": "parse_config",
                        "line": line, "line_count": line_count,
                        "error": "sandbox is locked", "errno": Errno::EAGAIN as i32);
                } else if let Some(inc) = line.strip_prefix("include_profile ") {
                    // Handle include_profile directive.

                    // Shell expand the include profile name.
                    let inc = shellexpand::full(inc)
                        .map_err(anyhow::Error::msg)
                        .with_context(|| {
                            format!("{name}: Failed to expand line {line_count}: `{line}'")
                        })?;

                    // Finally, parse the profile.
                    self.parse_profile(&inc).with_context(|| {
                        format!("{name}: Failed to load profile at line {line_count}!")
                    })?;
                } else if let Some(inc) = line.strip_prefix("include ") {
                    // Handle include directive.

                    // SAFETY: We do not parse include directives when loading from file descriptors.
                    if load {
                        bail!("{name}: include directive at line {line_count} is not permitted within this context!");
                    }

                    // Shell expand the include path.
                    let mut inc = shellexpand::full(inc)
                        .map_err(anyhow::Error::msg)
                        .with_context(|| {
                            format!("{name}: Failed to expand line {line_count}: `{line}'")
                        })?
                        .as_ref()
                        .to_string();

                    // Add .syd-3 extension if not exists for convenience.
                    if !inc.ends_with(&fext) {
                        inc = format!("{inc}{fext}");
                    }

                    // Handle relative files according to the directory of the current file.
                    let mut inc = PathBuf::from(inc);
                    if inc.is_relative() {
                        let dir = path.as_ref().parent().unwrap_or_else(|| Path::new("/"));
                        inc = dir.join(inc);
                    }

                    // Query file metadata.
                    let info = FileInformation::from_path(&inc)
                        .map_err(anyhow::Error::msg)
                        .with_context(|| {
                            format!("{name}: Failed to include file at line {line_count}!")
                        })?;

                    // Check for include loops and secure file permissions.
                    let permissions = Mode::from_bits_truncate(info.0.st_mode);
                    if !imap.insert(info) {
                        bail!("{name}: Include loop detected at line {line_count}!");
                    }
                    if permissions.contains(Mode::S_IWGRP | Mode::S_IWOTH) {
                        bail!(
                            "{name}: Insecure permissions for include file at line {line_count}!"
                        );
                    }

                    // Finally, open the file and parse it.
                    let file = File::open(&inc)
                        .map_err(anyhow::Error::msg)
                        .with_context(|| {
                            format!("{name}: Failed to open include file at line {line_count}!")
                        })?;
                    self.parse_config(BufReader::new(file), inc, imap)?;
                } else {
                    let line = shellexpand::full(line)
                        .map_err(anyhow::Error::msg)
                        .with_context(|| {
                            format!("{name}: Failed to expand line {line_count}: `{line}'")
                        })?;
                    self.config(&line)
                        .map_err(anyhow::Error::msg)
                        .with_context(|| {
                            format!("{name}: Failed to parse line {line_count}: `{line}'")
                        })?;
                }
                // Remove the processed line from temp storage.
                temp.drain(..=pos);
                line_count = line_count.saturating_add(1); // Increment line count after processing each line.
            }
        }

        Ok(())
    }

    /// Parses a configuration file and applies its configuration to the sandbox.
    ///
    /// This function reads the given configuration file line by line. It skips lines that are either
    /// empty or start with a '#' (treated as comments). For each valid line, it applies its
    /// configuration to the provided sandbox.
    ///
    /// # Arguments
    ///
    /// * `path` - A reference to the path of the configuration file. This can be any type that
    ///            implements the `AsRef<Path>` trait.
    ///
    /// # Returns
    ///
    /// * A Result indicating the success or failure of the operation.
    ///
    /// # Errors
    ///
    /// This function will return an error if:
    /// * There's an issue in opening the configuration file.
    /// * There's an error reading a line from the file.
    /// * There's an issue in parsing and applying a configuration line to the sandbox.
    ///
    /// # Examples
    ///
    /// ```no_run
    /// use syd::sandbox::Sandbox;
    ///
    /// let path = "/path/to/config/file";
    /// let mut sandbox = Sandbox::new();
    /// sandbox
    ///     .parse_config_file(&path)
    ///     .expect("Failed to parse configuration file");
    /// ```
    pub fn parse_config_file<P: AsRef<Path>>(&mut self, path: P) -> anyhow::Result<()> {
        let path = path
            .as_ref()
            .canonicalize()
            .map_err(anyhow::Error::msg)
            .context("Failed to canonicalize configuration file.")?;
        let file = BufReader::new(
            File::open(&path)
                .map_err(anyhow::Error::msg)
                .context("Failed to open configuration file.")?,
        );
        let mut imap = HashSet::new();
        self.parse_config(file, path, &mut imap)
    }

    /// Parses the given profile and applies its configuration to the sandbox.
    ///
    /// This function supports multiple predefined profiles such as "paludis", "noipv4", "noipv6", and "user".
    /// Each profile corresponds to a set of configuration lines which are applied to the sandbox.
    /// The "user" profile includes both static configurations and dynamic ones that depend on the
    /// environment and the existence of a user-specific configuration file.
    ///
    /// # Arguments
    ///
    /// * `name` - A string slice that holds the name of the profile to be parsed.
    /// * `sandbox` - A mutable reference to the sandbox where the profile configurations will be applied.
    ///
    /// # Returns
    ///
    /// * A Result indicating the success or failure of the operation.
    ///
    /// # Errors
    ///
    /// This function will return an error if:
    /// * The profile name is invalid.
    /// * There's an issue in parsing the configuration lines.
    /// * There's an issue in reading or parsing the user-specific configuration file for the "user" profile.
    pub fn parse_profile(&mut self, name: &str) -> anyhow::Result<()> {
        // Inner function to handle repetitive logic of applying configurations
        fn apply_config(sandbox: &mut Sandbox, profile: &[&str]) -> anyhow::Result<()> {
            for (idx, line) in profile.iter().enumerate() {
                let lcnt = idx.saturating_add(1);
                sandbox
                    .config(line)
                    .map_err(anyhow::Error::msg)
                    .context(format!("Failed to parse line {lcnt}: `{line}'."))?;
            }
            Ok(())
        }

        match name {
            "container" => apply_config(self, PROFILE_CONTAINER),
            "immutable" => apply_config(self, PROFILE_IMMUTABLE),
            "landlock" => apply_config(self, PROFILE_LANDLOCK),
            "paludis" => apply_config(self, PROFILE_PALUDIS),
            "pandora" => apply_config(self, PROFILE_PANDORA),
            "noipv4" => apply_config(self, PROFILE_NOIPV4),
            "noipv6" => apply_config(self, PROFILE_NOIPV6),
            "p" | "priv" | "privileged" => apply_config(self, PROFILE_PRIVILEGED),
            "silent" => apply_config(self, PROFILE_SILENT),
            "lib" => apply_config(self, PROFILE_LIB),
            "oci" => apply_config(self, PROFILE_OCI),
            "trace" => apply_config(self, PROFILE_TRACE),
            "user" => {
                // Step 1: Apply static user configuration defined at compile-time.
                apply_config(self, PROFILE_USER)?;

                // Step 2: Apply dynamic, user-specific configuration.
                let uid = Uid::effective();
                let gid = Gid::effective();
                let name = crate::get_user_name(uid);
                let home = crate::get_user_home(&name);

                // Save the user from some annoying warnings.
                if env::var("GIT_CEILING_DIRECTORIES").is_err() {
                    env::set_var("GIT_CEILING_DIRECTORIES", &home);
                }

                // /home
                // We allow read(/home/user) but not write(/home/user),
                // read|write(/home/user/**) is ok, i.e. the user can
                // not delete their home directory under the sandbox
                // which is a nice and funny protection.
                self.config(&format!("allow/lock/write+{home}"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/read+{home}/***"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/stat+{home}/***"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/write+{home}/**"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/exec+{home}/***"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/net/bind+{home}/**"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/net/connect+{home}/**"))
                    .map_err(anyhow::Error::msg)?;
                // /run/user/uid
                self.config(&format!("allow/read+/run/user/{uid}/**"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/write+/run/user/{uid}/**"))
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("allow/net/connect+/run/user/{uid}/**"))
                    .map_err(anyhow::Error::msg)?;
                // Trusted GID for TPE. FIXME: This is not a secure default.
                self.config("tpe/negate:1").map_err(anyhow::Error::msg)?;
                self.config("tpe/user_owned:1")
                    .map_err(anyhow::Error::msg)?;
                self.config(&format!("tpe/gid:{gid}"))
                    .map_err(anyhow::Error::msg)?;

                // Step 3: Parse the system & user configuration file if it exists.
                let rc = vec![
                    PathBuf::from(PATH_ETC).join(format!("user.syd-{API_VERSION}")),
                    Path::new(&home).join(format!(".user.syd-{API_VERSION}")),
                ];
                for p in rc {
                    if self.locked() {
                        info!("ctx": "config", "op": "parse_profile",
                            "errno": Errno::EAGAIN as i32, "error": "sandbox is locked",
                            "path": mask_path(p));
                    } else if path_exists(&p) {
                        self.parse_config_file(&p)?;
                    }
                }
                Ok(())
            }
            _ => bail!("Invalid profile `{name}'"),
        }
    }

    /// Configures the sandbox using a specified command.
    ///
    /// This method provides a central point for configuring the sandbox. It interprets and
    /// processes a variety of commands to adjust the sandbox's state, manage its tracing,
    /// handle regex-based configurations, and more.
    ///
    /// # Arguments
    ///
    /// * `command` - A string slice that represents the command to be executed.
    ///
    /// # Returns
    ///
    /// * A `Result` that indicates the success or failure of the operation. In the event of a
    ///   failure, it returns an appropriate error from the `Errno` enum.
    ///
    /// # Commands
    ///
    /// - If the command is empty or matches the API version, it simply returns `Ok(())`.
    /// - If the command starts with "lock", it attempts to set the sandbox's lock state.
    /// - If the command matches one of the supported commands, it applies the command to the sandbox.
    ///   See the ["Configuration" section in the README.md file](https://crates.io/crates/syd#configuration)
    ///   for a list of supported commands.
    /// - If none of the above conditions are met, it returns an error indicating invalid input.
    ///
    /// # Examples
    ///
    /// ```
    /// use syd::sandbox::Sandbox;
    ///
    /// let mut sandbox = Sandbox::new();
    /// sandbox
    ///     .config("lock:on")
    ///     .expect("Failed to lock the sandbox");
    /// ```
    #[allow(clippy::cognitive_complexity)]
    pub fn config(&mut self, command: &str) -> Result<(), Errno> {
        if command.is_empty() || command == API_VERSION {
            Ok(())
        } else if command == "reset" {
            self.reset()
        } else if command == "stat" {
            eprint!("{self}");
            Ok(())
        } else if let Some(state) = command.strip_prefix("lock:") {
            self.lock(LockState::from_str(state).map_err(|_| Errno::EINVAL)?)
        } else if let Some(value) = command.strip_prefix("time:") {
            self.time = Some(value.parse::<i64>().map_err(|_| Errno::EINVAL)?);
            Ok(())
        } else if let Some(command) = command.strip_prefix("cmd/") {
            self.handle_sandbox_command(command)
        } else if let Some(command) = command.strip_prefix("sandbox/") {
            self.handle_sandbox_config(command)
        } else if let Some(command) = command.strip_prefix("default/") {
            self.handle_default_config(command)
        } else if let Some(command) = command.strip_prefix("ioctl/") {
            self.handle_ioctl_config(command)
        } else if let Some(command) = command.strip_prefix("mem/") {
            self.handle_mem_config(command)
        } else if let Some(command) = command.strip_prefix("pid/") {
            self.handle_pid_config(command)
        } else if let Some(command) = command.strip_prefix("proxy/") {
            self.handle_proxy_config(command)
        } else if let Some(command) = command.strip_prefix("tpe/") {
            self.handle_tpe_config(command)
        } else if let Some(command) = command.strip_prefix("trace/") {
            self.handle_trace_config(command)
        } else if let Some(command) = command.strip_prefix("unshare/") {
            self.handle_unshare_config(command)
        } else if let Some(command) = command.strip_prefix("name/") {
            self.handle_name_config(command)
        } else if let Some(command) = command.strip_prefix("mask") {
            self.handle_mask_rules(command)
        } else if let Some(command) = command.strip_prefix("root") {
            self.handle_root_config(command)
        } else if let Some(captures) = RE_BIND.captures(command) {
            self.handle_bind_config(&captures)
        } else if let Some(command) = command.strip_prefix("crypt/") {
            self.handle_crypt_config(command)
        } else if let Some(command) = command.strip_prefix("crypt") {
            self.handle_crypt_rules(command)
        } else if let Some(command) = command.strip_prefix("segvguard/") {
            self.handle_segvguard_config(command)
        } else if let Some(captures) = RE_FORCE.captures(command) {
            self.handle_force_rule_config(&captures)
        } else if let Some(captures) = RE_SETID_0.captures(command) {
            self.handle_setid_rule_config(&captures)
        } else if let Some(captures) = RE_SETID_1.captures(command) {
            self.handle_setid_rule_config(&captures)
        } else if let Some(captures) = RE_NETALIAS.captures(command) {
            let alias = captures["alias"].to_ascii_lowercase();
            let command = captures["command"].to_ascii_lowercase();
            self.handle_netalias_config(&command, &alias)
        } else if let Some(captures) = RE_RULE_NETLINK.captures(command) {
            self.handle_netlink_config(&captures)
        } else if let Some(captures) = RE_RULE.captures(command) {
            self.handle_rule_config(&captures)
        } else {
            Err(Errno::EINVAL)
        }
    }

    fn handle_netalias_config(&mut self, command: &str, alias: &str) -> Result<(), Errno> {
        if let Some(addr_vec) = MAP_NETALIAS.get(&alias) {
            for addr in addr_vec {
                let c = command.replacen(alias, addr, 1);
                self.config(&c)?;
            }
            Ok(())
        } else {
            // This should never happen,
            // but let's handle it safely anyway.
            Err(Errno::EAFNOSUPPORT)
        }
    }

    fn handle_name_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("host:") {
            self.set_hostname(value)?;
        } else if let Some(value) = command.strip_prefix("domain:") {
            self.set_domainname(value)?;
        } else {
            return Err(Errno::EINVAL);
        }
        self.set_unshare_uts(true);
        Ok(())
    }

    fn handle_crypt_rules(&mut self, command: &str) -> Result<(), Errno> {
        // Check command length is at least 2 parameters.
        if command.len() < 2 {
            return Err(Errno::EINVAL);
        }
        // Check the first character
        match command.chars().nth(0) {
            Some('+') => self.rule_add_crypt(&command[1..]),
            Some('-') => self.rule_del_crypt(&command[1..]),
            Some('^') => self.rule_rem_crypt(&command[1..]),
            _ => Err(Errno::EINVAL),
        }
    }

    fn handle_mask_rules(&mut self, command: &str) -> Result<(), Errno> {
        // Check command length is at least 2 parameters.
        if command.len() < 2 {
            return Err(Errno::EINVAL);
        }
        // Check the first character
        match command.chars().nth(0) {
            Some('+') => self.rule_add_mask(&command[1..]),
            Some('-') => self.rule_del_mask(&command[1..]),
            Some('^') => self.rule_rem_mask(&command[1..]),
            _ => Err(Errno::EINVAL),
        }
    }

    fn handle_root_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(root) = command.strip_prefix(':') {
            let root = Path::new(root)
                .canonicalize()
                .map_err(|e| Errno::from_raw(e.raw_os_error().unwrap_or(nix::libc::EINVAL)))?;
            if self.is_running() {
                return Err(Errno::EBUSY);
            }
            self.root = Some(root.clone());
            self.proc = Some(root.join("proc"));
        } else if let Some(value) = command.strip_prefix("/map:") {
            self.set_flag(Flag::FL_MAP_ROOT, value)?;
        } else if let Some(value) = command.strip_prefix("/fake:") {
            self.set_flag(Flag::FL_FAKE_ROOT, value)?;
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    #[allow(clippy::cognitive_complexity)]
    fn handle_sandbox_config(&mut self, command: &str) -> Result<(), Errno> {
        #[allow(clippy::arithmetic_side_effects)]
        let (caps, state) = match command.chars().last() {
            Some('?') => (Capability::from_str(&command[..command.len() - 1])?, None),
            _ => {
                let mut splits = command.splitn(2, ':');
                (
                    Capability::from_str(splits.next().unwrap_or(""))?,
                    Some(strbool(splits.next().ok_or(Errno::EINVAL)?)?),
                )
            }
        };

        if let Some(state) = state {
            // Setting Capability with :{on,off}.
            if self.is_running() && caps.is_startup() {
                return Err(Errno::EBUSY);
            } else if state {
                self.state.insert(caps);
            } else {
                self.state.remove(caps);
            }
        } else if !self.state.contains(caps) {
            // Querying for Capability with ?.
            return Err(Errno::ENOENT);
        }

        Ok(()) // set|query successful.
    }

    fn handle_sandbox_command(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(command) = command.strip_prefix("exec!") {
            // Splitting the command using the Unit Separator character
            let parts: Vec<&str> = command.split('\x1F').collect();

            // Paranoid checks: Ensure the command and its arguments are not empty
            if parts.is_empty() || parts[0].is_empty() {
                error!("ctx": "config", "op": "check_sandbox_command",
                    "cfg": "cmd/exec", "errno": Errno::EINVAL as i32);
                return Err(Errno::EINVAL);
            }

            let program = parts[0];
            let args = &parts[1..];

            // SAFETY: We're spawning a child outside the sandbox here.
            // We should take some precautions so that the process to be
            // executed has a sane environment. That's why we change
            // the current directory to /, close the standard input, and
            // execute the process in a new process group.
            match Command::new(program)
                .args(args)
                .current_dir("/")
                .process_group(0)
                .stdin(Stdio::null())
                .spawn()
            {
                Ok(_) => Ok(()),
                Err(error) => {
                    let errstr = error.to_string();
                    let errno = err2no(error);
                    error!("ctx": "config", "op": "execute_sandbox_command", "cfg": "cmd/exec",
                        "errno": errno as i32, "error": errstr,
                        "cmd": program, "args": format!("{args:?}"));
                    Err(errno)
                }
            }
        } else {
            Err(Errno::ENOENT)
        }
    }

    fn handle_default_config(&mut self, command: &str) -> Result<(), Errno> {
        // Split the command into two parts: items and action.
        let mut parts = command.splitn(2, ':');

        let items = parts.next().ok_or(Errno::EINVAL)?;
        let action = Action::from_str(parts.next().ok_or(Errno::EINVAL)?)?;

        for item in items.split(',') {
            match item.trim() {
                "read" => self.read_act = action,
                "write" => self.write_act = action,
                "stat" => self.stat_act = action,
                "exec" => self.exec_act = action,
                "ioctl" => self.ioctl_act = action,
                "force" => self.force_act = action,
                "segvguard" => self.segvguard_act = action,
                "net/bind" => self.bind_act = action,
                "net/connect" => self.connect_act = action,
                "net" => {
                    self.bind_act = action;
                    self.connect_act = action;
                }
                "mem" => {
                    if action == Action::Allow {
                        return Err(Errno::EINVAL);
                    }
                    self.mem_act = action;
                }
                "pid" => {
                    if action == Action::Allow {
                        return Err(Errno::EINVAL);
                    }
                    self.pid_act = action;
                }
                "tpe" => {
                    if action == Action::Allow {
                        return Err(Errno::EINVAL);
                    }
                    self.tpe_act = action;
                }
                _ => return Err(Errno::EINVAL),
            }
        }

        Ok(())
    }

    fn handle_ioctl_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("deny+") {
            let value = value.to_ascii_lowercase();
            let request = if let Some(value) = value.strip_prefix("0x") {
                u64::from_str_radix(value, 16).map_err(|_| Errno::EINVAL)?
            } else if let Some(value) = value.strip_prefix("0o") {
                u64::from_str_radix(value, 8).map_err(|_| Errno::EINVAL)?
            } else {
                value.parse::<u64>().map_err(|_| Errno::EINVAL)?
            };
            self.add_ioctl(request);
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    fn handle_mem_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("max:") {
            match parse_size::Config::new().with_binary().parse_size(value) {
                Ok(value) => {
                    self.mem_max = value;
                }
                Err(_) => {
                    return Err(Errno::EINVAL);
                }
            }
        } else if let Some(value) = command.strip_prefix("vm_max:") {
            match parse_size::Config::new().with_binary().parse_size(value) {
                Ok(value) => {
                    self.mem_vm_max = value;
                }
                Err(_) => {
                    return Err(Errno::EINVAL);
                }
            }
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    fn handle_pid_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("max:") {
            self.pid_max = value.parse::<usize>().map_err(|_| Errno::EINVAL)?;
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    fn handle_proxy_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("port:") {
            self.proxy_port = value.parse::<u16>().map_err(|_| Errno::EINVAL)?;
        } else if let Some(value) = command.strip_prefix("ext/port:") {
            self.proxy_ext_port = value.parse::<u16>().map_err(|_| Errno::EINVAL)?;
        } else if let Some(value) = command.strip_prefix("ext/addr:") {
            self.proxy_ext_addr = value.parse::<IpAddr>().map_err(|_| Errno::EINVAL)?;
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    fn handle_tpe_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("gid:") {
            self.tpe_gid = if value == "none" {
                None
            } else {
                Some(Gid::from_raw(
                    value
                        .parse::<nix::libc::gid_t>()
                        .map_err(|_| Errno::EINVAL)?,
                ))
            };
        } else if let Some(value) = command.strip_prefix("negate:") {
            self.set_flag(Flag::FL_TPE_NEGATE, value)?;
        } else if let Some(value) = command.strip_prefix("root_owned:") {
            self.set_flag(Flag::FL_TPE_ROOT_OWNED, value)?;
        } else if let Some(value) = command.strip_prefix("user_owned:") {
            self.set_flag(Flag::FL_TPE_USER_OWNED, value)?;
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    #[allow(clippy::cognitive_complexity)]
    fn handle_trace_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("exit_wait_all:") {
            self.set_flag(Flag::FL_EXIT_WAIT_ALL, value)?;
        } else if let Some(value) = command.strip_prefix("allow_unsafe_bind:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_BIND, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_cpu:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_CPU, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_exec:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_EXEC, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_keyring:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_KEYRING, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_memory:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_MEMORY, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_pkey:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_PKEY, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_caps:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_CAPS, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_env:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_ENV, value)?
        } else if let Some(value) = command.strip_prefix("allow_safe_bind:") {
            self.set_flag(Flag::FL_ALLOW_SAFE_BIND, value)?
        } else if let Some(value) = command.strip_prefix("allow_safe_kcapi:") {
            self.set_flag(Flag::FL_ALLOW_SAFE_KCAPI, value)?
        } else if let Some(value) = command.strip_prefix("allow_safe_setid:") {
            self.set_flag(Flag::FL_ALLOW_SAFE_SETID, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsupp_exec:") {
            self.set_flag(Flag::FL_ALLOW_UNSUPP_EXEC, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsupp_socket:") {
            self.set_flag(Flag::FL_ALLOW_UNSUPP_SOCKET, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_socket:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_SOCKET, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_syslog:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_SYSLOG, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_prctl:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_PRCTL, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_prlimit:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_PRLIMIT, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_debug:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_DEBUG, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_namespace:") {
            self.set_nsflags(value)?;
        } else if let Some(value) = command.strip_prefix("allow_unsafe_time:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_TIME, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_open:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_OPEN, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_filename:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_FILENAME, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_uring:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_IOURING, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_kill:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_KILL, value)?
        } else if let Some(value) = command.strip_prefix("allow_unsafe_memfd:") {
            self.set_flag(Flag::FL_ALLOW_UNSAFE_MEMFD, value)?
        } else if let Some(value) = command.strip_prefix("force_cloexec:") {
            self.set_flag(Flag::FL_FORCE_CLOEXEC, value)?
        } else if let Some(value) = command.strip_prefix("deny_dotdot:") {
            self.set_flag(Flag::FL_DENY_DOTDOT, value)?
        } else if let Some(value) = command.strip_prefix("deny_script:") {
            self.set_flag(Flag::FL_DENY_SCRIPT, value)?
        } else if let Some(value) = command.strip_prefix("deny_tsc:") {
            self.set_flag(Flag::FL_DENY_TSC, value)?
        } else if let Some(value) = command.strip_prefix("lock_personality:") {
            self.set_flag(Flag::FL_LOCK_PERSONALITY, value)?
        } else if let Some(value) = command.strip_prefix("deny_elf32:") {
            if strbool(value)? {
                if cfg!(feature = "elf") {
                    self.flags.insert(Flag::FL_DENY_ELF32);
                } else {
                    return Err(Errno::ENOSYS);
                }
            } else if cfg!(feature = "elf") {
                self.flags.remove(Flag::FL_DENY_ELF32);
            } else {
                return Err(Errno::ENOSYS);
            }
        } else if let Some(value) = command.strip_prefix("deny_elf_dynamic:") {
            if strbool(value)? {
                if cfg!(feature = "elf") {
                    self.flags.insert(Flag::FL_DENY_ELF_DYNAMIC);
                } else {
                    return Err(Errno::ENOSYS);
                }
            } else if cfg!(feature = "elf") {
                self.flags.remove(Flag::FL_DENY_ELF_DYNAMIC);
            } else {
                return Err(Errno::ENOSYS);
            }
        } else if let Some(value) = command.strip_prefix("deny_elf_static:") {
            if strbool(value)? {
                if cfg!(feature = "elf") {
                    self.flags.insert(Flag::FL_DENY_ELF_STATIC);
                } else {
                    return Err(Errno::ENOSYS);
                }
            } else if cfg!(feature = "elf") {
                self.flags.remove(Flag::FL_DENY_ELF_STATIC);
            } else {
                return Err(Errno::ENOSYS);
            }
        } else if let Some(value) = command.strip_prefix("force_umask:") {
            // Note, 0 is a valid umask so we'd rather disable with -1 or strbools.
            let umask = match value {
                "-1" | "off" | "f" | "fa" | "fal" | "fals" | "false" | "✗" => None,
                _ => Some(
                    Mode::from_bits(
                        nix::libc::mode_t::from_str_radix(value, 8).map_err(|_| Errno::EINVAL)?,
                    )
                    .ok_or(Errno::EINVAL)?,
                ),
            };
            if self.is_running() {
                return Err(Errno::EBUSY);
            }
            self.umask = umask;
        } else if let Some(value) = command.strip_prefix("memory_access:") {
            let value = match value {
                "0" => false,
                "1" => true,
                _ => return Err(Errno::EINVAL),
            };
            if self.is_running() {
                return Err(Errno::EBUSY);
            }
            if value {
                env::set_var(ENV_NO_CROSS_MEMORY_ATTACH, "1");
            } else {
                env::remove_var(ENV_NO_CROSS_MEMORY_ATTACH);
            }
        } else {
            return Err(Errno::EINVAL);
        }
        Ok(())
    }

    fn handle_unshare_config(&mut self, command: &str) -> Result<(), Errno> {
        #[allow(clippy::arithmetic_side_effects)]
        let (flags, state) = match command.chars().last() {
            Some('?') => (
                Flag::ns_from_str(&command[..command.len() - 1], false)?,
                None,
            ),
            _ => {
                let mut splits = command.splitn(2, ':');
                (
                    Flag::ns_from_str(splits.next().unwrap_or(""), false)?,
                    Some(strbool(splits.next().ok_or(Errno::EINVAL)?)?),
                )
            }
        };

        if let Some(state) = state {
            // Setting Capability with :{on,off}.
            if self.is_running() {
                return Err(Errno::EBUSY);
            } else if state {
                self.flags.insert(flags);
            } else {
                self.flags.remove(flags);
            }
        } else if !self.flags.contains(flags) {
            // Querying for Capability with ?.
            return Err(Errno::ENOENT);
        }

        Ok(()) // set|query successful.
    }

    fn handle_bind_config(&mut self, captures: &Captures) -> Result<(), Errno> {
        let op = &captures["mod"];
        let mount = BindMount::try_from(captures)?;

        if self.is_running() {
            return Err(Errno::EBUSY);
        }

        match op {
            "+" => self.add_bind_mount(mount),
            "-" => self.del_bind_mount(mount),
            "^" => self.rem_bind_mount(mount),
            _ => return Err(Errno::EINVAL),
        };

        Ok(())
    }

    fn handle_segvguard_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("expiry:") {
            match value.parse::<u64>() {
                Ok(value) => {
                    self.segvguard_expiry = Duration::from_secs(value);
                }
                Err(_) => {
                    return Err(Errno::EINVAL);
                }
            }
        } else if let Some(value) = command.strip_prefix("suspension:") {
            match value.parse::<u64>() {
                Ok(value) => {
                    self.segvguard_suspension = Duration::from_secs(value);
                }
                Err(_) => {
                    return Err(Errno::EINVAL);
                }
            }
        } else if let Some(value) = command.strip_prefix("maxcrashes:") {
            match value.parse::<u8>() {
                Ok(value) => {
                    self.segvguard_maxcrashes = value;
                }
                Err(_) => {
                    return Err(Errno::EINVAL);
                }
            }
        } else if let Some(value) = command.strip_prefix("filter+") {
            self.rule_add_segvguard_filter(value)?;
        } else if let Some(value) = command.strip_prefix("filter-") {
            self.rule_del_segvguard_filter(value)?;
        } else if let Some(value) = command.strip_prefix("filter^") {
            self.rule_rem_segvguard_filter(value)?;
        } else {
            return Err(Errno::EINVAL);
        }

        Ok(())
    }

    fn handle_crypt_config(&mut self, command: &str) -> Result<(), Errno> {
        if let Some(value) = command.strip_prefix("key:") {
            // SAFETY: Do not allow changing crypt key at runtime.
            if self.is_running() {
                return Err(Errno::EBUSY);
            }
            self.crypt_id = Some(Secret::Key(Key::new(
                const_hex::decode(value)
                    .map_err(|_e| Errno::EINVAL)?
                    .try_into()
                    .map_err(|_| Errno::EINVAL)?,
            )));
        } else if let Some(value) = command.strip_prefix("tmp:") {
            if value.is_empty() {
                return Err(Errno::EINVAL);
            } else if value == "mem" {
                self.crypt_tmp = None;
            } else {
                let p = PathBuf::from(value);
                if p.is_relative() {
                    return Err(Errno::EINVAL);
                } else if !path_exists(&p) {
                    return Err(Errno::ENOENT);
                } else if !p.is_dir() {
                    return Err(Errno::ENOTDIR);
                }
                self.crypt_tmp = Some(p);
            }
        } else {
            return Err(Errno::EINVAL);
        }

        Ok(())
    }

    fn handle_force_rule_config(&mut self, captures: &Captures) -> Result<(), Errno> {
        let op = &captures["mod"];

        // Handle remove-all operator.
        if op == "^" {
            if ["src", "key", "act"]
                .iter()
                .any(|&name| captures.name(name).is_some())
            {
                return Err(Errno::EINVAL);
            }
            self.force_map.clear();
            return Ok(());
        }

        let src = if let Some(src) = captures.name("src") {
            PathBuf::from(src.as_str())
        } else {
            return Err(Errno::EINVAL);
        };
        match op {
            "-" => {
                self.force_map.remove(&src);
                return Ok(());
            }
            "+" => {}
            _ => unreachable!("fix force regex"),
        }

        // Addition requires key and action.
        let key = if let Some(key) = captures.name("key") {
            const_hex::decode(key.as_str()).map_err(|_e| Errno::EINVAL)?
        } else {
            return Err(Errno::EINVAL);
        };
        // Protect user from adding invalid checksums.
        if !matches!(key.len(), 16 | 20 | 32 | 48 | 64) {
            return Err(Errno::EINVAL);
        }

        let act = if let Some(act) = captures.name("act") {
            Action::from_str(act.as_str())?
        } else {
            return Err(Errno::EINVAL);
        };

        self.force_map.insert(src, (act, key));
        Ok(())
    }

    fn handle_setid_rule_config(&mut self, captures: &Captures) -> Result<(), Errno> {
        let id = captures["id"].chars().nth(0);
        let op = captures["mod"].chars().nth(0);

        match id {
            Some('u') => {
                match (op, captures.name("src")) {
                    (Some('+'), Some(src)) => {
                        let source_uid = parse_user(src.as_str())?;
                        let target_uid = parse_user(&captures["dst"])?;
                        self.add_uid_transit(source_uid, target_uid)?;
                    }
                    (Some('-'), Some(src)) => {
                        let source_uid = parse_user(src.as_str())?;
                        let target_uid = parse_user(&captures["dst"])?;
                        self.del_uid_transit(Some(source_uid), Some(target_uid));
                    }
                    (Some('^'), Some(src)) => {
                        let source_uid = parse_user(src.as_str())?;
                        self.del_uid_transit(Some(source_uid), None);
                    }
                    (Some('^'), None) => {
                        self.del_uid_transit(None, None);
                    }
                    _ => return Err(Errno::EINVAL),
                };
            }
            Some('g') => {
                match (op, captures.name("src")) {
                    (Some('+'), Some(src)) => {
                        let source_gid = parse_group(src.as_str())?;
                        let target_gid = parse_group(&captures["dst"])?;
                        self.add_gid_transit(source_gid, target_gid)?;
                    }
                    (Some('-'), Some(src)) => {
                        let source_gid = parse_group(src.as_str())?;
                        let target_gid = parse_group(&captures["dst"])?;
                        self.del_gid_transit(Some(source_gid), Some(target_gid));
                    }
                    (Some('^'), Some(src)) => {
                        let source_gid = parse_group(src.as_str())?;
                        self.del_gid_transit(Some(source_gid), None);
                    }
                    (Some('^'), None) => {
                        self.del_gid_transit(None, None);
                    }
                    _ => return Err(Errno::EINVAL),
                };
            }
            _ => return Err(Errno::EINVAL),
        };

        Ok(())
    }

    fn handle_netlink_config(&mut self, captures: &Captures) -> Result<(), Errno> {
        if self.is_running() {
            // Netlink family restrictions are applied at kernel level at startup.
            return Err(Errno::EBUSY);
        }

        if captures.name("clr").is_some() {
            self.netlink_families = NetlinkFamily::empty();
        } else if captures.name("add").is_some() {
            self.netlink_families
                .insert(NetlinkFamily::from_str(&captures["fml"])?);
        } else if captures.name("del").is_some() {
            self.netlink_families
                .remove(NetlinkFamily::from_str(&captures["fml"])?);
        } else {
            unreachable!();
        };

        Ok(())
    }

    fn handle_rule_config(&mut self, captures: &Captures) -> Result<(), Errno> {
        let act = Action::from(captures);
        let cap = Capability::from(captures);
        let op = &captures["mod"];
        let (pat, ip) = if let Some(addr) = captures.name("addr") {
            (
                addr.as_str(),
                cap.intersects(Capability::CAP_BIND | Capability::CAP_CONNECT),
            )
        } else if let Some(path) = captures.name("path") {
            (path.as_str(), false)
        } else {
            unreachable!("Invalid rule regex!");
        };

        // Landlock rules only define the Allow action.
        if cap.intersects(Capability::CAP_LOCK) && act != Action::Allow {
            return Err(Errno::EINVAL);
        }

        // SAFETY: Reject rules that must only be set at startup.
        if self.is_running() && cap.is_startup() {
            return Err(Errno::EBUSY);
        }

        match op {
            "+" => {
                // add rule
                if ip {
                    self.rule_add_cidr(act, cap, pat.as_ref())
                } else {
                    self.rule_add_glob(act, cap, pat.as_ref())
                }
            }
            "-" => {
                // remove rule
                if ip {
                    self.rule_del_cidr(act, cap, pat.as_ref())
                } else {
                    self.rule_del_glob(act, cap, pat.as_ref())
                }
            }
            "^" => {
                // remove all matching rules
                if ip {
                    self.rule_rem_cidr(act, cap, pat.as_ref())
                } else {
                    self.rule_rem_glob(act, cap, pat.as_ref())
                }
            }
            _ => Err(Errno::EINVAL),
        }
    }

    /// Remove CIDR with port range, removes all matching instances.
    pub fn rule_rem_cidr(&mut self, act: Action, cap: Capability, pat: &str) -> Result<(), Errno> {
        let rule_rem = CidrRule {
            act,
            cap,
            pat: pat.parse::<AddressPattern>()?,
        };
        self.cidr_rules.retain(|rule| *rule != rule_rem);
        Ok(())
    }

    /// Remove CIDR with port range, removes the first instance from the end for predictability.
    pub fn rule_del_cidr(&mut self, act: Action, cap: Capability, pat: &str) -> Result<(), Errno> {
        let rule_del = CidrRule {
            act,
            cap,
            pat: pat.parse::<AddressPattern>()?,
        };
        if let Some(index) = self.cidr_rules.iter().rposition(|rule| *rule == rule_del) {
            self.cidr_rules.remove(index);
        }
        Ok(())
    }

    /// Add CIDR with port range
    /// The rule is either a Unix shell style pattern, or
    /// a network address, one of the following formats:
    ///
    /// 1. GLOB-PATTERN
    /// 2. IP/NETMASK!$PORT
    ///
    /// - GLOB-PATTERN must start with a slash, `/`.
    /// - /NETMASK may be omitted.
    /// - PORT is a single integer or two in format port1-port2
    pub fn rule_add_cidr(&mut self, act: Action, cap: Capability, pat: &str) -> Result<(), Errno> {
        self.cidr_rules.push(CidrRule {
            act,
            cap,
            pat: pat.parse::<AddressPattern>()?,
        });
        Ok(())
    }

    /// Remove Unix shell style pattern, removes all matching instances.
    pub fn rule_rem_glob(&mut self, act: Action, cap: Capability, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_rem_glob(act, cap, if pat.is_empty() { "/" } else { pat })?;
            self.rule_rem_glob(act, cap, &pat_doublestar)?;
            return Ok(());
        }

        self.glob_rules
            .retain(|rule| !(act == rule.act && rule.cap.contains(cap) && pat == rule.pat.glob()));
        self.del_glob(act, cap); // Rebuilds the whole GlobBuilder.

        Ok(())
    }

    /// Remove Unix shell style pattern, removes the first instance from the end for predictability.
    pub fn rule_del_glob(&mut self, act: Action, cap: Capability, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_del_glob(act, cap, if pat.is_empty() { "/" } else { pat })?;
            self.rule_del_glob(act, cap, &pat_doublestar)?;
            return Ok(());
        }

        for &capability in &Capability::GLOB {
            if cap.contains(capability) {
                if let Some((index, _)) =
                    self.glob_rules.iter().enumerate().rev().find(|(_, rule)| {
                        act == rule.act && capability == rule.cap && pat == rule.pat.glob()
                    })
                {
                    self.glob_rules.remove(index);
                    self.del_glob(act, capability); // Rebuilds the whole GlobBuilder.
                }
            }
        }

        Ok(())
    }

    /// Add Unix shell style pattern.
    pub fn rule_add_glob(&mut self, act: Action, cap: Capability, pat: &str) -> Result<(), Errno> {
        if pat.chars().nth(0) != Some('/') {
            return Err(Errno::EINVAL);
        }

        // Landlock rules are not glob patterns but path beneath rules.
        if !cap.intersects(Capability::CAP_LOCK) {
            // Expand foo/*** to [foo, foo/**]
            if let Some(pat) = pat.strip_suffix("/***") {
                let pat_doublestar = format!("{pat}/**");
                self.rule_add_glob(act, cap, if pat.is_empty() { "/" } else { pat })?;
                self.rule_add_glob(act, cap, &pat_doublestar)?;
                return Ok(());
            }
        }

        let pat = GlobBuilder::new(pat)
            .backslash_escape(true)
            .empty_alternates(true)
            .literal_separator(true)
            .build()
            .map_err(|_| Errno::EINVAL)?;
        for &capability in &Capability::GLOB {
            if cap.contains(capability) {
                let pat = pat.clone();
                self.add_glob(act, capability, &pat);
                self.glob_rules.push(GlobRule {
                    act,
                    cap: capability,
                    pat,
                });
            }
        }

        Ok(())
    }

    /// Remove Unix shell style pattern from SegvGuard filter globset,
    /// removes all matching instances.
    pub fn rule_rem_segvguard_filter(&mut self, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_rem_segvguard_filter(if pat.is_empty() { "/" } else { pat })?;
            self.rule_rem_segvguard_filter(&pat_doublestar)?;
            return Ok(());
        }

        self.segvguard_filter_rules
            .retain(|rule| pat != rule.0.glob());
        self.del_segvguard_filter(); // Rebuilds the whole GlobBuilder.

        Ok(())
    }

    /// Remove Unix shell style pattern from SegvGuard filter globset,
    /// removes the first instance from the end for predictability.
    pub fn rule_del_segvguard_filter(&mut self, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_del_segvguard_filter(if pat.is_empty() { "/" } else { pat })?;
            self.rule_del_segvguard_filter(&pat_doublestar)?;
            return Ok(());
        }

        if let Some((index, _)) = self
            .segvguard_filter_rules
            .iter()
            .enumerate()
            .find(|(_, rule)| pat == rule.0.glob())
        {
            self.segvguard_filter_rules.remove(index);
            self.del_segvguard_filter(); // Rebuilds the whole GlobBuilder.
        }

        Ok(())
    }

    /// Add Unix shell style pattern to SegvGuard filter globset.
    pub fn rule_add_segvguard_filter(&mut self, pat: &str) -> Result<(), Errno> {
        // Ensure absolute path.
        if pat.chars().nth(0) != Some('/') {
            return Err(Errno::EINVAL);
        }

        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_add_segvguard_filter(if pat.is_empty() { "/" } else { pat })?;
            self.rule_add_segvguard_filter(&pat_doublestar)?;
            return Ok(());
        }

        let pat = GlobBuilder::new(pat)
            .empty_alternates(true)
            .build()
            .map_err(|_| Errno::EINVAL)?;
        self.add_segvguard_filter(&pat);
        self.segvguard_filter_rules.push(Rule(pat.clone()));

        Ok(())
    }

    /// Remove Unix shell style pattern from crypt globset, removes all matching instances.
    pub fn rule_rem_crypt(&mut self, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_rem_crypt(if pat.is_empty() { "/" } else { pat })?;
            self.rule_rem_crypt(&pat_doublestar)?;
            return Ok(());
        }

        self.crypt_rules.retain(|rule| pat != rule.0.glob());
        self.del_crypt(); // Rebuilds the whole GlobBuilder.

        Ok(())
    }

    /// Remove Unix shell style pattern from crypt globset, removes the first instance from the end for predictability.
    pub fn rule_del_crypt(&mut self, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_del_crypt(if pat.is_empty() { "/" } else { pat })?;
            self.rule_del_crypt(&pat_doublestar)?;
            return Ok(());
        }

        if let Some((index, _)) = self
            .crypt_rules
            .iter()
            .enumerate()
            .find(|(_, rule)| pat == rule.0.glob())
        {
            self.crypt_rules.remove(index);
            self.del_crypt(); // Rebuilds the whole GlobBuilder.
        }

        Ok(())
    }

    /// Add Unix shell style pattern to crypt globset.
    pub fn rule_add_crypt(&mut self, pat: &str) -> Result<(), Errno> {
        // Ensure absolute path.
        if pat.chars().nth(0) != Some('/') {
            return Err(Errno::EINVAL);
        }

        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_add_crypt(if pat.is_empty() { "/" } else { pat })?;
            self.rule_add_crypt(&pat_doublestar)?;
            return Ok(());
        }

        let pat = GlobBuilder::new(pat)
            .empty_alternates(true)
            .build()
            .map_err(|_| Errno::EINVAL)?;
        self.add_crypt(&pat);
        self.crypt_rules.push(Rule(pat.clone()));

        Ok(())
    }

    /// Remove Unix shell style pattern from mask globset, removes all matching instances.
    pub fn rule_rem_mask(&mut self, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_rem_mask(if pat.is_empty() { "/" } else { pat })?;
            self.rule_rem_mask(&pat_doublestar)?;
            return Ok(());
        }

        self.mask_rules.retain(|rule| pat != rule.0.glob());
        self.del_mask(); // Rebuilds the whole GlobBuilder.

        Ok(())
    }

    /// Remove Unix shell style pattern from mask globset, removes the first instance from the end for predictability.
    pub fn rule_del_mask(&mut self, pat: &str) -> Result<(), Errno> {
        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_del_mask(if pat.is_empty() { "/" } else { pat })?;
            self.rule_del_mask(&pat_doublestar)?;
            return Ok(());
        }

        if let Some((index, _)) = self
            .mask_rules
            .iter()
            .enumerate()
            .find(|(_, rule)| pat == rule.0.glob())
        {
            self.mask_rules.remove(index);
            self.del_mask(); // Rebuilds the whole GlobBuilder.
        }

        Ok(())
    }

    /// Add Unix shell style pattern to mask globset.
    pub fn rule_add_mask(&mut self, pat: &str) -> Result<(), Errno> {
        // Ensure absolute path.
        if pat.chars().nth(0) != Some('/') {
            return Err(Errno::EINVAL);
        }

        // Expand foo/*** to [foo, foo/**]
        if let Some(pat) = pat.strip_suffix("/***") {
            let pat_doublestar = format!("{pat}/**");
            self.rule_add_mask(if pat.is_empty() { "/" } else { pat })?;
            self.rule_add_mask(&pat_doublestar)?;
            return Ok(());
        }

        let pat = GlobBuilder::new(pat)
            .empty_alternates(true)
            .build()
            .map_err(|_| Errno::EINVAL)?;
        self.add_mask(&pat);
        self.mask_rules.push(Rule(pat.clone()));

        Ok(())
    }

    /// Extract the Landlock read-only and read-write path lists.
    /// Returns None if Landlock sandboxing is disabled.
    pub fn collect_landlock(&mut self) -> Option<LandlockPathPair> {
        if !self.landlocked() {
            return None;
        }
        // Use HashSets to avoid duplicate paths.
        let mut path_ro = HashSet::new();
        let mut path_rw = HashSet::new();
        self.glob_rules
            .retain(|rule| match (rule.act, rule.cap, &rule.pat) {
                (Action::Allow, Capability::CAP_LOCK_RO, pat) => {
                    path_ro.insert(pat.glob().to_string());
                    false
                }
                (Action::Allow, Capability::CAP_LOCK_RW, pat) => {
                    path_rw.insert(pat.glob().to_string());
                    false
                }
                _ => true,
            });

        Some((
            path_ro.iter().cloned().collect(),
            path_rw.iter().cloned().collect(),
        ))
    }

    /// Check if the given path is hidden (ie denylisted for stat sandboxing)
    pub fn is_hidden<P: AsRef<Path>>(&self, path: P) -> bool {
        if self.enabled(Capability::CAP_STAT) {
            let (action, _) = self.check_path(Capability::CAP_STAT, &path);
            action.is_denying()
        } else {
            false
        }
    }

    /// Check IPv{4,6} address for access.
    pub(crate) fn check_ip(&self, cap: Capability, addr: IpAddr, port: u16) -> (Action, bool) {
        for rule in self.cidr_rules.iter().rev() {
            // rev() because last matching rule wins.
            if cap != rule.cap {
                continue;
            }
            if rule.act == Action::Filter {
                continue;
            }
            let port_match = if rule.pat.port[0] == rule.pat.port[1] {
                port == rule.pat.port[0]
            } else {
                port >= rule.pat.port[0] && port <= rule.pat.port[1]
            };
            if port_match && rule.pat.addr.contains(addr) {
                if matches!(rule.act, Action::Allow | Action::Filter) {
                    // Filter is only checked for !(Allow|Filter).
                    return (rule.act, true);
                } else {
                    return (rule.act, self.filter_ip(cap, &addr, port));
                }
            }
        }

        // If no specific rule is found, return based on capability
        // being enabled or not.
        if self.enabled(cap) {
            let action = self.default_action(cap);
            if matches!(action, Action::Allow | Action::Filter) {
                (action, true)
            } else {
                (self.default_action(cap), self.filter_ip(cap, &addr, port))
            }
        } else {
            (Action::Allow, true)
        }
    }

    /// Check UNIX socket for access.
    pub(crate) fn check_unix<P: AsRef<Path>>(&self, cap: Capability, path: P) -> (Action, bool) {
        // First, see if there's a matching allow or deny rule for the path.
        if let Some(action) = self.match_action(cap, &path) {
            if matches!(action, Action::Allow | Action::Filter) {
                // Filter is only checked for !(Allow|Filter).
                return (action, true);
            } else {
                return (action, self.filter_path(cap, &path));
            }
        }

        // If no specific rule is found, return based on capability being enabled or not.
        if self.enabled(cap) {
            let action = self.default_action(cap);
            if matches!(action, Action::Allow | Action::Filter) {
                (action, true)
            } else {
                (action, self.filter_path(cap, &path))
            }
        } else {
            (Action::Allow, true)
        }
    }

    /// Check many paths for TPE.
    pub fn check_tpe_many<I, P>(&self, paths: I) -> Action
    where
        I: IntoIterator<Item = P>,
        P: AsRef<Path>,
    {
        if !self.apply_tpe() {
            return Action::Allow;
        }

        // Collect parent directories in a HashSet to avoid repetitive
        // stats.
        let mut parent_set = HashSet::new();
        for path in paths {
            // Fail if the parent is not a directory.
            parent_set.insert(match path.as_ref().parent() {
                Some(parent) => parent.to_path_buf(),
                None => return self.tpe_act,
            });
        }

        // Lookup UID once as necessary.
        static UID: Lazy<nix::libc::uid_t> = Lazy::new(|| Uid::current().as_raw());
        for parent in &parent_set {
            // SAFETY: Do not resolve symbolic links to avoid TOCTOU.
            let fd = match safe_open_path(parent, OFlag::O_DIRECTORY) {
                Ok(fd) => fd,
                Err(_) => return self.tpe_act,
            };
            let md = match fstat(fd.as_raw_fd()) {
                Ok(md) => md,
                Err(_) => return self.tpe_act,
            };

            // Check if the parent directory is root/user owned.
            // Check if the parent directory is writable only by the owner.
            let is0 = md.st_uid == 0;
            if (self.tpe_user_owned() && !is0 && md.st_uid != *UID)
                || (self.tpe_root_owned() && !is0)
                || (md.st_mode & (nix::libc::S_IWGRP | nix::libc::S_IWOTH) != 0)
            {
                return self.tpe_act;
            }
        }

        Action::Allow
    }

    /// Check path for TPE.
    pub(crate) fn check_tpe<P: AsRef<Path>>(&self, path: P) -> Action {
        if !self.apply_tpe() {
            return Action::Allow;
        }

        // Get the parent directory.
        let parent = match path.as_ref().parent() {
            Some(parent) => parent,
            None => return self.tpe_act,
        };

        // SAFETY:
        // 1. Do not resolve symbolic links.
        // 2. Fail if the parent is not a directory.
        // Both of these are necessary to avoid TOCTOU.
        let fd = match safe_open_path(parent, OFlag::O_DIRECTORY) {
            Ok(fd) => fd,
            Err(_) => return self.tpe_act,
        };
        let md = match fstat(fd.as_raw_fd()) {
            Ok(md) => md,
            Err(_) => return self.tpe_act,
        };

        // Check if the parent directory is root/user owned.
        // Check if the parent directory is writable only by the owner.
        let is0 = md.st_uid == 0;
        if (self.tpe_user_owned() && !is0 && md.st_uid != Uid::current().as_raw())
            || (self.tpe_root_owned() && !is0)
            || (md.st_mode & (nix::libc::S_IWGRP | nix::libc::S_IWOTH) != 0)
        {
            self.tpe_act
        } else {
            Action::Allow
        }
    }

    /// Check if TPE should be applied.
    fn apply_tpe(&self) -> bool {
        // Check if TPE should be applied.
        if let Some(tpe_gid) = self.tpe_gid {
            let egid = Gid::effective();
            let apply = if egid == tpe_gid {
                !self.tpe_negate()
            } else {
                match getgroups() {
                    Ok(gids) => {
                        if self.tpe_negate() {
                            !gids.contains(&tpe_gid)
                        } else {
                            gids.contains(&tpe_gid)
                        }
                    }
                    Err(_) => return true,
                }
            };

            if !apply {
                return false;
            }
        } // TPE is applied if tpe_gid=None.

        true
    }

    /// Check path for access.
    pub(crate) fn check_path<P: AsRef<Path>>(&self, cap: Capability, path: P) -> (Action, bool) {
        // SAFETY: Deny access to paths with control characters when
        // writing. They have little to no valid usecases and an
        // attacker could use paths with control characters in them to
        // make an attack against user's terminal.
        if cap.can_write() && !self.allow_unsafe_filename() && path_has_control_chars(&path) {
            // Use Action::Filter as default so access gets denied
            // even if sandbox is not enabled for the given capability.
            return self.check_path_nomatch(cap, &path, Action::Filter);
        }

        // Drop trailing slash which can cause inconsistencies with expectations.
        let path_bytes = path.as_ref().as_os_str().as_bytes();
        #[allow(clippy::arithmetic_side_effects)]
        let path = if path_bytes.ends_with(&[b'/']) && path_bytes.len() > 1 {
            // SAFETY: Since we're operating on valid path bytes, getting a slice is safe.
            // This excludes the root path "/" to avoid turning it into an empty path.
            PathBuf::from(OsStr::from_bytes(&path_bytes[..path_bytes.len() - 1]))
        } else {
            path.as_ref().to_path_buf()
        };

        // First, see if there's a matching allow or deny rule for the path.
        if let Some(action) = self.match_action(cap, &path) {
            match action {
                Action::Allow | Action::Filter => {
                    // Filter is only checked for !(Allow|Filter).
                    return (action, true);
                }
                Action::Deny if cap == Capability::CAP_STAT => {
                    // Filter is not checked if denying for stat (==hiding).
                    return (action, true);
                }
                action => {
                    // If the action is !(Allow|Filter), then we must check if it's filtered.
                    return (action, self.filter_path(cap, &path));
                }
            }
        }

        // If no specific rule is found, return based on capability being enabled or not.
        self.check_path_nomatch(cap, &path, Action::Allow)
    }

    // If no specific rule is found, return based on capability being enabled or not.
    fn check_path_nomatch<P: AsRef<Path>>(
        &self,
        cap: Capability,
        path: P,
        def: Action,
    ) -> (Action, bool) {
        if self.enabled(cap) {
            let action = self.default_action(cap);
            match action {
                Action::Allow | Action::Filter => {
                    // Filter is only checked for !(Allow|Filter).
                    (action, true)
                }
                Action::Deny if cap == Capability::CAP_STAT => {
                    // Filter is not checked if denying for stat (==hiding).
                    (action, true)
                }
                action => {
                    // If the action is !(Allow|Filter), then we must check if it's filtered.
                    (action, self.filter_path(cap, &path))
                }
            }
        } else {
            (def, true)
        }
    }

    /// Find a matching action (Allow or Deny) for the given path.
    pub fn match_action<P: AsRef<Path>>(&self, cap: Capability, path: P) -> Option<Action> {
        if let Some((set, _, ready)) = self.get_globset(Action::Deny, cap) {
            // matches is a Vec<usize> which has a
            // sequence number of every matching pattern
            // and last matching patterns wins.
            assert!(ready, "match_action called before building globsets!");
            set.matches(&path).last().map(|idx| {
                self.glob_rules
                    .iter()
                    .filter(|rule| cap == rule.cap && rule.act != Action::Filter)
                    .nth(*idx)
                    .map(|rule| rule.act)
                    .unwrap_or(self.default_action(cap))
            })
        } else {
            None
        }
    }

    /// Check if the given path should be encrypted.
    pub fn is_crypt<P: AsRef<Path>>(&self, path: P) -> bool {
        // matches is a Vec<usize> which has a
        // sequence number of every matching pattern.
        assert!(
            self.globset_crypt.2,
            "is_crypt called before building the mask globset!"
        );

        self.globset_crypt
            .0
            .matches(&path)
            .first()
            .map_or(false, |idx| self.crypt_rules.get(*idx).is_some())
    }

    /// Check if the given path is masked.
    pub fn is_masked<P: AsRef<Path>>(&self, path: P) -> bool {
        // matches is a Vec<usize> which has a
        // sequence number of every matching pattern.
        assert!(
            self.globset_mask.2,
            "is_masked called before building the mask globset!"
        );

        self.globset_mask
            .0
            .matches(&path)
            .first()
            .map_or(false, |idx| self.mask_rules.get(*idx).is_some())
    }

    /// Check if the ip address with the given port is filtered.
    fn filter_ip(&self, cap: Capability, addr: &IpAddr, port: u16) -> bool {
        self.cidr_rules
            .iter()
            .filter(|filter| filter.act == Action::Filter && filter.cap == cap)
            .any(|filter| {
                let port_match = if filter.pat.port[0] == filter.pat.port[1] {
                    port == filter.pat.port[0]
                } else {
                    port >= filter.pat.port[0] && port <= filter.pat.port[1]
                };
                port_match && filter.pat.addr.contains(*addr)
            })
    }

    /// Check if the path is filtered.
    fn filter_path<P: AsRef<Path>>(&self, cap: Capability, path: P) -> bool {
        let path_ref = path.as_ref();

        if cap.contains(Capability::CAP_READ) && self.globset_filter_r.0.is_match(path_ref) {
            return true;
        }
        if cap.contains(Capability::CAP_WRITE) && self.globset_filter_w.0.is_match(path_ref) {
            return true;
        }
        if cap.contains(Capability::CAP_EXEC) && self.globset_filter_x.0.is_match(path_ref) {
            return true;
        }
        if cap.contains(Capability::CAP_IOCTL) && self.globset_filter_i.0.is_match(path_ref) {
            return true;
        }
        if cap.contains(Capability::CAP_BIND) && self.globset_filter_nb.0.is_match(path_ref) {
            return true;
        }
        if cap.contains(Capability::CAP_CONNECT) && self.globset_filter_nc.0.is_match(path_ref) {
            return true;
        }

        false
    }

    /// Build `GlobSet`s from `GlobRule`s.
    /// This function must be called once before check functions.
    pub fn build_globsets(&mut self) -> Result<(), globset::Error> {
        fn build_globset(globset_data: &mut SandboxGlobSet) -> Result<(), globset::Error> {
            if !globset_data.2 {
                globset_data.0 = globset_data.1.clone().build()?;
                globset_data.2 = true;
            }
            Ok(())
        }

        build_globset(&mut self.globset_access_r)?;
        build_globset(&mut self.globset_access_s)?;
        build_globset(&mut self.globset_access_w)?;
        build_globset(&mut self.globset_access_x)?;
        build_globset(&mut self.globset_access_i)?;
        build_globset(&mut self.globset_access_nb)?;
        build_globset(&mut self.globset_access_nc)?;
        build_globset(&mut self.globset_filter_r)?;
        build_globset(&mut self.globset_filter_s)?;
        build_globset(&mut self.globset_filter_w)?;
        build_globset(&mut self.globset_filter_x)?;
        build_globset(&mut self.globset_filter_i)?;
        build_globset(&mut self.globset_filter_nb)?;
        build_globset(&mut self.globset_filter_nc)?;
        build_globset(&mut self.globset_crypt)?;
        build_globset(&mut self.globset_segvguard_filter)?;
        build_globset(&mut self.globset_mask)?;

        Ok(())
    }

    fn add_glob(&mut self, act: Action, cap: Capability, glob: &Glob) {
        if let Some(set) = self.get_globset_mut(act, cap) {
            set.1.add(glob.clone());
            set.2 = false;
        }
    }

    fn del_glob(&mut self, act: Action, cap: Capability) {
        // Temporarily take the rules out
        let rules = std::mem::take(&mut self.glob_rules);

        if let Some(set) = self.get_globset_mut(act, cap) {
            let mut glob_set = GlobSetBuilder::new();
            for rule in rules.iter().filter(|rule| {
                if !rule.cap.contains(cap) {
                    false
                } else {
                    match rule.act {
                        Action::Filter => act == Action::Filter,
                        Action::Kill => act == Action::Kill,
                        _ => act == Action::Allow || act == Action::Deny,
                    }
                }
            }) {
                glob_set.add(rule.pat.clone());
            }

            set.1 = glob_set;
            set.2 = false;
        }

        // Swap the rules back in
        self.glob_rules = rules;
    }

    fn get_globset(&self, act: Action, cap: Capability) -> Option<&SandboxGlobSet> {
        Some(match (cap, act) {
            (Capability::CAP_READ, Action::Filter) => &self.globset_filter_r,
            (Capability::CAP_READ, _) => &self.globset_access_r,
            (Capability::CAP_STAT, Action::Filter) => &self.globset_filter_s,
            (Capability::CAP_STAT, _) => &self.globset_access_s,
            (Capability::CAP_WRITE, Action::Filter) => &self.globset_filter_w,
            (Capability::CAP_WRITE, _) => &self.globset_access_w,
            (Capability::CAP_EXEC, Action::Filter) => &self.globset_filter_x,
            (Capability::CAP_EXEC, _) => &self.globset_access_x,
            (Capability::CAP_IOCTL, Action::Filter) => &self.globset_filter_i,
            (Capability::CAP_IOCTL, _) => &self.globset_access_i,
            (Capability::CAP_BIND, Action::Filter) => &self.globset_filter_nb,
            (Capability::CAP_BIND, _) => &self.globset_access_nb,
            (Capability::CAP_CONNECT, Action::Filter) => &self.globset_filter_nc,
            (Capability::CAP_CONNECT, _) => &self.globset_access_nc,
            _ => {
                /* Landlock et al. have no GlobSet. */
                return None;
            }
        })
    }

    fn get_globset_mut(&mut self, act: Action, cap: Capability) -> Option<&mut SandboxGlobSet> {
        Some(match (cap, act) {
            (Capability::CAP_READ, Action::Filter) => &mut self.globset_filter_r,
            (Capability::CAP_READ, _) => &mut self.globset_access_r,
            (Capability::CAP_STAT, Action::Filter) => &mut self.globset_filter_s,
            (Capability::CAP_STAT, _) => &mut self.globset_access_s,
            (Capability::CAP_WRITE, Action::Filter) => &mut self.globset_filter_w,
            (Capability::CAP_WRITE, _) => &mut self.globset_access_w,
            (Capability::CAP_EXEC, Action::Filter) => &mut self.globset_filter_x,
            (Capability::CAP_EXEC, _) => &mut self.globset_access_x,
            (Capability::CAP_IOCTL, Action::Filter) => &mut self.globset_filter_i,
            (Capability::CAP_IOCTL, _) => &mut self.globset_access_i,
            (Capability::CAP_BIND, Action::Filter) => &mut self.globset_filter_nb,
            (Capability::CAP_BIND, _) => &mut self.globset_access_nb,
            (Capability::CAP_CONNECT, Action::Filter) => &mut self.globset_filter_nc,
            (Capability::CAP_CONNECT, _) => &mut self.globset_access_nc,
            _ => {
                /* Landlock et al. have no GlobSet. */
                return None;
            }
        })
    }

    /// Returns the default action for the given capability.
    pub(crate) fn default_action(&self, cap: Capability) -> Action {
        match cap {
            Capability::CAP_READ => self.read_act,
            Capability::CAP_WRITE => self.write_act,
            Capability::CAP_STAT => self.stat_act,
            Capability::CAP_EXEC => self.exec_act,
            Capability::CAP_IOCTL => self.ioctl_act,
            Capability::CAP_BIND => self.bind_act,
            Capability::CAP_CONNECT => self.connect_act,
            Capability::CAP_MEM => self.mem_act,
            Capability::CAP_PID => self.pid_act,
            Capability::CAP_TPE => self.tpe_act,
            Capability::CAP_FORCE => self.force_act,
            _ => unreachable!(),
        }
    }

    /// Return true if Proxy is configured.
    pub fn has_proxy(&self) -> bool {
        self.enabled(Capability::CAP_PROXY)
    }

    /// Initialize the KCAPI connection.
    pub(crate) fn set_crypt(&mut self) -> Result<(), Errno> {
        if self.enabled(Capability::CAP_CRYPT) {
            match &self.crypt_id {
                Some(Secret::Fd(_)) => {}
                Some(Secret::Key(key)) => {
                    self.crypt_id = Some(Secret::Fd(aes_ctr_setup(key)?));
                }
                _ => return Err(Errno::ENOKEY),
            }
        }
        Ok(())
    }

    /// Check if request belongs to the denylist.
    pub(crate) fn has_ioctl(&self, request: &u64) -> bool {
        self.ioctl_set.contains(request)
    }

    /// Add an ioctl request to the denylist.
    fn add_ioctl(&mut self, request: u64) {
        self.ioctl_set.insert(request);
        if let Some(request) = extend_ioctl(request) {
            // musl compat, see documentation of extend_ioctl().
            self.ioctl_set.insert(request);
        }
    }

    fn add_crypt(&mut self, glob: &Glob) {
        self.globset_crypt.1.add(glob.clone());
        self.globset_crypt.2 = false;
    }

    fn del_crypt(&mut self) {
        // Temporarily take the rules out
        let rules = std::mem::take(&mut self.crypt_rules);

        let mut glob_set = GlobSetBuilder::new();
        for rule in &rules {
            glob_set.add(rule.0.clone());
        }
        self.globset_crypt.1 = glob_set;
        self.globset_crypt.2 = false;

        // Swap the rules back in
        self.crypt_rules = rules;
    }

    fn add_segvguard_filter(&mut self, glob: &Glob) {
        self.globset_segvguard_filter.1.add(glob.clone());
        self.globset_segvguard_filter.2 = false;
    }

    fn del_segvguard_filter(&mut self) {
        // Temporarily take the rules out
        let rules = std::mem::take(&mut self.segvguard_filter_rules);

        let mut glob_set = GlobSetBuilder::new();
        for rule in &rules {
            glob_set.add(rule.0.clone());
        }
        self.globset_segvguard_filter.1 = glob_set;
        self.globset_segvguard_filter.2 = false;

        // Swap the rules back in
        self.segvguard_filter_rules = rules;
    }

    fn add_mask(&mut self, glob: &Glob) {
        self.globset_mask.1.add(glob.clone());
        self.globset_mask.2 = false;
    }

    fn del_mask(&mut self) {
        // Temporarily take the rules out
        let rules = std::mem::take(&mut self.mask_rules);

        let mut glob_set = GlobSetBuilder::new();
        for rule in &rules {
            glob_set.add(rule.0.clone());
        }
        self.globset_mask.1 = glob_set;
        self.globset_mask.2 = false;

        // Swap the rules back in
        self.mask_rules = rules;
    }

    /// Get the process ID of the syd execve child.
    pub fn get_child_pid(&self) -> Pid {
        Pid::from_raw(self.cpid)
    }

    /// Set the process ID of the syd execve child.
    pub(crate) fn set_child(&mut self, pid: Pid, pid_fd: RawFd) {
        // SAFETY: Set the sandbox lock if the state is unspecified.
        // This is safer than the previous default LockState::Exec.
        // We set this post-exec to ensure the initial configuration
        // passes through (ie config file and CLI options).
        if self.lock.is_none() {
            // !self.is_running -> lock returns no errors.
            let _ = self.lock(LockState::Set);
        }

        self.cpid = pid.as_raw();
        self.fpid = pid_fd;
    }

    /// Returns true if syd execve child is still alive.
    fn child_is_alive(&self) -> bool {
        self.fpid == nix::libc::AT_FDCWD
            // SAFETY: fpid is a valid PID file descriptor.
            || unsafe { nix::libc::syscall(nix::libc::SYS_pidfd_send_signal, self.fpid, 0, 0, 0) }
                == 0
    }

    /// Return true of the sandbox is running.
    fn is_running(&self) -> bool {
        !self.fpid.is_negative()
    }

    /// Get map root flag.
    pub fn map_root(&self) -> bool {
        self.flags.contains(Flag::FL_MAP_ROOT)
    }

    /// Get fake root flag.
    pub fn fake_root(&self) -> bool {
        self.flags.contains(Flag::FL_FAKE_ROOT)
    }

    /// Return the enabled capabilities out of the given set of capabilities.
    pub fn getcaps(&self, caps: Capability) -> Capability {
        self.state & caps
    }

    /// Return true if the sandboxing is enabled for the given capability.
    pub fn enabled(&self, cap: Capability) -> bool {
        self.enabled_all(cap)
    }

    /// Return true if the sandboxing is enabled for the given capability.
    pub fn enabled_all(&self, caps: Capability) -> bool {
        self.state.contains(caps)
    }

    /// Return true if any of the sandboxing capabilities is enabled.
    pub fn enabled_any(&self, caps: Capability) -> bool {
        self.state.intersects(caps)
    }

    /// Lock sandbox.
    pub fn lock(&mut self, state: LockState) -> Result<(), Errno> {
        match state {
            LockState::Set if self.lock != Some(LockState::Set) => {
                self.lock = Some(LockState::Set);

                if self.is_running() {
                    // SAFETY: Sandbox lock is locked during runtime, which
                    // means execve(2) was not prevented for the Syd process
                    // at startup. Now that we don't need it anymore, we
                    // apply a seccomp filter here to prevent it.
                    let mut ctx = ScmpFilterContext::new_filter(ScmpAction::Allow)
                        .map_err(|e| e.errno().map(seccomp2errno).unwrap_or(Errno::EOPNOTSUPP))?;
                    // We kill for bad system call and bad arch.
                    let _ = ctx.set_act_badarch(ScmpAction::KillProcess);
                    // Synchronize filter to all threads.
                    let _ = ctx.set_ctl_tsync(true);

                    for sysname in ["execve", "execveat"] {
                        // SAFETY: Be swift and kill process as this attempt is most
                        // certainly malicious and the kill action cannot be misused
                        // to DOS the syd process.
                        ctx.add_rule(ScmpAction::KillProcess, ScmpSyscall::new(sysname))
                            .map_err(|e| {
                                e.errno().map(seccomp2errno).unwrap_or(Errno::EOPNOTSUPP)
                            })?;
                    }

                    // Load the filter.
                    ctx.load()
                        .map_err(|e| e.errno().map(seccomp2errno).unwrap_or(Errno::EOPNOTSUPP))?;
                }
            }
            _ => self.lock = Some(state),
        }

        Ok(())
    }

    /// Returns true if the sandbox is locked.
    pub fn locked(&self) -> bool {
        self.lock == Some(LockState::Set)
    }

    /// Returns true if the sandbox is locked for the given process ID.
    pub fn locked_for_pid(&self, pid: nix::libc::pid_t) -> bool {
        match self.lock {
            None => false, // same state as cpid==0.
            Some(LockState::Off) => false,
            Some(LockState::Exec) if self.cpid == 0 => false,
            Some(LockState::Exec) if self.cpid == pid => !self.child_is_alive(),
            _ => true,
        }
    }

    /// Returns true if Landlock sandboxing is enabled.
    pub fn landlocked(&self) -> bool {
        self.state.contains(Capability::CAP_LOCK)
    }

    /// Get the namespace settings of the Sandbox.
    pub fn namespaces(&self) -> nix::libc::c_int {
        let mut namespaces = 0;

        if self.flags.contains(Flag::FL_UNSHARE_MOUNT) {
            namespaces |= nix::libc::CLONE_NEWNS;
        }
        if self.flags.contains(Flag::FL_UNSHARE_UTS) {
            namespaces |= nix::libc::CLONE_NEWUTS;
        }
        if self.flags.contains(Flag::FL_UNSHARE_IPC) {
            namespaces |= nix::libc::CLONE_NEWIPC;
        }
        if self.flags.contains(Flag::FL_UNSHARE_USER) {
            namespaces |= nix::libc::CLONE_NEWUSER;
        }
        if self.flags.contains(Flag::FL_UNSHARE_PID) {
            namespaces |= nix::libc::CLONE_NEWPID;
        }
        if self.flags.contains(Flag::FL_UNSHARE_NET) {
            namespaces |= nix::libc::CLONE_NEWNET;
        }
        if self.flags.contains(Flag::FL_UNSHARE_CGROUP) {
            namespaces |= nix::libc::CLONE_NEWCGROUP;
        }
        if self.flags.contains(Flag::FL_UNSHARE_TIME) {
            namespaces |= CLONE_NEWTIME;
        }

        namespaces
    }

    /// Get the allowed namespaces of the Sandbox.
    pub(crate) fn allowed_namespaces(&self) -> nix::libc::c_int {
        let mut namespaces = 0;

        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_MOUNT) {
            namespaces |= nix::libc::CLONE_NEWNS;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_UTS) {
            namespaces |= nix::libc::CLONE_NEWUTS;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_IPC) {
            namespaces |= nix::libc::CLONE_NEWIPC;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_USER) {
            namespaces |= nix::libc::CLONE_NEWUSER;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_PID) {
            namespaces |= nix::libc::CLONE_NEWPID;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_NET) {
            namespaces |= nix::libc::CLONE_NEWNET;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_CGROUP) {
            namespaces |= nix::libc::CLONE_NEWCGROUP;
        }
        if self.flags.contains(Flag::FL_ALLOW_UNSAFE_UNSHARE_TIME) {
            namespaces |= CLONE_NEWTIME;
        }

        namespaces
    }

    /// Get the denied namespaces of the Sandbox.
    pub(crate) fn denied_namespaces(&self) -> nix::libc::c_int {
        let allowed = self.allowed_namespaces();
        NAMESPACE_FLAGS_ALL & !allowed
    }

    /// Set domainname, error if sandboxing is running.
    pub fn set_domainname(&mut self, value: &str) -> Result<(), Errno> {
        if self.is_running() {
            Err(Errno::EBUSY)
        } else {
            self.domainname = value.to_string();
            Ok(())
        }
    }

    /// Set hostname, error if sandboxing is running.
    pub fn set_hostname(&mut self, value: &str) -> Result<(), Errno> {
        if self.is_running() {
            Err(Errno::EBUSY)
        } else {
            self.hostname = value.to_string();
            Ok(())
        }
    }

    /// Get the value of the unshare-mount flag.
    pub fn unshare_mount(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_MOUNT)
    }

    /// Set the value of the unshare-mount flag.
    pub fn set_unshare_mount(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_MOUNT);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_MOUNT);
        }
    }

    /// Get the value of the unshare-uts flag.
    pub fn unshare_uts(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_UTS)
    }

    /// Set the value of the unshare-uts flag.
    pub fn set_unshare_uts(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_UTS);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_UTS);
        }
    }

    /// Get the value of the unshare-ipc flag.
    pub fn unshare_ipc(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_IPC)
    }

    /// Set the value of the unshare-ipc flag.
    pub fn set_unshare_ipc(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_IPC);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_IPC);
        }
    }

    /// Get the value of the unshare-user flag.
    pub fn unshare_user(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_USER)
    }

    /// Set the value of the unshare-user flag.
    pub fn set_unshare_user(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_USER);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_USER);
        }
    }

    /// Get the value of the unshare-pid flag.
    pub fn unshare_pid(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_PID)
    }

    /// Set the value of the unshare-pid flag.
    pub fn set_unshare_pid(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_PID);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_PID);
        }
    }

    /// Get the value of the unshare-net flag.
    pub fn unshare_net(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_NET)
    }

    /// Set the value of the unshare-net flag.
    pub fn set_unshare_net(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_NET);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_NET);
        }
    }

    /// Get the value of the unshare-cgroup flag.
    pub fn unshare_cgroup(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_CGROUP)
    }

    /// Set the value of the unshare-cgroup flag.
    pub fn set_unshare_cgroup(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_CGROUP);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_CGROUP);
        }
    }

    /// Get the value of the unshare-time flag.
    pub fn unshare_time(&self) -> bool {
        self.flags.contains(Flag::FL_UNSHARE_TIME)
    }

    /// Set the value of the unshare-time flag.
    pub fn set_unshare_time(&mut self, state: bool) {
        if state {
            self.flags.insert(Flag::FL_UNSHARE_TIME);
        } else {
            self.flags.remove(Flag::FL_UNSHARE_TIME);
        }
    }

    /// Returns true if exit wait flag is set.
    pub fn exit_wait_all(&self) -> bool {
        self.flags.contains(Flag::FL_EXIT_WAIT_ALL)
    }

    /// Returns the TPE Gid.
    pub fn tpe_gid(&self) -> Option<Gid> {
        self.tpe_gid
    }

    /// Returns true if tpe negate flag is set.
    pub fn tpe_negate(&self) -> bool {
        self.flags.contains(Flag::FL_TPE_NEGATE)
    }

    /// Returns true if tpe root owned flag is set.
    pub fn tpe_root_owned(&self) -> bool {
        self.flags.contains(Flag::FL_TPE_ROOT_OWNED)
    }

    /// Returns true if tpe user owned flag is set.
    pub fn tpe_user_owned(&self) -> bool {
        self.flags.contains(Flag::FL_TPE_USER_OWNED)
    }

    /// Returns true if force cloexec flag is set.
    pub fn force_cloexec(&self) -> bool {
        self.flags.contains(Flag::FL_FORCE_CLOEXEC)
    }

    /// Returns true if deny dotdot flag is set.
    pub fn deny_dotdot(&self) -> bool {
        self.flags.contains(Flag::FL_DENY_DOTDOT)
    }

    /// Returns true if deny elf32 flag is set.
    pub fn deny_elf32(&self) -> bool {
        self.flags.contains(Flag::FL_DENY_ELF32)
    }

    /// Returns true if deny elf-dynamic flag is set.
    pub fn deny_elf_dynamic(&self) -> bool {
        self.flags.contains(Flag::FL_DENY_ELF_DYNAMIC)
    }

    /// Returns true if deny elf-static flag is set.
    pub fn deny_elf_static(&self) -> bool {
        self.flags.contains(Flag::FL_DENY_ELF_STATIC)
    }

    /// Returns true if deny script flag is set.
    pub fn deny_script(&self) -> bool {
        self.flags.contains(Flag::FL_DENY_SCRIPT)
    }

    /// Returns true if CAP_NET_BIND_SERVICE capability is retained.
    pub fn allow_unsafe_bind(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_BIND)
    }

    /// Allow system calls for CPU emulation functionality.
    pub fn allow_unsafe_cpu(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_CPU)
    }

    /// Returns true if exec-TOCTOU mitigator is disabled.
    pub fn allow_unsafe_exec(&self) -> bool {
        // trace/allow_unsafe_debug implies trace/allow_unsafe_{exec,memory,perf,ptrace}
        self.flags
            .intersects(Flag::FL_ALLOW_UNSAFE_DEBUG | Flag::FL_ALLOW_UNSAFE_EXEC)
    }

    /// Returns true if access to the Kernel keyring is allowed.
    pub fn allow_unsafe_keyring(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_KEYRING)
    }

    /// Returns true if unsafe memory flag is set.
    pub fn allow_unsafe_memory(&self) -> bool {
        // trace/allow_unsafe_debug implies trace/allow_unsafe_{exec,memory,perf,ptrace}
        self.flags
            .intersects(Flag::FL_ALLOW_UNSAFE_DEBUG | Flag::FL_ALLOW_UNSAFE_MEMORY)
    }

    /// Returns true if system calls used for memory protection keys are allowed.
    pub fn allow_unsafe_pkey(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_PKEY)
    }

    /// Returns true if unsafe capabilities flag is set.
    pub fn allow_unsafe_caps(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_CAPS)
    }

    /// Returns true if unsafe environment flag is set.
    pub fn allow_unsafe_env(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_ENV)
    }

    /// Returns true if successful bind addresses should be allowed for subsequent connect calls.
    pub fn allow_safe_bind(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_SAFE_BIND)
    }

    /// Returns true if access to the Linux kernel crypto API is allowed.
    pub fn allow_safe_kcapi(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_SAFE_KCAPI)
    }

    /// Returns true if safe setid flag is set.
    pub fn allow_safe_setid(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_SAFE_SETID)
    }

    /// Returns true if:
    /// 1. Non-elf interpreters are allowed via binfmt (wine, java etc.)
    /// 2. Execution shared libraries directly is allowed  (ldd, ld-linux.so.1 etc.).
    pub fn allow_unsupp_exec(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSUPP_EXEC)
    }

    /// Returns true if unsupported socket families should be allowed.
    pub fn allow_unsupp_socket(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET)
    }

    /// Returns true if raw sockets should be allowed.
    pub fn allow_unsafe_socket(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_SOCKET)
    }

    /// Returns true if unsafe syslog should be allowed.
    pub fn allow_unsafe_syslog(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_SYSLOG)
    }

    /// Returns true if unsafe prctl calls are allowed.
    pub fn allow_unsafe_prctl(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_PRCTL)
    }

    /// Returns true if unsafe prlimit calls are allowed.
    pub fn allow_unsafe_prlimit(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_PRLIMIT)
    }

    /// Returns true if unsafe debug (perf, ptrace etc) are allowed.
    pub fn allow_unsafe_debug(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_DEBUG)
    }

    /// Returns true if unsafe perf calls are allowed.
    pub fn allow_unsafe_perf(&self) -> bool {
        // Note, we keep the unsafe_p{erf,trace} intact
        // outside this module such that in the future
        // they'll be easy to split again if need be.
        self.allow_unsafe_debug()
    }

    /// Returns true if unsafe ptrace calls are allowed.
    pub fn allow_unsafe_ptrace(&self) -> bool {
        // See the note to allow_unsafe_perf.
        self.flags
            .intersects(Flag::FL_ALLOW_UNSAFE_DEBUG | Flag::FL_ALLOW_UNSAFE_EXEC)
    }

    /// Returns true if unsafe adjtime calls are allowed.
    /// This also results in keeping the CAP_SYS_TIME capability.
    pub fn allow_unsafe_time(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_TIME)
    }

    /// Returns true if CONTINUE should be applied on O_PATH file
    /// descriptors rather than turning them into O_RDONLY.
    pub fn allow_unsafe_open(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_OPEN)
    }

    /// Returns true if filenames with control characters are allowed.
    pub fn allow_unsafe_filename(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_FILENAME)
    }

    /// Returns true if unsafe io_uring calls are allowed.
    pub fn allow_unsafe_iouring(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_IOURING)
    }

    /// Returns true if the deprecated tkill call is allowed.
    pub fn allow_unsafe_kill(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_KILL)
    }

    /// Returns true if secret memfds and executable memfds are allowed.
    pub fn allow_unsafe_memfd(&self) -> bool {
        self.flags.contains(Flag::FL_ALLOW_UNSAFE_MEMFD)
    }

    /// Returns true if personality(2) changes are locked.
    pub fn lock_personality(&self) -> bool {
        self.flags.contains(Flag::FL_LOCK_PERSONALITY)
    }

    /// Returns true if reading the timestamp counter is denied (x86 only).
    pub fn deny_tsc(&self) -> bool {
        self.flags.contains(Flag::FL_DENY_TSC)
    }

    /// Get SegvGuard entry expiry timeout.
    pub fn get_segvguard_expiry(&self) -> Duration {
        self.segvguard_expiry
    }

    /// Set SegvGuard entry expiry timeout.
    pub fn set_segvguard_expiry(&mut self, timeout: Duration) {
        self.segvguard_expiry = timeout;
    }

    /// SegvGuard entry suspension timeout.
    pub fn get_segvguard_suspension(&self) -> Duration {
        self.segvguard_suspension
    }

    /// SegvGuard entry suspension timeout.
    pub fn set_segvguard_suspension(&mut self, timeout: Duration) {
        self.segvguard_suspension = timeout;
    }

    /// Get SegvGuard max number of crashes before expiry.
    pub fn get_segvguard_maxcrashes(&self) -> u8 {
        self.segvguard_maxcrashes
    }

    /// Set SegvGuard max number of crashes before expiry.
    pub fn set_segvguard_maxcrashes(&mut self, maxcrashes: u8) {
        self.segvguard_maxcrashes = maxcrashes;
    }

    /// Record a crash in the SegvGuard map.
    /// Returns the current number of crashes for the given path,
    /// and a boolean which specify whether the number exceeded
    /// currently configured maxcrashes value.
    pub fn add_segvguard_crash(&mut self, path: PathBuf) -> (bool, u8) {
        match self.segvguard_map.entry(path.clone()) {
            crate::endorphin::map::Entry::Occupied(mut entry) => {
                let value = entry.get_mut();
                *value = value.saturating_add(1);
                let exceeded = *value >= self.segvguard_maxcrashes;
                if exceeded {
                    self.segvguard_set.insert(path, self.segvguard_suspension);
                }
                (exceeded, *value)
            }
            crate::endorphin::map::Entry::Vacant(entry) => {
                entry.insert(1, self.segvguard_expiry);
                let exceeded = self.segvguard_maxcrashes <= 1;
                if exceeded {
                    self.segvguard_set.insert(path, self.segvguard_suspension);
                }
                (exceeded, 1)
            }
        }
    }

    /// Check SegvGuard for access.
    pub fn check_segvguard<P: AsRef<Path>>(&self, path: P) -> Option<(Action, bool)> {
        // Setting SegvGuard default action to Allow disables SegvGuard.
        // Setting expiry timeout to 0 disables SegvGuard.
        if self.segvguard_act == Action::Allow
            || self.segvguard_expiry.is_zero()
            || !self.segvguard_set.contains(path.as_ref())
        {
            None
        } else if self.segvguard_act == Action::Filter {
            // Filter for path needn't be checked if default action is Filter.
            Some((Action::Filter, true))
        } else {
            Some((
                self.segvguard_act,
                self.globset_segvguard_filter.0.is_match(path.as_ref()),
            ))
        }
    }

    /// Return true if there're any SegvGuard suspended paths.
    pub fn has_segvguard(&self) -> bool {
        !self.segvguard_set.is_empty()
    }

    /// Add a UID transit.
    /// If the sandbox did not previously contain this transition, Ok(true) is returned.
    /// If the sandbox did contain this transition, Ok(false) is returned, sandbox is not modified.
    pub fn add_uid_transit(&mut self, source_uid: Uid, target_uid: Uid) -> Result<bool, Errno> {
        if target_uid == ROOT_UID || source_uid.as_raw() >= target_uid.as_raw() {
            return Err(Errno::EACCES);
        }

        let transit = (source_uid, target_uid);
        if self.transit_uids.contains(&transit) {
            return Ok(false);
        }

        if self
            .transit_uids
            .iter()
            .any(|&(s_uid, _)| s_uid == source_uid)
        {
            // Transition exists with same source UID but different target UID.
            // The user should remove the old transition to add this one.
            return Err(Errno::EEXIST);
        }

        self.transit_uids.push(transit);
        Ok(true)
    }

    /// Delete a UID transit.
    /// If both source and target is None, clears all transit UIDs.
    /// If source is not None and target is None, remove all transit UIDs with the source.
    /// If source is None and target is not None, remove all transit UIDs with the target.
    /// If both is not None, delete the matching UID.
    pub fn del_uid_transit(&mut self, source_uid: Option<Uid>, target_uid: Option<Uid>) {
        match (source_uid, target_uid) {
            (None, None) => self.transit_uids.clear(),
            (Some(source), None) => self.transit_uids.retain(|&(s_uid, _)| s_uid != source),
            (None, Some(target)) => self.transit_uids.retain(|&(_, t_uid)| t_uid != target),
            (Some(source), Some(target)) => self
                .transit_uids
                .retain(|&(s_uid, t_uid)| s_uid != source || t_uid != target),
        }
    }

    /// Check a UID transit.
    /// Returns true for allowed, false for not allowed.
    pub fn chk_uid_transit(&self, source_uid: Uid, target_uid: Uid) -> bool {
        if target_uid == ROOT_UID || source_uid.as_raw() >= target_uid.as_raw() {
            return false;
        }

        for (s_uid, t_uid) in &self.transit_uids {
            if source_uid == *s_uid && target_uid == *t_uid {
                return true;
            }
        }

        false
    }

    /// Add a GID transit.
    /// If the sandbox did not previously contain this transition, Ok(true) is returned.
    /// If the sandbox did contain this transition, Ok(false) is returned, sandbox is not modified.
    pub fn add_gid_transit(&mut self, source_gid: Gid, target_gid: Gid) -> Result<bool, Errno> {
        if target_gid == ROOT_GID || source_gid.as_raw() >= target_gid.as_raw() {
            return Err(Errno::EACCES);
        }

        let transit = (source_gid, target_gid);
        if self.transit_gids.contains(&transit) {
            return Ok(false);
        }

        if self
            .transit_gids
            .iter()
            .any(|&(s_gid, _)| s_gid == source_gid)
        {
            // Transition exists with same source GID but different target GID.
            // The user should remove the old transition to add this one.
            return Err(Errno::EEXIST);
        }

        self.transit_gids.push(transit);
        Ok(true)
    }

    /// Delete a GID transit.
    /// If both source and target is None, clears all transit GIDs.
    /// If source is not None and target is None, remove all transit GIDs with the source.
    /// If source is None and target is not None, remove all transit GIDs with the target.
    /// If both is not None, delete the matching GID.
    pub fn del_gid_transit(&mut self, source_gid: Option<Gid>, target_gid: Option<Gid>) {
        match (source_gid, target_gid) {
            (None, None) => self.transit_gids.clear(),
            (Some(source), None) => self.transit_gids.retain(|&(s_gid, _)| s_gid != source),
            (None, Some(target)) => self.transit_gids.retain(|&(_, t_gid)| t_gid != target),
            (Some(source), Some(target)) => self
                .transit_gids
                .retain(|&(s_gid, t_gid)| s_gid != source || t_gid != target),
        }
    }

    /// Check a GID transit.
    /// Returns true for allowed, false for not allowed.
    pub fn chk_gid_transit(&self, source_gid: Gid, target_gid: Gid) -> bool {
        if target_gid == ROOT_GID || source_gid.as_raw() >= target_gid.as_raw() {
            return false;
        }

        for (s_gid, t_gid) in &self.transit_gids {
            if source_gid == *s_gid && target_gid == *t_gid {
                return true;
            }
        }

        false
    }

    /// Add a bind mount action to the list of mount actions.
    pub fn add_bind_mount(&mut self, mount: BindMount) {
        if let Some(ref mut mounts) = self.bind_mounts {
            mounts.push(mount);
        } else {
            self.bind_mounts = Some(vec![mount]);
        }
    }

    /// Remove the first matching item from the end of the list of mount actions.
    pub fn del_bind_mount(&mut self, mount: BindMount) {
        if let Some(ref mut mounts) = self.bind_mounts {
            if let Some(pos) = mounts.iter().rposition(|m| m == &mount) {
                mounts.remove(pos);
            }
        }
    }

    /// Remove all matchign items from the list of mount actions.
    pub fn rem_bind_mount(&mut self, mount: BindMount) {
        if let Some(ref mut mounts) = self.bind_mounts {
            mounts.retain(|m| m != &mount);
        }
    }

    /// Extract the bind mount list.
    pub fn collect_bind_mounts(&mut self) -> Option<Vec<BindMount>> {
        std::mem::take(&mut self.bind_mounts)
    }

    /// Return true if there're any bind mounts defined.
    pub fn has_bind_mounts(&self) -> bool {
        !self
            .bind_mounts
            .as_ref()
            .map_or(true, |bind_mounts| bind_mounts.is_empty())
    }

    /// Check many paths against the Integrity Force map.
    pub fn check_force_many<I, P>(&self, paths: I) -> Result<(Action, Option<P>), IntegrityError>
    where
        I: IntoIterator<Item = P>,
        P: AsRef<Path> + Clone,
    {
        let mut action = Action::Allow;
        let mut mypath = None;
        for path in paths {
            let new_action = self.check_force(path.as_ref())?;
            if new_action > action {
                action = new_action;
                mypath = Some(path);
                if action == Action::Kill {
                    // Action with highest prio, cannot be overriden.
                    break;
                }
            }
        }

        Ok((action, mypath))
    }

    /// Check the path against the Integrity Force map.
    pub fn check_force<P: AsRef<Path>>(&self, path: P) -> Result<Action, IntegrityError> {
        if let Some((action, key)) = self.force_map.get(path.as_ref()) {
            let func = match HashAlgorithm::try_from(key.len()) {
                Ok(func) => func,
                _ => unreachable!("Hash with undefined length in Integrity Force map!"),
            };
            let file = match safe_open_file(path.as_ref()) {
                Ok((f, _)) => BufReader::new(f),
                Err(errno) => return Err(IntegrityError::Sys(errno)),
            };
            let hash = crate::hash::hash(file, func).map_err(IntegrityError::Sys)?;
            if hash != *key {
                return Err(IntegrityError::Hash {
                    action: *action,
                    expected: const_hex::encode(key),
                    found: const_hex::encode(hash),
                    path: path.as_ref().to_path_buf(),
                });
            } else {
                return Ok(Action::Allow);
            }
        }
        Ok(self.force_act)
    }

    /// Returns the AF_ALG setup socket FD.
    pub fn crypt_setup(&self) -> Result<RawFd, Errno> {
        match &self.crypt_id {
            Some(Secret::Fd(fd)) => Ok(fd.as_raw_fd()),
            _ => Err(Errno::ENOKEY),
        }
    }

    /// Panic if the sandbox state allows it.
    pub fn panic(&self) -> Result<(), Errno> {
        // SAFETY: Skip if Crypt Sandboxing is on.
        if self.enabled(Capability::CAP_CRYPT) {
            return Err(Errno::EBUSY);
        }
        #[allow(clippy::disallowed_methods)]
        std::process::exit(127)
    }

    /// Reset the sandbox to its default state, keeping the child pid information intact.
    pub fn reset(&mut self) -> Result<(), Errno> {
        // SAFETY: Skip if Crypt Sandboxing is on.
        if self.enabled(Capability::CAP_CRYPT) {
            return Err(Errno::EBUSY);
        }

        // SAFETY: We must preserve child pid{,fd} or lock:exec can be bypassed!
        *self = Self {
            cpid: self.cpid,
            fpid: self.fpid,
            ..Self::default()
        };

        Ok(())
    }

    /// Set namespace flags.
    fn set_nsflags(&mut self, value: &str) -> Result<(), Errno> {
        // Parse namespace flags.
        let nsflags = Flag::ns_from_str(value, true)?;

        // Reset flags for predictability.
        self.flags.remove(
            Flag::FL_ALLOW_UNSAFE_UNSHARE_MOUNT
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_UTS
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_IPC
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_USER
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_PID
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_NET
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_CGROUP
                | Flag::FL_ALLOW_UNSAFE_UNSHARE_TIME,
        );

        self.flags.insert(nsflags);

        Ok(())
    }

    /// Set or remove sandbox flag.
    fn set_flag(&mut self, flag: Flag, state: &str) -> Result<(), Errno> {
        // SAFETY: Reject flags that must only be set at startup.
        if self.is_running() && flag.is_startup() {
            return Err(Errno::EBUSY);
        }

        if strbool(state)? {
            self.flags.insert(flag);
        } else {
            self.flags.remove(flag);
        }

        Ok(())
    }

    /// Returns a new sandbox in default state.
    pub fn new() -> Self {
        Sandbox::default()
    }
}

// Note to self: To renumber the tests, do
// :let i=1 | g/sandbox_config_rules_\zs\d\+/s//\=i/ | let i+=1
// in VIM.
#[cfg(test)]
mod tests {
    use anyhow::Result as TestResult;
    use nix::unistd::Group;

    use super::*;

    fn get_nogroup() -> Result<(Gid, String), Errno> {
        // nogroup may be GID 65533 or 65534 depending on system.
        // nogroup may not even exist such as on Fedora (in which case nobody exists).
        // E.g. on Alpine it's 65533 meanwhile on Ubuntu it's 65534.
        let name = "nogroup".to_string();
        if let Ok(Some(group)) = Group::from_name(&name) {
            return Ok((group.gid, name));
        }

        let name = "nobody".to_string();
        if let Ok(Some(group)) = Group::from_name(&name) {
            return Ok((group.gid, name));
        }

        Err(Errno::ENOENT)
    }

    #[test]
    fn sandbox_config_api() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox.config("").map_err(anyhow::Error::msg)?;
        sandbox.config("3").map_err(anyhow::Error::msg)?;
        assert!(sandbox.config("2").is_err(), "{sandbox}");
        assert!(sandbox.config("1").is_err(), "{sandbox}");
        assert!(sandbox.config("0").is_err(), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_read() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(sandbox.state.contains(Capability::CAP_READ));
        assert!(sandbox.config("sandbox/read?").is_ok());
        sandbox
            .config("sandbox/read:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_READ));
        assert!(sandbox.config("sandbox/read?").is_err());
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_READ));
        assert!(sandbox.config("sandbox/read?").is_ok());

        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_stat() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(sandbox.state.contains(Capability::CAP_STAT));
        assert!(sandbox.config("sandbox/stat?").is_ok());
        sandbox
            .config("sandbox/stat:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_STAT));
        assert!(sandbox.config("sandbox/stat?").is_err());
        sandbox
            .config("sandbox/stat:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_STAT));
        assert!(sandbox.config("sandbox/stat?").is_ok());

        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_write() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(sandbox.state.contains(Capability::CAP_WRITE));
        assert!(sandbox.config("sandbox/write?").is_ok());
        sandbox
            .config("sandbox/write:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_WRITE));
        assert!(sandbox.config("sandbox/write?").is_err());
        sandbox
            .config("sandbox/write:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_WRITE));
        assert!(sandbox.config("sandbox/write?").is_ok());
        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_exec() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(sandbox.state.contains(Capability::CAP_EXEC));
        assert!(sandbox.config("sandbox/exec?").is_ok());
        sandbox
            .config("sandbox/exec:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_EXEC));
        assert!(sandbox.config("sandbox/exec?").is_err());
        sandbox
            .config("sandbox/exec:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_EXEC));
        assert!(sandbox.config("sandbox/exec?").is_ok());
        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_ioctl() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(sandbox.state.contains(Capability::CAP_IOCTL));
        assert!(sandbox.config("sandbox/ioctl?").is_ok());
        sandbox
            .config("sandbox/ioctl:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_IOCTL));
        assert!(sandbox.config("sandbox/ioctl?").is_err());
        sandbox
            .config("sandbox/ioctl:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_IOCTL));
        assert!(sandbox.config("sandbox/ioctl?").is_ok());
        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_crypt() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.state.contains(Capability::CAP_CRYPT));
        assert_eq!(sandbox.config("sandbox/crypt?"), Err(Errno::ENOENT));
        assert_eq!(sandbox.config("sandbox/crypt:on"), Ok(()));
        assert_eq!(sandbox.config("sandbox/crypt:off"), Ok(()));

        // Test setting crypt_key
        let good_key = "00112233445566778899aabbccddeeff00112233445566778899aabbccddeeff";
        assert!(sandbox.crypt_id.is_none());
        sandbox
            .config(&format!("crypt/key:{good_key}"))
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.crypt_id.is_some());
        if let Some(Secret::Key(key)) = &sandbox.crypt_id {
            assert_eq!(
                key.as_ref(),
                &const_hex::decode(good_key)
                    .unwrap()
                    .as_slice()
                    .try_into()
                    .unwrap() as &[u8; crate::hash::KEY_SIZE]
            );
        } else {
            panic!("Invalid key");
        }

        // Test invalid crypt_key length
        let bad_key = "00112233445566778899aabbccddeeff00112233445566778899aabbccddeef";
        assert_eq!(
            sandbox.config(&format!("crypt/key:{bad_key}")),
            Err(Errno::EINVAL)
        );

        sandbox
            .config("sandbox/crypt:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_CRYPT));
        assert!(sandbox.config("sandbox/crypt?").is_ok());
        sandbox
            .config("sandbox/crypt:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_CRYPT));
        assert!(sandbox.config("sandbox/crypt?").is_err());

        sandbox
            .config("sandbox/crypt:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_CRYPT));
        assert!(sandbox.config("sandbox/crypt?").is_ok());
        assert_eq!(sandbox.config(&format!("crypt/key:{good_key}")), Ok(()));

        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_network() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(sandbox.state.contains(Capability::CAP_BIND));
        assert!(sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(sandbox.config("sandbox/net?").is_ok());
        sandbox
            .config("sandbox/net:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.state.contains(Capability::CAP_BIND));
        assert!(!sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(sandbox.config("sandbox/net?").is_err());
        sandbox
            .config("sandbox/net:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_BIND));
        assert!(sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(sandbox.config("sandbox/net?").is_ok());

        sandbox
            .config("sandbox/net:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/net/bind:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_BIND));
        assert!(!sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(sandbox.config("sandbox/net?").is_err());
        assert!(sandbox.config("sandbox/net/bind?").is_ok());
        assert!(sandbox.config("sandbox/net/connect?").is_err());
        sandbox
            .config("sandbox/net/bind:off")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.config("sandbox/net?").is_err());
        assert!(sandbox.config("sandbox/net/bind?").is_err());
        assert!(sandbox.config("sandbox/net/connect?").is_err());

        sandbox
            .config("sandbox/net:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/net/connect:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(!sandbox.state.contains(Capability::CAP_BIND));
        assert!(sandbox.config("sandbox/net?").is_err());
        assert!(sandbox.config("sandbox/net/bind?").is_err());
        assert!(sandbox.config("sandbox/net/connect?").is_ok());
        sandbox
            .config("sandbox/net/connect:off")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.config("sandbox/net?").is_err());
        assert!(sandbox.config("sandbox/net/bind?").is_err());
        assert!(sandbox.config("sandbox/net/connect?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_many() -> TestResult<()> {
        let mut sandbox = Sandbox::default();

        sandbox.state = Capability::empty();
        sandbox
            .config("sandbox/read,stat:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/read,write,lock:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/lock,pid,mem,force,proxy,ioctl:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/force,mem,tpe,proxy:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/net,exec:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/net/bind,net/bind:off")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(sandbox.state.contains(Capability::CAP_EXEC));
        assert!(sandbox.state.contains(Capability::CAP_IOCTL));
        assert!(sandbox.state.contains(Capability::CAP_LOCK));
        assert!(sandbox.state.contains(Capability::CAP_PID));
        assert!(sandbox.state.contains(Capability::CAP_STAT));
        assert!(!sandbox.state.contains(Capability::CAP_BIND));
        assert!(!sandbox.state.contains(Capability::CAP_FORCE));
        assert!(!sandbox.state.contains(Capability::CAP_TPE));
        assert!(!sandbox.state.contains(Capability::CAP_PROXY));
        assert!(!sandbox.state.contains(Capability::CAP_MEM));
        assert!(!sandbox.state.contains(Capability::CAP_READ));
        assert!(!sandbox.state.contains(Capability::CAP_WRITE));

        Ok(())
    }

    #[test]
    fn sandbox_config_sandbox_many_lax() -> TestResult<()> {
        let mut sandbox = Sandbox::default();

        sandbox.state = Capability::empty();
        sandbox
            .config("sandbox/\treAd ,\t,\t\t,\t\t ,\t \t,  STaT :on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/  ReaD\t\t, write   , lock\t\t:OFF")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/   Lock  , pID, MeM,  ForcE,MeM,LOCK,IoCtL:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/ Force ,  ,\t,  MeM  , PROXY, TPe\t,  :off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/   Net ,   , ,  ,   ,    ,    \t,eXEc    :on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("sandbox/\t Net/BInd , net/bInd :off")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.state.contains(Capability::CAP_CONNECT));
        assert!(sandbox.state.contains(Capability::CAP_EXEC));
        assert!(sandbox.state.contains(Capability::CAP_IOCTL));
        assert!(sandbox.state.contains(Capability::CAP_LOCK));
        assert!(sandbox.state.contains(Capability::CAP_PID));
        assert!(sandbox.state.contains(Capability::CAP_STAT));
        assert!(!sandbox.state.contains(Capability::CAP_BIND));
        assert!(!sandbox.state.contains(Capability::CAP_FORCE));
        assert!(!sandbox.state.contains(Capability::CAP_TPE));
        assert!(!sandbox.state.contains(Capability::CAP_PROXY));
        assert!(!sandbox.state.contains(Capability::CAP_MEM));
        assert!(!sandbox.state.contains(Capability::CAP_READ));
        assert!(!sandbox.state.contains(Capability::CAP_WRITE));

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_mount() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_MOUNT));
        assert!(sandbox.config("unshare/mount?").is_err());
        sandbox
            .config("unshare/mount:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_MOUNT));
        assert!(sandbox.config("unshare/mount?").is_ok());
        sandbox
            .config("unshare/mount:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_MOUNT));
        assert!(sandbox.config("unshare/mount?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_uts() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_UTS));
        assert!(sandbox.config("unshare/uts?").is_err());
        sandbox
            .config("unshare/uts:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_UTS));
        assert!(sandbox.config("unshare/uts?").is_ok());
        sandbox
            .config("unshare/uts:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_UTS));
        assert!(sandbox.config("unshare/uts?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_ipc() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_IPC));
        assert!(sandbox.config("unshare/ipc?").is_err());
        sandbox
            .config("unshare/ipc:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_IPC));
        assert!(sandbox.config("unshare/ipc?").is_ok());
        sandbox
            .config("unshare/ipc:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_IPC));
        assert!(sandbox.config("unshare/ipc?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_user() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_USER));
        assert!(sandbox.config("unshare/user?").is_err());
        sandbox
            .config("unshare/user:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_USER));
        assert!(sandbox.config("unshare/user?").is_ok());
        sandbox
            .config("unshare/user:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_USER));
        assert!(sandbox.config("unshare/user?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_pid() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_PID));
        assert!(sandbox.config("unshare/pid?").is_err());
        sandbox
            .config("unshare/pid:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_PID));
        assert!(sandbox.config("unshare/pid?").is_ok());
        sandbox
            .config("unshare/pid:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_PID));
        assert!(sandbox.config("unshare/pid?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_net() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_NET));
        assert!(sandbox.config("unshare/net?").is_err());
        sandbox
            .config("unshare/net:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_NET));
        assert!(sandbox.config("unshare/net?").is_ok());
        sandbox
            .config("unshare/net:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_NET));
        assert!(sandbox.config("unshare/net?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_cgroup() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_CGROUP));
        assert!(sandbox.config("unshare/cgroup?").is_err());
        sandbox
            .config("unshare/cgroup:on")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_CGROUP));
        assert!(sandbox.config("unshare/cgroup?").is_ok());
        sandbox
            .config("unshare/cgroup:off")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_CGROUP));
        assert!(sandbox.config("unshare/cgroup?").is_err());

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_many() -> TestResult<()> {
        let mut sandbox = Sandbox::default();

        sandbox.flags = Flag::empty();
        sandbox
            .config("unshare/mount,uts:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/mount,uts,ipc:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/mount,uts,ipc,user:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/user,pid:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/net,cgroup:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/cgroup:off")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_MOUNT));
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_UTS));
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_IPC));
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_USER));
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_PID));
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_NET));
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_CGROUP));

        Ok(())
    }

    #[test]
    fn sandbox_config_unshare_flags_many_lax() -> TestResult<()> {
        let mut sandbox = Sandbox::default();

        sandbox.flags = Flag::empty();
        sandbox
            .config("unshare/\tmOunt ,\tuTS,\t\t,\t\t ,\t \t,   :on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/  moUnT\t\t, UTS   ,  \t\tIPc    ,,,,,,:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/  MoUnT\t , ,  ,,,,,,\tUtS  ,\t, IPC  \t,  USer:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/     user,  , , , , , , pId:off")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/\t\t,,,,net,,,,,,cgroup:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("unshare/,,,,,CGROUP,\t:off")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_MOUNT));
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_UTS));
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_IPC));
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_USER));
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_PID));
        assert!(sandbox.flags.contains(Flag::FL_UNSHARE_NET));
        assert!(!sandbox.flags.contains(Flag::FL_UNSHARE_CGROUP));

        Ok(())
    }

    #[test]
    fn sandbox_config_lock() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.lock, None);
        sandbox.config("lock:off").map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.lock, Some(LockState::Off));
        sandbox.config("lock:exec").map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.lock, Some(LockState::Exec));
        sandbox.config("lock:on").map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.lock, Some(LockState::Set));
        Ok(())
    }

    #[test]
    fn sandbox_config_trace() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));
        sandbox
            .config("trace/allow_safe_bind:true")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));
        sandbox
            .config("trace/allow_safe_bind:false")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));
        sandbox
            .config("trace/allow_safe_bind:t")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));
        sandbox
            .config("trace/allow_safe_bind:f")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));
        sandbox
            .config("trace/allow_safe_bind:1")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));
        sandbox
            .config("trace/allow_safe_bind:0")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_SAFE_BIND));

        assert_eq!(
            sandbox.config("trace/allow_safe_bind_invalid:t"),
            Err(Errno::EINVAL)
        );
        assert_eq!(
            sandbox.config("trace/allow_safe_bind!x"),
            Err(Errno::EINVAL)
        );
        assert_eq!(
            sandbox.config("trace/allow_safe_bind:x"),
            Err(Errno::EINVAL)
        );
        assert_eq!(
            sandbox.config("trace/allow_safe_bind:☮"),
            Err(Errno::EINVAL)
        );

        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));
        sandbox
            .config("trace/allow_unsupp_socket:true")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));
        sandbox
            .config("trace/allow_unsupp_socket:false")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));
        sandbox
            .config("trace/allow_unsupp_socket:t")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));
        sandbox
            .config("trace/allow_unsupp_socket:f")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));
        sandbox
            .config("trace/allow_unsupp_socket:1")
            .map_err(anyhow::Error::msg)?;
        assert!(sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));
        sandbox
            .config("trace/allow_unsupp_socket:0")
            .map_err(anyhow::Error::msg)?;
        assert!(!sandbox.flags.contains(Flag::FL_ALLOW_UNSUPP_SOCKET));

        assert_eq!(
            sandbox.config("trace/allow_unsupp_socket_invalid:t"),
            Err(Errno::EINVAL)
        );
        assert_eq!(
            sandbox.config("trace/allow_unsupp_socket!x"),
            Err(Errno::EINVAL)
        );
        assert_eq!(
            sandbox.config("trace/allow_unsupp_socket:x"),
            Err(Errno::EINVAL)
        );
        assert_eq!(
            sandbox.config("trace/allow_unsupp_socket:☮"),
            Err(Errno::EINVAL)
        );

        assert_eq!(sandbox.config("trace/memory_access:0"), Ok(()));
        assert_eq!(sandbox.config("trace/memory_access:1"), Ok(()));
        assert_eq!(
            sandbox.config("trace/memory_access_invalid:t"),
            Err(Errno::EINVAL)
        );
        assert_eq!(sandbox.config("trace/memory_access!x"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("trace/memory_access:x"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("trace/memory_access:t"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("trace/memory_access:☮"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("trace/memory_access:f"), Err(Errno::EINVAL));

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_1() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Invalid syntax
        assert_eq!(sandbox.config("setuid:0:65534"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setuid:0:0"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setgid:0:65534"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setgid:0:0"), Err(Errno::EINVAL));

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_2() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // ROOT not allowed in target.
        assert_eq!(sandbox.config("setuid+0:0"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+0:0"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setuid+1:0"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+1:0"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setuid+65534:0"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+65534:0"), Err(Errno::EACCES));

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_3() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // ID escalation is not allowed.
        assert_eq!(sandbox.config("setuid+65534:42"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+65534:42"), Err(Errno::EACCES));

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_4() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Same ID is meaningless and is disallowed.
        assert_eq!(sandbox.config("setuid+1:1"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+1:1"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setuid+65534:65534"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+65534:65534"), Err(Errno::EACCES));

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_5() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setuid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(65534), "{sandbox}");
        assert_eq!(sandbox.config("setuid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setuid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setuid-0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_6() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setgid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(65534), "{sandbox}");
        assert_eq!(sandbox.config("setgid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setgid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setgid-0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_7() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Adding a transition with same source but diff target fails.
        assert_eq!(sandbox.config("setuid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(65534), "{sandbox}");
        assert_eq!(sandbox.config("setuid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setuid+0:1"), Err(Errno::EEXIST));
        // Removing the transition and then readding with diff target is OK.
        assert_eq!(sandbox.config("setuid-0:65534"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.config("setuid+0:1"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(1), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_8() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Adding a transition with same source but diff target fails.
        assert_eq!(sandbox.config("setgid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(65534), "{sandbox}");
        assert_eq!(sandbox.config("setgid+0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setgid+0:1"), Err(Errno::EEXIST));
        // Removing the transition and then readding with diff target is OK.
        assert_eq!(sandbox.config("setgid-0:65534"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.config("setgid+0:1"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(1), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_9() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setuid+0:1"), Ok(()));
        assert_eq!(sandbox.config("setuid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setuid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].0, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].1, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].0, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].1, Uid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setuid^"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_10() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setgid+0:1"), Ok(()));
        assert_eq!(sandbox.config("setgid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setgid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].0, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].1, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].0, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].1, Gid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setgid^"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_11() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setuid+0:1"), Ok(()));
        assert_eq!(sandbox.config("setuid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setuid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].0, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].1, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].0, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].1, Uid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setuid^0:1"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setuid^0"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 2, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(2), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetid_12() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setgid+0:1"), Ok(()));
        assert_eq!(sandbox.config("setgid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setgid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].0, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].1, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].0, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].1, Gid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setgid^0:1"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setgid^0"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 2, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(2), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_1() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Invalid syntax
        assert_eq!(sandbox.config("setuid:root:nobody"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setuid:root:root"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setgid:root:nobody"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setgid:root:root"), Err(Errno::EINVAL));

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_2() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // ROOT not allowed in target.
        assert_eq!(sandbox.config("setuid+root:root"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setgid+root:root"), Err(Errno::EACCES));
        assert_eq!(sandbox.config("setuid+nobody:root"), Err(Errno::EACCES));
        if let Ok((_, nogroup)) = get_nogroup() {
            assert_eq!(
                sandbox.config(&format!("setgid+{nogroup}:root")),
                Err(Errno::EACCES)
            );
        }

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_3() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Same ID is meaningless and is disallowed.
        assert_eq!(sandbox.config("setuid+nobody:nobody"), Err(Errno::EACCES));
        if let Ok((_, nogroup)) = get_nogroup() {
            assert_eq!(
                sandbox.config(&format!("setgid+{nogroup}:{nogroup}")),
                Err(Errno::EACCES)
            );
        }

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_4() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setuid+root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(65534), "{sandbox}");
        assert_eq!(sandbox.config("setuid+root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setuid+root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setuid-root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_5() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        let (nogroup_gid, nogroup) = match get_nogroup() {
            Ok((nogroup_gid, nogroup)) => (nogroup_gid, nogroup),
            Err(_) => return Ok(()), // should not happen, skip.
        };

        assert_eq!(sandbox.config(&format!("setgid+root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, nogroup_gid, "{sandbox}");
        assert_eq!(sandbox.config(&format!("setgid+root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config(&format!("setgid+root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config(&format!("setgid-root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_6() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        // Adding a transition with same source but diff target fails.
        assert_eq!(sandbox.config("setuid+root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(65534), "{sandbox}");
        assert_eq!(sandbox.config("setuid+root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setuid+root:1"), Err(Errno::EEXIST));
        // Removing the transition and then readding with diff target is OK.
        assert_eq!(sandbox.config("setuid-root:nobody"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.config("setuid+root:1"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(1), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_7() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        let (nogroup_gid, nogroup) = match get_nogroup() {
            Ok((nogroup_gid, nogroup)) => (nogroup_gid, nogroup),
            Err(_) => return Ok(()), // should not happen, skip.
        };

        // Adding a transition with same source but diff target fails.
        assert_eq!(sandbox.config(&format!("setgid+root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, nogroup_gid, "{sandbox}");
        assert_eq!(sandbox.config(&format!("setgid+root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.config("setgid+root:1"), Err(Errno::EEXIST));
        // Removing the transition and then readding with diff target is OK.
        assert_eq!(sandbox.config(&format!("setgid-root:{nogroup}")), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.config("setgid+root:1"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 1, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(1), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_8() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setuid+root:1"), Ok(()));
        assert_eq!(sandbox.config("setuid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setuid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].0, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].1, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].0, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].1, Uid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setuid^"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_9() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setgid+root:1"), Ok(()));
        assert_eq!(sandbox.config("setgid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setgid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].0, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].1, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].0, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].1, Gid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setgid^"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_10() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setuid+root:1"), Ok(()));
        assert_eq!(sandbox.config("setuid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setuid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].0, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[1].1, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].0, Uid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_uids[2].1, Uid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setuid^root:1"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setuid^root"), Ok(()));
        assert_eq!(sandbox.transit_uids.len(), 2, "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].0, Uid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_uids[0].1, Uid::from_raw(2), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_safesetuser_11() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.transit_uids.len(), 0, "{sandbox}");
        assert_eq!(sandbox.transit_gids.len(), 0, "{sandbox}");

        assert_eq!(sandbox.config("setgid+root:1"), Ok(()));
        assert_eq!(sandbox.config("setgid+1:2"), Ok(()));
        assert_eq!(sandbox.config("setgid+2:3"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 3, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(0), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].0, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[1].1, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].0, Gid::from_raw(2), "{sandbox}");
        assert_eq!(sandbox.transit_gids[2].1, Gid::from_raw(3), "{sandbox}");
        assert_eq!(sandbox.config("setgid^root:1"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("setgid^root"), Ok(()));
        assert_eq!(sandbox.transit_gids.len(), 2, "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].0, Gid::from_raw(1), "{sandbox}");
        assert_eq!(sandbox.transit_gids[0].1, Gid::from_raw(2), "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_1() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        sandbox
            .config("allow/read-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_2() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/read+/usr")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");

        sandbox
            .config("allow/read^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_3() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("allow/write-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_4() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/write+/usr")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("allow/write^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_5() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("allow/exec-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_6() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/exec+/usr")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("allow/exec^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_7() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_8() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+/usr")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_9() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_10() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+/usr")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_11() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_12() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+127.0.0.0/8!1024")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_13() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_14() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+127.0.0.0/7!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_15() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_16() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+::1/8!1024-65525")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_17() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_18() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_19() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("deny/read-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_20() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/read+/usr/*")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("deny/read^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_21() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("deny/write-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_22() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/write+/usr/*")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("deny/write^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_23() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("deny/exec-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_24() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("deny/exec^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_25() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_26() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/net/bind+/usr")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+/usr/*")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("deny/net/bind-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_27() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_28() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+/usr/*")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+/usr")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_29() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_30() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+127.0.1.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+127.0.0.1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_31() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_32() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+127.0.0.1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+127.0.1.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 5, "{sandbox}");
        sandbox
            .config("deny/net/connect^127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_33() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_34() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+::1/16!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_35() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_36() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+::1!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_37() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("filter/read-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_38() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/read+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/read+/usr/*")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("filter/read^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_39() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("filter/write-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_40() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/write+/usr/**/")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/write+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("filter/write^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_41() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("filter/exec-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_42() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/exec+/usr/*")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/exec+/usr/./**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/exec+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 5, "{sandbox}");
        sandbox
            .config("filter/exec^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_43() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_44() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_45() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_46() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/net/connect+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+/usr/**/")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+/usr/*")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("filter/net/connect-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_47() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_48() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+127.0.0.0/8!1023-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+127.0.0.0/8!1023-65534")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("filter/net/bind^127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_49() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_50() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+127.0.0.0/8!1024-65534")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+127.0.0.1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+127.0.0.0/8!1025-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 5, "{sandbox}");
        sandbox
            .config("filter/net/connect^127.0.0.0/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_51() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_52() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_53() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_54() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^::1/8!1024-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_55() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_56() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_57() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_58() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_59() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_60() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_61() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_62() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+127.0.0.0/16!0-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+127.0.0.0/16!0-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+127.0.0.0/16!0-65535")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+127.0.0.0/8!0-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("deny/net/connect^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_63() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_64() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback4!1")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("filter/net/bind^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_65() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_66() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_67() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_68() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_69() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_70() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_71() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_72() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_73() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_74() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_75() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_76() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_77() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_78() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_79() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/bind-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_80() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/bind-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_81() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/bind^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_82() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/connect-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_83() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/connect^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_84() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("deny/net/bind-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_85() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("deny/net/bind^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_86() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("deny/net/connect-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_87() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("deny/net/connect^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_88() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("filter/net/bind-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_89() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/bind^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_90() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("filter/net/connect-loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_91() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/connect^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_92() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("allow/net/bind-local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_93() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("allow/net/bind^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_94() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("allow/net/connect-local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_95() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("allow/net/connect^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_96() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("deny/net/bind-local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_97() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("deny/net/bind^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_98() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("deny/net/connect-local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_99() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("deny/net/connect^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_100() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("filter/net/bind-local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_101() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/bind^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_102() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("filter/net/connect-local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_103() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/connect^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_104() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("allow/net/bind-local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_105() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("allow/net/bind^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_106() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("allow/net/connect-local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_107() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("allow/net/connect^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_108() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("deny/net/bind-local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_109() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("deny/net/bind^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_110() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("deny/net/connect-local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_111() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("deny/net/connect^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_112() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("filter/net/bind-local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_113() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/bind^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_114() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("filter/net/connect-local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_115() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/connect^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_116() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("filter/net/bind-local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_117() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 24, "{sandbox}");
        sandbox
            .config("filter/net/bind^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_118() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("filter/net/connect-local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_119() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 16, "{sandbox}");
        sandbox
            .config("filter/net/connect^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_120() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("allow/net/bind-local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_121() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+local!1")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 24, "{sandbox}");
        sandbox
            .config("allow/net/bind^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_122() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("allow/net/connect-local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_123() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+127.0.0.2!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+local!1")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 25, "{sandbox}");
        sandbox
            .config("allow/net/connect^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_124() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("deny/net/bind-local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_125() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+127.0.0.1!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 17, "{sandbox}");
        sandbox
            .config("deny/net/bind^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_126() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("deny/net/connect-local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_127() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 16, "{sandbox}");
        sandbox
            .config("deny/net/connect^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_128() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_129() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_130() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_131() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+127.0.0.3!7")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("allow/net/connect^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_132() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_133() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_134() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_135() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_136() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_137() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_138() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_139() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_140() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_141() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_142() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_143() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_144() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_145() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_146() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_147() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_148() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_149() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_150() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_151() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_152() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("filter/net/bind-linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_153() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/bind^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_154() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("filter/net/connect-linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_155() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/connect^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_156() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/bind-linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_157() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/bind^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_158() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/connect-linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_159() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/connect^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_160() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("deny/net/bind-linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_161() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("deny/net/bind^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_162() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("deny/net/connect-linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_163() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("deny/net/connect^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_164() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_165() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_166() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_167() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_168() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_169() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_170() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_171() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_172() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_173() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_174() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_175() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_176() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_177() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/bind^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_178() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_179() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_180() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_181() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/bind^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_182() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_183() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("deny/net/connect^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_184() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/bind-any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_185() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/bind^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_186() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("filter/net/connect-any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_187() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("filter/net/connect^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_188() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/bind-any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_189() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/bind^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_190() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/connect-any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_191() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/connect^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_192() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("deny/net/bind-any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_193() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("deny/net/bind^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_194() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("deny/net/connect-any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_195() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("deny/net/connect^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_196() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("filter/net/bind-any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_197() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/bind+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/bind^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_198() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("filter/net/connect-any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_199() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("filter/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/connect^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_200() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..8 {
            sandbox
                .config("allow/net/bind+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        for _ in 0..8 {
            sandbox
                .config("allow/net/bind-loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_201() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..8 {
            sandbox
                .config("allow/net/bind+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("allow/net/bind^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_202() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..9 {
            sandbox
                .config("allow/net/connect+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");
        for _ in 0..9 {
            sandbox
                .config("allow/net/connect-loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_203() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..9 {
            sandbox
                .config("allow/net/connect+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");
        sandbox
            .config("allow/net/connect^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_204() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..10 {
            sandbox
                .config("deny/net/bind+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        for _ in 0..10 {
            sandbox
                .config("deny/net/bind-loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_205() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..10 {
            sandbox
                .config("deny/net/bind+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        sandbox
            .config("deny/net/bind^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_206() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..11 {
            sandbox
                .config("deny/net/connect+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 11, "{sandbox}");
        for _ in 0..11 {
            sandbox
                .config("deny/net/connect-loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_207() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..11 {
            sandbox
                .config("deny/net/connect+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 11, "{sandbox}");
        sandbox
            .config("deny/net/connect^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_208() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..12 {
            sandbox
                .config("filter/net/bind+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        for _ in 0..12 {
            sandbox
                .config("filter/net/bind-loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_209() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..12 {
            sandbox
                .config("filter/net/bind+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/bind^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_210() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..13 {
            sandbox
                .config("filter/net/connect+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 13, "{sandbox}");
        for _ in 0..13 {
            sandbox
                .config("filter/net/connect-loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_211() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..13 {
            sandbox
                .config("filter/net/connect+loopback4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 13, "{sandbox}");
        sandbox
            .config("filter/net/connect^loopback4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_212() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..8 {
            sandbox
                .config("allow/net/bind+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        for _ in 0..8 {
            sandbox
                .config("allow/net/bind-loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_213() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..8 {
            sandbox
                .config("allow/net/bind+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("allow/net/bind^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_214() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..9 {
            sandbox
                .config("allow/net/connect+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");
        for _ in 0..9 {
            sandbox
                .config("allow/net/connect-loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_215() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..9 {
            sandbox
                .config("allow/net/connect+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");
        sandbox
            .config("allow/net/connect^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_216() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..10 {
            sandbox
                .config("deny/net/bind+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        for _ in 0..10 {
            sandbox
                .config("deny/net/bind-loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_217() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..10 {
            sandbox
                .config("deny/net/bind+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        sandbox
            .config("deny/net/bind^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_218() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..11 {
            sandbox
                .config("deny/net/connect+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 11, "{sandbox}");
        for _ in 0..11 {
            sandbox
                .config("deny/net/connect-loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_219() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..11 {
            sandbox
                .config("deny/net/connect+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 11, "{sandbox}");
        sandbox
            .config("deny/net/connect^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_220() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..12 {
            sandbox
                .config("filter/net/bind+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        for _ in 0..12 {
            sandbox
                .config("filter/net/bind-loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_221() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..12 {
            sandbox
                .config("filter/net/bind+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/bind^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_222() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..13 {
            sandbox
                .config("filter/net/connect+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 13, "{sandbox}");
        for _ in 0..13 {
            sandbox
                .config("filter/net/connect-loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_223() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..13 {
            sandbox
                .config("filter/net/connect+loopback6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 13, "{sandbox}");
        sandbox
            .config("filter/net/connect^loopback6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_224() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..2 {
            sandbox
                .config("allow/net/bind+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        for _ in 0..2 {
            sandbox
                .config("allow/net/bind-loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_225() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..2 {
            sandbox
                .config("allow/net/bind+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("allow/net/bind^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_226() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..3 {
            sandbox
                .config("allow/net/connect+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        for _ in 0..3 {
            sandbox
                .config("allow/net/connect-loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_227() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..3 {
            sandbox
                .config("allow/net/connect+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("allow/net/connect^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_228() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..4 {
            sandbox
                .config("deny/net/bind+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        for _ in 0..4 {
            sandbox
                .config("deny/net/bind-loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_229() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..4 {
            sandbox
                .config("deny/net/bind+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("deny/net/bind^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_230() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..5 {
            sandbox
                .config("deny/net/connect+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        for _ in 0..5 {
            sandbox
                .config("deny/net/connect-loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_231() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..5 {
            sandbox
                .config("deny/net/connect+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        sandbox
            .config("deny/net/connect^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_232() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..6 {
            sandbox
                .config("filter/net/bind+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        for _ in 0..6 {
            sandbox
                .config("filter/net/bind-loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_233() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..6 {
            sandbox
                .config("filter/net/bind+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/bind^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_234() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..7 {
            sandbox
                .config("filter/net/connect+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 14, "{sandbox}");
        for _ in 0..7 {
            sandbox
                .config("filter/net/connect-loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_235() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..7 {
            sandbox
                .config("filter/net/connect+loopback!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 14, "{sandbox}");
        sandbox
            .config("filter/net/connect^loopback!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_236() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..14 {
            sandbox
                .config("allow/net/bind+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 14, "{sandbox}");
        for _ in 0..14 {
            sandbox
                .config("allow/net/bind-local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_237() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..14 {
            sandbox
                .config("allow/net/bind+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 14, "{sandbox}");
        sandbox
            .config("allow/net/bind^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_238() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..15 {
            sandbox
                .config("allow/net/connect+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 15, "{sandbox}");
        for _ in 0..15 {
            sandbox
                .config("allow/net/connect-local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_239() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..15 {
            sandbox
                .config("allow/net/connect+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 15, "{sandbox}");
        sandbox
            .config("allow/net/connect^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_240() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..16 {
            sandbox
                .config("deny/net/bind+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 16, "{sandbox}");
        for _ in 0..16 {
            sandbox
                .config("deny/net/bind-local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_241() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..16 {
            sandbox
                .config("deny/net/bind+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 16, "{sandbox}");
        sandbox
            .config("deny/net/bind^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_242() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..17 {
            sandbox
                .config("deny/net/connect+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 17, "{sandbox}");
        for _ in 0..17 {
            sandbox
                .config("deny/net/connect-local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_243() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..17 {
            sandbox
                .config("deny/net/connect+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 17, "{sandbox}");
        sandbox
            .config("deny/net/connect^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_244() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..18 {
            sandbox
                .config("filter/net/bind+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 18, "{sandbox}");
        for _ in 0..18 {
            sandbox
                .config("filter/net/bind-local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_245() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..18 {
            sandbox
                .config("filter/net/bind+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 18, "{sandbox}");
        sandbox
            .config("filter/net/bind^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_246() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..19 {
            sandbox
                .config("filter/net/connect+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 19, "{sandbox}");
        for _ in 0..19 {
            sandbox
                .config("filter/net/connect-local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_247() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..19 {
            sandbox
                .config("filter/net/connect+local4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 19, "{sandbox}");
        sandbox
            .config("filter/net/connect^local4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_248() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 20, "{sandbox}");
        for _ in 0..20 {
            sandbox
                .config("allow/net/bind-local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_249() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 20, "{sandbox}");
        sandbox
            .config("allow/net/bind^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_250() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 21, "{sandbox}");
        for _ in 0..21 {
            sandbox
                .config("allow/net/connect-local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_251() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 21, "{sandbox}");
        sandbox
            .config("allow/net/connect^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_252() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 22, "{sandbox}");
        for _ in 0..22 {
            sandbox
                .config("deny/net/bind-local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_253() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 22, "{sandbox}");
        sandbox
            .config("deny/net/bind^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_254() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 23, "{sandbox}");
        for _ in 0..23 {
            sandbox
                .config("deny/net/connect-local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_255() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 23, "{sandbox}");
        sandbox
            .config("deny/net/connect^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_256() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 24, "{sandbox}");
        for _ in 0..24 {
            sandbox
                .config("filter/net/bind-local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_257() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 24, "{sandbox}");
        sandbox
            .config("filter/net/bind^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_258() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 25, "{sandbox}");
        for _ in 0..25 {
            sandbox
                .config("filter/net/connect-local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_259() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+local6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4 * 25, "{sandbox}");
        sandbox
            .config("filter/net/connect^local6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_260() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 20, "{sandbox}");
        for _ in 0..20 {
            sandbox
                .config("allow/net/bind-local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_261() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 20, "{sandbox}");
        sandbox
            .config("allow/net/bind^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_262() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 21, "{sandbox}");
        for _ in 0..21 {
            sandbox
                .config("allow/net/connect-local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_263() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 21, "{sandbox}");
        sandbox
            .config("allow/net/connect^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_264() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 22, "{sandbox}");
        for _ in 0..22 {
            sandbox
                .config("deny/net/bind-local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_265() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 22, "{sandbox}");
        sandbox
            .config("deny/net/bind^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_266() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 23, "{sandbox}");
        for _ in 0..23 {
            sandbox
                .config("deny/net/connect-local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_267() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 23, "{sandbox}");
        sandbox
            .config("deny/net/connect^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_268() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 24, "{sandbox}");
        for _ in 0..24 {
            sandbox
                .config("filter/net/bind-local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_269() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 24, "{sandbox}");
        sandbox
            .config("filter/net/bind^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_270() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 25, "{sandbox}");
        for _ in 0..25 {
            sandbox
                .config("filter/net/connect-local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_271() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+local!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8 * 25, "{sandbox}");
        sandbox
            .config("filter/net/connect^local!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_272() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..4 {
            sandbox
                .config("deny/net/bind+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        for _ in 0..4 {
            sandbox
                .config("deny/net/bind-any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_273() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..4 {
            sandbox
                .config("deny/net/bind+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 4, "{sandbox}");
        sandbox
            .config("deny/net/bind^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_274() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..5 {
            sandbox
                .config("deny/net/connect+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 5, "{sandbox}");
        for _ in 0..5 {
            sandbox
                .config("deny/net/connect-any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_275() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..5 {
            sandbox
                .config("deny/net/connect+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 5, "{sandbox}");
        sandbox
            .config("deny/net/connect^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_276() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..14 {
            sandbox
                .config("allow/net/bind+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 14, "{sandbox}");
        for _ in 0..14 {
            sandbox
                .config("allow/net/bind-linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_277() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..14 {
            sandbox
                .config("allow/net/bind+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 14, "{sandbox}");
        sandbox
            .config("allow/net/bind^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_278() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..15 {
            sandbox
                .config("allow/net/connect+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 15, "{sandbox}");
        for _ in 0..15 {
            sandbox
                .config("allow/net/connect-linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_279() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..15 {
            sandbox
                .config("allow/net/connect+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 15, "{sandbox}");
        sandbox
            .config("allow/net/connect^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_280() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..16 {
            sandbox
                .config("deny/net/bind+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 16, "{sandbox}");
        for _ in 0..16 {
            sandbox
                .config("deny/net/bind-linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_281() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..16 {
            sandbox
                .config("deny/net/bind+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 16, "{sandbox}");
        sandbox
            .config("deny/net/bind^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_282() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..17 {
            sandbox
                .config("deny/net/connect+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 17, "{sandbox}");
        for _ in 0..17 {
            sandbox
                .config("deny/net/connect-linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_283() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..17 {
            sandbox
                .config("deny/net/connect+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 17, "{sandbox}");
        sandbox
            .config("deny/net/connect^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_284() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..18 {
            sandbox
                .config("filter/net/bind+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 18, "{sandbox}");
        for _ in 0..18 {
            sandbox
                .config("filter/net/bind-linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_285() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..18 {
            sandbox
                .config("filter/net/bind+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 18, "{sandbox}");
        sandbox
            .config("filter/net/bind^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_286() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..19 {
            sandbox
                .config("filter/net/connect+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 19, "{sandbox}");
        for _ in 0..19 {
            sandbox
                .config("filter/net/connect-linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_287() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..19 {
            sandbox
                .config("filter/net/connect+linklocal4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 19, "{sandbox}");
        sandbox
            .config("filter/net/connect^linklocal4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_288() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 20, "{sandbox}");
        for _ in 0..20 {
            sandbox
                .config("allow/net/bind-linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_289() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 20, "{sandbox}");
        sandbox
            .config("allow/net/bind^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_290() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 21, "{sandbox}");
        for _ in 0..21 {
            sandbox
                .config("allow/net/connect-linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_291() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 21, "{sandbox}");
        sandbox
            .config("allow/net/connect^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_292() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 22, "{sandbox}");
        for _ in 0..22 {
            sandbox
                .config("deny/net/bind-linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_293() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 22, "{sandbox}");
        sandbox
            .config("deny/net/bind^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_294() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 23, "{sandbox}");
        for _ in 0..23 {
            sandbox
                .config("deny/net/connect-linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_295() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 23, "{sandbox}");
        sandbox
            .config("deny/net/connect^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_296() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 24, "{sandbox}");
        for _ in 0..24 {
            sandbox
                .config("filter/net/bind-linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_297() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 24, "{sandbox}");
        sandbox
            .config("filter/net/bind^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_298() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 25, "{sandbox}");
        for _ in 0..25 {
            sandbox
                .config("filter/net/connect-linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_299() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+linklocal6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 25, "{sandbox}");
        sandbox
            .config("filter/net/connect^linklocal6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_300() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 20, "{sandbox}");
        for _ in 0..20 {
            sandbox
                .config("allow/net/bind-linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_301() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..20 {
            sandbox
                .config("allow/net/bind+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 20, "{sandbox}");
        sandbox
            .config("allow/net/bind^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_302() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 21, "{sandbox}");
        for _ in 0..21 {
            sandbox
                .config("allow/net/connect-linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_303() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..21 {
            sandbox
                .config("allow/net/connect+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 21, "{sandbox}");
        sandbox
            .config("allow/net/connect^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_304() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 22, "{sandbox}");
        for _ in 0..22 {
            sandbox
                .config("deny/net/bind-linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_305() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..22 {
            sandbox
                .config("deny/net/bind+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 22, "{sandbox}");
        sandbox
            .config("deny/net/bind^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_306() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 23, "{sandbox}");
        for _ in 0..23 {
            sandbox
                .config("deny/net/connect-linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_307() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..23 {
            sandbox
                .config("deny/net/connect+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 23, "{sandbox}");
        sandbox
            .config("deny/net/connect^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_308() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 24, "{sandbox}");
        for _ in 0..24 {
            sandbox
                .config("filter/net/bind-linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_309() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..24 {
            sandbox
                .config("filter/net/bind+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 24, "{sandbox}");
        sandbox
            .config("filter/net/bind^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_310() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 25, "{sandbox}");
        for _ in 0..25 {
            sandbox
                .config("filter/net/connect-linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_311() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..25 {
            sandbox
                .config("filter/net/connect+linklocal!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2 * 25, "{sandbox}");
        sandbox
            .config("filter/net/connect^linklocal!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_312() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..2 {
            sandbox
                .config("allow/net/bind+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        for _ in 0..2 {
            sandbox
                .config("allow/net/bind-any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_313() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..2 {
            sandbox
                .config("allow/net/bind+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 2, "{sandbox}");
        sandbox
            .config("allow/net/bind^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_314() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..3 {
            sandbox
                .config("allow/net/connect+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        for _ in 0..3 {
            sandbox
                .config("allow/net/connect-any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_315() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..3 {
            sandbox
                .config("allow/net/connect+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 3, "{sandbox}");
        sandbox
            .config("allow/net/connect^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_316() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..6 {
            sandbox
                .config("filter/net/bind+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        for _ in 0..6 {
            sandbox
                .config("filter/net/bind-any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_317() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..6 {
            sandbox
                .config("filter/net/bind+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 6, "{sandbox}");
        sandbox
            .config("filter/net/bind^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_318() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..7 {
            sandbox
                .config("filter/net/connect+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 7, "{sandbox}");
        for _ in 0..7 {
            sandbox
                .config("filter/net/connect-any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_319() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..7 {
            sandbox
                .config("filter/net/connect+any4!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 7, "{sandbox}");
        sandbox
            .config("filter/net/connect^any4!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_320() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..8 {
            sandbox
                .config("allow/net/bind+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        for _ in 0..8 {
            sandbox
                .config("allow/net/bind-any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_321() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..8 {
            sandbox
                .config("allow/net/bind+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 8, "{sandbox}");
        sandbox
            .config("allow/net/bind^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_322() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..9 {
            sandbox
                .config("allow/net/connect+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");
        for _ in 0..9 {
            sandbox
                .config("allow/net/connect-any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_323() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..9 {
            sandbox
                .config("allow/net/connect+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 9, "{sandbox}");
        sandbox
            .config("allow/net/connect^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_324() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..10 {
            sandbox
                .config("deny/net/bind+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        for _ in 0..10 {
            sandbox
                .config("deny/net/bind-any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_325() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..10 {
            sandbox
                .config("deny/net/bind+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 10, "{sandbox}");
        sandbox
            .config("deny/net/bind^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_326() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..11 {
            sandbox
                .config("deny/net/connect+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 11, "{sandbox}");
        for _ in 0..11 {
            sandbox
                .config("deny/net/connect-any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_327() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..11 {
            sandbox
                .config("deny/net/connect+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 11, "{sandbox}");
        sandbox
            .config("deny/net/connect^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_328() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..12 {
            sandbox
                .config("filter/net/bind+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        for _ in 0..12 {
            sandbox
                .config("filter/net/bind-any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_329() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..12 {
            sandbox
                .config("filter/net/bind+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 12, "{sandbox}");
        sandbox
            .config("filter/net/bind^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_330() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..13 {
            sandbox
                .config("filter/net/connect+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 13, "{sandbox}");
        for _ in 0..13 {
            sandbox
                .config("filter/net/connect-any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_331() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..13 {
            sandbox
                .config("filter/net/connect+any6!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 13, "{sandbox}");
        sandbox
            .config("filter/net/connect^any6!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_332() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..14 {
            sandbox
                .config("allow/net/bind+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 14 * 2, "{sandbox}");
        for _ in 0..14 {
            sandbox
                .config("allow/net/bind-any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_333() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..14 {
            sandbox
                .config("allow/net/bind+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 14 * 2, "{sandbox}");
        sandbox
            .config("allow/net/bind^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_334() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..15 {
            sandbox
                .config("allow/net/connect+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 15 * 2, "{sandbox}");
        for _ in 0..15 {
            sandbox
                .config("allow/net/connect-any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_335() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..15 {
            sandbox
                .config("allow/net/connect+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 15 * 2, "{sandbox}");
        sandbox
            .config("allow/net/connect^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_336() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..16 {
            sandbox
                .config("deny/net/bind+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 16 * 2, "{sandbox}");
        for _ in 0..16 {
            sandbox
                .config("deny/net/bind-any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_337() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..16 {
            sandbox
                .config("deny/net/bind+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 16 * 2, "{sandbox}");
        sandbox
            .config("deny/net/bind^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_338() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..17 {
            sandbox
                .config("deny/net/connect+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 17 * 2, "{sandbox}");
        for _ in 0..17 {
            sandbox
                .config("deny/net/connect-any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_339() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..17 {
            sandbox
                .config("deny/net/connect+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 17 * 2, "{sandbox}");
        sandbox
            .config("deny/net/connect^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_340() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..18 {
            sandbox
                .config("filter/net/bind+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 18 * 2, "{sandbox}");
        for _ in 0..18 {
            sandbox
                .config("filter/net/bind-any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_341() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..18 {
            sandbox
                .config("filter/net/bind+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 18 * 2, "{sandbox}");
        sandbox
            .config("filter/net/bind^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_342() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..19 {
            sandbox
                .config("filter/net/connect+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 19 * 2, "{sandbox}");
        for _ in 0..19 {
            sandbox
                .config("filter/net/connect-any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_343() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        for _ in 0..19 {
            sandbox
                .config("filter/net/connect+any!0")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.cidr_rules.len(), 19 * 2, "{sandbox}");
        sandbox
            .config("filter/net/connect^any!0")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_344() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("allow/stat-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_345() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("allow/stat+/usr/*")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/stat+/usr")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 5, "{sandbox}");
        sandbox
            .config("allow/stat^/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_346() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("deny/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");
        sandbox
            .config("deny/stat-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("deny/stat-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_347() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.glob_rules.len();

        sandbox
            .config("filter/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("filter/stat+/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 3, "{sandbox}");
        sandbox
            .config("filter/stat-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 2, "{sandbox}");
        sandbox
            .config("filter/stat-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("filter/stat-/usr/**")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.glob_rules.len(), len, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_348() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/bind+1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/bind-1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/bind+1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/bind-1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("allow/net/connect+1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("allow/net/connect-1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        sandbox
            .config("deny/net/connect+1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 1, "{sandbox}");
        sandbox
            .config("deny/net/connect-1.1.1.1!80")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.cidr_rules.len(), 0, "{sandbox}");

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_349() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert_eq!(sandbox.config("mask"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("mask+"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("mask-"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("mask^"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("mask!"), Err(Errno::EINVAL));
        assert_eq!(sandbox.config("mask!/proc"), Err(Errno::EINVAL));

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(sandbox.is_masked("/proc/cmdline"));
        assert!(sandbox.is_masked("/dev/kmsg"));
        assert!(sandbox.is_masked("/proc/kmsg"));
        assert!(!sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_350() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.mask_rules.len();
        sandbox
            .config("mask-/proc/cmdline")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len - 1, "{sandbox}");
        sandbox
            .config("mask-/proc/kmsg")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len - 2, "{sandbox}");
        sandbox
            .config("mask-/dev/kmsg")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len - 3, "{sandbox}");

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(!sandbox.is_masked("/proc/cmdline"));
        assert!(!sandbox.is_masked("/dev/kmsg"));
        assert!(!sandbox.is_masked("/proc/kmsg"));
        assert!(!sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_351() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.mask_rules.len();
        sandbox
            .config("mask+/proc/vers?on")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len + 1, "{sandbox}");

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(sandbox.is_masked("/proc/cmdline"));
        assert!(sandbox.is_masked("/dev/kmsg"));
        assert!(sandbox.is_masked("/proc/kmsg"));
        assert!(sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_352() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.mask_rules.len();
        sandbox
            .config("mask+/proc/vers?on")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len + 1, "{sandbox}");
        sandbox
            .config("mask-/proc/vers?on")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len, "{sandbox}");

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(sandbox.is_masked("/proc/cmdline"));
        assert!(sandbox.is_masked("/dev/kmsg"));
        assert!(sandbox.is_masked("/proc/kmsg"));
        assert!(!sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_353() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.mask_rules.len();
        sandbox
            .config("mask+/proc/***")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len + 2, "{sandbox}");
        sandbox
            .config("mask-/proc/***")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len, "{sandbox}");

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(sandbox.is_masked("/proc/cmdline"));
        assert!(sandbox.is_masked("/dev/kmsg"));
        assert!(sandbox.is_masked("/proc/kmsg"));
        assert!(!sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_354() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.mask_rules.len();
        for _ in 0..7 {
            sandbox
                .config("mask+/proc/vers?on")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.mask_rules.len(), len + 7, "{sandbox}");
        sandbox
            .config("mask^/proc/vers?on")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len, "{sandbox}");

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(sandbox.is_masked("/proc/cmdline"));
        assert!(sandbox.is_masked("/dev/kmsg"));
        assert!(sandbox.is_masked("/proc/kmsg"));
        assert!(!sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_config_rules_355() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        let len = sandbox.mask_rules.len();
        for _ in 0..7 {
            sandbox
                .config("mask+/proc/***")
                .map_err(anyhow::Error::msg)?;
        }
        assert_eq!(sandbox.mask_rules.len(), len + 14, "{sandbox}");
        sandbox
            .config("mask^/proc/***")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(sandbox.mask_rules.len(), len, "{sandbox}");

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert!(sandbox.is_masked("/proc/cmdline"));
        assert!(sandbox.is_masked("/dev/kmsg"));
        assert!(sandbox.is_masked("/proc/kmsg"));
        assert!(!sandbox.is_masked("/proc/version"));

        Ok(())
    }

    #[test]
    fn sandbox_glob_doublestar_does_not_match_basename() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox
            .config("allow/read+/dev/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/dev"),
            None,
            "/dev =~ /dev/**, {sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_glob_doublestar_matches_basename_with_slash() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox
            .config("allow/read+/dev/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/dev/"),
            Some(Action::Allow),
            "/dev/ !~ /dev/**, {sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_glob_doublestar_matches_pathname() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox
            .config("allow/read+/dev/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/dev/null"),
            Some(Action::Allow),
            "/dev/null !~ /dev/**, {sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_glob_triplestar_matches_basename() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox
            .config("allow/read+/dev/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/dev"),
            Some(Action::Allow),
            "/dev =~ /dev/***, {sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_glob_triplestar_matches_basename_with_slash() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox
            .config("allow/read+/dev/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/dev/"),
            Some(Action::Allow),
            "/dev/ !~ /dev/***, {sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_glob_triplestar_matches_pathname() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox
            .config("allow/read+/dev/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/dev/null"),
            Some(Action::Allow),
            "/dev/null !~ /dev/***, {sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_access_last_matching_rule_wins() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/read:on")
            .map_err(anyhow::Error::msg)?;

        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            None,
            "{sandbox}"
        );
        sandbox
            .config("allow/read+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/read+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("allow/read+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/read+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("deny/read-/etc/passwd")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("allow/read-/etc/passwd")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_READ, "/etc/passwd"),
            None,
            "{sandbox}"
        );

        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            None,
            "{sandbox}"
        );
        sandbox
            .config("allow/write+/etc/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/write+/etc/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("allow/write+/etc/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/write+/etc/**")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("deny/write-/etc/**")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("allow/write-/etc/**")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_WRITE, "/etc/passwd"),
            None,
            "{sandbox}"
        );

        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            None,
            "{sandbox}"
        );
        sandbox
            .config("allow/exec+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/exec+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("allow/exec+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/exec+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("deny/exec-/etc/***")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("allow/exec-/etc/***")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_EXEC, "/etc/passwd"),
            None,
            "{sandbox}"
        );

        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            None,
            "{sandbox}"
        );
        sandbox
            .config("allow/stat+/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/stat+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("allow/stat+/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/stat+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("deny/stat-/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/stat-/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("allow/stat-/***")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_STAT, "/etc/passwd"),
            None,
            "{sandbox}"
        );

        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            None,
            "{sandbox}"
        );
        sandbox
            .config("allow/net/bind+/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/net/bind+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("allow/net/bind+/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/net/bind+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("deny/net/bind-/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/bind-/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("allow/net/bind-/***")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_BIND, "/etc/passwd"),
            None,
            "{sandbox}"
        );

        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            None,
            "{sandbox}"
        );
        sandbox
            .config("allow/net/connect+/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/net/connect+/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("allow/net/connect+/***")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        sandbox
            .config("deny/net/connect+/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            Some(Action::Deny),
            "{sandbox}"
        );
        sandbox
            .config("deny/net/connect-/etc/***")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("deny/net/connect-/etc/passwd")
            .map_err(anyhow::Error::msg)?;
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            Some(Action::Allow),
            "{sandbox}"
        );
        for _ in 0..2 {
            sandbox
                .config("allow/net/connect-/***")
                .map_err(anyhow::Error::msg)?;
        }
        sandbox.build_globsets().map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.match_action(Capability::CAP_CONNECT, "/etc/passwd"),
            None,
            "{sandbox}"
        );

        Ok(())
    }

    #[test]
    fn sandbox_check_filter_ip_port_range() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        sandbox
            .config("sandbox/net:on")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/connect+any!0")
            .map_err(anyhow::Error::msg)?;

        let addr = "127.0.0.1".parse::<IpAddr>().unwrap();
        assert_eq!(
            sandbox.check_ip(Capability::CAP_CONNECT, addr, 0),
            (Action::Allow, true)
        );
        for port in 1..=65535 {
            assert_eq!(
                sandbox.check_ip(Capability::CAP_CONNECT, addr, port),
                (Action::Deny, false),
                "{addr}!{port} {sandbox}"
            );
        }

        sandbox
            .config("filter/net/connect+any!1-65535")
            .map_err(anyhow::Error::msg)?;
        assert_eq!(
            sandbox.check_ip(Capability::CAP_CONNECT, addr, 0),
            (Action::Allow, true)
        );
        for port in 1..=65535 {
            assert_eq!(
                sandbox.check_ip(Capability::CAP_CONNECT, addr, port),
                (Action::Deny, true),
                "{addr}!{port} {sandbox}"
            );
        }

        Ok(())
    }

    #[test]
    fn sandbox_check_netlink_01() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(
            sandbox.netlink_families.is_empty(),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );

        sandbox
            .config("allow/net/link+route")
            .map_err(anyhow::Error::msg)?;
        assert!(
            !sandbox.netlink_families.is_empty(),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );
        assert!(
            sandbox
                .netlink_families
                .contains(NetlinkFamily::NETLINK_ROUTE),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );

        Ok(())
    }

    #[test]
    fn sandbox_check_netlink_02() -> TestResult<()> {
        let mut sandbox = Sandbox::default();
        assert!(
            sandbox.netlink_families.is_empty(),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );

        sandbox
            .config("allow/net/link+route,crypto,ecryptfs,inet_diag")
            .map_err(anyhow::Error::msg)?;
        sandbox
            .config("allow/net/link-ecryptfs,sock_diag")
            .map_err(anyhow::Error::msg)?;
        assert!(
            !sandbox.netlink_families.is_empty(),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );
        assert!(
            sandbox
                .netlink_families
                .contains(NetlinkFamily::NETLINK_ROUTE),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );
        assert!(
            sandbox
                .netlink_families
                .contains(NetlinkFamily::NETLINK_CRYPTO),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );
        assert!(
            !sandbox
                .netlink_families
                .contains(NetlinkFamily::NETLINK_ECRYPTFS),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );
        assert!(
            !sandbox
                .netlink_families
                .contains(NetlinkFamily::NETLINK_SOCK_DIAG),
            "netlink_families:{}",
            sandbox.netlink_families.bits()
        );

        Ok(())
    }
}
