//
// Syd: rock-solid unikernel
// src/test/test.rs: Integration tests
//
// Copyright (c) 2023, 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#![allow(non_snake_case)]
#![allow(clippy::disallowed_methods)]

use std::{
    env,
    fs::{create_dir_all, metadata, set_permissions, write, File},
    io::{BufReader, Read, Write},
    os::{
        fd::AsRawFd,
        unix::{
            fs::{symlink, PermissionsExt},
            process::ExitStatusExt,
        },
    },
    path::Path,
    process::{Command, Stdio},
    thread::sleep,
    time::Duration,
};

use nix::{
    sys::stat::{umask, Mode},
    unistd::{pipe, unlink, Pid, Uid},
};
use syd::{fs::path_split, hash::HashAlgorithm};

use crate::{
    assert, assert_eq, assert_status_aborted, assert_status_code, assert_status_code_matches,
    assert_status_killed, assert_status_not_ok, assert_status_not_supported, assert_status_ok,
    fixup, ignore, skip_unless_available, skip_unless_bitness, skip_unless_cap,
    skip_unless_coredumps, skip_unless_exists, skip_unless_feature, skip_unless_unshare, util::*,
};

const EX_SIGIOT: i32 = 128 + nix::libc::SIGIOT;
const EX_SIGSEGV: i32 = 128 + nix::libc::SIGSEGV;

const NONE: &[&str] = &[];

/// Represents a test case.
pub type Test<'a> = (&'a str, fn() -> TestResult);

macro_rules! test_entry {
    ($func:expr) => {
        (stringify!($func), $func)
    };
}

/// List of integration tests.
pub const TESTS: &[Test] = &[
    test_entry!(test_syd_version),
    test_entry!(test_syd_export),
    test_entry!(test_syd_true_returns_success),
    test_entry!(test_syd_true_returns_success_with_many_processes),
    test_entry!(test_syd_true_returns_success_with_many_threads),
    test_entry!(test_syd_false_returns_failure),
    test_entry!(test_syd_true_returns_failure_with_many_processes),
    test_entry!(test_syd_true_returns_failure_with_many_threads),
    test_entry!(test_syd_empty_file_returns_enoexec),
    test_entry!(test_syd_non_executable_file_returns_eacces_empty),
    test_entry!(test_syd_non_executable_file_returns_eacces_binary),
    test_entry!(test_syd_non_executable_file_returns_eacces_script),
    test_entry!(test_syd_sigint_returns_130),
    test_entry!(test_syd_sigabrt_returns_134),
    test_entry!(test_syd_sigkill_returns_137),
    test_entry!(test_syd_reap_zombies_bare),
    test_entry!(test_syd_reap_zombies_wrap),
    test_entry!(test_syd_whoami_returns_root_fake),
    test_entry!(test_syd_whoami_returns_root_user),
    test_entry!(test_syd_setuid_nobody_default),
    test_entry!(test_syd_setuid_nobody_safesetid_deny),
    test_entry!(test_syd_setuid_root_safesetid_deny),
    test_entry!(test_syd_setuid_nobody_safesetid_allow),
    test_entry!(test_syd_setgid_nobody_default),
    test_entry!(test_syd_setgid_nobody_safesetid_deny),
    test_entry!(test_syd_setgid_root_safesetid_deny),
    test_entry!(test_syd_setgid_nobody_safesetid_allow),
    test_entry!(test_syd_setreuid_nobody_default_1),
    test_entry!(test_syd_setreuid_nobody_default_2),
    test_entry!(test_syd_setreuid_nobody_default_3),
    test_entry!(test_syd_setreuid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setreuid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setreuid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setreuid_root_safesetid_deny_1),
    test_entry!(test_syd_setreuid_root_safesetid_deny_2),
    test_entry!(test_syd_setreuid_root_safesetid_deny_3),
    test_entry!(test_syd_setreuid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setreuid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setreuid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setregid_nobody_default_1),
    test_entry!(test_syd_setregid_nobody_default_2),
    test_entry!(test_syd_setregid_nobody_default_3),
    test_entry!(test_syd_setregid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setregid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setregid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setregid_root_safesetid_deny_1),
    test_entry!(test_syd_setregid_root_safesetid_deny_2),
    test_entry!(test_syd_setregid_root_safesetid_deny_3),
    test_entry!(test_syd_setregid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setregid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setregid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setresuid_nobody_default_1),
    test_entry!(test_syd_setresuid_nobody_default_2),
    test_entry!(test_syd_setresuid_nobody_default_3),
    test_entry!(test_syd_setresuid_nobody_default_4),
    test_entry!(test_syd_setresuid_nobody_default_5),
    test_entry!(test_syd_setresuid_nobody_default_6),
    test_entry!(test_syd_setresuid_nobody_default_7),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_4),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_5),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_6),
    test_entry!(test_syd_setresuid_nobody_safesetid_deny_7),
    test_entry!(test_syd_setresuid_root_safesetid_deny_1),
    test_entry!(test_syd_setresuid_root_safesetid_deny_2),
    test_entry!(test_syd_setresuid_root_safesetid_deny_3),
    test_entry!(test_syd_setresuid_root_safesetid_deny_4),
    test_entry!(test_syd_setresuid_root_safesetid_deny_5),
    test_entry!(test_syd_setresuid_root_safesetid_deny_6),
    test_entry!(test_syd_setresuid_root_safesetid_deny_7),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_4),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_5),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_6),
    test_entry!(test_syd_setresuid_nobody_safesetid_allow_7),
    test_entry!(test_syd_setresgid_nobody_default_1),
    test_entry!(test_syd_setresgid_nobody_default_2),
    test_entry!(test_syd_setresgid_nobody_default_3),
    test_entry!(test_syd_setresgid_nobody_default_4),
    test_entry!(test_syd_setresgid_nobody_default_5),
    test_entry!(test_syd_setresgid_nobody_default_6),
    test_entry!(test_syd_setresgid_nobody_default_7),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_1),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_2),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_3),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_4),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_5),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_6),
    test_entry!(test_syd_setresgid_nobody_safesetid_deny_7),
    test_entry!(test_syd_setresgid_root_safesetid_deny_1),
    test_entry!(test_syd_setresgid_root_safesetid_deny_2),
    test_entry!(test_syd_setresgid_root_safesetid_deny_3),
    test_entry!(test_syd_setresgid_root_safesetid_deny_4),
    test_entry!(test_syd_setresgid_root_safesetid_deny_5),
    test_entry!(test_syd_setresgid_root_safesetid_deny_6),
    test_entry!(test_syd_setresgid_root_safesetid_deny_7),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_1),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_2),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_3),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_4),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_5),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_6),
    test_entry!(test_syd_setresgid_nobody_safesetid_allow_7),
    test_entry!(test_syd_drop_cap_sys_ptrace_exec_default),
    test_entry!(test_syd_drop_cap_sys_ptrace_exec_unsafe_caps),
    test_entry!(test_syd_drop_cap_sys_ptrace_exec_unsafe_debug),
    test_entry!(test_syd_drop_cap_chown_exec_default),
    test_entry!(test_syd_drop_cap_chown_exec_unsafe),
    test_entry!(test_syd_drop_cap_chown_exec_safesetid),
    test_entry!(test_syd_drop_cap_setgid_exec_default),
    test_entry!(test_syd_drop_cap_setgid_exec_unsafe),
    test_entry!(test_syd_drop_cap_setgid_exec_safesetid),
    test_entry!(test_syd_drop_cap_setuid_exec_default),
    test_entry!(test_syd_drop_cap_setuid_exec_unsafe),
    test_entry!(test_syd_drop_cap_setuid_exec_safesetid),
    test_entry!(test_syd_drop_cap_net_bind_service_exec_default),
    test_entry!(test_syd_drop_cap_net_bind_service_exec_unsafe_caps),
    test_entry!(test_syd_drop_cap_net_bind_service_exec_unsafe_bind),
    test_entry!(test_syd_drop_cap_net_raw_exec_default),
    test_entry!(test_syd_drop_cap_net_raw_exec_unsafe_caps),
    test_entry!(test_syd_drop_cap_net_raw_exec_unsafe_socket),
    test_entry!(test_syd_drop_cap_sys_time_exec_default),
    test_entry!(test_syd_drop_cap_sys_time_exec_unsafe_caps),
    test_entry!(test_syd_drop_cap_sys_time_exec_unsafe_time),
    test_entry!(test_syd_drop_cap_syslog_exec_default),
    test_entry!(test_syd_drop_cap_syslog_exec_unsafe_caps),
    test_entry!(test_syd_drop_cap_syslog_exec_unsafe_syslog),
    test_entry!(test_syd_drop_cap_sys_ptrace_load_default),
    test_entry!(test_syd_drop_cap_sys_ptrace_load_unsafe_caps),
    test_entry!(test_syd_drop_cap_sys_ptrace_load_unsafe_debug),
    test_entry!(test_syd_drop_cap_chown_load_default),
    test_entry!(test_syd_drop_cap_chown_load_unsafe),
    test_entry!(test_syd_drop_cap_chown_load_safesetid),
    test_entry!(test_syd_drop_cap_setgid_load_default),
    test_entry!(test_syd_drop_cap_setgid_load_unsafe),
    test_entry!(test_syd_drop_cap_setgid_load_safesetid),
    test_entry!(test_syd_drop_cap_setuid_load_default),
    test_entry!(test_syd_drop_cap_setuid_load_unsafe),
    test_entry!(test_syd_drop_cap_setuid_load_safesetid),
    test_entry!(test_syd_drop_cap_net_bind_service_load_default),
    test_entry!(test_syd_drop_cap_net_bind_service_load_unsafe_caps),
    test_entry!(test_syd_drop_cap_net_bind_service_load_unsafe_bind),
    test_entry!(test_syd_drop_cap_net_raw_load_default),
    test_entry!(test_syd_drop_cap_net_raw_load_unsafe_caps),
    test_entry!(test_syd_drop_cap_net_raw_load_unsafe_socket),
    test_entry!(test_syd_drop_cap_sys_time_load_default),
    test_entry!(test_syd_drop_cap_sys_time_load_unsafe_caps),
    test_entry!(test_syd_drop_cap_sys_time_load_unsafe_time),
    test_entry!(test_syd_drop_cap_syslog_load_default),
    test_entry!(test_syd_drop_cap_syslog_load_unsafe_caps),
    test_entry!(test_syd_drop_cap_syslog_load_unsafe_syslog),
    test_entry!(test_syd_userns_drop_cap_sys_ptrace_exec_default),
    test_entry!(test_syd_userns_drop_cap_sys_ptrace_exec_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_sys_ptrace_exec_unsafe_debug),
    test_entry!(test_syd_userns_drop_cap_chown_exec_default),
    test_entry!(test_syd_userns_drop_cap_chown_exec_unsafe),
    test_entry!(test_syd_userns_drop_cap_chown_exec_safesetid),
    test_entry!(test_syd_userns_drop_cap_setgid_exec_default),
    test_entry!(test_syd_userns_drop_cap_setgid_exec_unsafe),
    test_entry!(test_syd_userns_drop_cap_setgid_exec_safesetid),
    test_entry!(test_syd_userns_drop_cap_setuid_exec_default),
    test_entry!(test_syd_userns_drop_cap_setuid_exec_unsafe),
    test_entry!(test_syd_userns_drop_cap_setuid_exec_safesetid),
    test_entry!(test_syd_userns_drop_cap_net_bind_service_exec_default),
    test_entry!(test_syd_userns_drop_cap_net_bind_service_exec_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_net_bind_service_exec_unsafe_bind),
    test_entry!(test_syd_userns_drop_cap_net_raw_exec_default),
    test_entry!(test_syd_userns_drop_cap_net_raw_exec_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_net_raw_exec_unsafe_socket),
    test_entry!(test_syd_userns_drop_cap_sys_time_exec_default),
    test_entry!(test_syd_userns_drop_cap_sys_time_exec_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_sys_time_exec_unsafe_time),
    test_entry!(test_syd_userns_drop_cap_syslog_exec_default),
    test_entry!(test_syd_userns_drop_cap_syslog_exec_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_syslog_exec_unsafe_syslog),
    test_entry!(test_syd_userns_drop_cap_sys_ptrace_load_default),
    test_entry!(test_syd_userns_drop_cap_sys_ptrace_load_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_sys_ptrace_load_unsafe_debug),
    test_entry!(test_syd_userns_drop_cap_chown_load_default),
    test_entry!(test_syd_userns_drop_cap_chown_load_unsafe),
    test_entry!(test_syd_userns_drop_cap_chown_load_safesetid),
    test_entry!(test_syd_userns_drop_cap_setgid_load_default),
    test_entry!(test_syd_userns_drop_cap_setgid_load_unsafe),
    test_entry!(test_syd_userns_drop_cap_setgid_load_safesetid),
    test_entry!(test_syd_userns_drop_cap_setuid_load_default),
    test_entry!(test_syd_userns_drop_cap_setuid_load_unsafe),
    test_entry!(test_syd_userns_drop_cap_setuid_load_safesetid),
    test_entry!(test_syd_userns_drop_cap_net_bind_service_load_default),
    test_entry!(test_syd_userns_drop_cap_net_bind_service_load_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_net_bind_service_load_unsafe_bind),
    test_entry!(test_syd_userns_drop_cap_net_raw_load_default),
    test_entry!(test_syd_userns_drop_cap_net_raw_load_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_net_raw_load_unsafe_socket),
    test_entry!(test_syd_userns_drop_cap_sys_time_load_default),
    test_entry!(test_syd_userns_drop_cap_sys_time_load_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_sys_time_load_unsafe_time),
    test_entry!(test_syd_userns_drop_cap_syslog_load_default),
    test_entry!(test_syd_userns_drop_cap_syslog_load_unsafe_caps),
    test_entry!(test_syd_userns_drop_cap_syslog_load_unsafe_syslog),
    test_entry!(test_syd_readlink_insecure_world_writable),
    test_entry!(test_syd_socket_domain_restrictions),
    test_entry!(test_syd_tkill_restrictions),
    test_entry!(test_syd_environment_filter),
    test_entry!(test_syd_environment_harden),
    test_entry!(test_syd_lock),
    test_entry!(test_syd_lock_exec),
    test_entry!(test_syd_lock_prevents_further_cli_args),
    test_entry!(test_syd_lock_prevents_further_cfg_items),
    test_entry!(test_syd_lock_prevents_further_inc_items),
    test_entry!(test_syd_chk),
    test_entry!(test_syd_cmd_exec_with_lock_default),
    test_entry!(test_syd_cmd_exec_with_lock_on),
    test_entry!(test_syd_cmd_exec_with_lock_off_1),
    test_entry!(test_syd_cmd_exec_with_lock_off_2),
    test_entry!(test_syd_cmd_exec_with_lock_exec_1),
    test_entry!(test_syd_cmd_exec_with_lock_exec_2),
    test_entry!(test_syd_parse_config),
    test_entry!(test_syd_include_config),
    test_entry!(test_syd_shellexpand_01),
    test_entry!(test_syd_shellexpand_02),
    test_entry!(test_syd_shellexpand_03),
    test_entry!(test_syd_shellexpand_04),
    test_entry!(test_syd_lock_personality),
    test_entry!(test_syd_mdwe_protections),
    test_entry!(test_syd_mdwe_relax),
    test_entry!(test_syd_mmap_fixed_null),
    test_entry!(test_syd_load_library),
    test_entry!(test_syd_load_library_noexec),
    test_entry!(test_syd_load_library_abort_after_load),
    test_entry!(test_syd_load_library_abort_at_startup),
    test_entry!(test_syd_load_library_check_fd_leaks_bare),
    test_entry!(test_syd_load_library_check_fd_leaks_wrap),
    test_entry!(test_syd_load_library_check_fd_leaks_init_bare),
    test_entry!(test_syd_load_library_check_fd_leaks_init_wrap),
    test_entry!(test_syd_read_sandbox_open_allow),
    test_entry!(test_syd_read_sandbox_open_deny),
    test_entry!(test_syd_stat_sandbox_chdir_allow),
    test_entry!(test_syd_stat_sandbox_chdir_hide),
    test_entry!(test_syd_stat_sandbox_stat_allow),
    test_entry!(test_syd_stat_sandbox_stat_hide),
    test_entry!(test_syd_stat_sandbox_getdents_allow),
    test_entry!(test_syd_stat_sandbox_getdents_hide),
    test_entry!(test_syd_stat_bypass_with_read),
    test_entry!(test_syd_stat_bypass_with_write),
    test_entry!(test_syd_stat_bypass_with_exec),
    test_entry!(test_syd_write_sandbox_open_allow),
    test_entry!(test_syd_write_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_open_allow),
    test_entry!(test_syd_exec_sandbox_open_deny),
    test_entry!(test_syd_exec_sandbox_deny_binfmt_script),
    test_entry!(test_syd_exec_sandbox_many_binfmt_script),
    test_entry!(test_syd_exec_sandbox_prevent_library_injection_dlopen_bare),
    test_entry!(test_syd_exec_sandbox_prevent_library_injection_dlopen_wrap),
    test_entry!(test_syd_exec_sandbox_prevent_library_injection_LD_LIBRARY_PATH),
    test_entry!(test_syd_exec_sandbox_prevent_library_injection_LD_PRELOAD),
    test_entry!(test_syd_network_sandbox_connect_ipv4_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv4_deny),
    test_entry!(test_syd_network_sandbox_connect_ipv6_allow),
    test_entry!(test_syd_network_sandbox_connect_ipv6_deny),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv4_failure),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv4_success),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv6_failure),
    test_entry!(test_syd_network_sandbox_allow_safe_bind_ipv6_success),
    test_entry!(test_syd_handle_toolong_unix_connect),
    test_entry!(test_syd_handle_toolong_unix_sendto),
    test_entry!(test_syd_handle_toolong_unix_sendmsg),
    test_entry!(test_syd_sendmsg_scm_rights_one),
    test_entry!(test_syd_sendmsg_scm_rights_many),
    test_entry!(test_syd_sendmmsg),
    test_entry!(test_syd_mask_simple),
    test_entry!(test_syd_truncate),
    test_entry!(test_syd_truncate64),
    test_entry!(test_syd_kcapi_hash_block),
    test_entry!(test_syd_kcapi_hash_stream),
    test_entry!(test_syd_kcapi_cipher_block),
    test_entry!(test_syd_crypt_sandboxing_bscan_append_cmp_mini_copy),
    test_entry!(test_syd_crypt_sandboxing_bscan_append_aes_mini_copy),
    test_entry!(test_syd_crypt_sandboxing_bscan_append_cmp_incr_copy),
    test_entry!(test_syd_crypt_sandboxing_bscan_append_aes_incr_copy),
    test_entry!(test_syd_crypt_sandboxing_bscan_append_cmp_decr_copy),
    test_entry!(test_syd_crypt_sandboxing_bscan_append_aes_decr_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_cmp_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_aes_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_cmp_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_aes_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_single_cmp_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_single_aes_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_append_cmp_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_append_aes_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_cmp_nano_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_aes_nano_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_cmp_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_aes_tiny_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_cmp_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_aes_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_cmp_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_sieve_append_aes_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_cmp_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_aes_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_cmp_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_aes_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_single_cmp_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_single_aes_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_append_cmp_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_append_aes_mild_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_cmp_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_aes_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_cmp_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_aes_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_single_cmp_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_single_aes_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_append_cmp_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_prime_append_aes_huge_copy),
    test_entry!(test_syd_crypt_sandboxing_single_cmp_rand_copy),
    test_entry!(test_syd_crypt_sandboxing_single_aes_rand_copy),
    test_entry!(test_syd_crypt_sandboxing_append_cmp_rand_copy),
    test_entry!(test_syd_crypt_sandboxing_append_aes_rand_copy),
    test_entry!(test_syd_crypt_sandboxing_append_cmp_fuzz_copy),
    test_entry!(test_syd_crypt_sandboxing_append_aes_fuzz_copy),
    test_entry!(test_syd_crypt_sandboxing_append_cmp_zero_copy),
    test_entry!(test_syd_crypt_sandboxing_append_aes_zero_copy),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_cmp_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_aes_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_cmp_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_aes_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_single_cmp_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_single_aes_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_append_cmp_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_append_aes_tiny_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_cmp_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_aes_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_cmp_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_aes_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_single_cmp_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_single_aes_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_append_cmp_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_append_aes_mild_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_cmp_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_single_aes_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_cmp_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_bsize_append_aes_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_single_cmp_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_single_aes_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_append_cmp_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_prime_append_aes_huge_pipe),
    test_entry!(test_syd_crypt_sandboxing_single_cmp_null_copy),
    test_entry!(test_syd_crypt_sandboxing_single_aes_null_copy),
    test_entry!(test_syd_crypt_sandboxing_single_cmp_null_pipe),
    test_entry!(test_syd_crypt_sandboxing_single_aes_null_pipe),
    test_entry!(test_syd_crypt_sandboxing_single_cmp_rand_pipe),
    test_entry!(test_syd_crypt_sandboxing_single_aes_rand_pipe),
    test_entry!(test_syd_crypt_sandboxing_append_cmp_rand_pipe),
    test_entry!(test_syd_crypt_sandboxing_append_aes_rand_pipe),
    test_entry!(test_syd_crypt_sandboxing_append_cmp_fuzz_pipe),
    test_entry!(test_syd_crypt_sandboxing_append_aes_fuzz_pipe),
    test_entry!(test_syd_crypt_sandboxing_append_cmp_zero_pipe),
    test_entry!(test_syd_crypt_sandboxing_append_aes_zero_pipe),
    test_entry!(test_syd_exit_wait_default),
    test_entry!(test_syd_exit_wait_default_unsafe_exec),
    test_entry!(test_syd_exit_wait_pid),
    test_entry!(test_syd_exit_wait_pid_unsafe_exec),
    test_entry!(test_syd_exit_wait_pid_with_runaway_cmd_exec_process),
    test_entry!(test_syd_exit_wait_pid_unsafe_exec_with_runaway_cmd_exec_process),
    test_entry!(test_syd_exit_wait_all),
    test_entry!(test_syd_exit_wait_all_unsafe_exec),
    test_entry!(test_syd_exit_wait_all_with_runaway_cmd_exec_process),
    test_entry!(test_syd_exit_wait_all_unsafe_exec_with_runaway_cmd_exec_process),
    test_entry!(test_syd_cli_args_override_user_profile),
    test_entry!(test_syd_ifconfig_lo_bare),
    test_entry!(test_syd_ifconfig_lo_wrap),
    test_entry!(test_syd_binfmt_malformed_too_short),
    test_entry!(test_syd_binfmt_malformed_bad_magic),
    test_entry!(test_syd_binfmt_malformed_invalid_elf_class),
    test_entry!(test_syd_deny_elf32),
    test_entry!(test_syd_deny_elf_dynamic),
    test_entry!(test_syd_deny_elf_static),
    test_entry!(test_syd_deny_script),
    test_entry!(test_syd_prevent_ld_linux_exec_break),
    test_entry!(test_syd_force_sandbox),
    test_entry!(test_syd_segvguard_core_safe),
    test_entry!(test_syd_segvguard_core_unsafe),
    test_entry!(test_syd_segvguard_suspension_safe),
    test_entry!(test_syd_segvguard_suspension_unsafe),
    test_entry!(test_syd_symlink_toctou),
    test_entry!(test_syd_symlinkat_toctou),
    test_entry!(test_syd_ptrmod_toctou_chdir),
    test_entry!(test_syd_ptrmod_toctou_exec_fail),
    test_entry!(test_syd_ptrmod_toctou_exec_success_quick),
    test_entry!(test_syd_ptrmod_toctou_exec_success_double_fork),
    test_entry!(test_syd_ptrmod_toctou_exec_success_quick_no_mitigation),
    test_entry!(test_syd_ptrmod_toctou_exec_success_double_fork_no_mitigation),
    test_entry!(test_syd_ptrmod_toctou_open),
    test_entry!(test_syd_ptrmod_toctou_creat),
    test_entry!(test_syd_ptrmod_toctou_opath_default),
    test_entry!(test_syd_ptrmod_toctou_opath_unsafe),
    test_entry!(test_syd_vfsmod_toctou_mmap),
    test_entry!(test_syd_io_uring_escape_strict),
    test_entry!(test_syd_io_uring_escape_unsafe),
    test_entry!(test_syd_opath_escape),
    test_entry!(test_syd_devfd_escape_chdir),
    test_entry!(test_syd_devfd_escape_chdir_relpath_1),
    test_entry!(test_syd_devfd_escape_chdir_relpath_2),
    test_entry!(test_syd_devfd_escape_chdir_relpath_3),
    test_entry!(test_syd_devfd_escape_chdir_relpath_4),
    test_entry!(test_syd_devfd_escape_chdir_relpath_5),
    test_entry!(test_syd_devfd_escape_chdir_relpath_6),
    test_entry!(test_syd_devfd_escape_chdir_relpath_7),
    test_entry!(test_syd_devfd_escape_chdir_relpath_8),
    test_entry!(test_syd_devfd_escape_chdir_relpath_9),
    test_entry!(test_syd_devfd_escape_chdir_relpath_10),
    test_entry!(test_syd_devfd_escape_chdir_relpath_11),
    test_entry!(test_syd_devfd_escape_chdir_relpath_12),
    test_entry!(test_syd_devfd_escape_chdir_relpath_13),
    test_entry!(test_syd_devfd_escape_chdir_relpath_14),
    test_entry!(test_syd_devfd_escape_chdir_relpath_15),
    test_entry!(test_syd_devfd_escape_chdir_relpath_16),
    test_entry!(test_syd_devfd_escape_chdir_relpath_17),
    test_entry!(test_syd_devfd_escape_chdir_relpath_18),
    test_entry!(test_syd_devfd_escape_chdir_relpath_19),
    test_entry!(test_syd_devfd_escape_chdir_relpath_20),
    test_entry!(test_syd_devfd_escape_open),
    test_entry!(test_syd_devfd_escape_open_relpath_1),
    test_entry!(test_syd_devfd_escape_open_relpath_2),
    test_entry!(test_syd_devfd_escape_open_relpath_3),
    test_entry!(test_syd_devfd_escape_open_relpath_4),
    test_entry!(test_syd_devfd_escape_open_relpath_5),
    test_entry!(test_syd_devfd_escape_open_relpath_6),
    test_entry!(test_syd_devfd_escape_open_relpath_7),
    test_entry!(test_syd_devfd_escape_open_relpath_8),
    test_entry!(test_syd_devfd_escape_open_relpath_9),
    test_entry!(test_syd_devfd_escape_open_relpath_10),
    test_entry!(test_syd_devfd_escape_open_relpath_11),
    test_entry!(test_syd_devfd_escape_open_relpath_12),
    test_entry!(test_syd_devfd_escape_open_relpath_13),
    test_entry!(test_syd_devfd_escape_open_relpath_14),
    test_entry!(test_syd_devfd_escape_open_relpath_15),
    test_entry!(test_syd_devfd_escape_open_relpath_16),
    test_entry!(test_syd_devfd_escape_open_relpath_17),
    test_entry!(test_syd_devfd_escape_open_relpath_18),
    test_entry!(test_syd_devfd_escape_open_relpath_19),
    test_entry!(test_syd_devfd_escape_open_relpath_20),
    test_entry!(test_syd_procself_escape_chdir),
    test_entry!(test_syd_procself_escape_chdir_relpath_1),
    test_entry!(test_syd_procself_escape_chdir_relpath_2),
    test_entry!(test_syd_procself_escape_chdir_relpath_3),
    test_entry!(test_syd_procself_escape_chdir_relpath_4),
    test_entry!(test_syd_procself_escape_chdir_relpath_5),
    test_entry!(test_syd_procself_escape_chdir_relpath_6),
    test_entry!(test_syd_procself_escape_chdir_relpath_7),
    test_entry!(test_syd_procself_escape_open),
    test_entry!(test_syd_procself_escape_open_relpath_1),
    test_entry!(test_syd_procself_escape_open_relpath_2),
    test_entry!(test_syd_procself_escape_open_relpath_3),
    test_entry!(test_syd_procself_escape_open_relpath_4),
    test_entry!(test_syd_procself_escape_open_relpath_5),
    test_entry!(test_syd_procself_escape_open_relpath_6),
    test_entry!(test_syd_procself_escape_open_relpath_7),
    test_entry!(test_syd_procself_escape_relpath),
    test_entry!(test_syd_procself_escape_symlink),
    test_entry!(test_syd_procself_escape_symlink_within_container),
    test_entry!(test_syd_umask_bypass_077),
    test_entry!(test_syd_umask_bypass_277),
    test_entry!(test_syd_emulate_opath),
    test_entry!(test_syd_emulate_otmpfile),
    test_entry!(test_syd_honor_umask_000),
    test_entry!(test_syd_honor_umask_022),
    test_entry!(test_syd_honor_umask_077),
    test_entry!(test_syd_force_umask_bypass_with_open),
    test_entry!(test_syd_force_umask_bypass_with_mknod),
    test_entry!(test_syd_force_umask_bypass_with_mkdir),
    test_entry!(test_syd_force_umask_bypass_with_fchmod),
    test_entry!(test_syd_open_utf8_invalid),
    test_entry!(test_syd_exec_in_inaccessible_directory),
    test_entry!(test_syd_fstat_on_pipe),
    test_entry!(test_syd_fstat_on_socket),
    test_entry!(test_syd_fstat_on_deleted_file),
    test_entry!(test_syd_fstat_on_temp_file),
    test_entry!(test_syd_fchmodat_on_proc_fd),
    test_entry!(test_syd_linkat_on_fd),
    test_entry!(test_syd_block_ioctl_tiocsti),
    test_entry!(test_syd_block_prctl_ptrace),
    test_entry!(test_syd_prevent_ptrace_detect),
    test_entry!(test_syd_block_dev_random),
    test_entry!(test_syd_block_dev_urandom),
    test_entry!(test_syd_kill_during_syscall),
    test_entry!(test_syd_open_toolong_path),
    test_entry!(test_syd_open_null_path),
    test_entry!(test_syd_utimensat_null),
    test_entry!(test_syd_normalize_path),
    test_entry!(test_syd_path_resolution),
    test_entry!(test_syd_remove_empty_path),
    test_entry!(test_syd_symlink_readonly_path),
    test_entry!(test_syd_open_trailing_slash),
    test_entry!(test_syd_openat_trailing_slash),
    test_entry!(test_syd_lstat_trailing_slash),
    test_entry!(test_syd_fstatat_trailing_slash),
    test_entry!(test_syd_mkdir_trailing_dot),
    test_entry!(test_syd_mkdirat_trailing_dot),
    test_entry!(test_syd_rmdir_trailing_slashdot),
    test_entry!(test_syd_mkdir_eexist_escape),
    test_entry!(test_syd_mkdirat_eexist_escape),
    test_entry!(test_syd_mknod_eexist_escape),
    test_entry!(test_syd_mknodat_eexist_escape),
    test_entry!(test_syd_fopen_supports_mode_e),
    test_entry!(test_syd_fopen_supports_mode_x),
    test_entry!(test_syd_link_no_symlink_deref),
    test_entry!(test_syd_link_posix),
    test_entry!(test_syd_linkat_posix),
    test_entry!(test_syd_getcwd_long),
    test_entry!(test_syd_creat_thru_dangling),
    test_entry!(test_syd_mkdirat_non_dir_fd),
    test_entry!(test_syd_blocking_udp4),
    test_entry!(test_syd_blocking_udp6),
    test_entry!(test_syd_close_on_exec),
    test_entry!(test_syd_open_exclusive_restart),
    test_entry!(test_syd_open_exclusive_repeat),
    test_entry!(test_syd_pty_io_rust),
    test_entry!(test_syd_pty_io_gawk),
    test_entry!(test_syd_diff_dev_fd),
    test_entry!(test_syd_fifo_multiple_readers),
    test_entry!(test_syd_bind_unix_socket),
    test_entry!(test_syd_signal_protection_simple),
    test_entry!(test_syd_signal_protection_killpg_0),
    test_entry!(test_syd_signal_protection_killpg_self),
    test_entry!(test_syd_signal_protection_killpg_syd),
    test_entry!(test_syd_signal_protection_mass_0),
    test_entry!(test_syd_signal_protection_mass_int),
    test_entry!(test_syd_emulate_open_fifo_1),
    test_entry!(test_syd_emulate_open_fifo_2),
    test_entry!(test_syd_deny_magiclinks),
    test_entry!(test_syd_lstat_magiclinks),
    test_entry!(test_syd_access_unsafe_paths),
    test_entry!(test_syd_access_proc_cmdline),
    test_entry!(test_syd_mkdir_with_control_chars_default),
    test_entry!(test_syd_mkdir_with_control_chars_unsafe),
    test_entry!(test_syd_touch_with_control_chars_default),
    test_entry!(test_syd_touch_with_control_chars_unsafe),
    test_entry!(test_syd_fanotify_mark_cwd_allow),
    test_entry!(test_syd_fanotify_mark_cwd_deny),
    test_entry!(test_syd_fanotify_mark_dir_allow),
    test_entry!(test_syd_fanotify_mark_dir_deny),
    test_entry!(test_syd_fanotify_mark_path_allow),
    test_entry!(test_syd_fanotify_mark_path_deny),
    test_entry!(test_syd_fanotify_mark_dir_path_allow),
    test_entry!(test_syd_fanotify_mark_dir_path_deny),
    test_entry!(test_syd_fanotify_mark_symlink_allow),
    test_entry!(test_syd_fanotify_mark_symlink_deny),
    test_entry!(test_syd_inotify_add_watch_path_allow),
    test_entry!(test_syd_inotify_add_watch_path_deny),
    test_entry!(test_syd_inotify_add_watch_symlink_allow),
    test_entry!(test_syd_inotify_add_watch_symlink_deny),
    test_entry!(test_syd_unshare_user_bypass_limit),
    test_entry!(test_syd_interrupt_mkdir),
    test_entry!(test_syd_interrupt_bind_ipv4),
    test_entry!(test_syd_interrupt_bind_unix),
    test_entry!(test_syd_interrupt_connect_ipv4),
    //FIXME: This test should be done better.
    //test_entry!(test_syd_repetitive_clone),
    test_entry!(test_syd_pid_fork_deny),
    test_entry!(test_syd_pid_thread_deny),
    test_entry!(test_syd_pid_fork_bomb),
    test_entry!(test_syd_pid_fork_bomb_asm),
    test_entry!(test_syd_pid_thread_bomb),
    test_entry!(test_syd_mem_alloc),
    test_entry!(test_syd_exp_mem_stress_ng_malloc_1),
    test_entry!(test_syd_exp_mem_stress_ng_malloc_2),
    test_entry!(test_syd_exp_mem_stress_ng_mmap),
    test_entry!(test_syd_exp_pid_stress_ng_kill),
    test_entry!(test_syd_exp_pid_stress_ng_allow),
    test_entry!(test_syd_exp_pid_stress_ng_fork),
    test_entry!(test_syd_exp_syscall_fuzz_bare),
    test_entry!(test_syd_exp_syscall_fuzz_wrap),
];

// Tests if syd -V and --version works.
fn test_syd_version() -> TestResult {
    let status = syd().arg("-V").status().expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .stdout(Stdio::null())
        .stderr(Stdio::null())
        .arg("--version")
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if syd -E works.
fn test_syd_export() -> TestResult {
    skip_unless_available!("true");

    let status = syd()
        .arg("-Ebpf")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-Epfc")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EBPF")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EPFC")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EbPf")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-EPfc")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .arg("-Eb")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .arg("-Ef")
        .stderr(Stdio::null())
        .stdout(Stdio::null())
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `true` returns success under sandbox.
fn test_syd_true_returns_success() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .do_("exit", ["0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `syd` returns success for a sandbox running many processes,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_processes() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .do_("fork", ["0", "8"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `syd` returns success for a sandbox running many threads,
// in case the execve child returns success.
fn test_syd_true_returns_success_with_many_threads() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .do_("thread", ["0", "8"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `false` returns failure under sandbox.
fn test_syd_false_returns_failure() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .argv(["false"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .p("lib")
        .argv(["false"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many processes,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_processes() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .do_("fork", ["7", "8"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 7);
    Ok(())
}

// Tests if `syd` returns failure for a sandbox running many threads,
// in case the execve child returns failure.
fn test_syd_true_returns_failure_with_many_threads() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .do_("thread", ["7", "8"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 7);
    Ok(())
}

fn test_syd_empty_file_returns_enoexec() -> TestResult {
    // Step 1: Create a file that's empty called "empty"
    let path = Path::new("empty");
    let file = File::create(path)?;

    // Set permissions to executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o755); // -rwxr-xr-x
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do_("exec", ["./empty"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOEXEC);

    Ok(())
}

fn test_syd_non_executable_file_returns_eacces_empty() -> TestResult {
    // Create a file that's non-executable called "non-executable"
    let path = Path::new("non-executable");
    let file = File::create(path)?;

    // Set permissions to non-executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o644); // -rw-r--r--
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do_("exec", ["./non-executable"])
        .status()
        .expect("execute syd");
    // empty & non-executable file must return EACCES not ENOEXEC!
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_non_executable_file_returns_eacces_binary() -> TestResult {
    // Create a file that's non-executable called "non-executable"
    let path = Path::new("non-executable");
    let mut file = File::create(path)?;

    writeln!(
        file,
        "Heavy is the root of light. Still is the master of moving."
    )?;

    // Set permissions to non-executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o644); // -rw-r--r--
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do_("exec", ["./non-executable"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_non_executable_file_returns_eacces_script() -> TestResult {
    // Create a file that's non-executable called "non-executable"
    let path = Path::new("non-executable");
    let mut file = File::create(path)?;

    writeln!(file, "#!/bin/sh\nexit 42")?;

    // Set permissions to non-executable
    let mut perms = file.metadata()?.permissions();
    perms.set_mode(0o644); // -rw-r--r--
    file.set_permissions(perms)?;
    drop(file); // close the file.

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do_("exec", ["./non-executable"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_sigint_returns_130() -> TestResult {
    skip_unless_available!("kill", "sh");

    let status = syd()
        .p("lib")
        .argv(["sh", "-cx"])
        .arg(r#"exec kill -INT $$"#)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 130);

    Ok(())
}

fn test_syd_sigabrt_returns_134() -> TestResult {
    skip_unless_available!("kill", "sh");

    let status = syd()
        .p("lib")
        .argv(["sh", "-cx"])
        .arg(r#"exec kill -ABRT $$"#)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 134);

    Ok(())
}

fn test_syd_sigkill_returns_137() -> TestResult {
    skip_unless_available!("kill", "sh");

    let status = syd()
        .p("lib")
        .argv(["sh", "-cx"])
        .arg(r#"exec kill -KILL $$"#)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 137);

    Ok(())
}

fn test_syd_reap_zombies_bare() -> TestResult {
    skip_unless_available!("bash", "sleep");

    let status = syd()
        .p("lib")
        .argv(["bash", "-cex"])
        .arg(
            r#"
for i in {1..10}; do
    ( sleep $i ) &
done
disown
exit 42
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    Ok(())
}

fn test_syd_reap_zombies_wrap() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .argv(["bash", "-c"])
        .arg(
            r#"
set -e
for i in {1..10}; do
    ( sleep $i ) &
done
echo >&2 "Spawned 10 processes in the background."
echo >&2 "Disowning and exiting..."
disown
exit 42
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    Ok(())
}

// Tests if `whoami` returns `root` with `root/fake:1`
fn test_syd_whoami_returns_root_fake() -> TestResult {
    let status = syd()
        .p("lib")
        .m("root/fake:1")
        .do_("getuid", ["0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if `whoami` returns `root` with `root/map:1`
fn test_syd_whoami_returns_root_user() -> TestResult {
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .m("root/map:1")
        .do_("getuid", ["0"])
        .stderr(Stdio::inherit())
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setuid_nobody_default() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setuid", ["65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setuid_nobody_safesetid_deny() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: See the comment to test setuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setuid", ["65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setuid_root_safesetid_deny() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setuid", ["0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setuid_nobody_safesetid_allow() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setuid", ["65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setgid_nobody_default() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setuid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setgid", ["65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setgid_nobody_safesetid_deny() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setgid", ["65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setgid_root_safesetid_deny() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setgid", ["0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setgid_nobody_safesetid_allow() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setgid", ["65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_default_1() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setreuid", ["-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_default_2() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setreuid", ["65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_default_3() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setreuid", ["65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setreuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["-1", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setreuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["65534", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setreuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["65534", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_root_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setreuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["-1", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_root_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setreuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["0", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_root_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setreuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["0", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_allow_1() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setreuid", ["-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_allow_2() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setreuid", ["65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setreuid_nobody_safesetid_allow_3() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setreuid", ["65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setregid_nobody_default_1() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setregid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setregid", ["-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setregid_nobody_default_2() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setregid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setregid", ["65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setregid_nobody_default_3() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setregid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setregid", ["65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setregid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setregid", ["-1", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setregid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setregid", ["65534", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setregid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setregid", ["65534", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_setregid_root_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setregid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setregid", ["-1", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setregid_root_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setregid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setregid", ["0", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setregid_root_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setregid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setregid", ["0", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_allow_1() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setregid", ["-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_allow_2() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setregid", ["65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setregid_nobody_safesetid_allow_3() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setregid", ["65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_1() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["-1", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_2() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["-1", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_3() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["65534", "-1", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_4() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["-1", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_5() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["65534", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_6() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["65534", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_default_7() -> TestResult {
    skip_unless_cap!("setuid");

    // Default filter is a no-op for setresuid.
    // Test must abort on uid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresuid", ["65534", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["-1", "-1", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["-1", "65534", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["65534", "-1", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_4() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["-1", "65534", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_5() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["65534", "65534", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_6() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["65534", "-1", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_deny_7() -> TestResult {
    skip_unless_cap!("setuid");

    // SAFETY: Set the comment to test setresuid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["65534", "65534", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["-1", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["-1", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["0", "-1", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_4() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["-1", "0", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_5() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["0", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_6() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["0", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_root_safesetid_deny_7() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled, but no UID transition defined.
    // The syscall must fail with EPERM even with UID=0.

    // SAFETY: Compared to the test setresuid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because UID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresuid", ["0", "0", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_1() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["-1", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_2() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["-1", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_3() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["65534", "-1", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_4() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["-1", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_5() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["65534", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_6() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["65534", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresuid_nobody_safesetid_allow_7() -> TestResult {
    skip_unless_cap!("setuid");

    // SafeSetID is enabled with UID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setuid+{uid}:nobody"))
        .do_("setresuid", ["65534", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_1() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["-1", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_2() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["-1", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_3() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["65534", "-1", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_4() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["-1", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_5() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["65534", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_6() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["65534", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_default_7() -> TestResult {
    skip_unless_cap!("setgid");

    // Default filter is a no-op for setresgid.
    // Test must abort on gid mismatch after successful call.
    let status = syd()
        .p("lib")
        .do_("setresgid", ["65534", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_aborted!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["-1", "-1", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["-1", "65534", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["65534", "-1", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_4() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["-1", "65534", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_5() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["65534", "65534", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_6() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["65534", "-1", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_deny_7() -> TestResult {
    skip_unless_cap!("setgid");

    // SAFETY: Set the comment to test setresgid_root_safesetid_deny.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM _and_ generate an acccess violation.
    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["65534", "65534", "65534"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_1() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["-1", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_2() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setreuid", ["-1", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_3() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["0", "-1", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_4() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["-1", "0", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_5() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["0", "0", "-1"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_6() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["0", "-1", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_root_safesetid_deny_7() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled, but no GID transition defined.
    // The syscall must fail with EPERM even with GID=0.

    // SAFETY: Compared to the test setresgid_nobody_safesetid_deny, this
    // test _must not_ generate a syd access violation, because GID
    // change to root must be blocked by the kernel-level parent seccomp
    // filter. This is crucial security-wise so we test it here by
    // overriding SYD_LOG and setting SYD_LOG_FD.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("setresgid", ["0", "0", "0"])
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; no access violation raised as expected.");
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
            return Err(TestError("Unexpected access violation logged.".to_string()));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    let status = child.wait().expect("wait for syd");
    assert_status_code!(status, nix::libc::EPERM);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_1() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["-1", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_2() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["-1", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_3() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["65534", "-1", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_4() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["-1", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_5() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["65534", "65534", "-1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_6() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["65534", "-1", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_setresgid_nobody_safesetid_allow_7() -> TestResult {
    skip_unless_cap!("setgid");

    // SafeSetID is enabled with GID transition defined.
    // The syscall must succeed.
    let uid = Uid::current();
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .m(&format!("setgid+{uid}:nobody"))
        .do_("setresgid", ["65534", "65534", "65534"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions over execve(2)
fn test_syd_drop_cap_sys_ptrace_exec_default() -> TestResult {
    skip_unless_cap!("sys_ptrace");

    // Ptrace must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["sys_ptrace"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions over execve(2)
fn test_syd_drop_cap_sys_ptrace_exec_unsafe_caps() -> TestResult {
    skip_unless_cap!("sys_ptrace");

    // Ptrace is kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["sys_ptrace"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions over execve(2)
fn test_syd_drop_cap_sys_ptrace_exec_unsafe_debug() -> TestResult {
    skip_unless_cap!("sys_ptrace");

    // Ptrace is kept with trace/allow_unsafe_debug:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_debug:1")
        .do_("hascap", ["sys_ptrace"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_CHOWN restrictions over execve(2)
fn test_syd_drop_cap_chown_exec_default() -> TestResult {
    skip_unless_cap!("chown");

    // CAP_CHOWN must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["chown"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_CHOWN restrictions over execve(2)
fn test_syd_drop_cap_chown_exec_unsafe() -> TestResult {
    skip_unless_cap!("chown");

    // CAP_CHOWN is not dropped with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["chown"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_CHOWN restrictions over execve(2)
fn test_syd_drop_cap_chown_exec_safesetid() -> TestResult {
    skip_unless_cap!("chown");

    // CAP_CHOWN is kept with trace/allow_safe_setid:1
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("hascap", ["chown"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETGID restrictions over execve(2)
fn test_syd_drop_cap_setgid_exec_default() -> TestResult {
    skip_unless_cap!("setgid");

    // CAP_SETGID must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["setgid"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SETGID restrictions over execve(2)
fn test_syd_drop_cap_setgid_exec_unsafe() -> TestResult {
    skip_unless_cap!("setgid");

    // CAP_SETGID is not dropped with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["setgid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETGID restrictions over execve(2)
fn test_syd_drop_cap_setgid_exec_safesetid() -> TestResult {
    skip_unless_cap!("setgid");

    // CAP_SETGID is not dropped with trace/allow_safe_setid:1
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("hascap", ["setgid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETUID restrictions over execve(2)
fn test_syd_drop_cap_setuid_exec_default() -> TestResult {
    skip_unless_cap!("setuid");

    // CAP_SETUID must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["setuid"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SETUID restrictions over execve(2)
fn test_syd_drop_cap_setuid_exec_unsafe() -> TestResult {
    skip_unless_cap!("setuid");

    // CAP_SETUID is not dropped with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["setuid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETUID restrictions over execve(2)
fn test_syd_drop_cap_setuid_exec_safesetid() -> TestResult {
    skip_unless_cap!("setuid");

    // CAP_SETUID is not dropped with trace/allow_safe_setid:1
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .do_("hascap", ["setuid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions over execve(2)
fn test_syd_drop_cap_net_bind_service_exec_default() -> TestResult {
    skip_unless_cap!("net_bind_service");

    // CAP_NET_BIND_SERVICE must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["net_bind_service"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions over execve(2)
fn test_syd_drop_cap_net_bind_service_exec_unsafe_caps() -> TestResult {
    skip_unless_cap!("net_bind_service");

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["net_bind_service"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions over execve(2)
fn test_syd_drop_cap_net_bind_service_exec_unsafe_bind() -> TestResult {
    skip_unless_cap!("net_bind_service");

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_bind:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_bind:1")
        .do_("hascap", ["net_bind_service"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_RAW restrictions over execve(2)
fn test_syd_drop_cap_net_raw_exec_default() -> TestResult {
    skip_unless_cap!("net_raw");

    // CAP_NET_RAW must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["net_raw"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_NET_RAW restrictions over execve(2)
fn test_syd_drop_cap_net_raw_exec_unsafe_caps() -> TestResult {
    skip_unless_cap!("net_raw");

    // CAP_NET_RAW must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["net_raw"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_RAW restrictions over execve(2)
fn test_syd_drop_cap_net_raw_exec_unsafe_socket() -> TestResult {
    skip_unless_cap!("net_raw");

    // CAP_NET_RAW must be kept with trace/allow_unsafe_socket:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_socket:1")
        .do_("hascap", ["net_raw"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_TIME restrictions over execve(2)
fn test_syd_drop_cap_sys_time_exec_default() -> TestResult {
    skip_unless_cap!("sys_time");

    // CAP_SYS_TIME must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["sys_time"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SYS_TIME restrictions over execve(2)
fn test_syd_drop_cap_sys_time_exec_unsafe_caps() -> TestResult {
    skip_unless_cap!("sys_time");

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["sys_time"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_TIME restrictions over execve(2)
fn test_syd_drop_cap_sys_time_exec_unsafe_time() -> TestResult {
    skip_unless_cap!("sys_time");

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_time:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_time:1")
        .do_("hascap", ["sys_time"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYSLOG restrictions over execve(2)
fn test_syd_drop_cap_syslog_exec_default() -> TestResult {
    skip_unless_cap!("syslog");

    // CAP_SYSLOG must be dropped by default.
    let status = syd()
        .p("lib")
        .do_("hascap", ["syslog"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SYSLOG restrictions over execve(2)
fn test_syd_drop_cap_syslog_exec_unsafe_caps() -> TestResult {
    skip_unless_cap!("syslog");

    // CAP_SYSLOG must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["syslog"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYSLOG restrictions over execve(2)
fn test_syd_drop_cap_syslog_exec_unsafe_syslog() -> TestResult {
    skip_unless_cap!("syslog");

    // CAP_SYSLOG must be kept with trace/allow_unsafe_syslog:1
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_syslog:1")
        .do_("hascap", ["syslog"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions on library load.
fn test_syd_drop_cap_sys_ptrace_load_default() -> TestResult {
    skip_unless_cap!("sys_ptrace");
    skip_unless_available!("cc", "sh");

    // Ptrace must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_PTRACE is permitted
    if (cap_get_flag(caps, CAP_SYS_PTRACE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_PTRACE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions on library load.
fn test_syd_drop_cap_sys_ptrace_load_unsafe_caps() -> TestResult {
    skip_unless_cap!("sys_ptrace");
    skip_unless_available!("cc", "sh");

    // Ptrace is kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_PTRACE is permitted
    if (cap_get_flag(caps, CAP_SYS_PTRACE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_PTRACE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions on library load.
fn test_syd_drop_cap_sys_ptrace_load_unsafe_debug() -> TestResult {
    skip_unless_cap!("sys_ptrace");
    skip_unless_available!("cc", "sh");

    // Ptrace is kept with trace/allow_unsafe_debug:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_PTRACE is permitted
    if (cap_get_flag(caps, CAP_SYS_PTRACE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_PTRACE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_debug:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_CHOWN restrictions on library load.
fn test_syd_drop_cap_chown_load_default() -> TestResult {
    skip_unless_cap!("chown");
    skip_unless_available!("cc", "sh");

    // CAP_CHOWN must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_CHOWN is permitted
    if (cap_get_flag(caps, CAP_CHOWN, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_CHOWN is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_CHOWN restrictions on library load.
fn test_syd_drop_cap_chown_load_safesetid() -> TestResult {
    skip_unless_cap!("chown");
    skip_unless_available!("cc", "sh");

    // CAP_CHOWN is kept with trace/allow_safe_setid:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_CHOWN is permitted
    if (cap_get_flag(caps, CAP_CHOWN, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_CHOWN is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_CHOWN restrictions on library load.
fn test_syd_drop_cap_chown_load_unsafe() -> TestResult {
    skip_unless_cap!("chown");
    skip_unless_available!("cc", "sh");

    // CAP_CHOWN is not dropped with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_CHOWN is permitted
    if (cap_get_flag(caps, CAP_CHOWN, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_CHOWN is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETGID restrictions on library load.
fn test_syd_drop_cap_setgid_load_default() -> TestResult {
    skip_unless_cap!("setgid");
    skip_unless_available!("cc", "sh");

    // CAP_SETGID must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETGID is permitted
    if (cap_get_flag(caps, CAP_SETGID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETGID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETGID restrictions on library load.
fn test_syd_drop_cap_setgid_load_safesetid() -> TestResult {
    skip_unless_cap!("setgid");
    skip_unless_available!("cc", "sh");

    // CAP_SETGID is not dropped with trace/allow_unsafe_safesetid:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETGID is permitted
    if (cap_get_flag(caps, CAP_SETGID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETGID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETGID restrictions on library load.
fn test_syd_drop_cap_setgid_load_unsafe() -> TestResult {
    skip_unless_cap!("setgid");
    skip_unless_available!("cc", "sh");

    // CAP_SETGID is not dropped with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETGID is permitted
    if (cap_get_flag(caps, CAP_SETGID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETGID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETUID restrictions on library load.
fn test_syd_drop_cap_setuid_load_default() -> TestResult {
    skip_unless_cap!("setuid");
    skip_unless_available!("cc", "sh");

    // CAP_SETUID must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETUID is permitted
    if (cap_get_flag(caps, CAP_SETUID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETUID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETUID restrictions on library load.
fn test_syd_drop_cap_setuid_load_safesetid() -> TestResult {
    skip_unless_cap!("setuid");
    skip_unless_available!("cc", "sh");

    // CAP_SETUID is not dropped with trace/allow_unsafe_safesetid:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETUID is permitted
    if (cap_get_flag(caps, CAP_SETUID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETUID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_safe_setid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETUID restrictions on library load.
fn test_syd_drop_cap_setuid_load_unsafe() -> TestResult {
    skip_unless_cap!("setuid");
    skip_unless_available!("cc", "sh");

    // CAP_SETUID is not dropped with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETUID is permitted
    if (cap_get_flag(caps, CAP_SETUID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETUID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions on library load.
fn test_syd_drop_cap_net_bind_service_load_default() -> TestResult {
    skip_unless_cap!("net_bind_service");
    skip_unless_available!("cc", "sh");

    // CAP_NET_BIND_SERVICE must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_BIND_SERVICE is permitted
    if (cap_get_flag(caps, CAP_NET_BIND_SERVICE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_BIND_SERVICE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions on library load.
fn test_syd_drop_cap_net_bind_service_load_unsafe_caps() -> TestResult {
    skip_unless_cap!("net_bind_service");
    skip_unless_available!("cc", "sh");

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_BIND_SERVICE is permitted
    if (cap_get_flag(caps, CAP_NET_BIND_SERVICE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_BIND_SERVICE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions on library load.
fn test_syd_drop_cap_net_bind_service_load_unsafe_bind() -> TestResult {
    skip_unless_cap!("net_bind_service");
    skip_unless_available!("cc", "sh");

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_bind:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_BIND_SERVICE is permitted
    if (cap_get_flag(caps, CAP_NET_BIND_SERVICE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_BIND_SERVICE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_bind:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_RAW restrictions on library load.
fn test_syd_drop_cap_net_raw_load_default() -> TestResult {
    skip_unless_cap!("net_raw");
    skip_unless_available!("cc", "sh");

    // CAP_NET_RAW must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_RAW is permitted
    if (cap_get_flag(caps, CAP_NET_RAW, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_RAW is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_RAW restrictions on library load.
fn test_syd_drop_cap_net_raw_load_unsafe_caps() -> TestResult {
    skip_unless_cap!("net_raw");
    skip_unless_available!("cc", "sh");

    // CAP_NET_RAW must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_RAW is permitted
    if (cap_get_flag(caps, CAP_NET_RAW, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_RAW is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_RAW restrictions on library load.
fn test_syd_drop_cap_net_raw_load_unsafe_socket() -> TestResult {
    skip_unless_cap!("net_raw");
    skip_unless_available!("cc", "sh");

    // CAP_NET_RAW must be kept with trace/allow_unsafe_socket:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_RAW is permitted
    if (cap_get_flag(caps, CAP_NET_RAW, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_RAW is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_socket:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_TIME restrictions on library load.
fn test_syd_drop_cap_sys_time_load_default() -> TestResult {
    skip_unless_cap!("sys_time");
    skip_unless_available!("cc", "sh");

    // CAP_SYS_TIME must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_TIME is permitted
    if (cap_get_flag(caps, CAP_SYS_TIME, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_TIME is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_TIME restrictions on library load.
fn test_syd_drop_cap_sys_time_load_unsafe_caps() -> TestResult {
    skip_unless_cap!("sys_time");
    skip_unless_available!("cc", "sh");

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_TIME is permitted
    if (cap_get_flag(caps, CAP_SYS_TIME, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_TIME is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_TIME restrictions on library load.
fn test_syd_drop_cap_sys_time_load_unsafe_time() -> TestResult {
    skip_unless_cap!("sys_time");
    skip_unless_available!("cc", "sh");

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_time:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_TIME is permitted
    if (cap_get_flag(caps, CAP_SYS_TIME, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_TIME is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_time:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYSLOG restrictions on library load.
fn test_syd_drop_cap_syslog_load_default() -> TestResult {
    skip_unless_cap!("syslog");
    skip_unless_available!("cc", "sh");

    // CAP_SYSLOG must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYSLOG is permitted
    if (cap_get_flag(caps, CAP_SYSLOG, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYSLOG is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYSLOG restrictions on library load.
fn test_syd_drop_cap_syslog_load_unsafe_caps() -> TestResult {
    skip_unless_cap!("syslog");
    skip_unless_available!("cc", "sh");

    // CAP_SYSLOG must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYSLOG is permitted
    if (cap_get_flag(caps, CAP_SYSLOG, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYSLOG is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYSLOG restrictions on library load.
fn test_syd_drop_cap_syslog_load_unsafe_syslog() -> TestResult {
    skip_unless_cap!("syslog");
    skip_unless_available!("cc", "sh");

    // CAP_SYSLOG must be kept with trace/allow_unsafe_syslog:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYSLOG is permitted
    if (cap_get_flag(caps, CAP_SYSLOG, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYSLOG is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_syslog:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions over execve(2)
fn test_syd_userns_drop_cap_sys_ptrace_exec_default() -> TestResult {
    skip_unless_unshare!();

    // Ptrace must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["sys_ptrace"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions over execve(2)
fn test_syd_userns_drop_cap_sys_ptrace_exec_unsafe_caps() -> TestResult {
    skip_unless_unshare!();

    // Ptrace is kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["sys_ptrace"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions over execve(2)
fn test_syd_userns_drop_cap_sys_ptrace_exec_unsafe_debug() -> TestResult {
    skip_unless_unshare!();

    // Ptrace is kept with trace/allow_unsafe_debug:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_debug:1")
        .do_("hascap", ["sys_ptrace"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_CHOWN restrictions over execve(2)
fn test_syd_userns_drop_cap_chown_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_CHOWN must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["chown"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_CHOWN restrictions over execve(2)
fn test_syd_userns_drop_cap_chown_exec_unsafe() -> TestResult {
    skip_unless_unshare!();

    // CAP_CHOWN is not dropped with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["chown"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_CHOWN restrictions over execve(2)
fn test_syd_userns_drop_cap_chown_exec_safesetid() -> TestResult {
    skip_unless_unshare!();

    // CAP_CHOWN is kept with trace/allow_safe_setid:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_safe_setid:1")
        .do_("hascap", ["chown"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETGID restrictions over execve(2)
fn test_syd_userns_drop_cap_setgid_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_SETGID must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["setgid"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SETGID restrictions over execve(2)
fn test_syd_userns_drop_cap_setgid_exec_unsafe() -> TestResult {
    skip_unless_unshare!();

    // CAP_SETGID is not dropped with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["setgid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETGID restrictions over execve(2)
fn test_syd_userns_drop_cap_setgid_exec_safesetid() -> TestResult {
    skip_unless_unshare!();

    // CAP_SETGID is not dropped with trace/allow_safe_setid:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_safe_setid:1")
        .do_("hascap", ["setgid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETUID restrictions over execve(2)
fn test_syd_userns_drop_cap_setuid_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_SETUID must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["setuid"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SETUID restrictions over execve(2)
fn test_syd_userns_drop_cap_setuid_exec_unsafe() -> TestResult {
    skip_unless_unshare!();

    // CAP_SETUID is not dropped with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["setuid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SETUID restrictions over execve(2)
fn test_syd_userns_drop_cap_setuid_exec_safesetid() -> TestResult {
    skip_unless_unshare!();

    // CAP_SETUID is not dropped with trace/allow_safe_setid:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_safe_setid:1")
        .do_("hascap", ["setuid"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions over execve(2)
fn test_syd_userns_drop_cap_net_bind_service_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_NET_BIND_SERVICE must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["net_bind_service"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions over execve(2)
fn test_syd_userns_drop_cap_net_bind_service_exec_unsafe_caps() -> TestResult {
    skip_unless_unshare!();

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["net_bind_service"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions over execve(2)
fn test_syd_userns_drop_cap_net_bind_service_exec_unsafe_bind() -> TestResult {
    skip_unless_unshare!();

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_bind:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_bind:1")
        .do_("hascap", ["net_bind_service"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_RAW restrictions over execve(2)
fn test_syd_userns_drop_cap_net_raw_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_NET_RAW must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["net_raw"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_NET_RAW restrictions over execve(2)
fn test_syd_userns_drop_cap_net_raw_exec_unsafe_caps() -> TestResult {
    skip_unless_unshare!();

    // CAP_NET_RAW must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["net_raw"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_NET_RAW restrictions over execve(2)
fn test_syd_userns_drop_cap_net_raw_exec_unsafe_socket() -> TestResult {
    skip_unless_unshare!();

    // CAP_NET_RAW must be kept with trace/allow_unsafe_socket:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_socket:1")
        .do_("hascap", ["net_raw"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_TIME restrictions over execve(2)
fn test_syd_userns_drop_cap_sys_time_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_SYS_TIME must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["sys_time"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SYS_TIME restrictions over execve(2)
fn test_syd_userns_drop_cap_sys_time_exec_unsafe_caps() -> TestResult {
    skip_unless_unshare!();

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["sys_time"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_TIME restrictions over execve(2)
fn test_syd_userns_drop_cap_sys_time_exec_unsafe_time() -> TestResult {
    skip_unless_unshare!();

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_time:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_time:1")
        .do_("hascap", ["sys_time"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYSLOG restrictions over execve(2)
fn test_syd_userns_drop_cap_syslog_exec_default() -> TestResult {
    skip_unless_unshare!();

    // CAP_SYSLOG must be dropped by default.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("hascap", ["syslog"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

// Check CAP_SYSLOG restrictions over execve(2)
fn test_syd_userns_drop_cap_syslog_exec_unsafe_caps() -> TestResult {
    skip_unless_unshare!();

    // CAP_SYSLOG must be kept with trace/allow_unsafe_caps:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .do_("hascap", ["syslog"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYSLOG restrictions over execve(2)
fn test_syd_userns_drop_cap_syslog_exec_unsafe_syslog() -> TestResult {
    skip_unless_unshare!();

    // CAP_SYSLOG must be kept with trace/allow_unsafe_syslog:1
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_syslog:1")
        .do_("hascap", ["syslog"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions on library load.
fn test_syd_userns_drop_cap_sys_ptrace_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // Ptrace must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_PTRACE is permitted
    if (cap_get_flag(caps, CAP_SYS_PTRACE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_PTRACE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions on library load.
fn test_syd_userns_drop_cap_sys_ptrace_load_unsafe_caps() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // Ptrace is kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_PTRACE is permitted
    if (cap_get_flag(caps, CAP_SYS_PTRACE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_PTRACE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_PTRACE restrictions on library load.
fn test_syd_userns_drop_cap_sys_ptrace_load_unsafe_debug() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // Ptrace is kept with trace/allow_unsafe_debug:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_PTRACE is permitted
    if (cap_get_flag(caps, CAP_SYS_PTRACE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_PTRACE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_debug:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_CHOWN restrictions on library load.
fn test_syd_userns_drop_cap_chown_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_CHOWN must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_CHOWN is permitted
    if (cap_get_flag(caps, CAP_CHOWN, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_CHOWN is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_CHOWN restrictions on library load.
fn test_syd_userns_drop_cap_chown_load_safesetid() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_CHOWN is kept with trace/allow_safe_setid:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_CHOWN is permitted
    if (cap_get_flag(caps, CAP_CHOWN, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_CHOWN is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_safe_setid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_CHOWN restrictions on library load.
fn test_syd_userns_drop_cap_chown_load_unsafe() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_CHOWN is not dropped with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_CHOWN is permitted
    if (cap_get_flag(caps, CAP_CHOWN, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_CHOWN is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETGID restrictions on library load.
fn test_syd_userns_drop_cap_setgid_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SETGID must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETGID is permitted
    if (cap_get_flag(caps, CAP_SETGID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETGID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETGID restrictions on library load.
fn test_syd_userns_drop_cap_setgid_load_safesetid() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SETGID is not dropped with trace/allow_unsafe_safesetid:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETGID is permitted
    if (cap_get_flag(caps, CAP_SETGID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETGID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_safe_setid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETGID restrictions on library load.
fn test_syd_userns_drop_cap_setgid_load_unsafe() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SETGID is not dropped with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETGID is permitted
    if (cap_get_flag(caps, CAP_SETGID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETGID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETUID restrictions on library load.
fn test_syd_userns_drop_cap_setuid_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SETUID must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETUID is permitted
    if (cap_get_flag(caps, CAP_SETUID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETUID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETUID restrictions on library load.
fn test_syd_userns_drop_cap_setuid_load_safesetid() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SETUID is not dropped with trace/allow_unsafe_safesetid:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETUID is permitted
    if (cap_get_flag(caps, CAP_SETUID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETUID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_safe_setid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SETUID restrictions on library load.
fn test_syd_userns_drop_cap_setuid_load_unsafe() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SETUID is not dropped with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SETUID is permitted
    if (cap_get_flag(caps, CAP_SETUID, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SETUID is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions on library load.
fn test_syd_userns_drop_cap_net_bind_service_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_NET_BIND_SERVICE must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_BIND_SERVICE is permitted
    if (cap_get_flag(caps, CAP_NET_BIND_SERVICE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_BIND_SERVICE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions on library load.
fn test_syd_userns_drop_cap_net_bind_service_load_unsafe_caps() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_BIND_SERVICE is permitted
    if (cap_get_flag(caps, CAP_NET_BIND_SERVICE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_BIND_SERVICE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_BIND_SERVICE restrictions on library load.
fn test_syd_userns_drop_cap_net_bind_service_load_unsafe_bind() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_NET_BIND_SERVICE must be kept with trace/allow_unsafe_bind:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_BIND_SERVICE is permitted
    if (cap_get_flag(caps, CAP_NET_BIND_SERVICE, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_BIND_SERVICE is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_bind:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_RAW restrictions on library load.
fn test_syd_userns_drop_cap_net_raw_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_NET_RAW must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_RAW is permitted
    if (cap_get_flag(caps, CAP_NET_RAW, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_RAW is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_RAW restrictions on library load.
fn test_syd_userns_drop_cap_net_raw_load_unsafe_caps() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_NET_RAW must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_RAW is permitted
    if (cap_get_flag(caps, CAP_NET_RAW, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_RAW is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_NET_RAW restrictions on library load.
fn test_syd_userns_drop_cap_net_raw_load_unsafe_socket() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_NET_RAW must be kept with trace/allow_unsafe_socket:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_NET_RAW is permitted
    if (cap_get_flag(caps, CAP_NET_RAW, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_NET_RAW is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_socket:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_TIME restrictions on library load.
fn test_syd_userns_drop_cap_sys_time_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SYS_TIME must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_TIME is permitted
    if (cap_get_flag(caps, CAP_SYS_TIME, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_TIME is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_TIME restrictions on library load.
fn test_syd_userns_drop_cap_sys_time_load_unsafe_caps() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_TIME is permitted
    if (cap_get_flag(caps, CAP_SYS_TIME, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_TIME is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYS_TIME restrictions on library load.
fn test_syd_userns_drop_cap_sys_time_load_unsafe_time() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SYS_TIME must be kept with trace/allow_unsafe_time:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYS_TIME is permitted
    if (cap_get_flag(caps, CAP_SYS_TIME, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYS_TIME is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_time:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYSLOG restrictions on library load.
fn test_syd_userns_drop_cap_syslog_load_default() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SYSLOG must be dropped by default.
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYSLOG is permitted
    if (cap_get_flag(caps, CAP_SYSLOG, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYSLOG is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYSLOG restrictions on library load.
fn test_syd_userns_drop_cap_syslog_load_unsafe_caps() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SYSLOG must be kept with trace/allow_unsafe_caps:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYSLOG is permitted
    if (cap_get_flag(caps, CAP_SYSLOG, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYSLOG is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_caps:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check CAP_SYSLOG restrictions on library load.
fn test_syd_userns_drop_cap_syslog_load_unsafe_syslog() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "sh");

    // CAP_SYSLOG must be kept with trace/allow_unsafe_syslog:1
    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/capability.h>

int syd_main(void) {
    cap_t caps;
    cap_flag_value_t cap_flag;

    // Get the capabilities of the current process
    caps = cap_get_proc();
    if (caps == NULL) {
        perror("cap_get_proc");
        return errno;
    }

    // Check if CAP_SYSLOG is permitted
    if (cap_get_flag(caps, CAP_SYSLOG, CAP_PERMITTED, &cap_flag) == -1) {
        perror("cap_get_flag");
        cap_free(caps);
        return errno;
    }

    // Free the capabilities structure
    cap_free(caps);

    if (cap_flag == CAP_SET) {
        return 0; // CAP_SYSLOG is set
    } else {
        return ENOENT; // Capability is not set.
    }
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC -lcap || exit 127
    "##,
        )
        .status()
        .expect("execute sh");
    let code = status.code().unwrap_or(127);
    if code == 127 {
        eprintln!("Failed to compile dynamic library, is libcap installed?");
        eprintln!("Skipping test!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .p("container")
        .m("trace/allow_unsafe_syslog:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Checks symlink insecure directory restrictions.
fn test_syd_readlink_insecure_world_writable() -> TestResult {
    let dir = "insecure_world_writable_dir";
    let dpt = Path::new(dir);
    create_dir_all(dpt)?;

    // Set the directory to be world-writable with sticky bit.
    let mut perms = metadata(dpt)?.permissions();
    perms.set_mode(0o1777); // world-writable with sticky bit
    set_permissions(dpt, perms.clone())?;

    // Create a valid file.
    let file = format!("{dir}/valid_file.txt");
    write(
        &file,
        b"Heavy is the root of light. Still is the master of moving.\n",
    )?;

    // Create a symlink in the world-writable directory pointing to the valid file.
    let link = format!("{dir}/symlink_to_valid_file");
    symlink(&file, &link)?;

    // Test 1: Symlink is the final path component.
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat:on")
        .m("allow/read,stat+/***")
        .do_("readlink", [&link])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ELOOP);

    // Test 2: Symlink is not the final path component.
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat:on")
        .m("allow/read,stat+/***")
        .do_("readlink", &[format!("{link}/invalid-file")])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ELOOP);

    perms.set_mode(0o700); // owner only
    set_permissions(dpt, perms)?;

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat:on")
        .m("allow/read,stat+/***")
        .do_("readlink", [&link])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Checks socket domain restrictions
fn test_syd_socket_domain_restrictions() -> TestResult {
    let allows = [
        (nix::libc::AF_UNIX, nix::libc::SOCK_DGRAM, 0),
        (nix::libc::AF_UNIX, nix::libc::SOCK_STREAM, 0),
        (nix::libc::AF_INET, nix::libc::SOCK_DGRAM, 0),
        (nix::libc::AF_INET, nix::libc::SOCK_STREAM, 0),
        (nix::libc::AF_INET6, nix::libc::SOCK_DGRAM, 0),
        (nix::libc::AF_INET6, nix::libc::SOCK_STREAM, 0),
    ];
    let denies = [
        // Do not add privileged sockets here.
        (
            nix::libc::AF_NETLINK,
            nix::libc::SOCK_DGRAM,
            nix::libc::NETLINK_GENERIC,
        ),
        (
            nix::libc::AF_NETLINK,
            nix::libc::SOCK_DGRAM,
            nix::libc::NETLINK_ROUTE,
        ),
    ];
    let kcapis = [(nix::libc::AF_ALG, nix::libc::SOCK_SEQPACKET, 0)];

    for (domain, ty, proto) in &allows {
        let status = syd()
            .p("lib")
            .do_(
                "socket",
                [&format!("{domain}"), &format!("{ty}"), &format!("{proto}")],
            )
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }

    for (domain, ty, proto) in &denies {
        let status = syd()
            .p("lib")
            .do_(
                "socket",
                [&format!("{domain}"), &format!("{ty}"), &format!("{proto}")],
            )
            .status()
            .expect("execute syd");
        assert_status_code!(status, nix::libc::EAFNOSUPPORT);

        let status = syd()
            .p("lib")
            .m("trace/allow_unsupp_socket:1")
            .do_(
                "socket",
                [&format!("{domain}"), &format!("{ty}"), &format!("{proto}")],
            )
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }

    for (domain, ty, proto) in &kcapis {
        let status = syd()
            .p("lib")
            .do_(
                "socket",
                [&format!("{domain}"), &format!("{ty}"), &format!("{proto}")],
            )
            .status()
            .expect("execute syd");
        assert_status_code!(status, nix::libc::EAFNOSUPPORT);

        let status = syd()
            .p("lib")
            .m("trace/allow_safe_kcapi:1")
            .do_(
                "socket",
                [&format!("{domain}"), &format!("{ty}"), &format!("{proto}")],
            )
            .status()
            .expect("execute syd");
        // Careful, kernel may return EAFNOSUPPORT
        // if CONFIG_CRYPTO_USER_API is either not
        // enabled or compiled as a module and the
        // module is not yet loaded.
        assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

        let status = syd()
            .p("lib")
            .m("trace/allow_unsupp_socket:1")
            .m("trace/allow_safe_kcapi:1")
            .do_(
                "socket",
                [&format!("{domain}"), &format!("{ty}"), &format!("{proto}")],
            )
            .status()
            .expect("execute syd");
        // Careful, kernel may return EAFNOSUPPORT
        // if CONFIG_CRYPTO_USER_API is either not
        // enabled or compiled as a module and the
        // module is not yet loaded.
        assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);
    }

    Ok(())
}

// Checks tkill restrictions.
fn test_syd_tkill_restrictions() -> TestResult {
    // tkill should return 0 by default.
    let status = syd()
        .p("lib")
        .do_("tkill", ["-42", "0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // trace/allow_unsafe_tkill:1 may be used to relax.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_kill:1")
        .do_("tkill", ["-42", "0"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EINVAL);

    // tkill(sydpid) does not propagate to syd.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_kill:1")
        .do__("tkill")
        .argv(["bash", "-cx", &format!("{} ${{PPID}} 9", *SYD_DO)])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

// Checks environment filtering
fn test_syd_environment_filter() -> TestResult {
    skip_unless_available!("sh");

    const ENV: &str = "SAFE";
    env::set_var(ENV, "/var/empty");

    // Step 1: Allow by default
    let output = syd()
        .p("lib")
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output1:{output}");

    // Step 2: Override with -evar=val
    let output = syd()
        .p("lib")
        .arg(&format!("-e{ENV}=/var/empty:/var/empty"))
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty:/var/empty", "output2:{output}");

    // Step 3: Unset with -evar
    let output = syd()
        .p("lib")
        .arg(&format!("-e{ENV}"))
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output3:{output}");

    // Step 4: Pass-through with -evar=
    let output = syd()
        .p("lib")
        .arg(&format!("-e{ENV}="))
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output4:{output}");

    env::remove_var(ENV);

    Ok(())
}

// Checks environment hardening and -e pass-through.
fn test_syd_environment_harden() -> TestResult {
    skip_unless_available!("sh");

    const ENV: &str = "LD_LIBRARY_PATH";
    env::set_var(ENV, "/var/empty");

    // Step 1: Deny by default
    let output = syd()
        .p("lib")
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output1:{output}");

    // Step 2: Override with -evar=val
    let output = syd()
        .p("lib")
        .arg(&format!("-e{ENV}=/var/empty:/var/empty"))
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty:/var/empty", "output2:{output}");

    // Step 3: Unset with -evar
    let output = syd()
        .p("lib")
        .arg(&format!("-e{ENV}"))
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output3:{output}");

    // Step 4: Pass-through with -evar=
    let output = syd()
        .p("lib")
        .arg(&format!("-e{ENV}="))
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output4:{output}");

    // Step 5: Allow with -m trace/allow_unsafe_env:1
    let output = syd()
        .p("lib")
        .m("trace/allow_unsafe_env:1")
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output == "/var/empty", "output:{output}");

    // Step 6: Toggle -m trace/allow_unsafe_env
    let output = syd()
        .p("lib")
        .m("trace/allow_unsafe_env:1")
        .m("trace/allow_unsafe_env:0")
        .argv(["sh", "-c", &format!("echo ${ENV}")])
        .output()
        .expect("execute syd");
    let output = String::from_utf8_lossy(&output.stdout);
    let output = output.trim_end();
    assert!(output.is_empty(), "output1:{output}");

    env::remove_var(ENV);

    Ok(())
}

// Tests if `lock:on` command disables access to `/dev/syd`.
fn test_syd_lock() -> TestResult {
    skip_unless_available!("sh");

    // Lib sets lock:exec!
    let status = syd()
        .p("lib")
        .argv(["sh", "-cx", "test -c /dev/syd"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    eprintln!("+ sh -c \"test -c /dev/syd\"");
    let status = syd()
        .p("lib")
        .m("lock:on")
        .argv(["sh", "-cex", "test -c /dev/syd"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `lock:exec` locks the sandbox for all except the exec child.
fn test_syd_lock_exec() -> TestResult {
    // Note, we use bash rather than sh,
    // because not all sh (e.g. busybox)
    // spawn a subsell on (cmd).
    skip_unless_available!("bash");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .argv(["bash", "-cex", "test -c /dev/syd"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .argv(["bash", "-cex", "( test -c /dev/syd )"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if `-mlock:on` prevents subsequent -m CLI args.
fn test_syd_lock_prevents_further_cli_args() -> TestResult {
    skip_unless_available!("true");

    let status = syd()
        .p("lib")
        .m("lock:on")
        .m("lock:off")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_exec:0")
        .m("lock:on")
        .m("lock:off")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .p("lib")
        .m("lock:on")
        .m("trace/allow_unsafe_exec:1")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    // Check for warning that paludis profile was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("lock:on")
        .p("paludis")
        .argv(["true"])
        .spawn()
        .expect("execute syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    // Check for warning that /dev/null file was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("lock:on")
        .P("/dev/null")
        .argv(["true"])
        .spawn()
        .expect("execute syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

// Tests if `-mlock:on` prevents subsequent configuration file edits.
fn test_syd_lock_prevents_further_cfg_items() -> TestResult {
    skip_unless_available!("true");

    let conf = "lock:on\nlock:off\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    // Check for warning that include line was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .P("./conf.syd-3")
        .argv(["true"])
        .spawn()
        .expect("execute syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

// Tests if `-mlock:on` prevents subsequent file includes.
fn test_syd_lock_prevents_further_inc_items() -> TestResult {
    skip_unless_available!("true");

    let conf = "sandbox/read:on\n";
    let mut file = File::create("conf-1.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let conf = "lock:on\ninclude conf-1.syd-3\n";
    let mut file = File::create("conf-2.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    // Check for warning that include line was not parsed.
    // Syd does not bail out in this case, just warns and continues.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .P("./conf-2.syd-3")
        .argv(["true"])
        .spawn()
        .expect("execute syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError("Expected parse error not logged.".to_string()));
        }
        Ok(_) => {
            // If any data was read, log it
            eprint!("Read data from pipe:\n{buffer}");
            if buffer.contains("sandbox is locked") {
                eprintln!("Sandbox was locked as expected.");
            } else {
                return Err(TestError("Expected parse error not logged".to_string()));
            }
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

// Tests if syd-chk works to check if process is running under syd.
fn test_syd_chk() -> TestResult {
    eprintln!("+ syd-chk");
    let status = std::process::Command::new(&*SYD_CHK)
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    let status = syd()
        .p("lib")
        .argv([&*SYD_CHK])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_default() -> TestResult {
    skip_unless_available!("bash", "sleep");

    // Lib sets lock:exec!
    let status = syd()
        .p("lib")
        .argv(["bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_on() -> TestResult {
    skip_unless_available!("bash", "chmod", "sleep", "true");

    let status = syd()
        .p("lib")
        .m("lock:on")
        .argv(["bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_off_1() -> TestResult {
    skip_unless_available!("bash", "chmod", "sleep", "true");

    let status = syd()
        .p("lib")
        .m("lock:off")
        .argv(["bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_off_2() -> TestResult {
    skip_unless_available!("bash", "cat", "chmod", "true");

    let status = syd()
        .p("lib")
        .m("lock:off")
        .argv(["bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
(
    test -c \"$({} $PWD/exec.sh)\"
) &
wait
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_exec_1() -> TestResult {
    skip_unless_available!("bash", "cat", "chmod", "true");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .argv(["bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
test -c \"$({} $PWD/exec.sh)\"
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_cmd_exec_with_lock_exec_2() -> TestResult {
    skip_unless_available!("bash", "cat", "chmod", "true");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .argv(["bash", "-cex"])
        .arg(format!(
            "
#!/bin/bash
: > test
cat > exec.sh <<EOF
#!/bin/sh -ex
# Careful here, cmd/exec changes CWD to /.
echo OK > $PWD/test
exit 42
EOF
chmod +x exec.sh
(
    test -c \"$({} $PWD/exec.sh)\"
) &
wait
sleep 5
test -s ./test
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_parse_config() -> TestResult {
    skip_unless_available!("true");

    let conf = "lock:on\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .p("lib")
        .P("./conf.syd-3")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .P("./conf.syd-3")
        .m("lock:exec")
        .args(["--", "true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_include_config() -> TestResult {
    skip_unless_available!("true");

    let idir = " Heavy\tis\tthe\troot\tof\tlight.\tStill\tis\tthe\tmaster\tof\tmoving ";
    create_dir_all(idir)?;

    let name = " Heavy is the root of light.\tStill is the master of moving";
    let conf = "lock:${SYD_LOCK_STATE}\n";
    let mut file = File::create(format!("./{idir}/{name}.syd-3"))?;
    write!(file, "{conf}")?;
    drop(file);

    let conf = format!("include {name}\n");
    let mut file = File::create(format!("./{idir}/conf.syd-3"))?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env("SYD_LOCK_STATE", "on")
        .p("lib")
        .P(&format!("./{idir}/conf.syd-3"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .env("SYD_LOCK_STATE", "on")
        .p("lib")
        .P(&format!("./{idir}/conf.syd-3"))
        .m("lock:exec")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_01() -> TestResult {
    skip_unless_available!("true");

    let conf = "allow/write+${SYD_TEST_OOPS}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env("SYD_TEST_OOPS", "/home")
        .p("lib")
        .P("./conf.syd-3")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_02() -> TestResult {
    skip_unless_available!("true");

    let conf = "allow/write+${SYD_TEST_OOPS}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .p("lib")
        .P("./conf.syd-3")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_03() -> TestResult {
    skip_unless_available!("true");

    let conf = "allow/write+${SYD_TEST_OOPS:-/home}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .p("lib")
        .P("./conf.syd-3")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_shellexpand_04() -> TestResult {
    skip_unless_available!("true");

    let conf = "allow/write+${SYD_TEST_OOPS:-}/***\n";
    let mut file = File::create("conf.syd-3")?;
    write!(file, "{conf}")?;
    drop(file);

    let status = syd()
        .env_remove("SYD_TEST_OOPS")
        .p("lib")
        .P("./conf.syd-3")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if personality(2) locking works with trace/lock_personality.
fn test_syd_lock_personality() -> TestResult {
    // Check if the target architecture is 32-bit and exit if true
    #[cfg(not(target_pointer_width = "64"))]
    {
        eprintln!("Test requires 64-bit target, skippping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .p("lib")
        .m("trace/lock_personality:1")
        .do_("personality", ["false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("trace/lock_personality:0")
        .m("trace/lock_personality:1")
        .do_("personality", ["false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .do_("personality", ["true"])
        .status()
        .expect("execute syd");
    // This and the rest of the asserts fail on GitLab CI.
    if !*GL_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .p("lib")
        .m("trace/lock_personality:0")
        .do_("personality", ["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("trace/lock_personality:1")
        .m("trace/lock_personality:0")
        .do_("personality", ["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if MDWE protections work as expected.
fn test_syd_mdwe_protections() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/mmap")
        .do_("mdwe", ["false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/mmap")
        .m("trace/allow_unsafe_memory:0")
        .do_("mdwe", ["false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/mmap")
        .m("trace/allow_unsafe_memory:1")
        .m("trace/allow_unsafe_memory:0")
        .do_("mdwe", ["false"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_memory:1")
        .do_("mdwe", ["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_memory:1")
        .m("trace/allow_unsafe_memory:0")
        .m("trace/allow_unsafe_memory:1")
        .do_("mdwe", ["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Test if MDWE can be relaxed as expected.
fn test_syd_mdwe_relax() -> TestResult {
    skip_unless_available!("luajit");

    // Execute with default restrictions.
    // Expect LuaJIT to segfault.
    let status = syd()
        .p("lib")
        .argv(["luajit", "-e", "for i=1,1e5 do local a=i*2 end"])
        .status()
        .expect("execute syd");
    // This segfaults on x86-64, but exits with 1 on arm64.
    assert_status_code_matches!(status, 1 | EX_SIGSEGV);

    // Relax restrictions.
    // Expect LuaJIT to succeed.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_memory:1")
        .argv(["luajit", "-e", "for i=1,1e5 do local a=i*2 end"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Test if mmap(NULL, MMAP_FIXED) is prevented.
fn test_syd_mmap_fixed_null() -> TestResult {
    let status = syd()
        .p("lib")
        .do_("mmap_fixed_null", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_load_library() -> TestResult {
    skip_unless_available!("cc");

    // Write code.
    let code = r#"
#include <stdio.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    {
        let mut file = File::create("load.c")?;
        write!(file, "{code}")?;
    }

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, 42);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(()); // Skip rest of the tests.
    }

    // Try to load nonexisting library.
    let status = syd()
        .p("lib")
        .argv(["./nolib.so"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    // Try to load a library without the syd_main symbol.
    let code = r"int oops(void) { return 42; }";
    {
        let mut file = File::create("load.c")?;
        write!(file, "{code}")?;
    }
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 128);

    Ok(())
}

fn test_syd_load_library_noexec() -> TestResult {
    skip_unless_available!("cc");
    skip_unless_unshare!();

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    {
        let mut file = File::create("load.c")?;
        write!(file, "{code}")?;
    }

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .m("bind+/:/:noexec")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_code!(status, 42);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_abort_after_load() -> TestResult {
    skip_unless_available!("cc");

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    abort();
}
"#;
    {
        let mut file = File::create("load.c")?;
        write!(file, "{code}")?;
    }

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_aborted!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_abort_at_startup() -> TestResult {
    skip_unless_available!("cc");

    // Write code.
    let code = r#"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

__attribute__((constructor))
void syd_init(void)
{
    printf("library initialized at pid %d\n", getpid());
    abort(); // This takes down syd with it...
}

int syd_main(void)
{
    printf("library loaded at pid %d\n", getpid());
    return 42;
}
"#;
    {
        let mut file = File::create("load.c")?;
        write!(file, "{code}")?;
    }

    // Compile code.
    Command::new("cc")
        .args([
            "-Wall", "-Wextra", "load.c", "-shared", "-o", "load.so", "-fPIC",
        ])
        .status()?;

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        let sign = status.signal().unwrap_or(0);
        assert!(
            matches!(sign, nix::libc::SIGABRT | nix::libc::SIGSEGV),
            "code:{code} status:{status:?}"
        );
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_bare() -> TestResult {
    skip_unless_available!("cc", "sh");

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

int syd_main(void) {
    DIR *dir;
    struct dirent *entry;
    int fd_leaks = 0, dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        return -1; // Return -1 in case of error
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        return -1; // Return -1 in case of error
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_wrap() -> TestResult {
    skip_unless_available!("cc", "sh");
    skip_unless_unshare!();

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

int syd_main(void) {
    DIR *dir;
    struct dirent *entry;
    int fd_leaks = 0, dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        return -1; // Return -1 in case of error
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        return -1; // Return -1 in case of error
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_init_bare() -> TestResult {
    skip_unless_available!("cc", "sh");

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int fd_leaks;

__attribute__((constructor))
void syd_init(void)
{
    DIR *dir;
    struct dirent *entry;
    int dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        fd_leaks = -1;
        return;
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        fd_leaks = -1;
        return;
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
}

int syd_main(void)
{
    if (fd_leaks > 0) {
        printf("Detected %d file descriptor leaks during init!\n", fd_leaks);
    } else {
        printf("No file descriptor leaks detected during init!\n");
    }
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_load_library_check_fd_leaks_init_wrap() -> TestResult {
    skip_unless_available!("cc", "sh");
    skip_unless_unshare!();

    let status = Command::new("sh")
        .arg("-cex")
        .arg(
            r##"
cat > load.c <<EOF
#include <dirent.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static int fd_leaks;

__attribute__((constructor))
void syd_init(void)
{
    DIR *dir;
    struct dirent *entry;
    int dir_fd;

    // Open the directory containing file descriptors
    dir = opendir("/proc/self/fd");
    if (!dir) {
        perror("Failed to open /proc/self/fd");
        fd_leaks = -1;
        return;
    }

    // Get the file descriptor for the directory stream
    dir_fd = dirfd(dir);
    if (dir_fd == -1) {
        perror("Failed to get file descriptor for directory");
        closedir(dir);
        fd_leaks = -1;
        return;
    }

    // Iterate over all entries in the directory
    while ((entry = readdir(dir)) != NULL) {
        int fd;
        char *end;

        // Convert the name of the entry to an integer
        fd = strtol(entry->d_name, &end, 10);
        if (*end != '\0' || entry->d_name == end) continue; // Skip non-integer entries

        // Build the path to the symbolic link for the file descriptor
        char link_path[4096];
        char target_path[4096];
        snprintf(link_path, sizeof(link_path), "/proc/self/fd/%d", fd);

        ssize_t len = readlink(link_path, target_path, sizeof(target_path) - 1);
        if (len > 0) {
            target_path[len] = '\0'; // Ensure null termination
            // We ignore standard input, output, and error which are 0, 1, and 2
            if (fd <= 2) {
                printf("Ignoring standard open fd %d -> %s...\n", fd, target_path);
            } else if (fd == dir_fd) {
                printf("Ignoring fd to current directory fd %d -> %s...\n", fd, target_path);
            } else {
                printf("!!! Leaked file descriptor %d -> %s !!!\n", fd, target_path);
                fd_leaks++;
            }
        } else {
            perror("Failed to read link");
            fd_leaks++;
        }
    }

    closedir(dir);
}

int syd_main(void)
{
    if (fd_leaks > 0) {
        printf("Detected %d file descriptor leaks during init!\n", fd_leaks);
    } else {
        printf("No file descriptor leaks detected during init!\n");
    }
    return fd_leaks;
}
EOF

cc -Wall -Wextra load.c -shared -o load.so -fPIC
    "##,
        )
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    // Load code.
    // ENOSYS = Dynamic linking not supported.
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .argv(["./load.so"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if code != nix::libc::ENOSYS {
        assert_status_ok!(status);
    } else {
        eprintln!("Dynamic linking not supported, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Tests if read sandboxing for open works to allow.
fn test_syd_read_sandbox_open_allow() -> TestResult {
    skip_unless_available!("dd");

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat:on")
        .m("allow/read,stat+/***")
        .m("deny/read+/dev/***")
        .m("allow/read+/dev/null")
        .argv(["dd", "if=/dev/null"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if read sandboxing for open works to deny.
fn test_syd_read_sandbox_open_deny() -> TestResult {
    skip_unless_available!("cat");

    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .m("deny/read+/dev/null")
        .argv(["cat", "/dev/null"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for chdir works to allow.
fn test_syd_stat_sandbox_chdir_allow() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/dev")
        .m("allow/stat+/dev")
        .argv(["sh", "-cex", "cd /dev"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_chdir_hide() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/dev")
        .argv(["sh", "-cx", "cd /dev || exit 42"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);
    Ok(())
}

// Tests if stat sandboxing for stat works to allow.
fn test_syd_stat_sandbox_stat_allow() -> TestResult {
    skip_unless_available!("ls");

    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/dev/null")
        .m("allow/stat+/dev/null")
        .argv(["ls", "/dev/null"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for stat works to hide.
fn test_syd_stat_sandbox_stat_hide() -> TestResult {
    skip_unless_available!("ls");

    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/dev/null")
        .argv(["ls", "/dev/null"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if stat sandboxing for getdents works to allow.
fn test_syd_stat_sandbox_getdents_allow() -> TestResult {
    skip_unless_available!("ls");

    let output = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/dev/null")
        .m("allow/stat+/dev/null")
        .argv(["ls", "/dev"])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window == b"null"),
        "Stdout:\n{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing for getdents works to hide.
fn test_syd_stat_sandbox_getdents_hide() -> TestResult {
    skip_unless_available!("ls");

    let output = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/dev/null")
        .argv(["ls", "/dev"])
        .output()
        .expect("execute syd");
    assert!(
        output
            .stdout
            .windows(b"null".len())
            .any(|window| window != b"null"),
        "Stdout:{:?}",
        output.stdout
    );

    Ok(())
}

// Tests if stat sandboxing can be bypassed by read attempt
fn test_syd_stat_bypass_with_read() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat:on")
        .m("allow/read,stat+/***")
        .m("deny/read,stat+/etc/***")
        .m("allow/read,stat+/etc/ld*/***")
        .do_("stat_bypass_with_read", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if stat sandboxing can be bypassed by write attempt
fn test_syd_stat_bypass_with_write() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/stat,write:on")
        .m("allow/stat,write+/***")
        .m("deny/stat,write+/etc/***")
        .m("allow/stat+/etc/ld*/***")
        .do_("stat_bypass_with_write", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if stat sandboxing can be bypassed by exec attempt
fn test_syd_stat_bypass_with_exec() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/exec,stat:on")
        .m("allow/exec,stat+/***")
        .m("deny/exec,stat+/**/z?sh")
        .m("deny/exec,stat+/**/[bd]ash")
        .m("deny/exec,stat+/**/busybox")
        .do_("stat_bypass_with_exec", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if write sandboxing for open works to allow.
fn test_syd_write_sandbox_open_allow() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("deny/write+/dev/***")
        .m("allow/write+/dev/null")
        .argv(["sh", "-cex", "echo welcome to the machine >> /dev/null"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if write sandboxing for open works to deny.
fn test_syd_write_sandbox_open_deny() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .m("deny/write+/dev/null")
        .argv(["sh", "-cex", "echo welcome to the machine >> /dev/null"])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Tests if exec sandboxing works to allow.
fn test_syd_exec_sandbox_open_allow() -> TestResult {
    skip_unless_available!("true");

    let bin = which("true")?;
    let status = syd()
            .p("lib")
            .m("sandbox/exec:on")
            .m("deny/exec+/***")
            .m("allow/exec+/**/*.so*")
            .m(&format!("allow/exec+{bin}"))
            .arg("-atrue") // this may be busybox
        .argv([
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if exec sandboxing works to deny.
fn test_syd_exec_sandbox_open_deny() -> TestResult {
    skip_unless_available!("true");

    let bin = which("true")?;
    let status = syd()
            .p("lib")
            .m("sandbox/exec:on")
            .m("allow/exec+/***")
            .m(&format!("deny/exec+{bin}"))
            .arg("-atrue") // this may be busybox
        .argv([
            &bin.to_string(),
        ])
        .status()
        .expect("execute syd");
    assert_status_not_ok!(status);

    Ok(())
}

// Check if #! interpreter path of scripts are properly sandboxed.
fn test_syd_exec_sandbox_deny_binfmt_script() -> TestResult {
    skip_unless_available!("sh");

    // Write script.
    let path = "./script.sh";
    let script = r#"#!/bin/sh -ex
exit 42
"#;
    let mut file = File::create(path)?;
    write!(file, "{script}")?;
    drop(file); // Close the file to avoid ETXTBUSY.

    syd::fs::chmod_x(path).expect("Failed to make file executable");

    // Step 1: Allow both the interpreter and the script.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["./script.sh"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    // Step 2: Allow the interpreter but disable the script.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/script.sh")
        .argv(["./script.sh"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    // Step 3: Allow the script but disable the interpreter.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("deny/exec+/***")
        .m("allow/exec+/**/*.so*")
        .m("allow/exec+/**/script.sh")
        .argv(["./script.sh"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    Ok(())
}

// Check if a script which has an interpreter that itself is a script is properly sandboxed.
fn test_syd_exec_sandbox_many_binfmt_script() -> TestResult {
    skip_unless_available!("sh");

    // Write script1 whose interpreter points to script2.
    let path1 = "./script1.sh";
    let script1 = r#"#!./script2.sh
"#;
    let mut file1 = File::create(path1)?;
    write!(file1, "{script1}")?;

    // Write script2 whole interpreter points to /bin/sh.
    let path2 = "./script2.sh";
    let script2 = r#"#!/bin/sh -ex
exit 42
"#;
    let mut file2 = File::create(path2)?;
    write!(file2, "{script2}")?;

    // Close the files to avoid ETXTBUSY.
    drop(file1);
    drop(file2);

    // Set permissions to make the scripts executable.
    for path in [path1, path2] {
        syd::fs::chmod_x(path).expect("Failed to set file executable");
    }

    // Step 1: Allow both the interpreter and the script.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["./script1.sh"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    // Step 2: Allow the scripts but disable the interpreter.
    // This will slip through the seccomp sandbox
    // but it's caught by the exec-TOCTOU mitigator.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("deny/exec+/***")
        .m("allow/exec+/**/*.so*")
        .m("allow/exec+/**/script[1-2].sh")
        .argv(["./script1.sh"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    // Step 3: Allow the interpreter but disable the script2.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/script2.sh")
        .argv(["./script1.sh"])
        .status()
        .expect("execute syd");
    fixup!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );

    Ok(())
}

// Check if a denylisted library can be injected using dlopen().
fn test_syd_exec_sandbox_prevent_library_injection_dlopen_bare() -> TestResult {
    skip_unless_available!("cc", "python");

    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/lib-bad/*.so")
        .argv(["python", "-c"])
        .arg(
            r##"
import ctypes, os, shutil, subprocess, sys

if os.path.exists("test.c"):
    os.remove("test.c")
if os.path.exists("lib-bad"):
    shutil.rmtree("lib-bad")

CODE = "int test() { return 0; }"
COMP = ["cc", "-Wall", "-Wextra", "-shared", "-o", "./lib-bad/libtest.so", "-fPIC", "test.c"]
try:
    with open("test.c", "w") as f:
        f.write(CODE)
    os.mkdir("lib-bad", 0o700)
    subprocess.run(COMP, check=True)
except Exception as e:
    sys.stderr.write("Exception during compile: %r\n" % e)
    sys.exit(128)

try:
    libtest = ctypes.CDLL(f'./lib-bad/libtest.so')
except OSError as e:
    # XXX: ctypes does not set errno!
    # e.errno and e.strerror are 0 and None respectively.
    # and the str(e) can be "permission denied" or
    # "failed to map" so it's not reliably either.
    sys.stderr.write("Exception during dlopen: %r\n" % e)
    sys.exit(13)
else:
    sys.exit(libtest.test())
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(
        code == nix::libc::EACCES || code == 128,
        "code:{code} status:{status:?}"
    );
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check if a denylisted library can be injected using dlopen().
fn test_syd_exec_sandbox_prevent_library_injection_dlopen_wrap() -> TestResult {
    skip_unless_unshare!();
    skip_unless_available!("cc", "python");

    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/lib-bad/*.so")
        .argv(["python", "-c"])
        .arg(
            r##"
import ctypes, os, shutil, subprocess, sys

if os.path.exists("test.c"):
    os.remove("test.c")
if os.path.exists("lib-bad"):
    shutil.rmtree("lib-bad")

CODE = "int test() { return 0; }"
COMP = ["cc", "-Wall", "-Wextra", "-shared", "-o", "./lib-bad/libtest.so", "-fPIC", "test.c"]
try:
    with open("test.c", "w") as f:
        f.write(CODE)
    os.mkdir("lib-bad", 0o700)
    subprocess.run(COMP, check=True)
except Exception as e:
    sys.stderr.write("Exception during compile: %r\n" % e)
    sys.exit(128)

try:
    libtest = ctypes.CDLL(f'./lib-bad/libtest.so')
except OSError as e:
    # XXX: ctypes does not set errno!
    # e.errno and e.strerror are 0 and None respectively.
    # and the str(e) can be "permission denied" or
    # "failed to map" so it's not reliably either.
    sys.stderr.write("Exception during dlopen: %r\n" % e)
    sys.exit(13)
else:
    sys.exit(libtest.test())
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(
        code == nix::libc::EACCES || code == 128,
        "code:{code} status:{status:?}"
    );
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

// Check if a denylisted library can be injected using LD_LIBRARY_PATH.
// Note the seccomp sandbox is not able to catch this.
// This is prevented by the TOCTOU-mitigator on exec(2) exit.
fn test_syd_exec_sandbox_prevent_library_injection_LD_LIBRARY_PATH() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("cc", "sh");

    let status = syd()
        .env("LD_TRACE_LOADED_OBJECTS", "YesPlease")
        .env("LD_VERBOSE", "YesPlease")
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/lib-bad/*.so")
        .argv(["sh", "-cex"])
        .arg(
            r##"
# Ensure syd's CWD does not match our CWD
mkdir -m700 -p foo
cd foo

cat > lib-good.c <<EOF
int func(void) { return 0; }
EOF

cat > lib-bad.c <<EOF
int func(void) { return 42; }
EOF

cat > bin.c <<EOF
extern int func(void);
int main(void) { return func(); }
EOF

mkdir -m700 -p lib-good lib-bad
cc -Wall -Wextra lib-good.c -shared -o lib-good/libext.so -fPIC
cc -Wall -Wextra lib-bad.c -shared -o lib-bad/libext.so -fPIC

cc -Wall -Wextra bin.c -L./lib-good -lext -obin
r=0
env LD_LIBRARY_PATH="./lib-good:$LD_LIBRARY_PATH" ./bin || r=$?
echo >&2 "Good returned: $r"
test $r -eq 0

r=0
env LD_LIBRARY_PATH="./lib-bad:$LD_LIBRARY_PATH" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 42; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
    true
fi
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Check if a denylisted library can be injected using LD_PRELOAD.
// Note the seccomp sandbox is not able to catch this.
// This is prevented by the TOCTOU-mitigator on exec(2) exit.
fn test_syd_exec_sandbox_prevent_library_injection_LD_PRELOAD() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("cc", "sh");

    let status = syd()
        .env("LD_TRACE_LOADED_OBJECTS", "YesPlease")
        .env("LD_VERBOSE", "YesPlease")
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/lib-bad/*.so")
        .argv(["sh", "-cex"])
        .arg(
            r##"
# Ensure syd's CWD does not match our CWD
mkdir -m700 -p foo
cd foo

cat > lib-good.c <<EOF
#include <sys/types.h>
pid_t getpid(void) { return 0; }
EOF

cat > lib-bad.c <<EOF
#include <sys/types.h>
pid_t getpid(void) { return 0; }
EOF

cat > bin.c <<EOF
#include <sys/types.h>
#include <unistd.h>
int main(void) { return getpid(); }
EOF

mkdir -m700 -p lib-good lib-bad
cc -Wall -Wextra lib-good.c -shared -o lib-good/libext.so -fPIC
cc -Wall -Wextra lib-bad.c -shared -o lib-bad/libext.so -fPIC

cc -Wall -Wextra bin.c -obin
r=0
env LD_PRELOAD="./lib-good/libext.so" ./bin || r=$?
echo >&2 "Good returned: $r"
test $r -eq 0

r=0
env LD_PRELOAD="./lib-bad/libext.so" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 1; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
fi

r=0
env LD_PRELOAD="foo bar baz ./lib-bad/libext.so" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 1; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
    true
fi

r=0
env LD_PRELOAD="foo:bar:baz:./lib-bad/libext.so:a:b:c" ./bin || r=$?
echo >&2 "Bad returned: $r"
if test $r -eq 1; then
    echo >&2 "Library injection succeded!"
    false
else
    echo >&2 "Library injection failed!"
    true
fi
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv4_allow() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:0")
        .m("allow/net/bind+127.0.0.1!4242")
        .m("allow/net/connect+127.0.0.1!4242")
        .do_("connect4", ["127.0.0.1", "4242"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv4_deny() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:0")
        .m("allow/net/bind+127.0.0.1!4242")
        .m("deny/net/connect+127.0.0.1!4242")
        .do_("connect4", ["127.0.0.1", "4242"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

// Tests if network connect sandboxing works to allow.
fn test_syd_network_sandbox_connect_ipv6_allow() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:0")
        .m("allow/net/bind+::1!4242")
        .m("allow/net/connect+::1!4242")
        .do_("connect6", ["::1", "4242"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

// Tests if network connect sandboxing works to deny.
fn test_syd_network_sandbox_connect_ipv6_deny() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:0")
        .m("allow/net/bind+::1!4242")
        .m("deny/net/connect+::1!4242")
        .do_("connect6", ["::1", "4242"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv4_failure() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:0")
        .m("allow/net/bind+127.0.0.1!0")
        .do_("connect4_0", ["127.0.0.1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv4_success() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:1")
        .m("allow/net/bind+127.0.0.1!0")
        .do_("connect4_0", ["127.0.0.1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv6_failure() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:0")
        .m("allow/net/bind+::1!0")
        .do_("connect6_0", ["::1"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ECONNREFUSED);
    Ok(())
}

fn test_syd_network_sandbox_allow_safe_bind_ipv6_success() -> TestResult {
    if !check_ipv6() {
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .m("allow/exec,read,stat,ioctl+/***")
        .m("trace/allow_safe_bind:1")
        .m("allow/net/bind+::1!0")
        .do_("connect6_0", ["::1"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_handle_toolong_unix_connect() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .m("allow/net/bind+/***")
        .m("trace/allow_safe_bind:1")
        .do_("toolong_unix_connect", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_sendmsg_scm_rights_one() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .m("allow/net/bind+/***")
        .m("trace/allow_safe_bind:1")
        .do_("sendmsg_scm_rights_one", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_sendmsg_scm_rights_many() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .m("allow/net/bind+/***")
        .m("trace/allow_safe_bind:1")
        .do_("sendmsg_scm_rights_many", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_sendmmsg() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .m("allow/net/bind+/***")
        .m("trace/allow_safe_bind:1")
        .do_("sendmmsg", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_handle_toolong_unix_sendto() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .m("allow/net/bind+/***")
        .m("trace/allow_safe_bind:1")
        .do_("toolong_unix_sendto", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_handle_toolong_unix_sendmsg() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .m("allow/net/bind+/***")
        .m("trace/allow_safe_bind:1")
        .do_("toolong_unix_sendmsg", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mask_simple() -> TestResult {
    skip_unless_available!("cmp", "readlink", "sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .argv(["sh", "-cx"])
        .arg(
            r##"
echo 'Heavy is the root of light. Still is the master of moving.' > tao.orig
echo 'Heavy is the root of light. Still is the master of moving.' > tao.mask
abs=$(readlink -f tao.mask)
test -f "$abs" || exit 1
test -c "/dev/syd/mask+${abs}" || exit 2
test -f "$abs" || exit 3
cat tao.mask || exit 4
echo > tao.mask || exit 5
cmp tao.orig tao.mask && exit 6
test -c "/dev/syd/mask-${abs}" || exit 7
cmp tao.orig tao.mask || exit 8
true
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_truncate() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .do_("truncate", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_truncate64() -> TestResult {
    let status = syd()
        .m("allow/exec,read,stat,write,ioctl+/***")
        .do_("truncate64", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_kcapi_hash_block() -> TestResult {
    let status = syd()
        .p("lib")
        .do_("kcapi_hash_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("trace/allow_safe_kcapi:1")
        .do_("kcapi_hash_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("sandbox/net:on")
        .do_("kcapi_hash_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_not_supported!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/net:on")
        .m("trace/allow_safe_kcapi:1")
        .do_("kcapi_hash_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    Ok(())
}

fn test_syd_kcapi_hash_stream() -> TestResult {
    let status = syd()
        .p("lib")
        .do_("kcapi_hash_stream", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("trace/allow_safe_kcapi:1")
        .do_("kcapi_hash_stream", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("sandbox/net:on")
        .do_("kcapi_hash_stream", ["0"])
        .status()
        .expect("execute syd");
    assert_status_not_supported!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/net:on")
        .m("trace/allow_safe_kcapi:1")
        .do_("kcapi_hash_stream", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    Ok(())
}

fn test_syd_kcapi_cipher_block() -> TestResult {
    let status = syd()
        .p("lib")
        .do_("kcapi_cipher_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("trace/allow_safe_kcapi:1")
        .do_("kcapi_cipher_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("sandbox/net:on")
        .do_("kcapi_cipher_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    let status = syd()
        .p("lib")
        .m("sandbox/net:on")
        .m("trace/allow_safe_kcapi:1")
        .do_("kcapi_cipher_block", ["0"])
        .status()
        .expect("execute syd");
    assert_status_code_matches!(status, 0 | nix::libc::EAFNOSUPPORT);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_cmp_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=2 count=8 status=progress | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_cmp_null_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_cmp_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=2 count=8 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_cmp_null_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/null | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_aes_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=2 count=8 status=progress | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_aes_null_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_aes_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=2 count=8 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_aes_null_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/null | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_cmp_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=16 count=1 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=32 count=2 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bscan_append_cmp_mini_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
for i in {1..8192}; do
    dd if=/dev/random bs=1 count=1 | tee -a ./test.plain >> ./test.crypt
done
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bscan_append_cmp_incr_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
for i in {1..8192}; do
    dd if=/dev/random bs=1 count=${i} | tee -a ./test.plain >> ./test.crypt
done
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bscan_append_cmp_decr_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
for i in {8192..1}; do
    dd if=/dev/random bs=1 count=${i} | tee -a ./test.plain >> ./test.crypt
done
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_cmp_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=16 count=1 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=32 count=2 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_aes_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=16 count=1 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=32 count=2 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bscan_append_aes_mini_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
for i in {1..8192}; do
    dd if=/dev/random bs=1 count=1 | tee -a ./test.plain >> ./test.crypt
done
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bscan_append_aes_incr_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
for i in {1..8192}; do
    dd if=/dev/random bs=1 count=${i} | tee -a ./test.plain >> ./test.crypt
done
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bscan_append_aes_decr_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
dd if=/dev/null | tee ./test.plain > ./test.crypt
for i in {8192..1}; do
    dd if=/dev/random bs=1 count=${i} | tee -a ./test.plain >> ./test.crypt
done
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_aes_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=16 count=1 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=32 count=2 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_cmp_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=2 count=7 status=progress | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_cmp_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=2 count=7 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_aes_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=2 count=7 status=progress | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_aes_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=2 count=7 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_cmp_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1 count=7 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=2 count=7 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_cmp_nano_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("python");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(64):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_cmp_tiny_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("python");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(4096):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_cmp_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1 count=7 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=2 count=7 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_aes_tiny_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1 count=7 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=2 count=7 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_aes_nano_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    skip_unless_available!("python");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(64):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_aes_tiny_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    skip_unless_available!("python");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(4096):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_aes_tiny_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1 count=7 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=2 count=7 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_cmp_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_cmp_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_cmp_rand_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}

dd_rand | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_cmp_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_cmp_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_cmp_rand_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_rand | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_aes_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_aes_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_aes_rand_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}

dd_rand | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_aes_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_single_aes_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_single_aes_rand_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_rand | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_cmp_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=2M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_cmp_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=16M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_cmp_rand_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}

dd_rand | tee ./test.plain > ./test.crypt
dd_rand | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_cmp_fuzz_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}

dd_rand | tee ./test.plain > ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_rand | tee -a ./test.plain >> ./test.crypt
done

cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_cmp_zero_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_zero() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/zero with random size and count
  dd if=/dev/zero bs=$random_size count=$random_count status=progress
}

dd_zero | tee ./test.plain > ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_zero | tee -a ./test.plain >> ./test.crypt
done

cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_cmp_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=2M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_cmp_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=16M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_cmp_rand_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_rand | tee ./test.plain | {syd_cp} -v ./test.crypt
dd_rand | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_cmp_fuzz_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_rand | tee ./test.plain | {syd_cp} -v ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_rand | tee -a ./test.plain >> ./test.crypt
done

cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_cmp_zero_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_zero() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/zero with random size and count
  dd if=/dev/zero bs=$random_size count=$random_count status=progress
}}

dd_zero | tee ./test.plain | {syd_cp} -v ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_zero | tee -a ./test.plain >> ./test.crypt
done

cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_aes_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=2M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_aes_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=16M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_aes_rand_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}

dd_rand | tee ./test.plain > ./test.crypt
dd_rand | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_aes_fuzz_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}

dd_rand | tee ./test.plain > ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_rand | tee -a ./test.plain >> ./test.crypt
done
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_aes_zero_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_zero() {
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/zero with random size and count
  dd if=/dev/zero bs=$random_size count=$random_count status=progress
}

dd_zero | tee ./test.plain > ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_zero | tee -a ./test.plain >> ./test.crypt
done
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_aes_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=2M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_bsize_append_aes_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=8M count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=16M count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_aes_rand_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_rand | tee ./test.plain | {syd_cp} -v ./test.crypt
dd_rand | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_aes_fuzz_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_rand() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/random with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_rand | tee ./test.plain | {syd_cp} -v ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_rand | tee -a ./test.plain >> ./test.crypt
done
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_append_aes_zero_pipe() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(&format!(
            r##"
# Simulates dd with random block sizes and count,
# with a maximum total size of 8MB.
dd_zero() {{
  # Generate random size between 1 and 1024 (inclusive).
  random_size=$((RANDOM % 1024 + 1))

  # Generate random count between 1 and 8192 (adjust for desired max size)
  # This ensures total size (count * block_size) won't exceed 8MB.
  max_count=$((8 * 1024 * 1024 / random_size)) # Adjust divisor for different max size.
  random_count=$((RANDOM % max_count + 1))

  # Read from /dev/zero with random size and count
  dd if=/dev/random bs=$random_size count=$random_count status=progress
}}

dd_zero | tee ./test.plain | {syd_cp} -v ./test.crypt

# Generate a random number between 3 and 7 (inclusive)
# for the number of iterations
num_iterations=$(( RANDOM % 5 + 3 ))
for (( i=0; i<$num_iterations; i++ )); do
    dd_zero | tee -a ./test.plain >> ./test.crypt
done
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_cmp_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_cmp_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain > ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_cmp_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_cmp_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_aes_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_aes_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain > ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_aes_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_single_aes_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_cmp_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=2097169 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_cmp_mild_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("python");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(2 * 4096):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_cmp_huge_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("python");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(4 * 4096):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_cmp_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=16000057 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_cmp_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=2097169 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_cmp_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=16000057 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
cmp test.plain test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_aes_mild_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=2097169 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_aes_mild_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    skip_unless_available!("python");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(2 * 4096):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_sieve_append_aes_huge_copy() -> TestResult {
    skip_unless_available!("bash");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    skip_unless_available!("python");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["bash", "-cex"])
        .arg(
            r##"
cat >primegen.py <<'EOF'
def primegen(limit):
  from math import sqrt
  primes = [2]
  for num in range(3, limit, 2):
    is_prime = True
    square_root = int(sqrt(num))
    for prime in primes:
      if num % prime == 0:
        is_prime = False
        break
      if prime > square_root:
        break
    if is_prime:
      yield num
for num in primegen(4 * 4096):
    print(num)
EOF

python primegen.py > primes.lst
dd if=/dev/null | tee ./test.plain > ./test.crypt
while read -r -d$'\n' num; do
    dd if=/dev/random bs=${num} count=1 | tee -a ./test.plain >> ./test.crypt
done < primes.lst
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("bash")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_aes_huge_copy() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain > ./test.crypt
dd if=/dev/random bs=16000057 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_aes_mild_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=1048573 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=2097169 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_crypt_sandboxing_prime_append_aes_huge_pipe() -> TestResult {
    skip_unless_available!("sh");
    skip_unless_available!("tee");
    skip_unless_available!("dd");
    skip_unless_available!("getfattr");
    build_openssl_aes_ctr();

    let key = syd::hash::Key::random().expect("getrandom").as_hex();
    let cwd = current_dir(false)?.display().to_string();

    let syd_cp = &SYD_CP.to_string();
    let status = syd()
        .p("lib")
        .m("sandbox/crypt:on")
        .m(&format!("crypt/key:{key}"))
        .m(&format!("crypt+{cwd}/*.crypt"))
        .argv(["sh", "-cex"])
        .arg(&format!(
            r##"
dd if=/dev/random bs=7999993 count=5 status=progress | tee ./test.plain | {syd_cp} -v ./test.crypt
dd if=/dev/random bs=16000057 count=3 status=progress | tee -a ./test.plain >> ./test.crypt
"##,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let syd_aes = &SYD_AES.to_string();
    let status = Command::new("sh")
        .arg("-cex")
        .arg(&format!(
            r##"
iv=$(getfattr -nuser.syd.crypt.iv --only-values ./test.crypt)
if test -x ./aes-ctr; then
    ./aes-ctr -d -k{key} -i${{iv}} < ./test.crypt > ./test-ssl.decrypt
    cmp test.plain test-ssl.decrypt
fi
{syd_aes} -v -d -k{key} -i${{iv}} < ./test.crypt > ./test.decrypt
cmp test.plain test.decrypt
"##,
        ))
        .status()
        .expect("execute sh");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_exit_wait_default() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .argv(["sh", "-cex"])
        .arg(
            r##"
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_default_unsafe_exec() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_exec:1")
        .argv(["sh", "-cex"])
        .arg(
            r##"
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("trace/exit_wait_all:0")
        .argv(["sh", "-cex"])
        .arg(
            r##"
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid_unsafe_exec() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("trace/exit_wait_all:0")
        .m("trace/allow_unsafe_exec:1")
        .argv(["sh", "-cex"])
        .arg(
            r##"
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid_with_runaway_cmd_exec_process() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/exit_wait_all:0")
        .argv(["sh", "-cex"])
        .arg(format!(
            "
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_pid_unsafe_exec_with_runaway_cmd_exec_process() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/exit_wait_all:0")
        .m("trace/allow_unsafe_exec:1")
        .argv(["sh", "-cex"])
        .arg(format!(
            "
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 0, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("trace/exit_wait_all:1")
        .argv(["sh", "-cex"])
        .arg(
            r##"
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all_unsafe_exec() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("trace/exit_wait_all:1")
        .m("trace/allow_unsafe_exec:1")
        .argv(["sh", "-cex"])
        .arg(
            r##"
: > test
cat > exec.sh <<'EOF'
#!/bin/sh -ex
sleep 5
echo OK > test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh &
true
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all_with_runaway_cmd_exec_process() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/exit_wait_all:1")
        .argv(["sh", "-cex"])
        .arg(format!(
            "
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_exit_wait_all_unsafe_exec_with_runaway_cmd_exec_process() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/exit_wait_all:1")
        .m("trace/allow_unsafe_exec:1")
        .argv(["sh", "-cex"])
        .arg(format!(
            "
: > test
chmod 600 test
cat > exec.sh <<EOF
#!/bin/sh -ex
sleep \\$1
echo \\$2 > $PWD/test
exit 42
EOF
chmod +x exec.sh
nohup ./exec.sh 5 OK &
# Careful here, cmd/exec changes CWD to /.
test -c \"$({} $PWD/exec.sh 15 RUN)\"
true
",
            *SYD_EXEC
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let path = Path::new("./test");
    assert!(path.exists());
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 3, "Unexpected file metadata: {data:?}");

    // Wait for the runaway process to change the file.
    eprintln!("Waiting for 20 seconds for the runaway process to write to the file");
    sleep(Duration::from_secs(20));
    let data = metadata(path).expect("Unable to access test file metadata");
    assert_eq!(data.len(), 4, "Unexpected file metadata: {data:?}");

    Ok(())
}

fn test_syd_cli_args_override_user_profile() -> TestResult {
    skip_unless_available!("sh");

    let _ = unlink(".user.syd-3");
    let mut file = File::create(".user.syd-3").expect("Failed to create .user.syd-3");
    file.write_all(b"mem/max:4242\npid/max:2525\n")
        .expect("Failed to write to .user.syd-3");

    let mut child = syd()
        .m("pid/max:4242")
        .m("stat")
        .c("true")
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");

    // Read the output from the child process
    let child_stderr = child.stderr.as_mut().expect("child stderr");
    let mut reader = BufReader::new(child_stderr);
    let mut output = String::new();
    if let Err(error) = reader.read_to_string(&mut output) {
        return Err(TestError(format!(
            "Failed to read output of child process: {error}"
        )));
    }
    print!("Child output:\n{output}");

    assert!(output.contains("Pid Max: 4242"));
    //This may fail if the site-wide config file has lock:on.
    //assert!(output.contains("Memory Max: 4242"));

    Ok(())
}

fn test_syd_ifconfig_lo_bare() -> TestResult {
    let status = syd()
        .p("lib")
        .do_("ifconfig_lo", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ifconfig_lo_wrap() -> TestResult {
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/net:1")
        .do_("ifconfig_lo", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_binfmt_malformed_too_short() -> TestResult {
    skip_unless_available!("python");

    let cwd = current_dir(true)?.display().to_string();

    syd::fs::cat("interpreter", "π")?;
    syd::fs::cat("script1", "#!interpreter")?;
    syd::fs::cat("script2", "#!./interpreter")?;
    syd::fs::cat("script3", "#!./././././././interpreter")?;
    syd::fs::cat("script4", "#! ./interpreter")?;
    syd::fs::cat("script5", "#!\t./interpreter")?;
    syd::fs::cat(
        "script6",
        format!("#! \t  \t   \t    \t../{cwd}/./../{cwd}/./././interpreter"),
    )?;
    syd::fs::chmod_x("interpreter")?;
    for i in 1..=6 {
        syd::fs::chmod_x(&format!("script{i}"))?;
    }

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["python", "-c"])
        .arg(
            r##"
import os
import sys
import errno

def say(msg):
    print(msg, file=sys.stderr)

def run(script):
    pid = os.fork()

    if pid == 0:
        # Child process
        say(f"Executing script {script}")
        try:
            os.execvp(script, [script])
        except OSError as e:
            exit(e.errno)
    elif pid > 0:
        # Parent process
        _, status = os.waitpid(pid, 0)
        code = os.WEXITSTATUS(status)
        say(f"Script {script} returned {code}")
        return code
    else:
        raise Exception("Failed to fork")

# syd delivering ENOEXEC
# TODO: Distinguish from kernel
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(1)

say("Turn allow_unsupp_exec on")
os.lstat("/dev/syd/trace/allow_unsupp_exec:1")

# kernel delivering ENOEXEC
# TODO: Distinguish from syd
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(2)
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_binfmt_malformed_bad_magic() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("python");

    // Number of bytes in an identifier.
    const SIZE_OF_IDENT: usize = 16;
    let cwd = current_dir(true)?.display().to_string();

    syd::fs::cat("interpreter", "#".repeat(SIZE_OF_IDENT))?;
    syd::fs::cat("script1", "#!interpreter")?;
    syd::fs::cat("script2", "#!./interpreter")?;
    syd::fs::cat("script3", "#!./././././././interpreter")?;
    syd::fs::cat("script4", "#! ./interpreter")?;
    syd::fs::cat("script5", "#!\t./interpreter")?;
    syd::fs::cat(
        "script6",
        format!("#! \t  \t   \t    \t../{cwd}/./../{cwd}/./././interpreter"),
    )?;
    syd::fs::chmod_x("interpreter")?;
    for i in 1..=6 {
        syd::fs::chmod_x(&format!("script{i}"))?;
    }

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["python", "-c"])
        .arg(
            r##"
import os
import sys
import errno

def say(msg):
    print(msg, file=sys.stderr)

def run(script):
    pid = os.fork()

    if pid == 0:
        # Child process
        say(f"Executing script {script}")
        try:
            os.execvp(script, [script])
        except OSError as e:
            exit(e.errno)
    elif pid > 0:
        # Parent process
        _, status = os.waitpid(pid, 0)
        code = os.WEXITSTATUS(status)
        say(f"Script {script} returned {code}")
        return code
    else:
        raise Exception("Failed to fork")

for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(1)

say("Turn allow_unsupp_exec on")
os.lstat("/dev/syd/trace/allow_unsupp_exec:1")

for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(2)

say("Turn allow_unsupp_exec off")
os.lstat("/dev/syd/trace/allow_unsupp_exec:0")

for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(3)
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_binfmt_malformed_invalid_elf_class() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("python");

    // Define a byte array `elf` to be used in testing ELF header parsing.
    // This array is structured to mimic an ELF header but includes an intentionally invalid ELF class to trigger an error.
    // The ELF header starts with a magic number 0x7F followed by "ELF" (45 4C 46 in hexadecimal),
    // which is correctly set to indicate that this is an ELF file.
    let elf = b"\x7FELF\x03\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00";
    // ELF header with a malformed class
    // The byte at index 4 (0-based index) represents the EI_CLASS:
    // - 0 (ELFCLASSNONE) means invalid class (not used),
    // - 1 (ELFCLASS32) means 32-bit objects,
    // - 2 (ELFCLASS64) means 64-bit objects.
    // Here, the byte at index 4 is set to 3, which is not a valid ELF class value,
    // as the defined valid values are 0 for ELFCLASSNONE, 1 for ELFCLASS32, and 2 for ELFCLASS64.
    // The value 3 does not correspond to any recognized or valid ELF class, which is intended to simulate
    // a corrupted or non-standard ELF file and is expected to trigger a 'Malformed' error in the parsing logic
    // indicating an "invalid ELF class".

    let cwd = current_dir(true)?.display().to_string();

    syd::fs::cat("interpreter", elf)?;
    syd::fs::cat("script1", "#!interpreter")?;
    syd::fs::cat("script2", "#!./interpreter")?;
    syd::fs::cat("script3", "#!./././././././interpreter")?;
    syd::fs::cat("script4", "#! ./interpreter")?;
    syd::fs::cat("script5", "#!\t./interpreter")?;
    syd::fs::cat(
        "script6",
        format!("#! \t  \t   \t    \t../{cwd}/./../{cwd}/./././interpreter"),
    )?;
    syd::fs::chmod_x("interpreter")?;
    for i in 1..=6 {
        syd::fs::chmod_x(&format!("script{i}"))?;
    }

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["python", "-c"])
        .arg(
            r##"
import os
import sys
import errno

def say(msg):
    print(msg, file=sys.stderr)

def run(script):
    pid = os.fork()

    if pid == 0:
        # Child process
        say(f"Executing script {script}")
        try:
            os.execvp(script, [script])
        except OSError as e:
            exit(e.errno)
    elif pid > 0:
        # Parent process
        _, status = os.waitpid(pid, 0)
        code = os.WEXITSTATUS(status)
        say(f"Script {script} returned {code}")
        return code
    else:
        raise Exception("Failed to fork")

# syd delivering ENOEXEC
# TODO: Distinguish from kernel
# One idea is to check JSON seccomp logs for flags:1 aka continue,
# which indicates it is the kernel doing the error return.
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(1)

say("Turn allow_unsupp_exec on")
os.lstat("/dev/syd/trace/allow_unsupp_exec:1")

# kernel delivering ENOEXEC
# TODO: Distinguish from syd
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(2)

say("Turn allow_unsupp_exec off")
os.lstat("/dev/syd/trace/allow_unsupp_exec:0")

# syd delivering ENOEXEC
# TODO: Distinguish from kernel
for i in range(1, 7):
    if run(f"./script{i}") != errno.ENOEXEC:
        exit(3)
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_deny_elf32() -> TestResult {
    skip_unless_bitness!("64");
    skip_unless_feature!("elf");
    skip_unless_available!("cc", "sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["sh", "-cx"])
        .arg(
            r##"
cat > exit.c <<EOF
#include <stdlib.h>
int main() {
    exit(0);
}
EOF

cc -m32 exit.c -o exit
test $? -eq 0 || exit 128
chmod +x ./exit || exit 128
./exit || exit 1
test -c /dev/syd/trace/deny_elf32:1 || exit 2
./exit && exit 3
test -c /dev/syd/trace/deny_elf32:0 || exit 2
./exit || exit 4
true
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(code == 0 || code == 128, "code:{code} status:{status:?}");
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_deny_elf_dynamic() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("cc", "sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .args(["sh", "-cx"])
        .arg(
            r##"
cat > exit.c <<EOF
#include <stdlib.h>
int main() {
    exit(0);
}
EOF

cc exit.c -o exit
test $? -eq 0 || exit 128
chmod +x ./exit || exit 128
./exit || exit 1
test -c /dev/syd/trace/deny_elf_dynamic:1 || exit 2
./exit && exit 3
test -c /dev/syd/trace/deny_elf_dynamic:0 || exit 2
./exit || exit 4
true
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(code == 0 || code == 128, "code:{code} status:{status:?}");
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_deny_elf_static() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("cc", "sh");

    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .args(["sh", "-cx"])
        .arg(
            r##"
cat > exit.c <<EOF
#include <stdlib.h>
int main() {
    exit(0);
}
EOF

cc exit.c -o exit -static
test $? -eq 0 || exit 128
chmod +x ./exit || exit 128
./exit || exit 1
test -c /dev/syd/trace/deny_elf_static:1 || exit 2
./exit && exit 3
test -c /dev/syd/trace/deny_elf_static:0 || exit 2
./exit || exit 4
true
    "##,
        )
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    assert!(code == 0 || code == 128, "code:{code} status:{status:?}");
    if code == 128 {
        // Compilation failed, test skipped.
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_deny_script() -> TestResult {
    syd::fs::cat("script", "#!/bin/sh\nexit 42")?;
    syd::fs::chmod_x("script")?;

    // Scripts are allowed by default.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["./script"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, 42);

    // Scripts are denied with deny_script:1.
    // syd uses ENOSYS to distinguish from a kernel deny.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("trace/deny_script:1")
        .argv(["./script"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOSYS);

    Ok(())
}

fn test_syd_prevent_ld_linux_exec_break() -> TestResult {
    skip_unless_feature!("elf");
    skip_unless_available!("bash");

    // Shared library execution is denied by default.
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["bash", "-cx"])
        .arg(
            r##"
#!/bin/bash
# Careful, ld-linux path differs on glibc and musl.
for f in /lib{64,}/ld-{musl,linux}-{aarch64,armv7,x86{-,_}64}.so.{1,2}; do
    if test -e "${f}"; then
        exec "${f}" /bin/true
        exit 127
    fi
done
echo >&2 "ld.so not found"
exit 127
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    // Shared library execution is allowed with allow_unsupp_exec.
    let status = syd()
        .p("lib")
        .m("trace/allow_unsupp_exec:1")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .argv(["bash", "-cx"])
        .arg(
            r##"
# Careful, ld-linux path differs on glibc and musl.
for f in /lib{64,}/ld-{musl,linux}-{aarch64,armv7,x86{-,_}64}.so.{1,2}; do
    if test -e "${f}"; then
        exec "${f}" /bin/true
        exit 127
    fi
done
echo >&2 "ld.so not found"
exit 127
"##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_force_sandbox() -> TestResult {
    skip_unless_available!("true");

    let bad_md5 = "a".repeat(32);
    let bad_sha1 = "a".repeat(40);
    let bad_sha256 = "a".repeat(64);
    let bad_sha384 = "a".repeat(96);
    let bad_sha512 = "a".repeat(128);

    // Note, `which" returns canonicalized path.
    let bin_true = which("true").expect("syd-test-do in PATH");

    let md5_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Md5,
    )?);
    let sha1_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha1,
    )?);
    let sha256_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha256,
    )?);
    let sha384_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha384,
    )?);
    let sha512_sum = const_hex::encode(syd::hash::hash(
        BufReader::new(File::open(&bin_true).expect("open syd-test-do")),
        HashAlgorithm::Sha512,
    )?);

    // Test 1: Force sandboxing defaults.
    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:allow")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:kill")
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    // Test 2: Force sandboxing allow with MD5 & SHA1,3-{256,384,512}
    // We set default/force:warn so as not to care about dynamic libraries.
    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{md5_sum}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{sha1_sum}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{sha256_sum}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{sha384_sum}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{sha512_sum}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{bad_md5}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{bad_sha1}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{bad_sha256}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{bad_sha384}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    let status = syd()
        .p("lib")
        .m("sandbox/force:on")
        .m("default/force:warn")
        .m(&format!("force+{bin_true}:{bad_sha512}:kill"))
        .argv(["true"])
        .status()
        .expect("execute syd");
    assert_status_killed!(status);

    Ok(())
}

fn test_syd_segvguard_core_safe() -> TestResult {
    skip_unless_available!("perl");
    skip_unless_coredumps!();

    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/allow_unsafe_prlimit:0")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do__("segv")
        .argv(["perl", "-e"])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :signal_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $sig;
my $code;
my $status;

# segvguard/maxcrashes is set to 5 by default.
my $coredumps = 0; # safe mode, may never trigger segvguard.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status)) {{
        warn "process $i was terminated by signal " . WTERMSIG($status) . "\n";
        if (wcoredump($status)) {{
            $coredumps += 1;
            warn "process $i dumped core.\n";
        }} else {{
            warn "process $i did not dump core.\n";
        }}
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

if (!$coredumps) {{
    warn "failed to trigger segvguard due to safe mode, skipping test!\n";
    exit 0;
}} else {{
    warn "caused $coredumps coredumps, proceeding...\n";
}}

# Now segvguard must block
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFSIGNALED($status)) {{
    $sig = WTERMSIG($status);
    if ($sig == SIGKILL) {{
        warn "execution was prevented by segvguard\n";
    }} else {{
        die "process was terminated with unexpected signal $sig\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Ensure segvguard allows everything else.
$pid = fork();
if ($pid == 0) {{
    exec('true') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    if ($code == 0) {{
        warn "execution was allowed by segvguard\n";
    }} else {{
        die "process exited with unexpected code $code\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Disable segvguard and retry!
if (-c '/dev/syd/segvguard/expiry:0') {{
    warn "segvguard disabled"
}} else {{
    die "failed to disable segvguard"
}}

# Now segvguard must allow
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process was not terminated but exited with code $code\n";
}} elsif (WIFSIGNALED($status)) {{
    warn "process was terminated by signal " . WTERMSIG($status) . "\n";
    if (wcoredump($status)) {{
        warn "process dumped core.\n";
    }} else {{
        warn "process did not dump core.\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_segvguard_core_unsafe() -> TestResult {
    skip_unless_available!("perl");
    skip_unless_coredumps!();

    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/allow_unsafe_prlimit:1")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do__("segv")
        .argv(["perl", "-e"])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :signal_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $sig;
my $code;
my $status;

# segvguard/maxcrashes is set to 5 by default.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
        warn "process $i dumped core as expected\n";
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

# Now segvguard must block
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFSIGNALED($status)) {{
    $sig = WTERMSIG($status);
    if ($sig == SIGKILL) {{
        warn "execution was prevented by segvguard\n";
    }} else {{
        die "process was terminated with unexpected signal $sig\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Ensure segvguard allows everything else.
$pid = fork();
if ($pid == 0) {{
    exec('true') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    if ($code == 0) {{
        warn "execution was allowed by segvguard\n";
    }} else {{
        die "process exited with unexpected code $code\n";
    }}
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

# Disable segvguard and retry!
if (-c '/dev/syd/segvguard/expiry:0') {{
    warn "segvguard disabled"
}} else {{
    die "failed to disable segvguard"
}}

# Now segvguard must allow
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process did not dump core but exited with code $code\n";
}} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
    warn "process dumped core as expected\n";
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_segvguard_suspension_safe() -> TestResult {
    skip_unless_available!("perl");
    skip_unless_coredumps!();

    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/allow_unsafe_prlimit:0")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do__("segv")
        .argv(["perl", "-e"])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $code;
my $status;

# Set segvguard suspension to 10 seconds.
if (-c '/dev/syd/segvguard/suspension:10') {{
    warn "segvguard suspension set to 10 seconds"
}} else {{
    die "failed to set segvguard suspension"
}}

my $coredumps = 0; # safe mode, may never trigger segvguard.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status)) {{
        warn "process $i was terminated by signal " . WTERMSIG($status) . "\n";
        if (wcoredump($status)) {{
            $coredumps += 1;
            warn "process $i dumped core.\n";
        }} else {{
            warn "process $i did not dump core.\n";
        }}
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

if (!$coredumps) {{
    warn "failed to trigger segvguard due to safe mode, skipping test!\n";
    exit 0;
}} else {{
    warn "caused $coredumps coredumps, proceeding...\n";
}}

# Ensure segvguard suspension expires.
warn "waiting 10 seconds for segvguard suspension to expire...";
sleep 10;

# Now segvguard must allow.
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process did not dump core but exited with code $code\n";
}} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
    warn "process dumped core as expected\n";
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_segvguard_suspension_unsafe() -> TestResult {
    skip_unless_available!("perl");
    skip_unless_coredumps!();

    let syd_do = &SYD_DO.to_string();
    let status = syd()
        .p("lib")
        .m("lock:exec")
        .m("trace/allow_unsafe_prlimit:1")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .do__("segv")
        .argv(["perl", "-e"])
        .arg(format!(
            r#"
use strict;
use warnings;
use POSIX qw(:errno_h :sys_wait_h);

# WCOREDUMP is not POSIX.
sub wcoredump {{
    my $status = shift;
    return ($status & 128) != 0;
}}

# Common variables.
my $pid;
my $code;
my $status;

# Set segvguard suspension to 10 seconds.
if (-c '/dev/syd/segvguard/suspension:10') {{
    warn "segvguard suspension set to 10 seconds"
}} else {{
    die "failed to set segvguard suspension"
}}

# segvguard/maxcrashes is set to 5 by default.
for my $i (1..5) {{
    $pid = fork();
    if ($pid == 0) {{ # Child process
        exec('{syd_do}') or exit($!); # Exit with the actual errno
    }}

    waitpid($pid, 0);
    $status = $?;
    if (WIFEXITED($status)) {{
        $code = WEXITSTATUS($status);
        die "process $i did not dump core but exited with code $code\n";
    }} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
        warn "process $i dumped core as expected\n";
    }} else {{
        die "process $i exited unexpectedly with status $status\n";
    }}
}}

# Ensure segvguard suspension expires.
warn "waiting 10 seconds for segvguard suspension to expire...";
sleep 10;

# Now segvguard must allow.
$pid = fork();
if ($pid == 0) {{
    exec('{syd_do}') or exit($!); # Exit with the actual errno
}}
waitpid($pid, 0);
$status = $?;
if (WIFEXITED($status)) {{
    $code = WEXITSTATUS($status);
    die "process did not dump core but exited with code $code\n";
}} elsif (WIFSIGNALED($status) && wcoredump($status)) {{
    warn "process dumped core as expected\n";
}} else {{
    die "process exited unexpectedly with status $status\n";
}}

1;
    "#,
        ))
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_symlink_toctou() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/stat+/etc/***")
        .m("allow/stat+/etc/ld*")
        .m("deny/read,write+/etc/passwd")
        .do_("symlink_toctou", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_symlinkat_toctou() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/stat+/etc/***")
        .m("allow/stat+/etc/ld*")
        .m("deny/read,write+/etc/passwd")
        .do_("symlinkat_toctou", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ptrmod_toctou_chdir() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .m("deny/stat+/var/empty/***")
        .do_("ptrmod_toctou_chdir", NONE)
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    // Note for some reason this test reliably passes on Ubuntu CI,
    // so we do not use fixup! here.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_fail() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/toctou_exec")
        .m("filter/exec+/**/toctou_exec")
        .do_("ptrmod_toctou_exec_fail", NONE)
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_quick() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/toctou_exec")
        .m("filter/exec+/**/toctou_exec")
        .m(&format!("deny/exec+{}", f.display()))
        .m(&format!("filter/exec+{}", f.display()))
        .do_("ptrmod_toctou_exec_success_quick", NONE)
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    assert!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_double_fork() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/toctou_exec")
        .m("filter/exec+/**/toctou_exec")
        .m(&format!("deny/exec+{}", f.display()))
        .m(&format!("filter/exec+{}", f.display()))
        .do_("ptrmod_toctou_exec_success_double_fork", NONE)
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    assert!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_quick_no_mitigation() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_exec:1")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/toctou_exec")
        .m("filter/exec+/**/toctou_exec")
        .m(&format!("deny/exec+{}", f.display()))
        .m(&format!("filter/exec+{}", f.display()))
        .do_("ptrmod_toctou_exec_success_quick", NONE)
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    ignore!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_exec_success_double_fork_no_mitigation() -> TestResult {
    // Test requires /bin/false to be denylisted.
    // false may point to various alternatives such
    // as gfalse, coreutils, busybox etc.
    let f = Path::new("/bin/false").canonicalize().expect("/bin/false");
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_exec:1")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/toctou_exec")
        .m("filter/exec+/**/toctou_exec")
        .m(&format!("deny/exec+{}", f.display()))
        .m(&format!("filter/exec+{}", f.display()))
        .do_("ptrmod_toctou_exec_success_double_fork", NONE)
        .status()
        .expect("execute syd");
    const EXKILL: i32 = 128 + nix::libc::SIGKILL;
    ignore!(
        matches!(status.code().unwrap_or(127), 0 | EXKILL),
        "status:{status:?}"
    );
    Ok(())
}

fn test_syd_ptrmod_toctou_open() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/stat+/etc/***")
        .m("allow/stat+/etc/ld*")
        .m("deny/read,write+/etc/passwd")
        .do_("ptrmod_toctou_open", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ptrmod_toctou_creat() -> TestResult {
    // FIXME: deny+filter is redundant here but test fails without deny???
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/stat+/etc/***")
        .m("allow/stat+/etc/ld*")
        .m("deny/write+/**/deny.syd-tmp")
        .m("filter/write+/**/deny.syd-tmp")
        .do_("ptrmod_toctou_creat", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_ptrmod_toctou_opath_default() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("allow/stat+/etc")
        .m("deny/stat+/etc/**")
        .m("allow/stat+/etc/ld*")
        .m("deny/read,stat,write+/etc/passwd")
        .do_("ptrmod_toctou_opath", NONE)
        .status()
        .expect("execute syd");
    // By default we turn O_PATH to O_RDONLY so there's no TOCTOU.
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_ptrmod_toctou_opath_unsafe() -> TestResult {
    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_open:1")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("allow/stat+/etc")
        .m("deny/stat+/etc/**")
        .m("allow/stat+/etc/ld*")
        .m("deny/read,stat,write+/etc/passwd")
        .do_("ptrmod_toctou_opath", NONE)
        .status()
        .expect("execute syd");
    // FIXME: https://bugzilla.kernel.org/show_bug.cgi?id=218501
    fixup!(status.success(), "status:{status:?}");

    Ok(())
}

fn test_syd_vfsmod_toctou_mmap() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/exec:on")
        .m("allow/exec+/***")
        .m("deny/exec+/**/lib-bad/*.so")
        .m("filter/exec+/**/lib-bad/*.so")
        .do_("vfsmod_toctou_mmap", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_io_uring_escape_strict() -> TestResult {
    #[cfg(feature = "uring")]
    {
        // Step 1: Default is strict.
        let status = syd()
            .p("lib")
            .m("sandbox/read,stat,write:on")
            .m("allow/read,stat,write+/***")
            .m("deny/stat+/etc/***")
            .m("allow/stat+/etc/ld*")
            .m("deny/read,write+/etc/passwd")
            .do_("io_uring_escape", ["0"])
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }
    Ok(())
}

fn test_syd_io_uring_escape_unsafe() -> TestResult {
    #[cfg(feature = "uring")]
    {
        // Step 2: Relax uring restriction.
        let status = syd()
            .p("lib")
            .m("sandbox/read,stat,write:on")
            .m("allow/read,stat,write+/***")
            .m("deny/stat+/etc/***")
            .m("allow/stat+/etc/ld*")
            .m("deny/read,write+/etc/passwd")
            .m("trace/allow_unsafe_uring:1")
            .do_("io_uring_escape", ["1"])
            .status()
            .expect("execute syd");
        assert_status_ok!(status);
    }

    Ok(())
}

fn test_syd_opath_escape() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/stat+/etc/***")
        .m("allow/stat+/etc/ld*")
        .m("deny/read,write+/etc/passwd")
        .do_("opath_escape", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_1() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_1", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_2() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_2", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_3() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_3", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_4", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_5() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_5", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_6() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_6", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_7() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_7", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_8() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_8", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_9() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_9", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_10() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_10", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_11() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_11", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_12() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_12", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_13() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_13", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_14() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_14", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_15() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_15", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_16() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_16", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_17() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_17", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_18() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_18", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_19() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_19", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_chdir_relpath_20() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_chdir_relpath_20", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_1() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_1", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_2() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_2", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_3() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_3", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_4", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_5() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_5", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_6() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_6", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_7() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_7", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_8() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_8", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_9() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_9", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_10() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_10", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_11() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_11", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_12() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_12", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_13() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_13", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_14() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_14", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_15() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_15", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_16() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_16", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_17() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_17", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_18() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_18", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_19() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_19", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_devfd_escape_open_relpath_20() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,write:on")
        .m("allow/read,write+/***")
        .do_("devfd_escape_open_relpath_20", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_1() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_1", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_2() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_2", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_3() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_3", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_4", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_5() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_5", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_6() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_6", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_chdir_relpath_7() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_chdir_relpath_7", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_1() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_1", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_2() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_2", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_3() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_3", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_4", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_5() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_5", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_6() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_6", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_open_relpath_7() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_open_relpath_7", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_relpath() -> TestResult {
    skip_unless_available!("grep");

    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .argv(["grep", "Name:[[:space:]]syd", "/proc/./self/status"])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 1,
        "code:{:?}",
        status.code()
    );
    Ok(())
}

fn test_syd_procself_escape_symlink() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_symlink", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_procself_escape_symlink_within_container() -> TestResult {
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("procself_escape_symlink", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_umask_bypass_077() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("umask_bypass_077", NONE)
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_umask_bypass_277() -> TestResult {
    // Set a liberal umask as the test expects.
    let prev_umask = umask(Mode::from_bits_truncate(0o022));
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("umask_bypass_277", NONE)
        .status()
        .expect("execute syd");
    let _ = umask(prev_umask);

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_opath() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .do_("emulate_opath", NONE)
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_otmpfile() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("emulate_otmpfile", NONE)
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_honor_umask_000() -> TestResult {
    let prev_umask = umask(Mode::from_bits_truncate(0));

    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("honor_umask", ["0666"])
        .status()
        .expect("execute syd");

    let _ = umask(prev_umask);

    assert_status_ok!(status);

    Ok(())
}

fn test_syd_honor_umask_022() -> TestResult {
    let prev_umask = umask(Mode::from_bits_truncate(0o022));

    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("honor_umask", ["0644"])
        .status()
        .expect("execute syd");

    let _ = umask(prev_umask);

    assert_status_ok!(status);

    Ok(())
}

fn test_syd_honor_umask_077() -> TestResult {
    let prev_umask = umask(Mode::from_bits_truncate(0o077));

    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("honor_umask", ["0600"])
        .status()
        .expect("execute syd");

    let _ = umask(prev_umask);

    assert_status_ok!(status);

    Ok(())
}

fn test_syd_force_umask_bypass_with_open() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_open", NONE)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    let status = syd()
        .p("lib")
        .m("trace/force_umask:7177")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_open", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_force_umask_bypass_with_mknod() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_mknod", NONE)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    let status = syd()
        .p("lib")
        .m("trace/force_umask:7177")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_mknod", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_force_umask_bypass_with_mkdir() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_mkdir", NONE)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    let status = syd()
        .p("lib")
        .m("trace/force_umask:7177")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_mkdir", NONE)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    Ok(())
}

fn test_syd_force_umask_bypass_with_fchmod() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_fchmod", NONE)
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    let status = syd()
        .p("lib")
        .m("trace/force_umask:7177")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("force_umask_bypass_with_fchmod", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_open_utf8_invalid() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/write:on")
        .m("allow/write+/***")
        .do_("open_utf8_invalid", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_exec_in_inaccessible_directory() -> TestResult {
    skip_unless_available!("bash");

    let status = syd()
        .p("lib")
        .m("sandbox/exec,write:on")
        .m("allow/exec,write+/***")
        .do_("exec_in_inaccessible_directory", NONE)
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_pipe() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .do_("fstat_on_pipe", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_socket() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/stat:on")
        .m("allow/stat+/***")
        .do_("fstat_on_socket", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_deleted_file() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("fstat_on_deleted_file", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstat_on_temp_file() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("fstat_on_temp_file", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fchmodat_on_proc_fd() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("fchmodat_on_proc_fd", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_linkat_on_fd() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("linkat_on_fd", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_ioctl_tiocsti() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("block_ioctl_tiocsti", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_prctl_ptrace() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("block_prctl_ptrace", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_prevent_ptrace_detect() -> TestResult {
    let status = syd()
        .p("lib")
        .do_("detect_ptrace", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_dev_random() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/read+/dev/random")
        .do_("block_dev_random", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_block_dev_urandom() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/read+/dev/urandom")
        .do_("block_dev_urandom", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_kill_during_syscall() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("kill_during_syscall", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_toolong_path() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("open_toolong_path", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_null_path() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("open_null_path", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_utimensat_null() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("utimensat_null", NONE)
        .status()
        .expect("execute syd");

    // Fails on GitLab CI.
    if !*GL_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }
    Ok(())
}

fn test_syd_normalize_path() -> TestResult {
    skip_unless_available!("sh");

    const NORMALIZE_PATH_TESTS: &[&str] = &[
        "null",
        "./null",
        ".////null",
        ".///.////.///./null",
        "./././././././null",
        "./././.././././dev/null",
        "../dev/././../dev/././null",
    ];

    for path in NORMALIZE_PATH_TESTS {
        let status = syd()
            .p("lib")
            .m("sandbox/write:on")
            .m("deny/write+/***")
            .m("allow/write+/dev/null")
            .argv(["sh", "-cx", &format!("cd /dev; :> {path}")])
            .status()
            .expect("execute syd");
        assert_eq!(
            status.code().unwrap_or(127),
            0,
            "path:{path}, status:{status:?}"
        );
    }

    Ok(())
}

fn test_syd_path_resolution() -> TestResult {
    let cwd = syd::fs::canonicalize(
        Pid::this(),
        "/proc/self/cwd",
        None,
        syd::fs::MissingHandling::Normal,
        true,
        false,
        false,
    )?
    .to_string_lossy()
    .to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("allow/write+{cwd}/***"))
        .do_("path_resolution", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_symlink_readonly_path() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/write+/")
        .argv([
            "sh",
            "-c",
            "ln -s / test_syd_symlink_readonly_path && unlink test_syd_symlink_readonly_path",
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_remove_empty_path() -> TestResult {
    skip_unless_available!("sh");

    let status = syd()
            .p("lib")
            .m("sandbox/read,stat,write:on")
            .m("allow/read,stat,write+/***")
        .argv([
            "sh",
            "-c",
            "env LC_ALL=C LANG=C LANGUAGE=C rm '' 2>&1 | tee /dev/stderr | grep -qi 'No such file or directory'"
        ])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_open_trailing_slash() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("open_trailing_slash", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_openat_trailing_slash() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("openat_trailing_slash", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_lstat_trailing_slash() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("lstat_trailing_slash", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fstatat_trailing_slash() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("fstatat_trailing_slash", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdir_trailing_dot() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("mkdir_trailing_dot", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdirat_trailing_dot() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("mkdirat_trailing_dot", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_rmdir_trailing_slashdot() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("rmdir_trailing_slashdot", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdir_eexist_escape() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/read,stat,write+/boot/***")
        .do_("mkdir_eexist_escape", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdirat_eexist_escape() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/read,stat,write+/boot/***")
        .do_("mkdirat_eexist_escape", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mknod_eexist_escape() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/read,stat,write+/boot/***")
        .do_("mknod_eexist_escape", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mknodat_eexist_escape() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .m("deny/read,stat,write+/boot/***")
        .do_("mknodat_eexist_escape", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fopen_supports_mode_x() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("fopen_supports_mode_x", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fopen_supports_mode_e() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("fopen_supports_mode_e", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_link_no_symlink_deref() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("link_no_symlink_deref", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_link_posix() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("link_posix", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_linkat_posix() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("linkat_posix", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_getcwd_long() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("getcwd_long", NONE)
        .status()
        .expect("execute syd");

    // FIXME: Fails on musl, possibly not something to fix.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_creat_thru_dangling() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("creat_thru_dangling", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_mkdirat_non_dir_fd() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("mkdirat_non_dir_fd", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_blocking_udp4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("trace/allow_safe_bind:0")
        .m("allow/read,stat,write+/***")
        .m("allow/net/bind+loopback!65432")
        .m("allow/net/connect+loopback!65432")
        .do_("blocking_udp4", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_blocking_udp6() -> TestResult {
    // Gitlab CI uses docker which has no IPv6.
    if *GL_BUILD {
        eprintln!("IPv6 not available on CI!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("trace/allow_safe_bind:0")
        .m("allow/read,stat,write+/***")
        .m("allow/net/bind+loopback6!65432")
        .m("allow/net/connect+loopback6!65432")
        .do_("blocking_udp6", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_close_on_exec() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .do_("close_on_exec", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_open_exclusive_restart() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .do_("open_exclusive_restart", NONE)
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_open_exclusive_repeat() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .do_("open_exclusive_repeat", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_pty_io_rust() -> TestResult {
    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .do_("pty_io_rust", NONE)
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_pty_io_gawk() -> TestResult {
    skip_unless_available!("gawk");

    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .do_("pty_io_gawk", NONE)
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_diff_dev_fd() -> TestResult {
    skip_unless_exists!("/dev/fd");
    skip_unless_available!("diff");

    env::set_var("SYD_TEST_TIMEOUT", "30s");
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("diff_dev_fd", NONE)
        .status()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_fifo_multiple_readers() -> TestResult {
    skip_unless_available!("bash", "nproc");

    env::set_var("SYD_TEST_TIMEOUT", "15s");
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .argv(["bash", "-c"])
        .arg(
            r#"
# Attempt to DOS syd by spawning multiple FIFO readers in the background.
set -ex
nreaders=$(expr $(nproc) '*' 2)
fifo=$(env TMPDIR=. mktemp -u)
mkfifo "$fifo"
for i in $(eval echo {1..${nreaders}}); do
    cat "$fifo" &
done
# Give the cats a little time to settle and potentially block syd.
sleep 3
# Execute a system call that syd must intervene, this must not block.
touch "$fifo".done
rm -f "$fifo".done
# All good, unblock the cats and wait.
:>"$fifo"
wait
rm -f "$fifo" || true
"#,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    env::remove_var("SYD_TEST_TIMEOUT");

    Ok(())
}

fn test_syd_bind_unix_socket() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .m("allow/net/bind+/***")
        .do_("bind_unix_socket", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_signal_protection_simple() -> TestResult {
    skip_unless_available!("bash", "kill");

    let status = syd()
        .p("lib")
        .argv(["bash", "-cx"])
        .arg(
            r#"
pid=$PPID
r=0

# Sending signal 0 must work.
kill -0 ${pid} || r=1
sleep 1

# Sending other signals are not permitted.
for sig in INT ABRT STOP KILL; do
    kill -${sig} ${pid} && r=3
    sleep 1
done

exit $r
"#,
        )
        .status()
        .expect("execute syd");

    assert_status_ok!(status);
    Ok(())
}

fn test_syd_signal_protection_killpg_0() -> TestResult {
    // killpg(exec process) does not propagate to syd.
    let status = syd()
        .p("lib")
        .do_("kill", ["0", "9"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_signal_protection_killpg_self() -> TestResult {
    // killpg syd's process group with signal 0 is OK.
    let status = syd()
        .p("lib")
        .do_("killpg_self", ["0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // killpg syd's process group with signal!=0 is denied.
    let status = syd()
        .p("lib")
        .do_("killpg_self", ["9"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_signal_protection_killpg_syd() -> TestResult {
    skip_unless_available!("bash");

    // kill(-sydpid) does not propagate to syd.
    let status = syd()
        .p("lib")
        .do__("kill")
        .argv(["bash", "-cx", &format!("{} -${{PPID}} 9", *SYD_DO)])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_signal_protection_mass_0() -> TestResult {
    // mass signaling is not permitted unless sig=0.
    let status = syd()
        .p("lib")
        .do_("kill", ["-1", "0"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_signal_protection_mass_int() -> TestResult {
    skip_unless_unshare!();

    // mass signaling is not permitted unless sig=0.
    let status = syd()
        .p("lib")
        .p("container")
        .do_("kill", ["-1", "2"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::EACCES);

    Ok(())
}

fn test_syd_emulate_open_fifo_1() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("emulate_open_fifo_1", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_emulate_open_fifo_2() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .do_("emulate_open_fifo_2", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);
    Ok(())
}

fn test_syd_deny_magiclinks() -> TestResult {
    skip_unless_unshare!();

    // Check protections with stat sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .p("paludis")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .do_("deny_magiclinks", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .do_("deny_magiclinks", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .m("lock:on")
        .do_("deny_magiclinks", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_lstat_magiclinks() -> TestResult {
    skip_unless_unshare!();

    // Check protections with stat sandboxing on.
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .p("paludis")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .do_("lstat_magiclinks", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .do_("lstat_magiclinks", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with stat sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .p("lib")
        .m("unshare/user:1")
        .m("unshare/pid:1")
        .m("lock:on")
        .do_("lstat_magiclinks", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_access_unsafe_paths() -> TestResult {
    // Check protections with read, stat, write sandboxing on.
    // As of version 3.19.0 we pass allow_unsafe_debug:1 in
    // the paludis profile which allows access to unsafe paths,
    // hence we explicitly disable it here...
    eprintln!("\x1b[36m<<< paludis >>>\x1b[0m");
    let status = syd()
        .p("paludis")
        .m("trace/allow_unsafe_debug:0")
        .do_("access_unsafe_paths", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with sandboxing off.
    eprintln!("\x1b[36m<<< lib >>>\x1b[0m");
    let status = syd()
        .p("lib")
        .do_("access_unsafe_paths", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    // Check protections with sandboxing off and lock on.
    eprintln!("\x1b[36m<<< lib with lock on >>>\x1b[0m");
    let status = syd()
        .p("lib")
        .m("lock:on")
        .do_("access_unsafe_paths", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_access_proc_cmdline() -> TestResult {
    skip_unless_available!("cat", "sh");

    let status = syd()
        .p("lib")
        .argv(["sh", "-cx"])
        .arg(
            r#"
cmdline=$(cat /proc/cmdline)
if test -n "$cmdline"; then
    echo >&2 "/proc/cmdline leaked with sandboxing off."
    false
else
    echo >&2 "/proc/cmdline is empty as expected."
    true
fi
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_code!(status, 1);

    let status = syd()
        .p("lib")
        .m("sandbox/read:on")
        .m("allow/read+/***")
        .argv(["sh", "-cx"])
        .arg(
            r#"
cmdline=$(cat /proc/cmdline)
if test -n "$cmdline"; then
    echo >&2 "/proc/cmdline leaked with sandboxing off."
    false
else
    echo >&2 "/proc/cmdline is empty as expected."
    true
fi
        "#,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_mkdir_with_control_chars_default() -> TestResult {
    skip_unless_available!("bash");

    // Create pipe for communication and check if syd raises an access violation.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .argv(["bash", "-cx"])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir mccd || exit 127
mkdir $'./mccd/test_alert_dir\a' && r=1
test -e $'./mccd/test_alert_dir\a' && r=2
mkdir $'./mccd/test_vertical_tab_dir\v' && r=3
test -e $'./mccd/test_vertical_tab_dir\v' && r=4
mkdir $'./mccd/test_form_feed_dir\f' && r=5
test -e $'./mccd/test_form_feed_dir\f' && r=6
mkdir $'./mccd/test_multi_control_dir\x01\x02\x03' && r=7
test -e $'./mccd/test_multi_control_dir\x01\x02\x03' && r=8
mkdir $'./mccd/test_\x1F_unit_sep_dir' && r=9
test -e $'./mccd/test_\x1F_unit_sep_dir' && r=10
exit $r
    "##,
        )
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) if buffer.contains("\"access\"") => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Ok(_) => {
            eprint!("Unexpected data, expected access violation:\n{buffer}");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_mkdir_with_control_chars_unsafe() -> TestResult {
    skip_unless_available!("bash");

    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_filename:1")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .argv(["bash", "-cx"])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir mccu || exit 127
mkdir $'./mccu/test_alert_dir\a' || r=1
test -e $'./mccu/test_alert_dir\a' || r=2
mkdir $'./mccu/test_vertical_tab_dir\v' || r=3
test -e $'./mccu/test_vertical_tab_dir\v' || r=4
mkdir $'./mccu/test_form_feed_dir\f' || r=5
test -e $'./mccu/test_form_feed_dir\f' || r=6
mkdir $'./mccu/test_multi_control_dir\x01\x02\x03' || r=7
test -e $'./mccu/test_multi_control_dir\x01\x02\x03' || r=8
mkdir $'./mccu/test_\x1F_unit_sep_dir' || r=9
test -e $'./mccu/test_\x1F_unit_sep_dir' || r=10
exit $r
    "##,
        )
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_touch_with_control_chars_default() -> TestResult {
    skip_unless_available!("bash");

    // Create pipe for communication and check if syd raises an access violation.
    let (fd_rd, fd_rw) = match pipe() {
        Ok((fd_rd, fd_rw)) => (fd_rd, fd_rw),
        Err(errno) => return Err(TestError(format!("error creating pipe: {errno}!"))),
    };

    let mut child = syd()
        .log("warn")
        .log_fd(fd_rw.as_raw_fd())
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .argv(["bash", "-cx"])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir tccd || exit 127
touch $'./tccd/test_alert_file\a' && r=1
test -e $'./tccd/test_alert_file\a' && r=2
touch $'./tccd/test_vertical_tab_file\v' && r=3
test -e $'./tccd/test_vertical_tab_file\v' && r=4
touch $'./tccd/test_form_feed_file\f' && r=5
test -e $'./tccd/test_form_feed_file\f' && r=6
touch $'./tccd/test_multi_control_file\x01\x02\x03' && r=7
test -e $'./tccd/test_multi_control_file\x01\x02\x03' && r=8
touch $'./tccd/test_\x1F_unit_sep_file' && r=9
test -e $'./tccd/test_\x1F_unit_sep_file' && r=10
exit $r
    "##,
        )
        .spawn()
        .expect("spawn syd");
    drop(fd_rw);

    let status = child.wait().expect("wait for syd");
    assert_status_ok!(status);

    // Convert raw file descriptor to File, then to BufReader
    let file = File::from(fd_rd);
    let mut reader = BufReader::new(file);

    let mut buffer = String::new();
    match reader.read_to_string(&mut buffer) {
        Ok(0) => {
            eprintln!("No data read from pipe; expected access violation.");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Ok(_) if buffer.contains("\"access\"") => {
            // If any data was read, log it
            eprint!("Access violation logged:\n{buffer}");
        }
        Ok(_) => {
            eprint!("Unexpected data, expected access violation:\n{buffer}");
            return Err(TestError(
                "Expected access violation not logged.".to_string(),
            ));
        }
        Err(e) => {
            // Error reading from the buffer
            eprintln!("Error reading from pipe: {e}");
            return Err(TestError(format!("Error reading from pipe: {e}")));
        }
    }

    Ok(())
}

fn test_syd_touch_with_control_chars_unsafe() -> TestResult {
    skip_unless_available!("bash");

    let status = syd()
        .p("lib")
        .m("trace/allow_unsafe_filename:true")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat,write+/***")
        .argv(["bash", "-cx"])
        .arg(
            r##"
#!/bin/bash
r=0
mkdir tccu || exit 127
touch $'./tccu/test_alert_file\a' || r=1
test -e $'./tccu/test_alert_file\a' || r=2
touch $'./tccu/test_vertical_tab_file\v' || r=3
test -e $'./tccu/test_vertical_tab_file\v' || r=4
touch $'./tccu/test_form_feed_file\f' || r=5
test -e $'./tccu/test_form_feed_file\f' || r=6
touch $'./tccu/test_multi_control_file\x01\x02\x03' || r=7
test -e $'./tccu/test_multi_control_file\x01\x02\x03' || r=8
touch $'./tccu/test_\x1F_unit_sep_file' || r=9
test -e $'./tccu/test_\x1F_unit_sep_file' || r=10
exit $r
    "##,
        )
        .status()
        .expect("spawn syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_unshare_user_bypass_limit() -> TestResult {
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .do_("unshare_user_bypass_limit", NONE)
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_fanotify_mark_cwd_allow() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("fanotify_mark", ["0", "0"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_ok!(status);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_cwd_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("fanotify_mark", ["0", "0"])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_dir_allow() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("fanotify_mark", &[cwd, "0".to_string()])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_ok!(status);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_dir_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("fanotify_mark", &[cwd, "0".to_string()])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_path_allow() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("fanotify_mark", &["0".to_string(), cwd])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_ok!(status);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_path_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("fanotify_mark", &["0".to_string(), cwd])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_dir_path_allow() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?;
    let (dir, path) = path_split(&cwd);
    let dir = dir.display().to_string();
    let path = path.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("fanotify_mark", &[dir, path])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_ok!(status);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_dir_path_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?;
    let (dir, path) = path_split(&cwd);
    let dir = dir.display().to_string();
    let path = path.display().to_string();
    let cwd = cwd.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("fanotify_mark", &[dir, path])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_symlink_allow() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    if let Err(error) = symlink("/var/empty/foo", "symlink") {
        eprintln!("Failed to create symbolic link, skipping: {error}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("fanotify_mark", &[cwd, "symlink".to_string()])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_ok!(status);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_fanotify_mark_symlink_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    if let Err(error) = symlink("/var/empty/foo", "symlink") {
        eprintln!("Failed to create symbolic link, skipping: {error}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("fanotify_mark", &[cwd, "symlink".to_string()])
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    if !matches!(code, nix::libc::ENOSYS | nix::libc::EPERM) {
        assert_status_code!(status, nix::libc::ENOENT);
    } else {
        eprintln!("fanotify API not supported or permitted, skipping!");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
    }

    Ok(())
}

fn test_syd_inotify_add_watch_path_allow() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("inotify_add_watch", &[cwd])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_inotify_add_watch_path_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("inotify_add_watch", &[cwd])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

fn test_syd_inotify_add_watch_symlink_allow() -> TestResult {
    if let Err(error) = symlink("/var/empty/foo", "symlink") {
        eprintln!("Failed to create symbolic link, skipping: {error}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .do_("inotify_add_watch", ["symlink"])
        .status()
        .expect("execute syd");
    assert_status_ok!(status);

    Ok(())
}

fn test_syd_inotify_add_watch_symlink_deny() -> TestResult {
    let cwd = current_dir(false)?.canonicalize()?.display().to_string();

    if let Err(error) = symlink("/var/empty/foo", "symlink") {
        eprintln!("Failed to create symbolic link, skipping: {error}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    // Inotify is disabled by default.
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write:on")
        .m("allow/read,stat+/***")
        .m(&format!("deny/stat+{cwd}/***"))
        .do_("inotify_add_watch", ["symlink"])
        .status()
        .expect("execute syd");
    assert_status_code!(status, nix::libc::ENOENT);

    Ok(())
}

fn test_syd_interrupt_mkdir() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .do_("interrupt_mkdir", NONE)
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_interrupt_bind_ipv4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .m("allow/net/bind+loopback!65432")
        .do_("interrupt_bind_ipv4", NONE)
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, they have a POC.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_interrupt_bind_unix() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .m("allow/net/bind+/***")
        .do_("interrupt_bind_unix", NONE)
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, check dev/seccomp_poc_no_lib.c
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_interrupt_connect_ipv4() -> TestResult {
    let status = syd()
        .p("lib")
        .m("sandbox/read,stat,write,net:on")
        .m("allow/read,stat,write+/***")
        .m("allow/net/bind+loopback!65432")
        .m("allow/net/connect+loopback!65432")
        .do_("interrupt_connect_ipv4", NONE)
        .status()
        .expect("execute syd");
    // FIXME: This is a kernel bug, mixi will report it, they have a POC.
    ignore!(status.success(), "status:{status:?}");
    Ok(())
}

fn test_syd_exp_syscall_fuzz_bare() -> TestResult {
    let epoch = std::time::Instant::now();
    let status = syd()
        .p("lib")
        .m("lock:off")
        .m("sandbox/read,stat,write,exec,net:on")
        .m("allow/read,stat,exec,write+/***")
        .do_("syscall_fuzz", NONE)
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_exp_syscall_fuzz_wrap() -> TestResult {
    skip_unless_unshare!();

    let epoch = std::time::Instant::now();
    let status = syd()
        .p("lib")
        .p("container")
        .m("lock:off")
        .m("sandbox/read,stat,write,exec,net:on")
        .m("allow/read,stat,exec,write+/***")
        .do_("syscall_fuzz", NONE)
        .status()
        .expect("execute syd");
    let code = status.code().unwrap_or(127);
    let time = format_duration(epoch.elapsed());
    println!("# fuzz completed in {time} with code {code}.");
    ignore!(code == 0, "status:{status:?}");
    Ok(())
}

fn test_syd_pid_thread_deny() -> TestResult {
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:deny")
        .m("pid/max:1")
        .do_("thread", ["0", "24"])
        .status()
        .expect("execute syd");
    // Early fails get EACCES ie 13. However rust panics have arbitrary
    // exit values on different systems and on different build profiles,
    // e.g. we get 101 in debug mode, 134 in release mode, and 139 on
    // Alpine Linux... Let's check for error to keep things simple and stable.
    assert_status_not_ok!(status);

    Ok(())
}

fn test_syd_pid_fork_deny() -> TestResult {
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:deny")
        .m("pid/max:16")
        .do_("fork", ["0", "24"])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == nix::libc::EACCES,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_pid_fork_bomb() -> TestResult {
    skip_unless_unshare!();

    eprintln!("Running the fork bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:deny")
        .m("pid/max:16")
        .do_("fork_bomb", NONE)
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_pid_fork_bomb_asm() -> TestResult {
    skip_unless_unshare!();

    eprintln!("Running the fork bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:deny")
        .m("pid/max:16")
        .do_("fork_bomb_asm", NONE)
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_pid_thread_bomb() -> TestResult {
    skip_unless_unshare!();

    eprintln!("Running the thread bomb test for 10 seconds...");
    env::set_var("SYD_TEST_TIMEOUT", "10s");
    let child = syd()
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:deny")
        .m("pid/max:16")
        .do_("thread_bomb", NONE)
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("execute syd");
    env::remove_var("SYD_TEST_TIMEOUT");

    eprintln!("Done spawning test, waiting test to complete...");
    let output = child.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");

    // Check if the output contains the specified string and assert based on it
    assert!(
        output.contains(r#""cap":"p""#),
        "syd has not raised a pid violation!"
    );

    Ok(())
}

fn test_syd_exp_pid_stress_ng_kill() -> TestResult {
    skip_unless_available!("stress-ng");
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:kill")
        .m("pid/max:1")
        .argv(["stress-ng", "-c", "1", "-t", "7"])
        .status()
        .expect("execute syd");
    assert!(
        status.code().unwrap_or(127) == 128 + nix::libc::SIGKILL,
        "status:{status:?}"
    );

    Ok(())
}

fn test_syd_exp_pid_stress_ng_allow() -> TestResult {
    skip_unless_available!("stress-ng");
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:deny")
        .m("pid/max:2")
        .argv(["stress-ng", "--log-file", "log", "-c", "1", "-t", "7"])
        .status()
        .expect("execute syd");
    // Fails on CI.
    if !*CI_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let mut file = File::open("log")?;
    let mut logs = String::new();
    file.read_to_string(&mut logs)?;

    assert!(!logs.contains("errno="), "logs:{logs:?}");

    Ok(())
}

fn test_syd_exp_pid_stress_ng_fork() -> TestResult {
    skip_unless_available!("stress-ng");
    skip_unless_unshare!();

    let status = syd()
        .p("lib")
        .p("container")
        .m("sandbox/pid:on")
        .m("default/pid:filter")
        .m("pid/max:128")
        .argv([
            "stress-ng",
            "--log-file",
            "log",
            "-f",
            "4",
            "-t",
            "15",
            "--fork-max",
            "1024",
        ])
        .status()
        .expect("execute syd");
    // Fails on CI.
    if !*CI_BUILD {
        assert_status_ok!(status);
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    let mut file = File::open("log")?;
    let mut logs = String::new();
    file.read_to_string(&mut logs)?;
    assert!(!logs.contains("errno="), "logs:{logs:?}");

    Ok(())
}

fn test_syd_mem_alloc() -> TestResult {
    let status = syd()
        .env("SYD_TEST_FORCE", "IKnowWhatIAmDoing")
        .p("lib")
        .m("sandbox/mem:on")
        .m("mem/max:128M")
        .m("mem/vm_max:256M")
        .do_("alloc", NONE)
        .status()
        .expect("execute syd");
    // This test times out on GITLAB CI.
    // TODO: Investigate, see: #166.
    if !*GL_BUILD {
        // Segmentation fault is expected.
        // IOT is confusing but happens on alpine+musl.
        // Otherwise we require ENOMEM.
        assert!(
            matches!(
                status.code().unwrap_or(127),
                nix::libc::ENOMEM | EX_SIGIOT | EX_SIGSEGV
            ),
            "status:{status:?}"
        );
    } else {
        ignore!(status.success(), "status:{status:?}");
        env::set_var("SYD_TEST_SOFT_FAIL", "1");
        return Ok(());
    }

    Ok(())
}

fn test_syd_exp_mem_stress_ng_malloc_1() -> TestResult {
    skip_unless_available!("stress-ng");
    skip_unless_unshare!();

    let command = syd()
        .p("lib")
        .p("container")
        .m("sandbox/mem:on")
        .m("mem/max:32M")
        .m("mem/vm_max:256M")
        .argv([
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--malloc",
            "4",
            "--malloc-bytes",
            "128M",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    assert!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

fn test_syd_exp_mem_stress_ng_malloc_2() -> TestResult {
    skip_unless_available!("stress-ng");
    skip_unless_unshare!();

    let command = syd()
        .p("lib")
        .p("container")
        .m("sandbox/mem:on")
        .m("mem/max:32M")
        .m("mem/vm_max:256M")
        .argv([
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--malloc",
            "4",
            "--malloc-bytes",
            "128M",
            "--malloc-touch",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    assert!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

fn test_syd_exp_mem_stress_ng_mmap() -> TestResult {
    skip_unless_available!("stress-ng");
    skip_unless_unshare!();

    let command = syd()
        .p("lib")
        .p("container")
        .m("sandbox/mem:on")
        .m("mem/max:16M")
        .m("mem/vm_max:64M")
        .argv([
            "stress-ng",
            "-v",
            "-t",
            "5",
            "--mmap",
            "4",
            "--mmap-bytes",
            "1G",
        ])
        .stdout(Stdio::inherit())
        .stderr(Stdio::piped())
        .spawn()
        .expect("spawn syd");

    let output = command.wait_with_output().expect("wait syd");
    let output = String::from_utf8_lossy(&output.stderr);
    eprintln!("{output}");
    fixup!(output.contains(r#""cap":"m""#), "out:{output:?}");

    Ok(())
}

const OPENSSL_AES_CTR_CODE: &str = r##"
/*
 * aes-ctr: AES-CTR Encryption and Decryption Utility
 * Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
 * SPDX-License-Identifier: GPL-3.0-or-later
 *
 * This program uses OpenSSL to encrypt and decrypt data using AES-CTR
 * mode. It supports both encryption and decryption operations, with
 * key and IV provided as hexadecimal strings.
 *
 * Compile:
 *        cc aes-ctr.c -o aes-ctr -lssl -lcrypto
 *
 * Usage: aes-ctr [-hV] -e|-d -k <hex-encoded key> -i <hex-encoded iv>
 *
 * Options:
 *   -h        Print this help message and exit
 *   -V        Print version information and exit
 *   -e        Encrypt the input data
 *   -d        Decrypt the input data
 *   -k <key>  Hex-encoded key (256 bits for AES-CTR)
 *   -i <iv>   Hex-encoded IV (128 bits for AES-CTR)
 *
 * Examples:
 *   Encrypt: echo -n "data" | ./aes-ctr -e -k <key> -i <iv>
 *   Decrypt: ./aes-ctr -d -k <key> -i <iv> < encrypted_data.bin
 */

#include <openssl/evp.h>
#include <openssl/err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>

#define VERSION "0.1.0"
#define KEY_LEN 32
#define IV_LEN 16
#define BUFFER_SIZE 4096

void handleErrors(void)
{
	ERR_print_errors_fp(stderr);
	exit(127);
}

/**
 * Memory allocation with error checking.
 * @param size The size of memory to allocate.
 * @return Pointer to the allocated memory.
 */
void *xmalloc(size_t size)
{
	void *ptr = malloc(size);
	if (!ptr) {
		fprintf(stderr, "Memory allocation failed\n");
		exit(127);
	}
	return ptr;
}

/**
 * Memory reallocation with error checking.
 * @param ptr The original memory pointer.
 * @param size The new size of memory to allocate.
 * @return Pointer to the reallocated memory.
 */
void *xrealloc(void *ptr, size_t size)
{
	ptr = realloc(ptr, size);
	if (!ptr) {
		fprintf(stderr, "Memory reallocation failed\n");
		exit(127);
	}
	return ptr;
}

/**
 * Convert a hexadecimal string to a byte array.
 * @param hex The hexadecimal string.
 * @param bytes The output byte array.
 * @param bytes_len The length of the output byte array.
 */
void hex_to_bytes(const char *hex, unsigned char *bytes, size_t bytes_len)
{
	for (size_t i = 0; i < bytes_len; i++) {
		sscanf(hex + 2 * i, "%2hhx", &bytes[i]);
	}
}

/**
 * Encrypt or decrypt data using AES-CTR.
 * @param encrypt Set to 1 for encryption, 0 for decryption.
 * @param key The AES key.
 * @param iv The IV (tweak) value.
 */
void process_data(int encrypt, const unsigned char *key,
                  const unsigned char *iv)
{
	EVP_CIPHER_CTX *ctx = EVP_CIPHER_CTX_new();
	if (!ctx) handleErrors();

	if (encrypt) {
		if (1 != EVP_EncryptInit_ex(ctx, EVP_aes_256_ctr(), NULL, key,
		                            iv)) handleErrors();
	} else {
		if (1 != EVP_DecryptInit_ex(ctx, EVP_aes_256_ctr(), NULL, key,
		                            iv)) handleErrors();
	}

	unsigned char inbuf[BUFFER_SIZE];
	unsigned char outbuf[BUFFER_SIZE + EVP_CIPHER_block_size(EVP_aes_256_ctr())];
	int inlen, outlen;

	while ((inlen = fread(inbuf, 1, BUFFER_SIZE, stdin)) > 0) {
		if (encrypt) {
			if (1 != EVP_EncryptUpdate(ctx, outbuf, &outlen, inbuf, inlen))
				handleErrors();
		} else {
			if (1 != EVP_DecryptUpdate(ctx, outbuf, &outlen, inbuf, inlen))
				handleErrors();
		}
		fwrite(outbuf, 1, outlen, stdout);
	}

	if (encrypt) {
		if (1 != EVP_EncryptFinal_ex(ctx, outbuf, &outlen))
			handleErrors();
	} else {
		if (1 != EVP_DecryptFinal_ex(ctx, outbuf, &outlen))
			handleErrors();
	}
	fwrite(outbuf, 1, outlen, stdout);

	EVP_CIPHER_CTX_free(ctx);
}

int main(int argc, char *argv[])
{
	int opt;
	int encrypt = -1;
	char *key_hex = NULL;
	char *iv_hex = NULL;

	while ((opt = getopt(argc, argv, "hVdek:i:")) != -1) {
		switch (opt) {
		case 'h':
			printf("Usage: aes-ctr [-hV] -e|-d -k <hex-encoded key> -i <hex-encoded iv>\n");
			printf("  -h        Print this help message and exit\n");
			printf("  -V        Print version information and exit\n");
			printf("  -e        Encrypt the input data\n");
			printf("  -d        Decrypt the input data\n");
			printf("  -k <key>  Hex-encoded key (256 bits for AES-CTR)\n");
			printf("  -i <iv>   Hex-encoded IV (128 bits for AES-CTR)\n");
			return EXIT_SUCCESS;
		case 'V':
			printf("%s v%s\n", argv[0], VERSION);
			return EXIT_SUCCESS;
		case 'e':
			encrypt = 1;
			break;
		case 'd':
			encrypt = 0;
			break;
		case 'k':
			key_hex = optarg;
			break;
		case 'i':
			iv_hex = optarg;
			break;
		default:
			fprintf(stderr,
			        "Usage: aes-ctr [-hV] -e|-d -k <hex-encoded key> -i <hex-encoded iv>\n");
			return EXIT_FAILURE;
		}
	}

	if (encrypt == -1 || key_hex == NULL || iv_hex == NULL) {
		fprintf(stderr, "Error: -e or -d and both -k and -i options are required.\n");
		fprintf(stderr,
		        "Usage: aes-ctr [-hV] -e|-d -k <hex-encoded key> -i <hex-encoded iv>\n");
		return EXIT_FAILURE;
	}

	if (strlen(key_hex) != 2 * KEY_LEN) {
		fprintf(stderr, "Error: Key must be 256 bits (32 bytes) in length.\n");
		return 1;
	}

	if (strlen(iv_hex) != 2 * IV_LEN) {
		fprintf(stderr, "Error: IV must be 128 bits (16 bytes) in length.\n");
		return 1;
	}

	// AES-256-CTR requires a 256-bit key (32 bytes)
	unsigned char key[KEY_LEN];
	// CTR mode uses a 128-bit tweak (16 bytes)
	unsigned char iv[IV_LEN];
	hex_to_bytes(key_hex, key, sizeof(key));
	hex_to_bytes(iv_hex, iv, IV_LEN);

	process_data(encrypt, key, iv);

	return EXIT_SUCCESS;
}
"##;

fn build_openssl_aes_ctr() {
    // Write the C code to a temporary file.
    match File::create("aes-ctr.c") {
        Ok(mut file) => {
            if let Err(e) = file.write_all(OPENSSL_AES_CTR_CODE.as_bytes()) {
                eprintln!("Failed to write to file aes-ctr.c: {e}");
                return;
            }
        }
        Err(e) => {
            eprintln!("Failed to create file aes-ctr.c: {e}");
            return;
        }
    }

    // Compile the C code into a binary.
    let status = Command::new("cc")
        .args([
            "aes-ctr.c",
            "-o",
            "aes-ctr",
            "-lssl",
            "-lcrypto",
            "-Wall",
            "-Wextra",
        ])
        .stdin(Stdio::null())
        .stdout(Stdio::inherit())
        .stderr(Stdio::inherit())
        .status();

    match status {
        Ok(status) => {
            if !status.success() {
                eprintln!("Compilation of aes-ctr failed with status: {status}");
            }
        }
        Err(e) => {
            eprintln!("Failed to execute aes-ctr compile command: {e}");
        }
    }
}
